#include "config.h"
#ifndef OLDGCC
#pragma implementation
#endif

#include "AuxMethodMap.h"
#include "Instantiator.h"
#include <stdio.h>
#include "find_leaks.h"

//
// Method name : (constructor)
// Description : Creates a new, empty AuxMethodMap
//
CAuxMethodMap::CAuxMethodMap(const unicode_string &name)
  : fClassName(name)
{
  REF_PTR(this);
}

//
// Method name : (copy constructor)
// Description : The copy constructor.
//
CAuxMethodMap::CAuxMethodMap(const CAuxMethodMap& other)
  : fClassName(other.fClassName)
{
  REF_PTR(this);
  for (amm_rep::const_iterator i = other.fMap.begin();
       i != other.fMap.end(); ++i) {
    fMap.insert(amm_entry((*i).first, 
			  new unicode_string(*(*i).second)));
  }
}

//
// Method name : (destructor)
//
CAuxMethodMap::~CAuxMethodMap()
{
  UNREF_PTR(this);
  for(amm_rep::iterator i = fMap.begin();
      i != fMap.end(); ++i) {
    delete (*i).second;
  }
}

//
// Method name : RequireMethod
// Description : Add an entry to this AuxMethodMap, but do not name it yet.
//    Copies Name, Method, and Throws if they're given.
//   
void
CAuxMethodMap::RequireMethod(CAuxMethodSignature::AuxMethodType Type,
	   const CJavaTypeSignature &Class,
	   const unicode_string *Name = 0,
	   const CJavaMethodSignature *Method = 0,
	   const deque<CJavaTypeSignature> *Throws = 0,
	   CAuxMethodSignature::WhereType WType = CAuxMethodSignature::Virtual,
	   unsigned Width = 0, unsigned Dim = 0)
{  
  unicode_string *real_name = Name ? new unicode_string(*Name) : 0;
  CJavaMethodSignature *real_method = 
    Method ? new CJavaMethodSignature(*Method) : 0;
  deque<CJavaTypeSignature> *real_throws =
    Throws ? new deque<CJavaTypeSignature>(*Throws) : 0;
  CAuxMethodSignature ams(Type,Class,real_name,real_method,real_throws,
			  WType, Width, Dim);
  amm_rep::iterator i = fMap.find(ams);
  if (i == fMap.end()) {
    fMap.insert(amm_entry(ams,(const unicode_string *)0));
  }
}

//
// Method Name : NameMethods
// Description : Assign a unique auxmethod name to every as-yet-unnamed entry
//    in this map, where currentClass is the name of the current class, to
//    be used for name-mangling.
//
void 
CAuxMethodMap::NameMethods()
{
  for(amm_rep::iterator m = fMap.begin(); m != fMap.end(); ++m) {
    if ((*m).second == 0) {
      unicode_string *result = new unicode_string;
      NameForMethod((*m).first, *result);
      (*m).second = result;
    }
  }
}

//
// Method name : AuxMethodName
// Description : Returns the name of the AuxMethod encoding a given operation.
//       Adds the auxmethod to the map if it is not there already.
// Requires    : All parent AMM's have been adopted.
//
unicode_string
CAuxMethodMap::AuxMethodName(CAuxMethodSignature::AuxMethodType Type,
			     const CJavaTypeSignature &Class,
			     const unicode_string *Name,
			     const CJavaMethodSignature *Method ,
			     const deque<CJavaTypeSignature> *Throws ,
			     CAuxMethodSignature::WhereType WType,
			     unsigned Width, unsigned Dim)
{
  // #define AMNDEBUG
  unicode_string *real_name = Name ? new unicode_string(*Name) : 0;
  CJavaMethodSignature *real_method = 
    Method ? new CJavaMethodSignature(*Method) : 0;
  deque<CJavaTypeSignature> *real_throws =
    Throws ? new deque<CJavaTypeSignature>(*Throws) : 0;
  if (!real_throws && Type == CAuxMethodSignature::Where) 
    real_throws = new deque<CJavaTypeSignature>();
  CAuxMethodSignature ams(Type,Class,real_name,real_method,real_throws,
			  WType, Width, Dim);
  
  amm_rep::iterator i = fMap.find(ams);

  if (i == fMap.end()) {
    unicode_string name;
    NameForMethod(ams,name);
    fMap.insert(amm_entry(ams,new unicode_string(name)));
    return name;
  } else {
    assert((*i).second);
    return *((*i).second);
  }
}

//
// Method name : AuxMethodName
// Description : Returns the name of an AuxMethod with this type and name.
//     Used for accessing static members.
// Requires    : All parent AMM's have been adopted, and NameMethods have been
//               called. The auxmethod in question exists.
//
unicode_string
CAuxMethodMap::AuxMethodName(CAuxMethodSignature::AuxMethodType Type,
			     const unicode_string &Name)
{
  for (amm_rep::iterator i = fMap.begin(); i != fMap.end(); ++i) {    
    if (Type == (*i).first.fType) {
      if ((*i).first.fName && *(*i).first.fName == Name) {
	return *(*i).second;
      }
    }
  }

  // This should never happen. 
  cerr << "Couldn't find name for named auxmethod for "<< Name << endl
       << "Possibilities include:" << endl;
  for (amm_rep::iterator j = fMap.begin(); j != fMap.end(); ++j) {
    cerr << (*j).first << endl;
  }
  assert(0);
}

//
//  Method name: ASCParse
//  Description: Parse a single ascii-formatted auxmethod entry and add it
//     to the map. Return true if successful.
//
bool
CAuxMethodMap::ASCParse(const deque<unicode_string> &s)
{
  unicode_string *am_name = new unicode_string(s[0]);
  
  string am_type = ::UnicodeToUTF(s[1]);
  CAuxMethodSignature::AuxMethodType type;
  if (am_type == "Where") {
    type = CAuxMethodSignature::Where;
  } else if (am_type == "ArrayStore") {
    type = CAuxMethodSignature::ArrayStore;
  } else if (am_type == "ArrayFetch") {
    type = CAuxMethodSignature::ArrayFetch;
  } else if (am_type == "ArrayLength") {
    type = CAuxMethodSignature::ArrayLength;
  } else if (am_type == "ArrayNew") {
    type = CAuxMethodSignature::ArrayNew;
  } else if (am_type == "GetWhere") {
    type = CAuxMethodSignature::GetWhere;
  } else if (am_type == "FetchStatic") {
    type = CAuxMethodSignature::FetchStatic;
  } else if (am_type == "SetStatic") {
    type = CAuxMethodSignature::SetStatic;  
  } else if (am_type == "InstanceOf") {
    type = CAuxMethodSignature::InstanceOf;
  } else if (am_type == "Cast") {
    type = CAuxMethodSignature::Cast;
  } else if (am_type == "PSig") {
    type = CAuxMethodSignature::PSig;
  } else if (am_type == "Default") {
    type = CAuxMethodSignature::Default;
  } else {
    cerr << "Bad AMM entry: unknown method type." << endl;
    return false;
  }

  CJavaTypeSignature am_class;
  if (!am_class.ParseSignatureString(s[2])) {
    cerr << "Bad AMM entry: bad 'Class' field." << endl;
    return false;
  }

  CAuxMethodSignature::WhereType wt = CAuxMethodSignature::Virtual;
  unicode_string *name = 0;
  CJavaMethodSignature *method = 0;
  CJavaMethodSignature *firstMethod = 0;
  CJavaTypeSignature *firstClass = 0;
  deque<CJavaTypeSignature> *throws = 0;
  unsigned dim = 0;
  unsigned width = 0;

  switch(type) 
    {
    case CAuxMethodSignature::Where: {      
      string str = ::UnicodeToUTF(s[3]);
      if (s.size() != 8) {
	cerr << "Bad AMM entry: bad 'Where' line." << endl;
	return false;
      }
      if (str == "C") 
	wt = CAuxMethodSignature::Constructor;
      else if (str == "S")
	wt = CAuxMethodSignature::Static;
      else if (str == "V")
	wt = CAuxMethodSignature::Virtual;
      else {
	cerr << "Bad AMM entry: invalid constraint method type." << endl;
	return false;
      }
      method = new CJavaMethodSignature;
      firstMethod = new CJavaMethodSignature;
      name = new unicode_string(s[7]);
      if (!method->Initialize(*name,s[4])) {
	cerr << "Bad AMM entry: bad 'method' field." << endl;
	return false;
      }
      if (!firstMethod->Initialize(*name,s[5])) {
	cerr << "Bad AMM entry: bad 'name' field." << endl;
	return false;      
      }
      throws = new deque<CJavaTypeSignature>;
      unicode_string::const_iterator i = s[6].begin();
      while (i != s[6].end()) {
	CJavaTypeSignature t;
	if (!t.ParseSignatureString(i,s[6].end())) {
	  cerr << "Bad AMM entry: bad 'throws' field." << endl;
	  return false;
	}
	throws->push_back(t);
      }
    }
    break;
    case CAuxMethodSignature::ArrayNew: {
      if (s.size() != 5) {
	cerr << "Bad AMM entry: bad 'ArrayNew' line." << endl;
	return false;
      }
      width = atoi(::UnicodeToUTF(s[3]).c_str());
      dim = atoi(::UnicodeToUTF(s[4]).c_str());
    }
    break;
    case CAuxMethodSignature::SetStatic:
    case CAuxMethodSignature::FetchStatic: {
      if (s.size() != 5) {
	cerr << "Bad AMM entry: bad '*Static' line." << endl;
	return false;
      }
      firstClass = new CJavaTypeSignature;
      if (!firstClass->ParseSignatureString(s[3])) {
	cerr << "Bad AMM entry: bad 'firstClass' field." << endl;
	return false;
      }
      name = new unicode_string(s[4]);
    }
    break;
    case CAuxMethodSignature::Cast: {
      if (s.size() !=  4) {
	cerr << "Bad AMM entry: bad 'Cast' line." << endl;	
	return false;
      }
      firstClass = new CJavaTypeSignature;
      if (!firstClass->ParseSignatureString(s[3])) {
	cerr << "Bad AMM entry: bad 'firstClass' field." << endl;	
	return false;
      }
    }
    break;
    default:
      break;
    }  
  
  CAuxMethodSignature ams(type,
			  am_class,
			  name,
			  method,
			  throws,
			  wt,
			  width, dim,
			  firstMethod,
			  firstClass);
  fMap.insert(amm_entry(ams, am_name));

  return true;
}


//
// Method Name : NameForMethod
// Description : Computes a unique name for the auxmethod m.  Stores
//    that name in setName.
// Checks      : The class name for this AMM has been set.
//
void 
CAuxMethodMap::NameForMethod(const CAuxMethodSignature &m,
			     unicode_string &setName) const
{
  assert(fClassName.length() > 0);

  unicode_string result = ::UTFToUnicode("$");
  switch(m.fType) 
    {
    case CAuxMethodSignature::Where:
      result += ::UTFToUnicode("Where");
      break;
    case CAuxMethodSignature::ArrayStore:
      result += ::UTFToUnicode("AStore");
      break;     
    case CAuxMethodSignature::ArrayFetch:
      result += ::UTFToUnicode("AFetch");
      break;     
    case CAuxMethodSignature::ArrayLength:
      result += ::UTFToUnicode("ALen");
      break;     
    case CAuxMethodSignature::ArrayNew:
      result += ::UTFToUnicode("ANew");
      break;     
    case CAuxMethodSignature::GetWhere:
      result += ::UTFToUnicode("GetWhere");
      break;     
    case CAuxMethodSignature::SetStatic:
      result += ::UTFToUnicode("SetS");
      break;     
    case CAuxMethodSignature::FetchStatic:
      result += ::UTFToUnicode("FetchS");
      break;     
    case CAuxMethodSignature::InstanceOf:
      result += ::UTFToUnicode("InstanceOf");
      break;
    case CAuxMethodSignature::Cast:
      result += ::UTFToUnicode("Cast");
      break;
    case CAuxMethodSignature::PSig:
      result += ::UTFToUnicode("PSig");
      break;
    case CAuxMethodSignature::Default:
      result += ::UTFToUnicode("Default");
      break;
    default:
      assert(0);
    }
  result += ::UTFToUnicode("$");
  // Find the lowest unassigned auxmethod number
  int i = 1;
  unicode_string result_nn = result;
  bool exists;
  do {
    char buf[10];
    sprintf(buf, "%d", i);
    result = result_nn + ::UTFToUnicode(buf);
    ++i;
    exists = false;
    for(amm_rep::const_iterator it = fMap.begin(); it != fMap.end(); ++it) {
      if ((*it).second && (*(*it).second == result)) {
	exists = true;
	break;	    
      }
    }
  } while (exists);
  setName = result;
}
