
#ifndef _AuxMethodSignature_h
#define _AuxMethodSignature_h
#pragma interface

#include "JavaTypeSignature.h"
#include <set>

class srcstream;
class ostream;
class CJavaMethodSignature;

typedef set<CJavaTypeSignature, less<CJavaTypeSignature> > TypeSet;

//
// Class name : CAuxMethodSignature
// Description: This class encapsulates the signature for an AuxMethod,
//    and provides a "less" function so we can use it as a key in a map.
//    It's basically for the use of CAuxMethodMap.
//
class CAuxMethodSignature {
public:
  typedef enum {
    UNK,
    Where,       // Constraint method
    ArrayStore,  // Store an element in an array
    ArrayFetch,  // Fetch an element in an array
    ArrayLength, // Get the length of an array
    ArrayNew,    // Create a new array
    GetWhere,    // Obtain a WObject for another class.
    SetStatic,   // Set a static variable
    FetchStatic, // Fetch a static variable
    PSig,        // Returns a signature string for a given parameter.
    Cast,        // Cast to a type.
    InstanceOf,  // Test for "instanceof"
    Default      // Returns T.default for some T.
  } AuxMethodType;
  typedef enum {
    Constructor,
    Static,
    Virtual
  } WhereType;
  
  CAuxMethodSignature();
  CAuxMethodSignature(const CAuxMethodSignature& other);

  CAuxMethodSignature(AuxMethodType type, 
		      const CJavaTypeSignature &Class,
		      unicode_string *name = 0,
		      CJavaMethodSignature *method = 0,
		      deque<CJavaTypeSignature> *throws = 0,
		      WhereType wtype = Virtual,
		      unsigned width = 0, unsigned dim = 0, 
		      CJavaMethodSignature *firstMethod = 0,
		      CJavaTypeSignature *firstClass = 0);

  ~CAuxMethodSignature();

  int compare(const CAuxMethodSignature& other) const;
  bool operator<(const CAuxMethodSignature& other) const
     { return (compare(other) < 0); }
  bool operator==(const CAuxMethodSignature& other) const
     { return (compare(other) == 0); }
  bool operator!=(const CAuxMethodSignature& other) const
     { return (compare(other) != 0); }

  string ASCCompile() const;
  bool IsInstantiated() const;

  void Unparse(const unicode_string& name, 
	       srcstream &sOut,
	       TypeSet& instantiationList,
	       CTypeResolver& tRes,
	       const CJavaTypeSignature &inType,
	       bool headerOnly) const;

  void Disassemble(ostream &out) const;
protected:
  friend class CAuxMethodMap;
  friend class CInstantiator;
  friend class CCompiler;

  void UnparseArgsToCall(srcstream &sOut,
			 const CJavaMethodSignature &transMethod,
			 CTypeResolver& tRes) const;
  void UnparsePrimitiveMethod(srcstream &sOut) const;
  void UnparsePrimitiveConstructor(srcstream &sOut) const;
  string UnparseUnwrapArg(const string& unwrap,
			  const CJavaTypeSignature& firstType,
			  const CJavaTypeSignature& type) const;
 
  // For *
  AuxMethodType fType;
  CJavaTypeSignature fClass;
  // For Where
  WhereType fWType;
  CJavaMethodSignature *fMethod;
  CJavaMethodSignature *fFirstMethod; 
  deque<CJavaTypeSignature>* fThrows;
  // For Where, *Static.
  unicode_string *fName;
  // For ArrayNew
  unsigned fWidth;
  unsigned fDim;
  // For *Static
  CJavaTypeSignature *fFirstClass;
  bool repOK() const;
};

ostream& operator<<(ostream &o, const CAuxMethodSignature &ams);

#endif // _AuxMethodSignature_h

/*
 * WHEN THE PARSER SEES                      ADD THE A.M.S 
 *
 * (where S,T are parameters,           < Type,class,(name,method) >
 *  a,b,c are types, p,q are 
 *  expressions, idn is an identifier,
 *  and C is the current class.)
 *
 * ---------------------                ------------------------
 * where T { a idn(c,d); }              < WhereV, T, idn, a idn(c,d) >
 * where T { static a idn(c,d); }       < WhereS, T, idn, a idn(c,d) >
 * where T ( T(c,d); }                  < WhereC, T, <init>, T(c,d) >
 * T[]                                  < ArryStore, T >
 *                                      < ArryFetch, T >
 *                                      < ArryLength, T >
 * new T[p][q][][][][]                  < ArryNew, T, "2,4" >
 * new a[S,b](c)                        < NewOther, a[S,b], "<init>", a(c) >
 * a[S,T,b].idn(...)                    < GetWhere, a[S,T,b] > 
 * static a idn; (in a PClass)          < SetVar, a, idn, C void() >
 *                                      < FetchVar, a, idn, C void() >
 * p instanceof T                       < InstanceOf, T >
 * p instanceof a[S,T,b]                < InstanceOf, a[S,T,b] >
 * class X[S]                           < PMatch, S >
 * (type[S,b]) x                        < TCast, type[S,b] >
 * (S) x                                < PCast, S >
 */
