// $Id: JavaMethodSignature.C,v 1.15 1998/08/20 18:28:14 nickm Exp $
// Copyright (c) 1995  David Engberg  All rights reserved
#pragma implementation

#include "JavaMethodSignature.h"
#include "Srcstream.h"
#include "find_leaks.h"

const unicode_string CJavaMethodSignature::kConstructorName 
    = ::StringToUnicode("<init>");

//
//  Method name : CJavaMethodSignature
//  Description : Default constructor.
//
CJavaMethodSignature::CJavaMethodSignature()
{
  REF_PTR(this);
}

//
//  Method name : CJavaMethodSignature
//  Description : Constructs a method signature out of a the constituent
//    parts.
//
CJavaMethodSignature::CJavaMethodSignature(
			      const CJavaTypeSignature& returnType,
			      const unicode_string& name,
			      const deque<CJavaTypeSignature>& argumentTypes)
  : fMethodName(name),
    fArgumentTypes(argumentTypes),
    fReturnType(returnType)
{
  REF_PTR(this);
  SetHashKey();
}

//
//  Method name : CJavaMethodSignature
//  Description : Copy constructor.
//
CJavaMethodSignature::CJavaMethodSignature(const CJavaMethodSignature& source)
  : fMethodName(source.fMethodName),
    fArgumentTypes(source.fArgumentTypes),
    fReturnType(source.fReturnType),   
    fHashKey(source.fHashKey)
{
  REF_PTR(this);
}

//
//  Method name : ~CJavaMethodSignature
//  Description : Destructor
//
CJavaMethodSignature::~CJavaMethodSignature()
{
  UNREF_PTR(this);
}

//
//  Method name : operator=
//  Description : Assignment operator.
//
CJavaMethodSignature&
CJavaMethodSignature::operator=(const CJavaMethodSignature& source)
{
  if (&source != this) {
    fReturnType = source.fReturnType;
    fArgumentTypes = source.fArgumentTypes;
    fHashKey = source.fHashKey;
    fMethodName = source.fMethodName;
  }
  return *this;
}

//
//  Method name : EqualsIgnoreReturn
//  Description : Returns true if the name and arguments of this signature
//    match the name and arguments of the provided signature.  The return
//    types do not affect the result of this function.
//
bool
CJavaMethodSignature::EqualsIgnoreReturn(
				const CJavaMethodSignature& other) const
{
  bool equal = fMethodName == other.fMethodName &&
    fArgumentTypes.size() == other.fArgumentTypes.size();
  if (equal) {
    deque<CJavaTypeSignature>::const_iterator i1 = fArgumentTypes.begin();
    deque<CJavaTypeSignature>::const_iterator i2 =
      other.fArgumentTypes.begin();
    for (; equal && !(i1 == fArgumentTypes.end()); ++i1, ++i2) {
      equal = (*i1) == (*i2);
    }
  }
  return equal;
}

//
//  Method name : Compare
//  Description : Compares the arguments, names, and return types of two 
//    CJavaMethodSignature objects and returns 0 if they are equal, -1 if 
//    this is greater than the other and 1 if the other is greater than this.
//
int
CJavaMethodSignature::Compare(const CJavaMethodSignature& other) const
{
  int result = other.fHashKey - fHashKey;
  if (result == 0) {
    result = fMethodName.compare(other.fMethodName);
    if (result == 0) {
      result = fReturnType.Compare(other.fReturnType);
      if (result == 0)
	result = (other.fArgumentTypes.size() - fArgumentTypes.size());
      if (result == 0) {
	deque<CJavaTypeSignature>::const_iterator i1 = fArgumentTypes.begin();
	deque<CJavaTypeSignature>::const_iterator i2 =
	  other.fArgumentTypes.begin();
	for (; !(i1 == fArgumentTypes.end()); ++i1, ++i2) {
	  if ((result = (*i1).Compare(*i2)) != 0) {
	    break;
	  }
	}
      }
    }
  }
  return result;
}

//
//  Method name : operator==
//  Description : Compares two method signatures for equality.
//
bool
CJavaMethodSignature::operator==(const CJavaMethodSignature& other) const
{
  return Compare(other) == 0;
}

//
//  Method name : operator<
//  Description : Returns true if this method signature does not match the
//    argument and the hash value for this signature is lower than the other.
//
bool
CJavaMethodSignature::operator<(const CJavaMethodSignature& other) const
{
  return Compare(other) > 0;
}


//
//  Method name : SetHashKey
//  Description : Internal method to set the hash key for this signature.
//
void
CJavaMethodSignature::SetHashKey()
{
  fHashKey = ::Hash(fMethodName) + fReturnType.Hash();
  for (deque<CJavaTypeSignature>::iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end()); ++i) {
    fHashKey += (*i).Hash();
  }
}


//
//  Method name : Initialize
//  Description : This function chews over the input strings and tries to
//     interpret it as a method signature.  If it is successful, then
//     this value is set appropriately and 'true' is returned.  If
//     not, then this value becomes invalid and a false value is given.
//
bool
CJavaMethodSignature::Initialize(const unicode_string& methodName,
				 const unicode_string& signatureString)
{
  fMethodName = methodName;
  unicode_string::const_iterator stringPointer = signatureString.begin();
  unicode_string::const_iterator stringEnd = signatureString.end();
  bool result = stringPointer < stringEnd && *stringPointer++ == '(';
  if (result) {
    fArgumentTypes.erase(fArgumentTypes.begin(), fArgumentTypes.end());
    while (stringPointer < stringEnd && *stringPointer != ')') {
      CJavaTypeSignature argument;
      bool valid = argument.ParseSignatureString(stringPointer, stringEnd);
      if (valid) {
	fArgumentTypes.push_back(argument);
      } else {
	result = false;
	break;
      }
    }
    result = result && *stringPointer++ == ')' && stringPointer < stringEnd;
    if (result) {
      result = fReturnType.ParseSignatureString(stringPointer, stringEnd);
    }
  }
  SetHashKey();
  return result;
}

//
//  Method name : Disassemble
//  Description : Dumps out a human-readable description of this method to
//    the provided stream.  For debugging purposes.
//
void
CJavaMethodSignature::Disassemble(ostream& toStream) const
{
  if (fMethodName == kConstructorName) {
    toStream << "<constructor>(";
  } else {
    fReturnType.Disassemble(toStream);
    toStream << " " << ::UnicodeToUTF(fMethodName) << "(";
  }
  for (deque<CJavaTypeSignature>::const_iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end());) {
    (*i).Disassemble(toStream);
    i++;
    if (!(i == fArgumentTypes.end())) {
      toStream << ", ";
    }
  }
  toStream << ")";
}


//
//  Method name : Disassemble
//  Description : Returns a human-readable description of this signature in the
//    form of a string.
//
unicode_string
CJavaMethodSignature::Disassemble() const
{
  unicode_string result;
  if (fMethodName == kConstructorName) {
    result = ::StringToUnicode("<constructor>(");
  } else {
    result = fReturnType.Disassemble();
    result += (unicode_char)' ';
    result += fMethodName;
    result += (unicode_char)'(';
  }
  for (deque<CJavaTypeSignature>::const_iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end());) {
    result += (*i).Disassemble();
    i++;
    if (!(i == fArgumentTypes.end())) {
      result += (unicode_char)',';
      result += (unicode_char)' ';
    }
  }
  result += (unicode_char)')';
  return result;
}

//
//  Method name : CompileType
//  Description : This method is used to produce the method type signature
//    string used in the Java VM class file format.  Spits out an appropriate
//    string, ready for use.
//
string
CJavaMethodSignature::CompileType() const
{
  string result;
  result += '(';
  for (deque<CJavaTypeSignature>::const_iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end()); ++i) {
    result += (*i).Compile();
  }
  result += ')';
  result += fReturnType.Compile();
  return result;
}

//
//  Method name : ParametersBegin
//  Description : Returns an iterator for the beginning of the list of
//    parameter types.
//
deque<CJavaTypeSignature>::const_iterator
CJavaMethodSignature::ParametersBegin() const
{
  return fArgumentTypes.begin();
}

//
//  Method name : ParametersEnd
//  Description : Returns an iterator for the end of the list of parameter
//    types.
//
deque<CJavaTypeSignature>::const_iterator
CJavaMethodSignature::ParametersEnd() const
{
  return fArgumentTypes.end();
}

// 
//  Method name : IsInstantiable
//  Description : Returns false if any of the arguments or return types of 
//     this method are parameters or parameterized.
//
bool
CJavaMethodSignature::IsInstantiable() const
{
  
  if (!fReturnType.IsInstantiable()) 
    return false;
  for (deque<CJavaTypeSignature>::const_iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end()); ++i) {
    if (! (*i).IsInstantiable())
      return false;
  }
  return true;
}

// 
//  Method name : IsParameterized
//  Description : Returns true if any of the arguments or return types of 
//     this method are parameters or parameterized.
//
bool
CJavaMethodSignature::IsParameterized() const
{
  
  if (fReturnType.IsParameterized()) 
    return true;
  for (deque<CJavaTypeSignature>::const_iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end()); ++i) {
    if ((*i).IsParameterized())
      return true;
  }
  return false;
}


//  Method name : PostTransCompare
//  Description : Compares the arguments, names, and return types of two 
//    CJavaMethodSignature objects (after PJava type translation) and returns 
//    0 if they are equal, -1 if this is greater than the other and 1 if the
//    other is greater than this. If (otherPStatic), then this is taken to be
//    be an untranslated static function, and other to be a posttranslated
//    (possibly static) function.
//
int
CJavaMethodSignature::PostTransCompare(const CJavaMethodSignature& other,
				       bool otherPStatic) const
{
  CJavaMethodSignature a(*this);
  CJavaMethodSignature b(other);
  if (otherPStatic) {
    a.Objectify();
    if ((b.fArgumentTypes.size()>0) && 
	(*(b.fArgumentTypes.begin())).IsWhereType()) {
      b.fArgumentTypes.pop_front();
      b.SetHashKey();
    }
  } else {
    a.Objectify();
    b.Objectify();
  }
  return a.Compare(b);
}

//  Method name : Objectify
//  Description : Objectify all returns and arguments
void
CJavaMethodSignature::Objectify()
{
  fReturnType.Objectify();
  for (deque<CJavaTypeSignature>::iterator i = fArgumentTypes.begin();
       !(i == fArgumentTypes.end()); ++i) {
    (*i).Objectify();
  }
  SetHashKey();
}

//
//  Method name : SObjectify
//  Description : Objectify this signature for a static method, and prepend
//     a where-object argument for the appropriate class.
//
void
CJavaMethodSignature::SObjectify(const unicode_string & className)
{
  Objectify();
  CJavaTypeSignature t(className + ::UTFToUnicode("$$$W"));
  fArgumentTypes.push_front(t);
  SetHashKey();
}

//
// Method name: operator<< (ostream, JavaMethodSignature)
//  Description : Dumps out a human-readable description of this signature to
//    the provided stream.  For debugging purposes.
//
ostream& operator<<(ostream &o, const CJavaMethodSignature &sig)
{
  sig.Disassemble(o);
  return o;
}

//
//  Method name : AddArgument
//  Description : Add a new argument this method
//
void
CJavaMethodSignature::AddArgument(const CJavaTypeSignature &t)
{
  fArgumentTypes.push_back(t);
  SetHashKey();
}
