// $Id: JavaTypeSignature.C,v 1.26 1998/08/25 14:38:04 nickm Exp $
// Copyright (c) 1995  David Engberg  All rights reserved
#pragma implementation

#include "JavaTypeSignature.h"
#include <cstdlib>
#include <cassert>
#include <iostream.h>
#include <deque>
#include "Srcstream.h"
#include "find_leaks.h"
#include "unicode_string.h"

const CJavaTypeSignature
           CJavaTypeSignature::kString(::StringToUnicode("java/lang/String"));
const unicode_string kEmptyString;

const CJavaTypeSignature CJavaTypeSignature::kVoid(CJavaTypeSignature::Void);
const CJavaTypeSignature CJavaTypeSignature::kByte(CJavaTypeSignature::Byte);
const CJavaTypeSignature
                   CJavaTypeSignature::kInteger(CJavaTypeSignature::Integer);
const CJavaTypeSignature
                  CJavaTypeSignature::kLong(CJavaTypeSignature::LongInteger);
const CJavaTypeSignature CJavaTypeSignature::kNullType(kEmptyString);
const CJavaTypeSignature
           CJavaTypeSignature::kBoolean(CJavaTypeSignature::Boolean);
const CJavaTypeSignature
           CJavaTypeSignature::kDouble(CJavaTypeSignature::Double);
const CJavaTypeSignature
           CJavaTypeSignature::kFloat(CJavaTypeSignature::Float);
const CJavaTypeSignature
           CJavaTypeSignature::kShort(CJavaTypeSignature::Short);
const CJavaTypeSignature
           CJavaTypeSignature::kCharacter(CJavaTypeSignature::Character);

const unicode_string CJavaTypeSignature::kJLOName 
     = ::StringToUnicode("java/lang/Object");
const CJavaTypeSignature
           CJavaTypeSignature::kObject(CJavaTypeSignature::kJLOName);

const unicode_string kJLName = ::StringToUnicode("java/lang");
const unicode_string kOName = ::StringToUnicode("Object");
const unicode_string kWString = ::StringToUnicode("$$$W");

//
//  Method name : CJavaTypeSignature
//  Description : Default constructor.
//
CJavaTypeSignature::CJavaTypeSignature()
  : fType(Void), fArrayBounds(0), fClassName(), fParameters(0)
{
  REF_PTR(this);
}

//
//  Method name : CJavaTypeSignature
//  Description : Copy constructor.
//
CJavaTypeSignature::CJavaTypeSignature(const CJavaTypeSignature& source)
  : fType(source.fType),
    fArrayBounds(source.fArrayBounds),
    fClassName(source.fClassName),
    fParameters((source.fParameters == 0) ? 0 :
		new deque<CJavaTypeSignature>(*source.fParameters))
{
  REF_PTR(this);
}

//
//  Method name : CJavaTypeSignature
//  Description : Constructs a type signature from a provided element from
//    the base type enumeration.  This should never be constructed using the
//    'Class' value.  The second argument is used to specify how many array
//    indices are included in this type.  So int[][] would be represented
//    by atom 'Integer' and arrayBounds = 2
//
CJavaTypeSignature::CJavaTypeSignature(JavaType baseType,
				       unsigned long arrayBounds)
  : fType(baseType), fArrayBounds(arrayBounds), 
    fClassName(), fParameters(0)
{
  REF_PTR(this);
}

//
//  Method name : CJavaTypeSignature
//  Description : Constructs a type signature corresponding to a class type
//    or an array of classes.  The user provides the base class type name as
//    the first argument, and then the number of array indices as the second
//    argument.  If this type isn't an array, this should be left at 0.
//
CJavaTypeSignature::CJavaTypeSignature(const unicode_string& className,
				       unsigned long arrayBounds,
				       deque<CJavaTypeSignature>* parameters,
				       bool isParameter)
  : fType(isParameter ? Parameter : Class), fArrayBounds(arrayBounds), 
    fClassName(className), 
    fParameters((parameters == 0) ? 0 : 
		new deque<CJavaTypeSignature>(*parameters))
{
  REF_PTR(this);
}

//
//  Method name : ~CJavaTypeSignature
//  Description : Destructor
//
CJavaTypeSignature::~CJavaTypeSignature()
{
  delete fParameters;
  UNREF_PTR(this);
}

//
//  Method name : operator=
//  Description : Assignment operator.
//
CJavaTypeSignature&
CJavaTypeSignature::operator=(const CJavaTypeSignature& source)
{
  if (&source != this) {
    fType = source.fType;
    fArrayBounds = source.fArrayBounds;
    fClassName = source.fClassName;
    if (fParameters) delete fParameters;
    if (source.fParameters == 0) {
      fParameters = 0;
    } else {
      fParameters = new deque<CJavaTypeSignature>(*source.fParameters);
    }
  }
  return *this;
}


//
//  Method name : Compare
//  Description : Compares two java type signatures.  If they are equal, 0
//    is returned.  If this is greater than the other, a negative is returned,
//    otherwise a positive is returned.
//
int
CJavaTypeSignature::Compare(const CJavaTypeSignature& other) const
{
  int result = other.fType - fType;
  if (result == 0) {
    result = other.fArrayBounds - fArrayBounds;
    if (result == 0 && (fType == Class || fType == Parameter)) {
      result = fClassName.compare(other.fClassName);
      if (result == 0) {
	if (fParameters == 0) {
	  if (other.fParameters != 0) return 1;
	  return 0;
	}
 	if (other.fParameters == 0) { return -1; }
	result = other.fParameters->size() - fParameters->size() ;
	if (result == 0) {
	  deque<CJavaTypeSignature>::iterator i = fParameters->begin();
	  deque<CJavaTypeSignature>::iterator j = other.fParameters->begin();
	  while (result == 0 && i != fParameters->end()) {
	    result = (*i).Compare(*j);
	    ++i; ++j;
	  }
	}
      }
    }
  }
  return result;
}


//
//  Method name : operator==
//  Description : Compares two signatures for equality
//
bool
CJavaTypeSignature::operator==(const CJavaTypeSignature& other) const
{
  return Compare(other) == 0;
}

//
//  Method name : ParseSignatureString
//  Description : Attempts to parse the provided string as if it started with
//    the Java VM-standard signature for a type.  If it does, then this type
//    is configured to match that spec and true is returned, otherwise false
//    is returned and this class is left with the value 'void'
//
bool
CJavaTypeSignature::ParseSignatureString(
			       unicode_string::const_iterator& stringPointer,
			       unicode_string::const_iterator stringEnd)
{
  fArrayBounds = 0;
  fParameters = 0;

  fClassName = unicode_string();
  fType = Void;
  bool result = stringPointer < stringEnd;
  if (result) {
    switch (*stringPointer++) {
    case 'V':
      break;
    case 'B':
      fType = Byte;
      break;
    case 'C':
      fType = Character;
      break;
    case 'D':
      fType = Double;
      break;
    case 'F':
      fType = Float;
      break;
    case 'I':
      fType = Integer;
      break;
    case 'J':
      fType = LongInteger;
      break;
    case 'P':
      {
	fType = Parameter;
	unicode_string::const_iterator marker = stringPointer;
	while (marker < stringEnd && *marker != ';') {
	  marker++;      
	}
	if (marker == stringEnd) {
	  result = false;
	} else {
	  fClassName.replace(fClassName.begin(), fClassName.end(),
			     stringPointer, marker - stringPointer);
	  fType = Parameter;
	  stringPointer = marker + 1;	  
	}
      }
      break;
    case 'L':
      {
	unicode_string::const_iterator marker = stringPointer;
	while (marker < stringEnd && *marker != ';' && *marker != '<') {
	  marker++;	 
	}
	if (marker == stringEnd) {
	  result = false;
	} else {
	  fClassName.replace(fClassName.begin(), fClassName.end(),
			     stringPointer, marker - stringPointer);
	  if (*marker == '<') {
	    fParameters = new deque<CJavaTypeSignature>;
	    ++marker;
	    while (marker < stringEnd && result && *marker != '>') {
	      CJavaTypeSignature t;
	      result = t.ParseSignatureString(marker,stringEnd);
	      if (result) 
		fParameters->push_back(t);
	    }
	    if (marker == stringEnd) {
	      result = false;
	    } else {
	      ++marker;
	      if (*marker != ';')
		result = false;
	    }
	  }	  
	  if (result) {
	    fType = Class;
	    stringPointer = ++marker;
	  }
	}
      }
      break;
    case 'S':
      fType = Short;
      break;
    case 'Z':
      fType = Boolean;
      break;
    case '[':
      result = ParseSignatureString(stringPointer, stringEnd);
      fArrayBounds++;
      break;
    default:
      result = false;      
    }
  }
  return result;
}

bool
CJavaTypeSignature::ParseSignatureString(const unicode_string& uString)
{
  unicode_string::const_iterator i = uString.begin();
  return ParseSignatureString(i, uString.end());
}

//
//  Method name : ParseSignatureString
//  Description : Attempts to parse the provided string as if it started with
//    the Java VM-standard signature for a type.  If it does, then this type
//    is configured to match that spec and true is returned, otherwise false
//    is returned and this class is left with the value 'void'
//
bool
CJavaTypeSignature::ParseSignatureString(const string& utfString)
{
  fArrayBounds = 0;
  fParameters = 0;

  fClassName = unicode_string();
  fType = Void;
  bool result = utfString.length() > 0;
  if (result) {
    switch (utfString[0]) {
    case 'V':
      break;
    case 'B':
      fType = Byte;
      break;
    case 'C':
      fType = Character;
      break;
    case 'D':
      fType = Double;
      break;
    case 'F':
      fType = Float;
      break;
    case 'I':
      fType = Integer;
      break;
    case 'J':
      fType = LongInteger;
      break;
    case 'P':
      result = utfString[utfString.length() - 1] == ';';
      if (result) {
	fClassName =
	  ::UTFToUnicode(utfString.substr(1, utfString.length() - 2));
	fType = Parameter;
      }
      break;
    case 'L':
      {
	result = utfString[utfString.length() - 1] == ';';
	if (result) {
	  fType = Class;
	  unsigned idx = utfString.find_first_of('<');
	  if (idx >= utfString.length()) 
	    idx = utfString.length()-1;
	  fClassName =
	    ::UTFToUnicode(utfString.substr(1, idx - 1));
	  if (idx > 0 && idx < utfString.length()) {
	    ++idx;
	    fParameters = new deque<CJavaTypeSignature>;
	    while(result && idx < utfString.length()
		  && utfString[idx] != '>') {
	      int start = idx;
	      while(idx < utfString.length() && utfString[idx] == '[')
		++idx;
	      if (utfString[idx] == 'L' || utfString[idx] == 'P') {
		int brace_depth = 0;
		while(idx < utfString.length() &&
		      (brace_depth != 0 || 
		       (utfString[idx] != '>' && utfString[idx] != ';'))) {
		  if (utfString[idx] == '<') ++brace_depth;
		  if (utfString[idx] == '>') --brace_depth;  
		  ++idx;
		}
		if (brace_depth != 0 || idx >= utfString.length()) {
		  result = false;
		}
	      }
	      if (result) {
		CJavaTypeSignature t;
		string s(utfString, start, idx-start+1);
		result = t.ParseSignatureString(s);
		if (result)
		  fParameters->push_back(t);
	      }
	      ++idx;
	    }
	  }
	}
	if (!result) fType = Void;
	if (fParameters && fParameters->size() == 0) {
	  delete fParameters;
	  fParameters = 0;
	}
      }
      break;
    case 'S':
      fType = Short;
      break;
    case 'Z':
      fType = Boolean;
      break;
    case '[':
      {
	string remainder(utfString, 1);
	result = ParseSignatureString(remainder);
	fArrayBounds++;
	break;
      }
    default:
      result = false;      
    }
  }
  return result;
}

//
//  Method name : Disassemble
//  Description : Returns a human-readable description of this signature in the
//    form of a string.
//
unicode_string
CJavaTypeSignature::Disassemble() const
{
  unicode_string result;
  switch (fType) {
  case Void:
    result = ::StringToUnicode("void");
    break;
  case Byte:
    result = ::StringToUnicode("byte");
    break;
  case Character:
    result = ::StringToUnicode("char");
    break;
  case Double:
    result = ::StringToUnicode("double");
    break;
  case Float:
    result = ::StringToUnicode("float");
    break;
  case Integer:
    result = ::StringToUnicode("int");
    break;
  case LongInteger:
    result = ::StringToUnicode("long");
    break;
  case Short:
    result = ::StringToUnicode("short");
    break;
  case Boolean: 
    result = ::StringToUnicode("boolean");
    break;
  case Parameter:
    if (fClassName.size() > 0) {
      result = fClassName;
    } else {
      result = ::StringToUnicode("<null-parameter-type>");
    }
    break;
  case Class:    
    if (fClassName.size() > 0) {
      result = ::StringToUnicode(::SymbolToString(fClassName));
    } else {
      result = ::StringToUnicode("<null-reference-type>");
    }
    break;
  default:
    result = ::StringToUnicode("<**BAD REFERENCE TYPE**>");
    return result;
  }
  if (fParameters != 0) {
    result += ::StringToUnicode("[");
    deque<CJavaTypeSignature>::const_iterator end = fParameters->end();
    for (deque<CJavaTypeSignature>::const_iterator i = fParameters->begin();
	 i != end; ) {
      result += (*i).Disassemble();
      ++i;
      if (i != end)
	result += ::StringToUnicode(",");
    }
    result += ::StringToUnicode("]");
  }
  for (int arrayCount = fArrayBounds; arrayCount > 0; arrayCount--) {
    result += ::StringToUnicode("[]");
  }
  return result;
}

//
//  Method name : Disassemble
//  Description : Dumps out a human-readable description of this class to
//    the provided stream.  For debugging purposes.
//
void
CJavaTypeSignature::Disassemble(ostream& toStream) const
{
  toStream << ::UnicodeToString(Disassemble());
}

//
//  Method name : Unparse
//  Description : Dumps the type to sOut in canonical form.
//
void
CJavaTypeSignature::Unparse(srcstream& sOut) const
{
  UnparseBaseType(sOut);
  int arrayCount = fArrayBounds;
  if (fType == Parameter)
    arrayCount--;
  for (; arrayCount > 0; arrayCount--) {
    sOut.out() << "[]";
  }
}

//
//  Method name : UnparseBaseType
//  Description : Dumps the type to sOut, in canonical form, but without
//     any array braces.
//
void 
CJavaTypeSignature::UnparseBaseType(srcstream& sOut) const
{
  switch (fType) {
  case Void:
    sOut.out() << "void";
    break;
  case Byte:
    sOut.out() << "byte";
    break;
  case Character:
    sOut.out() << "char";
    break;
  case Double:
    sOut.out() << "double";
    break;
  case Float:
    sOut.out() << "float";
    break;
  case Integer:
    sOut.out() << "int";
    break;
  case LongInteger:
    sOut.out() << "long";
    break;
  case Short:
    sOut.out() << "short";
    break;
  case Boolean:
    sOut.out() << "boolean";
    break;
  case Parameter:
    sOut.out() << "Object";
    break;
  case Class:
    if (fClassName.size() > 0) {
      sOut.out() << ::SymbolToString(fClassName);
    } else {
      // assert(0);
      sOut.out() << "**NULL-TYPE**";
    }
    break;
  default:
    assert(0);
  }
}

//
//  Method name : Hash
//  Description : This function returns an unsigned long value that will
//    compare equal for any two identical types.
//
unsigned long
CJavaTypeSignature::Hash() const
{
  unsigned long result = (unsigned long)fType;
  if (fType == Class) {
    result = ::Hash(fClassName);
  }
  for (unsigned i=0; i < fArrayBounds; i++) {
    result = result * 13;
  }
  return result;
}

//
//  Method name : GetWidth
//  Description : This annoying bit of cruft is here because some values have
//    a width of two slots in some locations.  Ick, whine, etc.
//
unsigned short
CJavaTypeSignature::GetWidth() const
{
  assert(fType != Void);
  if (fArrayBounds > 0) {
    return 1;
  } else {
    return ((fType == LongInteger || fType == Double) ? 2 : 1);
  }
}

//
//  Method name : SetBaseClassName
//  Description : Sets the base class name of this type to the provided name.
//
void
CJavaTypeSignature::SetBaseClassName(const unicode_string& name)
{
  assert(fType == Class || fType == Parameter);
  fClassName = name;
}

//
//  Method name : IsNumeric
//  Description : Returns true if this type is a numeric type.  This includes
//    bytes, doubles, floats, integers, shorts and longs
//
bool
CJavaTypeSignature::IsNumeric() const
{
  switch (fType) {
  case Byte: case Double: case Float:
  case Integer: case LongInteger: case Short: case Character:
    return fArrayBounds == 0;
  default:
    return false;
  }
}

//
//  Method name : IsOrdinal
//  Description : Returns true if this type is an ordinal type.  This includes
//     bytes, shorts, integers, and longs.
//
bool
CJavaTypeSignature::IsOrdinal() const
{
  switch (fType) {
  case Byte: case Integer: case LongInteger: case Short: case Character:
    return fArrayBounds == 0;
  default:
    return false;
  }
}

//
//  Method name : IsFloatingPoint
//  Description : Returns true if this type is a non-ordinal numeric value.
//    This includes doubles and floats.
//
bool
CJavaTypeSignature::IsFloatingPoint() const
{
  switch (fType) {
  case Double: case Float:
    return fArrayBounds == 0;
  default:
    return false;
  }
}

//
//  Method name : IsReference
//  Description : Returns true if this type indicates a reference to a class
//    or array expression.
//
bool
CJavaTypeSignature::IsReference() const
{
  return fType == Class || fType == Parameter || fArrayBounds != 0;
}

//
//  Method name : IsParameterized
//  Description : Returns true iff this type is a parameterized type.
//
bool  
CJavaTypeSignature::IsParameterized() const
{
  return ((fParameters != 0) && (fParameters->size() > 0));
    
}

//
//  Method name : IsInstantiable
//  Description : Returns true iff this type can be instantiated. (I.e.,
//      is not a parameter, and is not parameterized on any parameters.)
//
bool 
CJavaTypeSignature::IsInstantiable() const
{
  if (fType == Parameter)
    return false;
  if (fType != Class)
    return true;
  if (fParameters == 0)
    return true;

  bool result = true;
  deque<CJavaTypeSignature>::const_iterator end = fParameters->end();
  for (deque<CJavaTypeSignature>::const_iterator i = fParameters->begin();
       i != end && result; ++i ) {
    result = result && (*i).IsInstantiable();
  }
  return result;
}

//
//  Method name : GetBaseClassName
//  Description : If the base type of this type signature is Class, this will
//    return true and fill the provided reference with the name of the class.
//    Otherwise, returns false.  Note that an array of a class type will have
//    a class base type, so Foo[] will return true and put "Foo" into fillName.
//
bool
CJavaTypeSignature::GetBaseClassName(unicode_string& fillName) const
{
  if (fType == Class || fType == Parameter) {
    fillName = fClassName;
    return true;
  }
  return false;
}

//
//  Method name : Compile
//  Description : Writes out this type in the Java VM Class format.
//
string
CJavaTypeSignature::Compile() const
{
  string result;
  for (unsigned i = 0; i < GetArrayBounds(); ++i) {
    result += '[';
  }
  switch (GetBaseType()) {
  case Void:
    result += 'V';
    break;
  case Byte:
    result += 'B';
    break;
  case Character:
    result += 'C';
    break;
  case Double:
    result += 'D';
    break;
  case Float:
    result += 'F';
    break;
  case Integer:
    result += 'I';
    break;
  case LongInteger:
    result += 'J';
    break;
  case Parameter:
    result += 'P';
    result += ::UnicodeToUTF(fClassName);
    result += ';';
    break;
  case Class:
    result += 'L';
    result += ::UnicodeToUTF(fClassName);
    if (fParameters != 0) {
      result += '<';
      deque<CJavaTypeSignature>::const_iterator end = fParameters->end();
      for (deque<CJavaTypeSignature>::const_iterator i = fParameters->begin();
	   i != end; ++i ) {
	result += (*i).Compile();
      }
      result += '>';
    }
    result += ';';
    break;
  case Short:
    result += 'S';
    break;
  case Boolean:
    result += 'Z';
    break;
  default:
    assert(0);
    break;
  }
  return result;
}

const unicode_string dollar_1 = ::UTFToUnicode("$1");
const unicode_string dollar_2 = ::UTFToUnicode("$2");
const unicode_string dollar_3 = ::UTFToUnicode("$3");
const unicode_string dollar_d = ::UTFToUnicode("$$");

//
//  Method name: EncodeParams
//  Description: Returns a unicode string representing the encoding of this
//      type's parameter list.  The encoding is obtained by 
//      fully qualifying all typenames, then replacing "." by "$",
//      "," by "$$", "[" by "$1", "]" by "$2", and "[]" by "$3" wherever
//      they occur.
//
unicode_string 
CJavaTypeSignature::EncodeParams() const
{
  unicode_string result;
  assert(fParameters != 0);
  deque<CJavaTypeSignature>::const_iterator end = fParameters->end();
  for (deque<CJavaTypeSignature>::const_iterator i = fParameters->begin();
       i != end; ) {
    switch((*i).fType) {
    case Byte:
      result += ::UTFToUnicode("byte");
      break;
    case Character:
      result += ::UTFToUnicode("char");
      break;
    case Double:
      result += ::UTFToUnicode("double");
      break;
    case Float:
      result += ::UTFToUnicode("float");
      break;
    case Integer:
      result += ::UTFToUnicode("int");
      break;
    case LongInteger:
      result += ::UTFToUnicode("long");
      break;
    case Short:
      result += ::UTFToUnicode("short");
      break;
    case Boolean:
      result += ::UTFToUnicode("boolean");
      break;
    case Class:
      if ((*i).fClassName.size() > 0) {
	result += ::SlashToDollar((*i).fClassName);
      } else {
	assert(0);
      }
      break;
    case Void:
    default:
      assert(0);      
    }
    if ((*i).IsParameterized()) {
      result += dollar_1;
      result += (*i).EncodeParams();
      result += dollar_2;
    }
    for(unsigned j = 0; j < (*i).fArrayBounds; ++j) {
      result += dollar_3;
    }

    ++i;
    if (i != end) {
      result += dollar_d;
    }
  }
  return result;
} 

//
// Method name: Objectify
// Description: Converts all parameters of dimension N to java.lang.Objects
//     of dimension N-1.
//
void
CJavaTypeSignature::Objectify()
{
  if (fType == Parameter) {
    fType = Class;
    fClassName = kJLOName;
    if (fArrayBounds > 0)
      --fArrayBounds;    
  } else if (fType == Class && IsParameterized()) {
    delete fParameters;
    fParameters = 0;
  }
}

//
// Method name: GetNParameters
// Description: Return the number of parameters of this type.
//
int
CJavaTypeSignature::GetNParameters() const
{ 
  return (fParameters == 0) ? 0 : (fParameters->size()); 
}

//
// Method name: PostTransCompare
// Description: Same as Compare, but returns <0 if this is < other after
//      translation, >0 if this is > other after translation, and 0
//      if this is equal to other after translation. Faster than
//      Objectify(); other.Objectify(); Compare(other);
//
int 
CJavaTypeSignature::PostTransCompare(const CJavaTypeSignature& other) const
{
  CJavaTypeSignature a(*this);
  CJavaTypeSignature b(other);
  a.Objectify();
  b.Objectify();
  return a.Compare(b);
}

// 
//  Method name : UnparseInstantiation
//  Description : Dumps the name of this type's instantiation class (WClass) 
//      to sOut.
//
void
CJavaTypeSignature::UnparseInstantiation(srcstream& sOut,
					 CTypeResolver& tr) const
{
  unicode_string package;
  if (fType == Class) 
    package = tr.InstantiationPackage(*this);

  if (package.length() > 0)
    sOut.out() << ::SymbolToString(package) << ".";
  sOut.out() << ::SymbolToString(::SlashToDollar(fClassName));
  
  if (IsParameterized()) {
    sOut.out() << "$$";
    sOut.out() << ::UnicodeToUTF(EncodeParams());
  }
  int arrayCount = fArrayBounds;
  if (fType == Parameter)
    arrayCount--;
  for (; arrayCount > 0; arrayCount--) {
    sOut.out() << "[]";
  }
}

// 
//  Method name : UnparseWType
//  Description : Dumps the name of this type's instantiation type (WType) 
//      to sOut.
//
void
CJavaTypeSignature::UnparseWType(srcstream& sOut) const
{
  assert(IsParameterized() && fType == Class);
  UnparseBaseType(sOut);
  sOut.out() << "$$$W";
}

// 
//  Method name : UnparseTokenClass
//  Description : Dumps the name of this type's Token class (TClass) 
//      to sOut.
//
void
CJavaTypeSignature::UnparseTokenClass(srcstream& sOut) const
{
  unicode_string::size_type lastSlash = fClassName.find_last_of('/');
  if (lastSlash == unicode_string::npos) 
    lastSlash = 0;
  unicode_string packageName(fClassName,0,lastSlash);
  if (packageName.size() > 0) {
    sOut.out() << ::SymbolToString(packageName) << ".";
  }
  sOut.out() << "$token";
}

//
// Method name : IsWhereType
// Description : Returns true if the string "$$$W" appears anywhere in
//    this type's name.
//
bool
CJavaTypeSignature::IsWhereType() const
{ 
  return (fType == Class) && 
    (fClassName.find(kWString) != unicode_string::npos);
}

// kBoolWrapperName...kShortWrapperName: the names of the class files for
// the various wrappers.
static const unicode_string kBoolWrapperName = 
    ::StringToUnicode("polyj.lang.BoolWrap");
static const unicode_string kByteWrapperName = 
    ::StringToUnicode("polyj.lang.ByteWrap");
static const unicode_string kCharWrapperName = 
    ::StringToUnicode("polyj.lang.CharWrap");
static const unicode_string kDoubleWrapperName = 
    ::StringToUnicode("polyj.lang.DoubleWrap");
static const unicode_string kFloatWrapperName = 
    ::StringToUnicode("polyj.lang.FloatWrap");
static const unicode_string kIntWrapperName = 
    ::StringToUnicode("polyj.lang.IntWrap");
static const unicode_string kLongWrapperName = 
    ::StringToUnicode("polyj.lang.LongWrap");
static const unicode_string kShortWrapperName = 
    ::StringToUnicode("polyj.lang.ShortWrap");

//
// Method name : GetWrapperName
// Description : Returns a reference to a unicode string representing the
//    name of the wrapper class for this type.
//
const unicode_string& 
CJavaTypeSignature::GetWrapperName() const
{
  switch(fType) {
  case Boolean:
    return kBoolWrapperName;
  case Byte:
    return kByteWrapperName;
  case Character:
    return kCharWrapperName;
  case Double:    
    return kDoubleWrapperName;
  case Float:
    return kFloatWrapperName;
  case Integer:
    return kIntWrapperName;
  case LongInteger:
    return kLongWrapperName;
  case Short:
    return kShortWrapperName;

  case Class:
  case Void:
  case Parameter:
  default:
    cerr << "**Tried to get wrapper type for non-wrappable type: ";
    Disassemble(cerr);
    cerr << endl;
    assert(0);
  } 
}

//
// Method name : IsWrapper
// Description : Returns true iff this type is one of the wrapper classes.
//
bool
CJavaTypeSignature::IsWrapper() const
{
  return (fType == Class) && (fArrayBounds == 0) && 
    ((fClassName == kBoolWrapperName) ||
     (fClassName == kByteWrapperName) ||
     (fClassName == kCharWrapperName) ||
     (fClassName == kDoubleWrapperName) ||
     (fClassName == kIntWrapperName) ||
     (fClassName == kLongWrapperName) ||
     (fClassName == kShortWrapperName));
}

//
// Method name : IsObject
// Description : Returns true iff this type is "object".
//
bool
CJavaTypeSignature::IsObject() const 
{
  return (fType == Class) && (fArrayBounds == 0) &&
    ((fClassName == kJLOName) ||
     (fClassName == kOName));
}

//
// Method name  : operator<< (ostream, JavaTypeSignature)
// Description  : Dumps out a human-readable description of this class to
//    the provided stream.  For debugging purposes.
//
ostream& operator<<(ostream &o, const CJavaTypeSignature &sig)
{
  return (o << ::UnicodeToString(sig.Disassemble()));
}
