// $Id: JavaTypeSignature.h,v 1.16 1998/07/28 20:36:21 nickm Exp $
// Copyright (c) 1995  David Engberg  All rights reserved
#ifndef _JavaTypeSignature_h
#define _JavaTypeSignature_h
#pragma interface

#include "dynamic_cast.h"
#include "unicode_string.h"
#include "config.h"

#ifdef OLDGCC
template<class T> class deque;
#else
#include <deque>
using namespace std;
#endif

class srcstream;
class unicode_string;
class CTypeResolver;

//
//  Class name : CJavaTypeSignature
//  Description : Java represents the type of a variable or parameter using
//    a sequence of characters that can be unpacked to give the type.  For
//    example, the integer type would be represented by "I" and an array
//    of 72 integers would be represented by "[72I"
//    This class attempts to encapsulate that in a little bit more friendly
//    form.  The one quirk is that the type 'Void' is only valid when used
//    as the return value for a function.  In any other context, a 'Void'
//    type signature can be taken as an error.
//
class CJavaTypeSignature {
public:
  typedef enum {
    Void,
    Byte,
    Character,
    Double,
    Float,
    Integer,
    LongInteger,
    Class,
    Short,
    Boolean,
    Parameter
  } JavaType;

  CJavaTypeSignature();
  CJavaTypeSignature(JavaType baseType, 
                     unsigned long arrayBounds = 0);
  CJavaTypeSignature(const unicode_string& className,
		     unsigned long arrayBounds = 0,
                     deque<CJavaTypeSignature>* parameters = 0,
		     bool isParameter = false);
  CJavaTypeSignature(const CJavaTypeSignature& source);
  ~CJavaTypeSignature();
  CJavaTypeSignature& operator=(const CJavaTypeSignature& source);
  bool operator==(const CJavaTypeSignature& other) const;
  
  bool ParseSignatureString(unicode_string::const_iterator& stringPointer,
			    unicode_string::const_iterator stringEnd);
  bool ParseSignatureString(const unicode_string &uString);
  bool ParseSignatureString(const string& utfString);
  void Disassemble(ostream& toStream) const;
  unicode_string Disassemble() const;
  void Unparse(srcstream& sOut) const;
  void UnparseBaseType(srcstream& sOut) const;
  void UnparseInstantiation(srcstream& sOut, CTypeResolver& tr) const;
  void UnparseWType(srcstream& sOut) const;
  void UnparseTokenClass(srcstream& sOut) const;

  string Compile() const;

  unicode_string EncodeParams() const;
  void Objectify();

  unsigned long GetArrayBounds() const { return fArrayBounds; }
  void SetArrayBounds(unsigned long bounds) { fArrayBounds = bounds; }
  JavaType GetBaseType() const { return fType; }

  unsigned short GetWidth() const;
  unsigned long Hash() const;
  int Compare(const CJavaTypeSignature& other) const;
  bool operator<(const CJavaTypeSignature& other) const 
    { return Compare(other) < 0; }

  bool IsNumeric() const;
  bool IsOrdinal() const;
  bool IsFloatingPoint() const;
  bool IsReference() const;
  bool IsParameterized() const;
  bool IsInstantiable() const;
  bool IsParameter() const { return fType == Parameter; }
  bool IsPrimitive() const { return (fType != Class) &&			       
			            (fType != Parameter) &&
			            (fType != Void) &&
			            (fArrayBounds == 0); }
  bool IsArray() const     { return (fArrayBounds != 0); }
  bool IsObject() const;

  bool GetBaseClassName(unicode_string& fillName) const;
  void SetBaseClassName(const unicode_string& name);

  static const CJavaTypeSignature kVoid;
  static const CJavaTypeSignature kInteger;
  static const CJavaTypeSignature kLong;
  static const CJavaTypeSignature kString;
  static const CJavaTypeSignature kNullType;
  static const CJavaTypeSignature kBoolean;
  static const CJavaTypeSignature kByte;
  static const CJavaTypeSignature kDouble;
  static const CJavaTypeSignature kFloat;
  static const CJavaTypeSignature kShort;
  static const CJavaTypeSignature kCharacter;
  static const CJavaTypeSignature kObject;
  static const unicode_string kJLOName;

  int GetNParameters() const;
  const deque<CJavaTypeSignature>* GetParameters() const
    { return fParameters; }
  int PostTransCompare(const CJavaTypeSignature& other) const;

  bool IsWhereType() const;
  bool IsWrapper() const;

  const unicode_string& GetWrapperName() const;

private:
  friend class CInstantiator;

  JavaType fType;
  unsigned long fArrayBounds;
  unicode_string fClassName;
  deque<CJavaTypeSignature>* fParameters;  
};

ostream& operator<<(ostream &o, const CJavaTypeSignature &sig); 

class CJavaMethodSignature;

//
// The CTypeResolver class encapsulates a stub function to determine
// the package in which to unparse a given type.
//
class CTypeResolver {
public:
  virtual unicode_string InstantiationPackage(const CJavaTypeSignature &t) 
    const = 0;
  virtual bool GetDeclaredMethodType(const CJavaTypeSignature &onType,
				     const CJavaMethodSignature &findMethod,
				     CJavaMethodSignature &setMethod,
				     CJavaMethodSignature &transMethod)
    const = 0;
  virtual bool IsVarProtected(const CJavaTypeSignature &onType,
			      const unicode_string &varName,
			      bool &setProt) const = 0;
};
 
#endif
