#pragma implementation
#include <map>
#include <set>
#include "JavaClassFile.h"
#include "PJavaClassInfo.h"
#include "PJavaConstraint.h"
#include "JavaCodeAttribute.h"
#include "JavaExceptionsTable.h"
#include "Instantiator.h"
#include "AuxMethodMap.h"
#include "unicode_string.h"
#include <stdio.h>
#include "find_leaks.h"

#define MAX_INT_STR_SIZE 128

#define PJAUX_MAJOR 0
#define PJAUX_MAJOR_STR "0"

#define PJAUX_MINOR 0
#define PJAUX_MINOR_STR "0"

const unicode_string kEmptyString;

// 
// Constructor
// Description: Create a "blank" PJavaClassInfo object.
//
CPJavaClassInfo::CPJavaClassInfo(bool finished)
  : fParams(0),
    fSuper(0),
    fInterfaces(0),
    fHasStinit(false),
    fPTable(0),    
    fAMM(new CAuxMethodMap(kEmptyString)),        
    fMethods(new MethodNameMap),
    fFinished(finished),
    fProcessing(false)
{
  REF_PTR(this);
}

//
// Constructor
// Description : Constructor used by StartClass
//
CPJavaClassInfo::CPJavaClassInfo(const unicode_string &className,
				 deque<unicode_string> *adoptParameters,
				 CJavaTypeSignature *super,
				 deque<CJavaTypeSignature>* interfaces,
				 deque<CPJavaConstraint*>* adoptConstraints,
				 bool finished)
  : fParams(0),
    fSuper(super),
    fInterfaces(interfaces),
    fHasStinit(false),
    fPTable(0),
    fAMM(new CAuxMethodMap(className)),
    fMethods(new MethodNameMap),       
    fFinished(finished),
    fProcessing(false)
{
  REF_PTR(this);
  if (!adoptParameters) {
    adoptParameters = new deque<unicode_string>;
  }
  if (!adoptConstraints) {
    adoptConstraints = new deque<CPJavaConstraint*>;
  }
  SetParameters(adoptParameters, adoptConstraints);
}
    

//
// Constructor
// Description : the copy constructor
//
CPJavaClassInfo::CPJavaClassInfo(CPJavaClassInfo &other)
  : fParams(new deque<unicode_string>(*other.fParams)),
    fSuper(new CJavaTypeSignature(*other.fSuper)),
    fInterfaces(new deque<CJavaTypeSignature>(*other.fInterfaces)),
    fHasStinit(other.fHasStinit),
    fPTable(new ParamTable),
    fAMM(new CAuxMethodMap(*other.fAMM)),    
    fMethods(new MethodNameMap(*other.fMethods)),
    fFinished(other.fFinished),
    fProcessing(other.fProcessing)
{
  REF_PTR(this); 
  for(ParamTable::iterator i = other.fPTable->begin();
      i != other.fPTable->end(); ++i) {
    CPJavaConstraint *c =
      (*i).second ? new CPJavaConstraint(*(*i).second) : 0;
    fPTable->insert(ParamTable::value_type((*i).first, c));
  } 
}


//
//  Method name: SetParameters
//  Description: Sets the parameters of this to be the strings named in
//     "parameters".  Sets the constraints on those parameters to be
//     those represented by "constraints". No error checking is performed.
//
void
CPJavaClassInfo::SetParameters(deque<unicode_string>* adoptParameters,
			       deque<CPJavaConstraint*>* adoptConstraints)
{
  assert (adoptParameters);
  assert (adoptConstraints);

  delete fParams;
  delete fPTable;
  fParams = adoptParameters;
  fPTable = new ParamTable;

  for(deque<unicode_string>::iterator i = fParams->begin();
      i != fParams->end(); ++i) {
    ParamTable::iterator found = fPTable->find((*i));
    if (!(found != fPTable->end())) {
      fPTable->insert(ParamTable::value_type((*i), new CPJavaConstraint(*i)));
    }
  }

  for(deque<CPJavaConstraint*>::iterator i = adoptConstraints->begin();
      i != adoptConstraints->end(); ++i) {
    assert((*i) != 0);
    ParamTable::iterator found = fPTable->find((*i)->GetParameter());
    if (found == fPTable->end()) {
      fPTable->insert(ParamTable::value_type((*i)->GetParameter(), (*i)));
    } else if ((*found).second == 0) {
      (*found).second = new CPJavaConstraint(*(*i));
    } else {
      (*found).second->Merge(*(*i));
    }
  }
#ifndef NDEBUG
  for (ParamTable::iterator i = fPTable->begin(); i != fPTable->end(); ++i) {
    assert((*i).second != 0);
  }
#endif
  delete adoptConstraints;
}

//
//  Method name: CheckConstraints
//  Description: Checks to see whether the parameters and constraints are
//     semanticly correct. (i.e., no duplicate parameter names, no where
//     clauses for nonexistent parameters.) Returns 0 on success,
//     an error message on failure.
//
string *
CPJavaClassInfo::CheckConstraints()
{
  set<unicode_string, less<unicode_string> > parameters;
  string *result = 0;
  if (fParams != 0) {
    for(deque<unicode_string>::iterator i = fParams->begin();
	(result == 0) && i != fParams->end(); ++i) {
      if (parameters.find(*i) != parameters.end()) {
	result = new string("Duplicate parameter ");
	*result += ::UnicodeToUTF(*i);
      } else {
	parameters.insert(*i);
      }
    }
  }
  if (result == 0 && fPTable != 0) {
    for(ParamTable::iterator i = fPTable->begin();
	(result == 0) && i != fPTable->end(); ++i) {
      if (!(parameters.find((*i).first) != parameters.end())) {
	result = new string("Where clause with no corresponding parameter ");
	*result += ::UnicodeToUTF((*i).first);	
      }
    }
  }
  for (ParamTable::iterator i = fPTable->begin(); i != fPTable->end(); ++i) {
    assert((*i).second != 0);
  }
  return result;
}
  
//
//  Method name: (destructor)
//
CPJavaClassInfo::~CPJavaClassInfo()
{
  UNREF_PTR(this);
  delete fParams;
  if (fPTable) {
    for(ParamTable::iterator i = fPTable->begin(); i != fPTable->end(); ++i) {
      delete (*i).second;
    }
    delete fPTable;
  }
  delete fAMM;
  delete fSuper;
  delete fInterfaces;
  delete fMethods;
}


//
// Method name: WriteToString
// Description: Write the PJavaClassInfo from a string.
//
string
CPJavaClassInfo::WriteToString(CJavaClassFile& classFile)
{
  string s;

  // HEADER
  s += PJAUX_MAJOR_STR;
  s += ":";
  s += PJAUX_MINOR_STR;
  s += "\n";

  // Section 1: Class information
  s += "--- CLASS\nParameters";
  //    Line 1: Parameter names
  for(deque<unicode_string>::iterator i = fParams->begin();
      i != fParams->end(); ++i) {
    s += ":";
    s += ::UnicodeToUTF(*i);
  }
  s += "\nExtends:";    
      
  //    Line 2: Supertype
  if (fSuper) {
    s += fSuper->Compile();
  }
  s += "\nImplements";

  //    Line 3: Interfaces
  if (fInterfaces) {
    for(deque<CJavaTypeSignature>::iterator i = fInterfaces->begin();
	i != fInterfaces->end(); ++i) {
      s += ":";
      s += (*i).Compile();
    }       
  }
  s += "\nStInit:";

  //    Line 4: Static initializer
  s += (fHasStinit) ? "Yes\n" : "No\n";
  
  //    Line 5: Access modifiers
  s += "Access:";
  s += classFile.GetAccessFlags().ASCCompile();
  s += "\n";

  // Section 2:  fields
  s += "--- FIELDS\n";
  for (FieldTable::iterator i = classFile.fFields.begin();
       i != classFile.fFields.end(); ++i) {
    assert((*i).second);
    s += ::UnicodeToUTF((*i).first.GetFieldName());
    s += ":";
    s += (*i).first.GetType().Compile();
    s += ":";
    s += (*i).second->GetModifiers().ASCCompile();
    s += "\n";
  }

  // Section 3: Method signatures
  s += "--- METHODS\n";
  assert(fMethods);
  for (MethodNameMap::iterator i = fMethods->begin();
       i != fMethods->end(); ++i) {
    char buf[MAX_INT_STR_SIZE];
    s += ::UnicodeToUTF((*i).first.GetName());
    s += ":";
    s += (*i).first.CompileType();
    s += ":";
    s += (*i).second.fTranslatedSignature.CompileType();
    s += ":";
    sprintf(buf, "%d", (*i).second.fIndex);
    s += buf;
    s += ":";
    s += (*i).second.fAccessFlags.ASCCompile();
    s += "\n";
  }

  // Section 4: Constraints
  s += "--- CONSTRAINTS\n";
  if (fPTable) {
    for(deque<unicode_string>::iterator p = fParams->begin(); 
	p != fParams->end(); ++p) {
      ParamTable::iterator ent = fPTable->find(*p);
      assert(ent != fPTable->end());
      if ((*ent).second != 0) {
	for(deque<CPJavaConstraint::ConstraintMethod*>::const_iterator m 
	      = (*ent).second->MethodsBegin();
	    m != (*ent).second->MethodsEnd(); ++m) {
	  assert((*m)->fMethod);
	  s += ::UnicodeToUTF(*p);
	  s += ":";
	  s += ::UnicodeToUTF((*m)->fMethod->GetName());
	  s += ":";
	  s += (*m)->fMethod->CompileType();
	  s += ":";
	  if ((*m)->fThrows) {
	    for (deque<CJavaTypeSignature>::iterator i=(*m)->fThrows->begin();
		 i != (*m)->fThrows->end(); ++i) {
	      s += (*i).Compile();
	    }
	  }
	  s +=  ((*m)->fStatic) ? ":S\n" : ":\n";
	}
      }
    }
  }

  // Section 5: AuxMethod map
  s += "--- AUXMETHODS\n";
  if (fAMM) {    
    for(amm_rep::const_iterator m = fAMM->Begin(); m != fAMM->End(); ++m) {
      assert((*m).second);
      s += ::UnicodeToUTF(*(*m).second);
      s += ":";
      string a = (*m).first.ASCCompile();
      s += a;
      s += "\n";
    }
  }

  return s;
}

// 
// Method name: LookupConstraint
// Description: If the given parameter has a constraint, return it. Otherwise
//     return 0.
//
const CPJavaConstraint* 
CPJavaClassInfo::LookupConstraint(const unicode_string& name) const
{
  assert(fPTable);
  ParamTable::iterator found = fPTable->find(name);
  if (found != fPTable->end()) {
    return (*found).second;
  } else {
    return 0;
  }
}

//
// Method name: ParseFromString
// Description: Read the PJavaClassInfo from a string. 
//              Returns true iff successful.
//
bool 
CPJavaClassInfo::ParseFromString(CJavaClassFile& classFile,
				 const string& s)
{
  // Hold parsed data for each section.
  long major;
  long minor;
  deque<deque<unicode_string> >      classSection;
  deque<deque<unicode_string> >      fieldSection;
  deque<deque<unicode_string> >     methodSection;
  deque<deque<unicode_string> > constraintSection;
  deque<deque<unicode_string> >  auxmethodSection;
  deque<deque<unicode_string> >   *currentSection;

  // Indexes into string.
  string::size_type pos = 0;  
  string::size_type idx, colon, eol;

  delete fParams;
  delete fSuper;
  delete fInterfaces;
  delete fPTable;

  // macro: advance 'pos' past/to whitespace.
#define ADVANCE_POS_W \
  for(char ch = s[pos]; pos < s.length() && isspace(ch); ch = s[++pos]) ;
#define ADVANCE_POS_NW \
  for(char ch = s[pos]; pos < s.length() && !isspace(ch); ch = s[++pos]) ;
#define ADVANCE_POS_WNOEOL \
  for(char ch = s[pos]; pos < s.length() && (isspace(ch) && ch != '\n'); \
      ch = s[++pos]) ;

#define AT_END(ptr) ((ptr) >= s.length())
#define CHECKEND(what) \
  if (AT_END(pos)) { \
    cerr << "Error in pjaux file: premature EOF while " \
	 << what << "." << endl; \
    return false; \
  }
 
  /////////
  // Parse Step: separate the string into sections, records, and fields.
  /////////

  ////
  // First, find version numbers
  ////
  ADVANCE_POS_W;
  idx = pos;
  CHECKEND("looking for major number");
  ADVANCE_POS_NW;
  eol = pos;
  CHECKEND("looking for end of minor number");
  colon = s.find_first_of(':', idx);
  if (colon <= idx || colon >= eol) {
    cerr << "Error in pjaux file: Couldn't find minor number" << endl;
    return false;
  }
  major = strtol(s.substr(idx,colon-idx).c_str(), NULL, 10);
  minor = strtol(s.substr(colon+1,eol-colon-1).c_str(), NULL, 10);
  if (major != PJAUX_MAJOR) {
    cerr << "Error in pjaux file: Unrecognized major number." << endl;
    return false;
  }
  if (minor < 0 || minor > 1024) {
    cerr << "Error in pjaux file: Bad minor number." << endl;
    return false;
  }
  
  ////
  // Now, parse all the sections.
  ////
  ADVANCE_POS_W;
  while(!AT_END(pos)) {
    if (s[pos] != '-' || s[++pos] != '-') {
      cerr << "Error in pjaux file: Couldn't find section.";
    }
    pos = s.find_first_not_of('-',pos);
    ADVANCE_POS_W;
    CHECKEND("looking for section name");
    idx = pos;
    ADVANCE_POS_NW;
    CHECKEND("looking for end of section name");
    string sectionName = s.substr(idx, pos-idx);
    if (sectionName == "CLASS")
      currentSection = &classSection;
    else if (sectionName == "FIELDS")
      currentSection = &fieldSection;
    else if (sectionName == "METHODS")
      currentSection = &methodSection;
    else if (sectionName == "CONSTRAINTS")
      currentSection = &constraintSection;
    else if (sectionName == "AUXMETHODS")
      currentSection = &auxmethodSection;
    else {
      cerr << "Warning: unrecognized section " << sectionName 
	   << " in pjaux file" << endl;
      currentSection = 0;
    }
    if (currentSection && currentSection->size() != 0) {
      cerr << "Warning: duplicate section " << sectionName
	   << " in pjaux file" << endl;
    }
    ADVANCE_POS_W;
    // ForEach line...
    while(!AT_END(pos) && s[pos] != '-') {
      deque<unicode_string> *record = new deque<unicode_string>;
      idx = pos;
      ADVANCE_POS_NW;
      
      eol = pos;
      while(idx <= eol) {
	colon = s.find_first_of(':', idx);
	if (colon > eol) colon = eol;
	record->push_back(::UTFToUnicode(s.substr(idx, colon-idx)));
	idx = colon+1;
      }
      if (currentSection)
	currentSection->push_back(*record);
      delete record;
      ADVANCE_POS_WNOEOL;
      if (! AT_END(pos) && s[pos] != '\n') {
	cerr << "Error in pjaux file while looking for eol." << endl;
	return false;
      }
      ADVANCE_POS_W;
    }
  }

  /////////
  // Now, go through each section of the pjaux file.
  /////////
  
  ////
  // CLASS
  ////
  bool foundParams = false, foundExtends = false, foundImplements = false,
       foundStInit = false, foundAccess = false;
  for(deque<deque<unicode_string> >::iterator rec = classSection.begin();
      rec != classSection.end(); ++rec) {
    deque<unicode_string>::iterator f = (*rec).begin();
    string name = ::UnicodeToUTF(*f);
    ++f;
    if (name == "Parameters") {
      if (foundParams)
	cerr << "Warning in pjaux file: duplicate parameter lines." << endl;
      fParams = new deque<unicode_string>;      
      for(; f != (*rec).end(); ++f) {
	fParams->push_back(*f);
      }
      foundParams = true;
    } else if (name == "Extends") {
      if (foundExtends)
	cerr << "Warning in pjaux file: duplicate extends lines." << endl;
      if ((*rec).size() != 2)
	cerr << "Error in pjaux file: bad supertype line." << endl;
      fSuper = new CJavaTypeSignature;
      if (f == (*rec).end()) {
	cerr << "Error in pjaux file: no supertype." << endl;
	return false;
      }
      if (!fSuper->ParseSignatureString((*f))) {
	cerr << "Error in pjaux file: bad supertype." << endl;
	return false;
      }
      foundExtends = true;
    } else if (name == "Implements") {
      if (foundImplements)
	cerr << "Warning in pjaux file: duplicate implements lines." << endl;
      fInterfaces = new deque<CJavaTypeSignature>;
      for (; f != (*rec).end(); ++f) {
	CJavaTypeSignature type;
	if (!type.ParseSignatureString((*f))) {
	  cerr << "Error in pjaux file: bad interface." << endl;
	  return false;
	}
	fInterfaces->push_back(type);
      }
      foundImplements = true;
    } else if (name == "StInit") {
      if (foundStInit)
	cerr << "Warning in pjaux file: duplicate stinit lines." << endl;
      if ((*rec).size() != 2)
	cerr << "Error in pjaux file: bad supertype line." << endl;
      string name = ::UnicodeToUTF((*f));
      if (name == "Yes") 
	fHasStinit = true;
      else if (name == "No")
	fHasStinit = false;
      else {
	cerr << "Error in pjaux file: bad stinit entry" << endl;
	return false;
      }
      foundStInit = true;
    } else if (name == "Access") {
      if (foundAccess)
	cerr << "Warning in pjaux file: duplicate access lines." << endl;
      if ((*rec).size() != 2)
	cerr << "Error in pjaux file: bad supertype line." << endl;
      if (!classFile.fAccessFlags.ASCParse(::UnicodeToUTF(*f))) {
	cerr << "Error in pjaux file: bad access flags" << endl;
	return false;
      }
      foundAccess = true;
    } else {
      cerr << "Warning: Unrecognized line in class section of pjaux file"
	   << endl;
    }
  }

  if (! (foundParams && foundExtends && foundImplements &&
	 foundStInit && foundAccess)) {
    cerr << "Warning: Line missing from class section of pjaux file."
	 << endl;
  }

  ////
  // FIELDS
  ////
  for(deque<deque<unicode_string> >::iterator rec = fieldSection.begin();
      rec != fieldSection.end(); ++rec) {
    if ((*rec).size() != 3) {
      cerr << "Error in pjaux file: bad field record" << endl;
      return false;
    }
    CJavaTypeSignature sig;
    if (!sig.ParseSignatureString((*rec)[1])) {
      cerr << "Error in pjaux file: bad field signature" << endl;
      return false;
    }
    CJavaAccessFlags flags;
    flags.ASCParse(::UnicodeToUTF((*rec)[2]));
    unicode_string name= (*rec)[0];
    bool found = false;
    for(FieldTable::iterator i = classFile.fFields.begin();
	i != classFile.fFields.end(); ++i) {
      if ((*i).first.GetFieldName() == name) {
	found = true;
	if ((*i).first.GetType().Compare(sig) != 0) {
	  FieldTable::value_type v(
		   CJavaFieldSignature(sig, name),
      		   new CJavaFieldInfo(flags,
		                      CJavaFieldSignature(sig,name)));
	  delete (*i).second;
	  classFile.fFields.erase(i);
	  classFile.fFields.insert(v);
	}
	break;
      }
    }
    if (flags.fStatic && !found) {
      FieldTable::value_type v(
		CJavaFieldSignature(sig, name),
	  	new CJavaFieldInfo(flags,
				   CJavaFieldSignature(sig,name)));
      classFile.fFields.insert(v);
    } else if (! found) {
      cerr << "Error in pjaux file: Couldn't find field " << name 
	   << " in the class file." << endl;
      return false;
    }
  }

  ////
  // METHODS
  ////
  for(deque<deque<unicode_string> >::iterator rec = methodSection.begin();
      rec != methodSection.end(); ++rec) {
    if ((*rec).size() != 5) {
      cerr << "Error in pjaux file: bad method record" << endl;
      return false;
    }
    unicode_string methodname = (*rec)[0];
    CJavaAccessFlags flags;
    CJavaMethodSignature type;
    CJavaMethodSignature transType;
    if (!type.Initialize(methodname, (*rec)[1])) {
      cerr << "Error in pjaux file: bad method signature" << endl;
      return false;
    }
    unicode_string index = (*rec)[3];
    unicode_string mangledname = methodname;
    if ((index[0] != (unicode_char) '0') && 
	methodname != CJavaMethodSignature::kConstructorName) {
      mangledname += (unicode_char) '$';
      mangledname += index;
    }
    if (!transType.Initialize(mangledname, (*rec)[2])) {
      cerr << "Error in pjaux file: bad translated signature" << endl;
      return false;
    }
    flags.ASCParse(::UnicodeToUTF((*rec)[4]));
    for(MethodTable::iterator i = classFile.fMethods.begin();
	i != classFile.fMethods.end(); ++i) {     
      if ((*i).first.GetName() == mangledname) { 
	if ((*i).first.Compare(transType) == 0) {
	  if ((*i).first.Compare(type) != 0) {
	    (*i).second->SetSignature(type);
	    MethodTable::value_type vt(type, (*i).second);
	    (*i).second = 0;
	    classFile.fMethods.erase(i);
	    classFile.fMethods.insert(vt);
	  }
	  break;
	}
      }
    }
    int idx = atoi(::UnicodeToUTF(index).c_str());
    if (idx == INT_MAX || idx == INT_MIN)
      return false;
    TranslatedMethod trans;
    trans.fIndex = idx;
    trans.fTranslatedSignature = transType;
    trans.fAccessFlags = flags;
    fMethods->insert(MethodNameMap::value_type(type, trans));
  }
  
  ////
  // CONSTRAINTS
  ////
  fPTable = new ParamTable;
  for (deque<unicode_string>::iterator i = fParams->begin();
       i != fParams->end(); ++i) {
    fPTable->insert(ParamTable::value_type((*i),new CPJavaConstraint(*i)));
  }
  for(deque<deque<unicode_string> >::iterator rec = constraintSection.begin();
      rec != constraintSection.end(); ++rec) {
    if ((*rec).size() != 5) {
      cerr << "Error in pjaux file: Bad constraint record." << endl;
      return false;
    }
    unicode_string pname = (*rec)[0];
    ParamTable::iterator i = fPTable->find(pname);
    if (i == fPTable->end()) {
      cerr << "Error in pjaux file: Constraint for nonexistent parameter." 
	   << endl;
      return false;
    }
    CJavaMethodSignature m;
    CJavaAccessFlags flags;
    if (! m.Initialize((*rec)[1], (*rec)[2])) {
      cerr << "Error in pjaux file: Bad constraint method signature." << endl;
      return false;
    }
    deque<CJavaTypeSignature> *throws = new deque<CJavaTypeSignature>;
    unicode_string::const_iterator it = (*rec)[3].begin();
    while (it != (*rec)[3].end()) {
      CJavaTypeSignature t;
      if (!t.ParseSignatureString(it,(*rec)[3].end())) {
	cerr << "Error in pjaux file: Bad constraint method throws." << endl;
	return false;
      }
      throws->push_back(t);
    }
    flags.ASCParse(::UnicodeToUTF((*rec)[4]));
    (*i).second->AddMethod(m,flags.fStatic,throws);
  }

  ////
  // AUXMETHODS
  ////
  fAMM = new CAuxMethodMap(classFile.GetClassName());
  for(deque<deque<unicode_string> >::iterator rec = auxmethodSection.begin();
      rec != auxmethodSection.end(); ++rec) {
    if (! fAMM->ASCParse(*rec)) {
      cerr << "Error in pjaux file: bad AMM entry." << endl;
      return false;
    }
  }
  // Remove extra methods from .class file. ($cast,$instanceof,$set_,$fetch_)
  bool more = true;
  static const unicode_string kCastPrefix  = ::UTFToUnicode("$cast");
  static const unicode_string kIOPrefix    = ::UTFToUnicode("$instanceof");
  static const unicode_string kSetPrefix   = ::UTFToUnicode("$set_");
  static const unicode_string kFetchPrefix = ::UTFToUnicode("$fetch_");
  while (more) {
    MethodTable::iterator it = classFile.fMethods.begin(); 
    bool erase = false;
    for( ; it != classFile.fMethods.end(); ++it) {
      if ((*it).first.GetName().length() > 5
	  && ((*it).first.GetName().compare(kCastPrefix,0,5) == 0)) {	
	erase = true;	
	break;
      } else if ((*it).first.GetName().length() > 11
		 && ((*it).first.GetName().compare(kIOPrefix,0,11) == 0)) {
	erase = true;
	break;
      } else if ((*it).first.GetName().length() > 5
		 && ((*it).first.GetName().compare(kSetPrefix,0,5) == 0)) {
	erase = true;
	break; 
      } else if ((*it).first.GetName().length() > 7
		 && ((*it).first.GetName().compare(kFetchPrefix,0,7) == 0)) {
	erase = true;
	break;
      }
    }
    if (erase) {
      (*it).second->RemoveReference();
      classFile.fMethods.erase(it);
    }
    more = erase;
  }

  fFinished = true;
  return true;
}

//
// Method name : RegisterConstraints
// Description : Register our constraints on the AMM.  Should be called after
//    all _adopt's_ to avoid collision.
//
void
CPJavaClassInfo::RegisterConstraints()
{
  assert(fPTable);
  assert(fParams);
  assert(fAMM);
  // Register PSig methods
  for(deque<unicode_string>::iterator i = fParams->begin(); 
      i != fParams->end(); ++i) {
    CJavaTypeSignature type(*i,0,0,true);
    fAMM->RequireMethod(CAuxMethodSignature::PSig, type);
  }
}


//
// Method name : Disassemble
// Description : Write a human-readable summary of this class's PClassInfo to
//     out.
//
void 
CPJavaClassInfo::Disassemble(ostream &out) const
{
  out << "  Parameters: ";
  if (fParams != 0 && fParams->size() > 0) {
    out << "[";
    for (deque<unicode_string>::iterator i = fParams->begin();
	 i != fParams->end(); ) {
      out << ::SymbolToString(*i);
      if ((++i) != fParams->end())
	out << ",";
    }
    out << "]" << endl;
  } else {
    out << "<none>" << endl;
  }

  out << "  Super: ";
  if (fSuper) {
    out << (*fSuper) << endl;
  } else {
    out << "<none>" << endl;
  }

  out << "  Interfaces: {";
  if (fInterfaces) {
    for(deque<CJavaTypeSignature>::iterator i = fInterfaces->begin();
	i != fInterfaces->end(); ) {
      out << (*i);
      if ((++i) != fInterfaces->end())
	out << ", ";      
    }
  }
  out << " }" << endl;

  out << (fHasStinit ? "  Has $stinit" : "  Has no $stinit") << endl;
  
  out << "  Constraints:" << endl;
  if (fPTable) {
    for (ParamTable::iterator i = fPTable->begin(); 
	 i != fPTable->end(); ++i) {
      out << "    " << ::SymbolToString((*i).first) << " -> "; 
      if ((*i).second)
	(*i).second->Unparse(out);
      out << endl;
    }
  }

  out << "  AuxMethods:" << endl;
  if (fAMM) {
    for (amm_rep::const_iterator i = fAMM->Begin(); i != fAMM->End(); ++i) {
      out << "    " << (*i).first;
      if ((*i).second) 
	out << " -> " << ::SymbolToString(*(*i).second);
      out << endl;
    }
  }

  out << "  Methods: " << endl;
  if (fMethods) {
    for (MethodNameMap::iterator i = fMethods->begin(); 
	 i != fMethods->end(); ++i) {
      out << "    " << (*i).first 
	  << " (" << (*i).second.fIndex << ") -> "
	  << (*i).second.fTranslatedSignature
	  << endl;
    }
  }
}

