#pragma implementation

#include "PJavaConstraint.h"
#include "AuxMethodMap.h"
#include "Srcstream.h"
#include "find_leaks.h"

//
// Constructor name : CPJavaConstraint
// Description : Constructs a new, empty CPJavaConstraint.
//
CPJavaConstraint::CPJavaConstraint()
{
  REF_PTR(this);
}

//
// Constructor name : CPJavaConstraint
// Description : Constructs a new, empty CPJavaConstraint for a given 
//     parameter.
//
CPJavaConstraint::CPJavaConstraint(const unicode_string &param)
  : fParam(param)
{
  REF_PTR(this);
}

//
// Constructor name : CPJavaConstraint
// Description      : Copy constructor
//
CPJavaConstraint::CPJavaConstraint(const CPJavaConstraint &source)
  : fParam(source.fParam)
{  
  REF_PTR(this);
  for(deque<ConstraintMethod*>::const_iterator i = source.fMethods.begin();
      i != source.fMethods.end(); ++i) {
    assert(*i);
    assert((*i)->fMethod);
    assert((*i)->fThrows);
    ConstraintMethod* cns = new ConstraintMethod;
    cns->fStatic = (*i)->fStatic;
    cns->fMethod = new CJavaMethodSignature(* (*i)->fMethod);
    cns->fThrows = (*i)->fThrows ? 
      new deque<CJavaTypeSignature>(* (*i)->fThrows) : 0;
    fMethods.push_back(cns);
  }
}

//
// Destructor name : ~CPJavaConstraint
//
CPJavaConstraint::~CPJavaConstraint()
{
  UNREF_PTR(this);
  for(deque<ConstraintMethod*>::iterator i = fMethods.begin();
      i != fMethods.end(); ++i) {
    delete((*i)->fMethod);
    delete((*i)->fThrows);
    delete(*i);
  }
}

//
// Method name : operator=
// Description : Assignment operator
//
CPJavaConstraint&
CPJavaConstraint::operator=(const CPJavaConstraint &source)
{  
  if (&source == this) 
    return *this;  

  fParam = source.fParam;
  for(deque<ConstraintMethod*>::iterator i = fMethods.begin();
      i != fMethods.end(); ++i) {
    delete (*i)->fMethod;
    delete (*i)->fThrows;
    delete (*i);
  }
  fMethods.erase(fMethods.begin(), fMethods.end());
  for(deque<ConstraintMethod*>::const_iterator i = source.fMethods.begin();
      i != source.fMethods.end(); ++i) {
    assert(*i);
    assert((*i)->fMethod);
    assert((*i)->fThrows);
    ConstraintMethod* cns = new ConstraintMethod;
    cns->fStatic = (*i)->fStatic;
    cns->fMethod = new CJavaMethodSignature(* (*i)->fMethod);
    cns->fThrows = (*i)->fThrows ? 
      new deque<CJavaTypeSignature>(* (*i)->fThrows) : 0;
    fMethods.push_back(cns);
  }
  return *this;  
}


//
// Method name : AddMethod
// Description : Add a method to this constraint.
//
void
CPJavaConstraint::AddMethod(CJavaMethodSignature &method, 
			    bool isStatic,
			    deque<CJavaTypeSignature> *adoptThrows)
{
  ConstraintMethod* cns = new ConstraintMethod;
  cns->fStatic = isStatic;
  cns->fMethod = new CJavaMethodSignature(method);
  cns->fThrows = adoptThrows ? adoptThrows : new deque<CJavaTypeSignature>;

  fMethods.push_back(cns);
}

//
// Method name : AddMethod
// Description : Add a method to this constraint.
//
void
CPJavaConstraint::AddMethod(ConstraintMethod *method)
{
  assert(method);
  assert(method->fMethod);
  if (!method->fThrows)
    method->fThrows = new deque<CJavaTypeSignature>;
  fMethods.push_back(method);
}

//
// Method name : Merge
// Description : Add the methods of "other" to this.
//
void
CPJavaConstraint::Merge(const CPJavaConstraint &other)
{
  for(deque<ConstraintMethod*>::const_iterator i = other.fMethods.begin();
      i != other.fMethods.end(); ++i) {
    assert(*i);
    assert((*i)->fMethod);
    ConstraintMethod* cns = new ConstraintMethod;
    cns->fStatic = (*i)->fStatic;
    cns->fMethod = new CJavaMethodSignature(* (*i)->fMethod);
    cns->fThrows = ((*i)->fThrows) ?
      new deque<CJavaTypeSignature>(* (*i)->fThrows) : 0;
    fMethods.push_back(cns);
  }
}

//
// Method name : Register
// Description : Register the auxmethods corresponding to this constraint
//    in the AuxMethodMap AMM.
//
void
CPJavaConstraint::Register(CAuxMethodMap& AMM)
{
  CJavaTypeSignature p(fParam,0,0,true);
  for(deque<ConstraintMethod*>::iterator i = fMethods.begin();
      i != fMethods.end(); ++i) {
    assert((*i) != 0);
    assert((*i)->fMethod != 0);
    assert((*i)->fThrows != 0);
    unicode_string name((*i)->fMethod->GetName());
    CAuxMethodSignature::WhereType wt;
    if ((*i)->fStatic)
      wt = CAuxMethodSignature::Static;
    else if ((*i)->fMethod->IsConstructor())
      wt = CAuxMethodSignature::Constructor;
    else
      wt = CAuxMethodSignature::Virtual;

    AMM.RequireMethod(CAuxMethodSignature::Where,
		      p, &name, (*i)->fMethod, (*i)->fThrows, wt);
  }
}

//
// Method name: Unparse
// Description: Writes the code for this constraint to sOut.
//
void
CPJavaConstraint::Unparse(ostream &out) const
{
  if (fMethods.size() == 0)
    return;
  out << " where "
      << ::SymbolToString(fParam)
      << " { ";
  for(deque<ConstraintMethod*>::const_iterator i = fMethods.begin();
      i != fMethods.end(); ++i) {
    assert((*i)->fMethod);
    if ((*i)->fStatic) 
      out << "static ";
    (*i)->fMethod->Disassemble(out);
    if ((*i)->fThrows && (*i)->fThrows->size() > 0) {
      out << " throws ";
      for (deque<CJavaTypeSignature>::iterator t =(*i)->fThrows->begin();
	   t != (*i)->fThrows->end(); ++t) {
	(*t).Disassemble(out);
      }
    }
    out << "; ";
  }
  out << "}";
}

ostream& 
operator<<(ostream& o, const CPJavaConstraint& c)
{
  c.Unparse(o);
  return o;
}
