#include "PrimitiveMethods.h"
#include "JavaTypeSignature.h"
#include "JavaMethodSignature.h"
#include "JavaClassFile.h"
#include "JavaMethodInfo.h"
#include "Instantiator.h"
#include "PJavaConstraint.h"
#include "find_leaks.h"

// Bits to determine whether a type has a method or not.
enum PMMask {
  BYTE    = 0x001,
  SHORT   = 0x002,
  INT     = 0x004,
  LONG    = 0x008,
  FLOAT   = 0x010,
  DOUBLE  = 0x020,
  CHAR    = 0x040,
  BOOLEAN = 0x080,
  ARRAY   = 0x100,
  INTEGER = BYTE|SHORT|INT|LONG|CHAR,
  LOGICAL = BYTE|SHORT|INT|LONG|CHAR|BOOLEAN,
  NUMERIC = BYTE|SHORT|INT|LONG|CHAR|FLOAT|DOUBLE,
  ALL     = BYTE|SHORT|INT|LONG|FLOAT|DOUBLE|CHAR|BOOLEAN|ARRAY
};

// Entry in table of primitive methods.
struct PMTEntry {
  PMMask fMask;     // What types have this method?
  char *fName;      // What's this method's name?
  char *fSignature; // What's this method's signature? 
                    // (parameter T refers to this class, minus one [].)
};

const PMTEntry kPrimitiveMethods[] = {
  { (PMMask) (NUMERIC|BOOLEAN),"<init>",    "()V"    },
  { ALL,     "clone",        "()Ljava/lang/Object;"  },
  { ALL,     "equals",       "(Ljava/lang/Object;)Z" },
  { NUMERIC, "equals",       "(J)Z"                  },
  { NUMERIC, "equals",       "(D)Z"                  },
  { BOOLEAN, "equals",       "(Z)Z"                  },
  { CHAR,    "equals",       "(C)Z"                  },
  { ALL,     "toString",     "()Ljava/lang/String;"  },
  { ALL,     "hashCode",     "()I"                   },
  { BYTE,    "<init>",       "(B)V"                  },
  { SHORT,   "<init>",       "(S)V"                  },
  //{ (PMMask) (INT|CHAR),       "<init>",       "(I)V"},
  { INT,     "<init>",       "(I)V"                  },
  { CHAR,    "<init>",       "(C)V"                  },
  //{ (PMMask) (LONG|FLOAT|DOUBLE), "<init>",    "(J)V"},
  { LONG,    "<init>",       "(J)V"                  },
  { FLOAT,   "<init>",       "(F)V"                  },
  { DOUBLE,  "<init>",       "(D)V"                  },
  { NUMERIC, "byteValue",    "()B"                   },  
  { NUMERIC, "shortValue",   "()S"                   },  
  { NUMERIC, "intValue",     "()I"                   },  
  { NUMERIC, "longValue",    "()J"                   },  
  { NUMERIC, "floatValue",   "()F"                   },  
  { NUMERIC, "doubleValue",  "()D"                   },  
  { NUMERIC, "max",          "(PT;)PT;"              },
  { NUMERIC, "min",          "(PT;)PT;"              },
  { NUMERIC, "signum",       "()I"                   },
  { NUMERIC, "add",          "(PT;)PT;"              },
  { NUMERIC, "subtract",     "(PT;)PT;"              },
  { NUMERIC, "multiply",     "(PT;)PT;"              },
  { NUMERIC, "divide",       "(PT;)PT;"              },
  { (PMMask) (NUMERIC-CHAR), "negate",  "()PT;"        },
  { NUMERIC, "abs",          "()PT;"                 },
  { NUMERIC, "less",         "(PT;)Z"                },
  { NUMERIC, "greater",      "(PT;)Z"                },
  { LOGICAL, "and",          "(PT;)PT;"              },
  { LOGICAL, "or",           "(PT;)PT;"              },
  { LOGICAL, "xor",          "(PT;)PT;"              },
  { LOGICAL, "not",          "()PT;"                 },
  { INTEGER, "mod",          "(PT;)PT;"              },
  { NUMERIC, "compareTo",    "(PT;)I"                },
  { (PMMask) (BYTE|SHORT|INT|CHAR),  "charValue", "()C"},
  { BOOLEAN, "booleanValue", "()Z"                   },
  { BOOLEAN, "<init>",       "(Z)V"                  },
  { ARRAY,   "<init>",       "([PT;)V"               }, 
  { ARRAY,   "<init>",       "(I)V"                  }, 
  { ARRAY,   "size",         "()I"                   },
  { ARRAY,   "setElementAt", "(PT;I)V"               },
  { ARRAY,   "set",          "(IPT;)V"               },
  { ARRAY,   "elementAt",    "(I)PT;"                },
  { ARRAY,   "get",          "(I)PT;"                },
  { (PMMask) 0, (char*) 0, (char*) 0                 }
};

//
//  Function name : CreateClassForType
//  Description   : Return a newly created JavaClassFile with the stub
//     methods appropriate for the type "type".  Not appropriate for
//     anything but interface information.  "Void" is used instead of
//     "array of T";
//   
CJavaClassFile* CreateClassForType(const CJavaTypeSignature& type,
				   const unicode_string& className)
{
  PMMask mask;          
  assert(!type.IsArray()); // We take "void" instead of arrays.
  CJavaTypeSignature t = type;
  switch(type.GetBaseType()) 
    {
    case CJavaTypeSignature::Byte:
      mask = BYTE;
      break;
    case CJavaTypeSignature::Character:
      mask = CHAR;
      break;
    case CJavaTypeSignature::Double:
      mask = DOUBLE;
      break;
    case CJavaTypeSignature::Float:
      mask = FLOAT;
      break;
    case CJavaTypeSignature::Integer:
      mask = INT;
      break;
    case CJavaTypeSignature::LongInteger:
      mask = LONG;
      break;
    case CJavaTypeSignature::Short:
      mask = SHORT;
      break;
    case CJavaTypeSignature::Boolean:
      mask = BOOLEAN;
      break;
    case CJavaTypeSignature::Void:
      // Really T[]
      t = CJavaTypeSignature(::StringToUnicode("T"),0,0,true);
      mask = ARRAY;
      break;
    case CJavaTypeSignature::Parameter:
    case CJavaTypeSignature::Class:
    default:
      assert(0);
      break;
    }
  CInstantiator ins(::StringToUnicode("T"),t);
  CJavaClassFile* Class = new CJavaClassFile();
  
  Class->fVersion = 0x3002d;
  Class->fAccessFlags.fPublic   = 1;
  Class->fAccessFlags.fAbstract = 1;
  Class->fThisClassName         = className;
  Class->fDeprecated            = false;
  // Leave Class->fSuperclassName empty.
  for(int i = 0; kPrimitiveMethods[i].fMask != 0; ++i) {
    if (kPrimitiveMethods[i].fMask & mask) {
      string sigStr(kPrimitiveMethods[i].fSignature);
      string nameStr(kPrimitiveMethods[i].fName);
      unicode_string uSig = ::UTFToUnicode(sigStr);
      unicode_string uName = ::UTFToUnicode(nameStr);
      CJavaMethodSignature in,out;
      if (!in.Initialize(uName,uSig)) {
	cerr << "*** Couldn't initialize method named "
	     << nameStr << " with type " << sigStr << endl;
	assert(0);      
      }
      ins.Instantiate(out,in);
			 
      CJavaMethodInfo* m = new CJavaMethodInfo;
      m->SetSignature(out);
      m->fAccessFlags.fPublic   = 1;
      m->fAccessFlags.fAbstract = 1;
      Class->InsertMethod(m);
    }
  }
  if (type.GetBaseType() == CJavaTypeSignature::Void) {
    deque<unicode_string> *params = new deque<unicode_string>;
    params->push_back(::UTFToUnicode("T"));
    Class->fPClassInfo = new CPJavaClassInfo(::UTFToUnicode(""),params,
					     0,0,0,true);
  }
  return Class;
}
