// Copyright (c) 1996  David Engberg  All rights reserved
// $Id: BinaryExpression.C,v 1.18 1998/08/20 20:04:37 nickm Exp $
#pragma implementation
#include "find_leaks.h"
#include "BinaryExpression.h"
#include "Expression.h"
#include "parser_decls.h"
#include "parser.h"  // for token values
#include "Compiler.h"
#include "JavaCodeAttribute.h"
#include "CompileContext.h"
#include "Srcstream.h"
#include <cmath>

static const unicode_string kStringBufferName =
               ::StringToUnicode("java/lang/StringBuffer");
static const unicode_string kAppendName = ::StringToUnicode("append");
static const unicode_string kConcatName = ::StringToUnicode("concat");
static const unicode_string kToStringName = ::StringToUnicode("toString");
static const unicode_string kValueOfName = ::StringToUnicode("valueOf");

//
//  Method name : CBinaryExpression
//  Description : Constructs a binary expression from the two child expressions
//    and the operator.  The operator is stored as an unsigned short, which is
//    meant to hold characters (in the case of '+', '-', etc...) and token
//    values generated from the parser.y yacc specification.  These constants
//    are found in parser.h.
//
CBinaryExpression::CBinaryExpression(CExpression* leftExpression,
                   unsigned short binaryOperator, CExpression* rightExpression)
  : CExpression(CJavaTypeSignature::Void, false),
    fLeft(leftExpression),
    fRight(rightExpression),
    fOperator(binaryOperator)
{
  assert(leftExpression != 0 && rightExpression != 0);
  rightExpression->SetLValue(false);
  fLineNumber = fLeft->GetLineNumber();
  switch (fOperator)
    {
    case INCR                   : 
    case DECR                   : fPrecedence = 1;  break;
      // Cast, other unaries : 1
    case '*'                    : fPrecedence = 2;  break;
    case '/'                    : 
    case '%'                    : fPrecedence = 2;  break;
    case '+'                    : 
    case '-'                    : fPrecedence = 3;  break;
    case BITSHIFT_RIGHT         : 
    case FILL_SHIFT_RIGHT       : 
    case SHIFT_LEFT             : fPrecedence = 4;  break;
    case '<'                    : 
    case '>'                    : 
    case LTEQ                   : 
    case GTEQ                   : fPrecedence = 5;  break;
      // instanceof : 5
    case EQUAL_COMPARE          : 
    case NOT_EQUAL              : fPrecedence = 6;  break; 
    case '&'                    : fPrecedence = 7;  break;
    case '^'                    : fPrecedence = 8;  break;
    case '|'                    : fPrecedence = 9;  break;
    case AND                    : fPrecedence = 10; break;
    case OR                     : fPrecedence = 11; break;
      // ?: : 12
    case '='                    :
    case SHIFT_RIGHT_EQUALS     : 
    case FILL_SHIFT_RIGHT_EQUALS: 
    case SHIFT_LEFT_EQUALS      : 
    case ADD_EQUALS             : 
    case SUB_EQUALS             : 
    case MUL_EQUALS             : 
    case DIV_EQUALS             : 
    case MOD_EQUALS             : 
    case AND_EQUALS             : 
    case OR_EQUALS              : 
    case XOR_EQUALS             : fPrecedence = 13; break;
    default                     : assert(0); break;
      assert(0);
    }

}

//
//  Method name : CBinaryExpression
//  Description : Copy constructor.
//
CBinaryExpression::CBinaryExpression(const CBinaryExpression& source)
  : CExpression(source), fLeft(0), fRight(0), fOperator(source.fOperator)
{
  if (source.fLeft != 0) {
    fLeft = source.fLeft->Clone();
  }
  if (source.fRight != 0) {
    fRight = source.fRight->Clone();
  }
}

//
//  Method name : ~CBinaryExpression
//  Description : Destructor.
//
CBinaryExpression::~CBinaryExpression()
{
  delete fLeft;
  delete fRight;
}

//
//  Method name : HandleEvaluateType
//  Description : This is an overrideable hook that is used by EvaluateType to
//    polymorphically force an expression to calculate its type.  It either
//    fills in 'setType' based on the evaluated type of this expression and
//    then returns 0, or it returns an error object to the user explaining why
//    the type couldn't be determined.
//
CCompileError*
CBinaryExpression::HandleEvaluateType(CCompileContext& context,
				      CJavaTypeSignature& setType)
{
  assert(fLeft != 0 && fRight != 0);
  CJavaTypeSignature leftType;
  bool isAssignment = false;
  switch(fOperator) {
  case '=': case SHIFT_RIGHT_EQUALS: case FILL_SHIFT_RIGHT_EQUALS:
  case SHIFT_LEFT_EQUALS: case AND_EQUALS: case XOR_EQUALS:
  case OR_EQUALS: case ADD_EQUALS: case SUB_EQUALS:
  case MUL_EQUALS: case DIV_EQUALS:  case MOD_EQUALS:
    isAssignment = true;
  }
  CCompileError* error = EvaluateType(fLeft, context, leftType, isAssignment);
  if (error == 0 && 
      (isAssignment && fOperator != '=') && 
      fLeft->IsLValueAlias()) {
    string msg = "Cannot use compound assignment operator on ";
    msg += DYNAMIC_CAST(CClassFieldExpression, fLeft) ?
      "static field of PClass" :
      "array of parameter type";
    error = MakeError(msg);
  }
  bool leftOk = !error;
  if (error == 0) {
    CJavaTypeSignature rightType;
    error = EvaluateType(fRight, context, rightType);
    if (error == 0 && (leftType == CJavaTypeSignature::kVoid ||
		       rightType == CJavaTypeSignature::kVoid)) {
      error = MakeError("Invalid use of void-typed expression");
    }
    if (error == 0) {
      switch (fOperator) {
      case OR:
      case AND:
	if (rightType != CJavaTypeSignature::kBoolean ||
	    leftType != CJavaTypeSignature::kBoolean) {
	  error =
	    MakeError("Non-boolean value used in conditional expression");
	} else {
	  setType = CJavaTypeSignature::kBoolean;
	}
	break;
      case BITSHIFT_RIGHT:
      case FILL_SHIFT_RIGHT:
      case SHIFT_LEFT:
	UnaryPromote(fLeft, context);
	UnaryPromote(fRight, context);
	leftType = fLeft->GetType();
	rightType = fRight->GetType();
	if (leftType != CJavaTypeSignature::kInteger &&
	    leftType != CJavaTypeSignature::kLong) {
	  unicode_string errorMessage = leftType.Disassemble();
	  errorMessage +=
	    ::StringToUnicode(" value used in a shift expression");
	  error = MakeError(errorMessage);
	}
	if (rightType == CJavaTypeSignature::kLong) {
	  error = CCastExpression::CastPointer(fRight,
			  CJavaTypeSignature::kInteger, context);
	} else if (rightType != CJavaTypeSignature::kInteger) {
	  unicode_string errorMessage = rightType.Disassemble();
	  errorMessage +=
	    ::StringToUnicode(" value used in a shift expression");
	  error = MakeError(errorMessage);
	}
	setType = leftType;
	break;
      case '+':
	if (rightType == CJavaTypeSignature::kString ||
	    leftType == CJavaTypeSignature::kString) {
	  setType = CJavaTypeSignature::kString;
	} else {
	  if (!BinaryPromote(fLeft, fRight, context)) {
	    error = MakeError("Invalid types used with '+' operator.");
	  } else {
	    setType = fLeft->GetType();
	  }
	}
	break;
      case '/':
      case '%':
	{
	  COrdinalLiteral* integerLiteral =
	    DYNAMIC_CAST(COrdinalLiteral, fRight);
	  if (integerLiteral != 0 && integerLiteral->GetLong() == 0) {
	    error = MakeError("Invalid division by zero");
	  }
	}
	// fall through ...
      case '-':
      case '*':
	if (error == 0 && !BinaryPromote(fLeft, fRight, context)) {
	  error = MakeError("Invalid types used in arithmetic expression");
	}
	setType = fLeft->GetType();
	break;
      case LTEQ:
      case GTEQ:
      case '<':
      case '>':
	if (!BinaryPromote(fLeft, fRight, context)) {
	  error = MakeError("Incompatible types used in comparison");
	}
	setType = CJavaTypeSignature(CJavaTypeSignature::Boolean);
	break;
      case '|':
      case '&':
      case '^':
	// 9.16 -- both boolean or binary promotion on both, any integral
	if (rightType == CJavaTypeSignature::kBoolean &&
	    leftType == CJavaTypeSignature::kBoolean) {
	  setType = CJavaTypeSignature::kBoolean;
	} else {
	  if (!BinaryPromote(fLeft, fRight, context)) {
	    error = MakeError("Invalid types used in bitwise expression");
	  } else {
	    if (!fLeft->GetType().IsOrdinal() ||
		!fRight->GetType().IsOrdinal()) {
	      error = MakeError("Non-integer type used in bitwise expression");
	    } else {
	      setType = fLeft->GetType();
	    }
	  }
	}
	break;
      case '=':
	if (!context.GetCompiler().SameType(leftType, rightType,
					    context.GetSource())) {
	  error =
	    CCastExpression::ImplicitCastPointer(fRight, leftType, context);
	  rightType = fRight->GetType();
	}
	if (error == 0) {
	  if (! fLeft->IsLValue()) {
	    error = MakeError("Invalid value on left-hand side of assignment");
	  } else {
	    setType = leftType;
	  }
	}
	break;
      case SHIFT_RIGHT_EQUALS:
      case FILL_SHIFT_RIGHT_EQUALS:
      case SHIFT_LEFT_EQUALS:
	if (!leftType.IsOrdinal()) {
	  error = MakeError("Invalid shift-assignment on non-integral value");
	} else {
	  if (!UnaryPromote(fRight, context)) {
	    error =
	     MakeError("Cannot convert right side of shift-assignment to int");
	  } else {
	    setType = leftType;
	  }
	}
	break;
      case AND_EQUALS:
      case XOR_EQUALS:
      case OR_EQUALS:
	if (rightType == CJavaTypeSignature::kBoolean &&
	    leftType == CJavaTypeSignature::kBoolean) {
	  setType = CJavaTypeSignature::kBoolean;
	} else {
	  UnaryPromote(fRight, context);
	  if (!leftType.IsOrdinal() ||
	      !fRight->GetType().IsOrdinal()) {
	    error = MakeError("Non-integer type used in bitwise assignment");
	  } else {
	    if (!context.GetCompiler().SameType(leftType, rightType,
						context.GetSource())) {
	      error = CCastExpression::CastPointer(fRight, leftType, context);
	    }
	    setType = leftType;
	  }
	}
	break;
      case ADD_EQUALS:
	if (leftType == CJavaTypeSignature::kString) {
	  setType = CJavaTypeSignature::kString;
	} else if (!leftType.IsNumeric() || !rightType.IsNumeric()) {
	  error = MakeError("Non-numeric value used in compound assignment.");
	} else {
	  if (!context.GetCompiler().SameType(leftType, rightType, 
					      context.GetSource())) {
	    error = CCastExpression::CastPointer(fRight, leftType, context);
	  }
	  setType = leftType;
	}
	break;
      case SUB_EQUALS:
      case MUL_EQUALS:
      case DIV_EQUALS:
      case MOD_EQUALS:
	if (!leftType.IsNumeric() || !rightType.IsNumeric()) {
	  error = MakeError("Non-numeric value used in compound assignment.");
	} else {
	  if (!context.GetCompiler().SameType(leftType, rightType,
					      context.GetSource())) {
	    error = CCastExpression::CastPointer(fRight, leftType, context);
	  }
	  setType = leftType;
	}
	break;
      case EQUAL_COMPARE:
      case NOT_EQUAL:
	// 9.15 -- boolean both, reference both, or binary promote
	// if reference, one must descend from other (or be same type)
	// don't forget about null
	setType = CJavaTypeSignature::kBoolean;
	if (!context.GetCompiler().SameType(leftType, rightType,
					    context.GetSource())) {
	  if (leftType.IsReference() && rightType.IsReference()) {
	    if (leftType == CJavaTypeSignature::kNullType) {
	      if (rightType != CJavaTypeSignature::kNullType) {
		error =
		  CCastExpression::CastPointer(fLeft, rightType, context);
	      }
	    } else if (rightType == CJavaTypeSignature::kNullType) {
	      error = CCastExpression::CastPointer(fRight, leftType, context);
	    } else if (context.GetCompiler().AssignableSubtype(leftType, 
					    rightType,context.GetSource())) {
	      error = CCastExpression::CastPointer(fLeft, rightType, context);
	    } else if (context.GetCompiler().AssignableSubtype(rightType, 
					   leftType, context.GetSource())) {
	      error = CCastExpression::CastPointer(fRight, leftType, context);
	    } else {
	      unicode_string errorString =
		::StringToUnicode("Can't compare reference of ");
	      errorString += leftType.Disassemble();
	      errorString += ::StringToUnicode(" to a reference of ");
	      errorString += rightType.Disassemble();
	      error = MakeError(errorString);
	    }
	  } else if (leftType.IsNumeric() && rightType.IsNumeric()) {
	    error = CCastExpression::CastPointer(fRight, leftType, context);
	  } else {
	    unicode_string errorString =
	      ::StringToUnicode("Can't compare ");
	    errorString += leftType.Disassemble();
	    errorString += ::StringToUnicode(" value to a ");
	    errorString += rightType.Disassemble();
	    errorString += ::StringToUnicode(" value.");
	    error = MakeError(errorString);
	  }
	}
	break;
      default:
	assert(0);
      }
    }
  }
  // If the lhs is a local variable, we should consider it initialized from
  // now on, whether there was an error or not.
  if (isAssignment && leftOk) {
    CLocalVariableExpression *localVar =
      DYNAMIC_CAST(CLocalVariableExpression, fLeft);
    if (localVar) {
      context.InitializeVariable(localVar->GetVariableIndex());
    }
  }

#if 0
  if (isAssignment && error == 0) {
    switch(fOperator) {
    case SHIFT_RIGHT_EQUALS: case FILL_SHIFT_RIGHT_EQUALS:
    case SHIFT_LEFT_EQUALS: case AND_EQUALS: case XOR_EQUALS:
    case OR_EQUALS: case ADD_EQUALS: case SUB_EQUALS:
    case MUL_EQUALS: case DIV_EQUALS:  case MOD_EQUALS:
      if (fLeft->IsLValueAlias()) {
	string errMsg("Attempt to use OP= operator on ");
	if (DYNAMIC_CAST(CArrayIndex,fLeft) != 0)
	  errMsg += "array-of-parameters";
	else
	  errMsg += "parameterized static field";
	error = MakeError(errMsg);
      }
      break;
    }
  }
#endif
  return error;
}

//
//  Method name : CreateConstantExpression
//  Description : This method asks the expression if it can generate a single
//    constant-valued equivalent of itself.  This is needed in several
//    several places in Java, but it also provides a simple constant-folding
//    optimization.  If this operation can be performed, a new literal
//    expression is allocated and assigned to 'result,' otherwise it is set to
//    0, indicating no constant folding was possible.  In any case,
//    this expression is not modified or deleted during this check.
//    If any errors are found during compilation, they are returned to the
//    caller, otherwise 0 is returned.
//
CCompileError*
CBinaryExpression::CreateConstantExpression(CExpression*& result,
					    CCompileContext& context) const
{
  CCompileError* error = 0;
  result = 0;
return error;			// Let javac do the work!
  assert(fLeft != 0 && fRight != 0);
  bool bothLiterals = fLeft->IsLiteral() && fRight->IsLiteral();
  COrdinalLiteral* intLeft = DYNAMIC_CAST(COrdinalLiteral, fLeft);
  COrdinalLiteral* intRight = DYNAMIC_CAST(COrdinalLiteral, fRight);
  CFloatLiteral* floatLeft = DYNAMIC_CAST(CFloatLiteral, fLeft);
  CFloatLiteral* floatRight = DYNAMIC_CAST(CFloatLiteral, fRight);
  switch (fOperator) {
  case OR:
    if (intLeft != 0) {
      if (intLeft->GetBoolean()) {
	result = new COrdinalLiteral(true);
      } else if (intRight != 0) {
	result = new COrdinalLiteral(intRight->GetBoolean());
      }
    }
    break;
  case AND:
    if (intLeft != 0) {
      if (!intLeft->GetBoolean()) {
	result = new COrdinalLiteral(false);
      } else if (intRight != 0) {
	result = new COrdinalLiteral(intRight->GetBoolean());
      }
    }
    break;
  case BITSHIFT_RIGHT:
    if (bothLiterals) {
      CJavaTypeSignature leftType = fLeft->GetType();
      if (leftType == CJavaTypeSignature::kInteger) {
	result = new COrdinalLiteral((JavaSInt)intLeft->GetInteger() >>
				     (intRight->GetInteger() & 0x1f));
      } else if (leftType != CJavaTypeSignature::kLong) {
	result = new COrdinalLiteral((JavaSLong)intLeft->GetLong() >>
				     (intRight->GetInteger() & 0x3f));
      }
    }
    break;
  case FILL_SHIFT_RIGHT:
    if (bothLiterals) {
      CJavaTypeSignature leftType = fLeft->GetType();
      if (leftType == CJavaTypeSignature::kInteger) {
	result = new COrdinalLiteral(intLeft->GetInteger() >>
				     (intRight->GetInteger() & 0x1f));
      } else if (leftType != CJavaTypeSignature::kLong) {
	result = new COrdinalLiteral(intLeft->GetLong() >>
				     (intRight->GetInteger() & 0x3f));
      }
    }
    break;
  case SHIFT_LEFT:
    if (bothLiterals) {
      CJavaTypeSignature leftType = fLeft->GetType();
      if (leftType == CJavaTypeSignature::kInteger) {
	result = new COrdinalLiteral(intLeft->GetInteger() <<
				     (intRight->GetInteger() & 0x1f));
      } else if (leftType != CJavaTypeSignature::kLong) {
	result = new COrdinalLiteral(intLeft->GetLong() <<
				     (intRight->GetInteger() & 0x3f));
      }
    }
    break;
  case '+':
    if (intLeft != 0 && intRight != 0) {
      switch (intLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Integer:
	result = new COrdinalLiteral((JavaSInt)intLeft->GetInteger() +
				     (JavaSInt)intRight->GetInteger());
	break;
      case CJavaTypeSignature::Byte:
	result = new COrdinalLiteral(
				     (u8)((JavaSByte)intLeft->GetByte() +
					  (JavaSByte)intRight->GetByte()));
	break;
      case CJavaTypeSignature::Short:
	result = new COrdinalLiteral((u16)(intLeft->GetShort() +
					   intRight->GetShort()));
	break;
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral((JavaSLong)intLeft->GetLong() +
				     (JavaSLong)intRight->GetLong());
	break;
      default: break;
      }      
    } else if (floatLeft != 0 && floatRight != 0) {
      switch (floatLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Float:
	result = new CFloatLiteral(floatLeft->GetFloat() +
				   floatRight->GetFloat());
	break;
      case CJavaTypeSignature::Double:
	result = new CFloatLiteral(floatLeft->GetDouble() +
				   floatRight->GetDouble());
	break;
      default: break;
      }
    } else if (bothLiterals) {
      CStringLiteral* stringLeft = DYNAMIC_CAST(CStringLiteral, fLeft);
      CStringLiteral* stringRight = DYNAMIC_CAST(CStringLiteral, fRight);
      if (stringLeft != 0 && stringRight != 0) {
	result = new CStringLiteral(stringLeft->GetString() +
				    stringRight->GetString());
      }
    }
    break;
  case '-':
    if (intLeft != 0 && intRight != 0) {
      switch (intLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Integer:
	result = new COrdinalLiteral((JavaSInt)intLeft->GetInteger() -
				     (JavaSInt)intRight->GetInteger());
	break;
      case CJavaTypeSignature::Byte:
	result = new COrdinalLiteral(
				     (u8)((JavaSByte)intLeft->GetByte() -
					  (JavaSByte)intRight->GetByte()));
	break;
      case CJavaTypeSignature::Short:
	result = new COrdinalLiteral((u16)(intLeft->GetShort() -
					   intRight->GetShort()));
	break;
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral((JavaSLong)intLeft->GetLong() -
				     (JavaSLong)intRight->GetLong());
	break;
      default: break;
      }
    } else if (floatLeft != 0 && floatRight != 0) {
      switch (floatLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Float:
	result = new CFloatLiteral(floatLeft->GetFloat() -
				   floatRight->GetFloat());
	break;
      case CJavaTypeSignature::Double:
	result = new CFloatLiteral(floatLeft->GetDouble() -
				   floatRight->GetDouble());
	break;
      default: break;
      }
    }
    break;
  case '*':
    if (intLeft != 0 && intRight != 0) {
      switch (intLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Integer:
	result = new COrdinalLiteral((JavaSInt)intLeft->GetInteger() *
				     (JavaSInt)intRight->GetInteger());
	break;
      case CJavaTypeSignature::Byte:
	result = new COrdinalLiteral(
				     (u8)((JavaSByte)intLeft->GetByte() *
					  (JavaSByte)intRight->GetByte()));
	break;
      case CJavaTypeSignature::Short:
	result = new COrdinalLiteral((u16)(intLeft->GetShort() *
					   intRight->GetShort()));
	break;
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral((JavaSLong)intLeft->GetLong() *
				     (JavaSLong)intRight->GetLong());
	break;
      default: break;
      }
    } else if (floatLeft != 0 && floatRight != 0) {
      switch (floatLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Float:
	result = new CFloatLiteral(floatLeft->GetFloat() *
				   floatRight->GetFloat());
	break;
      case CJavaTypeSignature::Double:
	result = new CFloatLiteral(floatLeft->GetDouble() *
				   floatRight->GetDouble());
	break;
      default: break;
      }
    }
    break;
  case '/':
    if (intLeft != 0 && intRight != 0) {
      if (intRight->GetLong() == 0) {
	error = MakeError("Division by zero");
      } else {
	switch (intLeft->GetType().GetBaseType()) {
	case CJavaTypeSignature::Integer:
	  result = new COrdinalLiteral((JavaSInt)intLeft->GetInteger() /
				       (JavaSInt)intRight->GetInteger());
	  break;
	case CJavaTypeSignature::Byte:
	  result = new COrdinalLiteral(
				       (u8)((JavaSByte)intLeft->GetByte() /
					    (JavaSByte)intRight->GetByte()));
	  break;
	case CJavaTypeSignature::Short:
	  result = new COrdinalLiteral((u16)(intLeft->GetShort() /
					     intRight->GetShort()));
	  break;
	case CJavaTypeSignature::LongInteger:
	  result = new COrdinalLiteral((JavaSLong)intLeft->GetLong() /
				       (JavaSLong)intRight->GetLong());
	  break;
	default: break;
	}
      }
    } else if (floatLeft != 0 && floatRight != 0) {
      switch (floatLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Float:
	result = new CFloatLiteral(floatLeft->GetFloat() /
				   floatRight->GetFloat());
	break;
      case CJavaTypeSignature::Double:
	result = new CFloatLiteral(floatLeft->GetDouble() /
				   floatRight->GetDouble());
	break;
      default: break;
      }
    }
    break;
  case '%':
    if (intLeft != 0 && intRight != 0) {
      if (intRight->GetLong() == 0) {
	error = MakeError("Division by zero");
      } else {
	switch (intLeft->GetType().GetBaseType()) {
	case CJavaTypeSignature::Integer:
	  result = new COrdinalLiteral((JavaSInt)intLeft->GetInteger() %
				       (JavaSInt)intRight->GetInteger());
	  break;
	case CJavaTypeSignature::Byte:
	  result = new COrdinalLiteral(
				       (u8)((JavaSByte)intLeft->GetByte() %
					    (JavaSByte)intRight->GetByte()));
	  break;
	case CJavaTypeSignature::Short:
	  result = new COrdinalLiteral((u16)(intLeft->GetShort() %
					     intRight->GetShort()));
	  break;
	case CJavaTypeSignature::LongInteger:
	  result = new COrdinalLiteral((JavaSLong)intLeft->GetLong() %
				       (JavaSLong)intRight->GetLong());
	  break;
	default: break;
	}
      }
    } else if (floatLeft != 0 && floatRight != 0) {
      switch (floatLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Float:
	result = new CFloatLiteral((float)::fmod(floatLeft->GetFloat(),
						 floatRight->GetFloat()));
	break;
      case CJavaTypeSignature::Double:
	result = new CFloatLiteral((double)::fmod(floatLeft->GetDouble(),
						  floatRight->GetDouble()));
	break;
      default: break;
      }
    }
    break;
  case '|':
    if (intLeft != 0 && intRight != 0) {
      switch (intLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Integer:
	result = new COrdinalLiteral(intLeft->GetInteger() |
				     intRight->GetInteger());
	break;
      case CJavaTypeSignature::Boolean:
	result = new COrdinalLiteral(intLeft->GetBoolean() ||
				     intRight->GetBoolean());
	break;
      case CJavaTypeSignature::Byte:
	result = new COrdinalLiteral((u8)(intLeft->GetByte() |
					  intRight->GetByte()));
	break;
      case CJavaTypeSignature::Short:
	result = new COrdinalLiteral((unsigned short)(intLeft->GetShort() |
						      intRight->GetShort()));
	break;
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral(intLeft->GetLong() |
				     intRight->GetLong());
	break;
      default: break;
      }
    }
    break;
  case '&':
    if (intLeft != 0 && intRight != 0) {
      switch (intLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Integer:
	result = new COrdinalLiteral(intLeft->GetInteger() &
				     intRight->GetInteger());
	break;
      case CJavaTypeSignature::Boolean:
	result = new COrdinalLiteral(intLeft->GetBoolean() &&
				     intRight->GetBoolean());
	break;
      case CJavaTypeSignature::Byte:
	result = new COrdinalLiteral((u8)(intLeft->GetByte() &
						     intRight->GetByte()));
	break;
      case CJavaTypeSignature::Short:
	result = new COrdinalLiteral((u16)(intLeft->GetShort() &
						      intRight->GetShort()));
	break;
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral(intLeft->GetLong() &
				     intRight->GetLong());
	break;
      default: break;
      }
    }
    break;
  case '^':
    if (intLeft != 0 && intRight != 0) {
      switch (intLeft->GetType().GetBaseType()) {
      case CJavaTypeSignature::Integer:
	result = new COrdinalLiteral(intLeft->GetInteger() ^
				     intRight->GetInteger());
	break;
      case CJavaTypeSignature::Boolean:
	result = new COrdinalLiteral(
		     (intLeft->GetBoolean() || intRight->GetBoolean()) &&
		     !(intLeft->GetBoolean() && intRight->GetBoolean()));
	break;
      case CJavaTypeSignature::Byte:
	result = new COrdinalLiteral((u8)(intLeft->GetByte() ^
						     intRight->GetByte()));
	break;
      case CJavaTypeSignature::Short:
	result = new COrdinalLiteral((u16)(intLeft->GetShort() ^
						      intRight->GetShort()));
	break;
      case CJavaTypeSignature::LongInteger:
	result = new COrdinalLiteral(intLeft->GetLong() ^
				     intRight->GetLong());
	break;
      default: break;
      }
    }
    break;
  case LTEQ:
    if (intLeft != 0 && intRight != 0) {
      result = new COrdinalLiteral(intLeft->GetLong() <=
				   intRight->GetLong());
    } else if (floatLeft != 0 && floatRight != 0) {
      result = new COrdinalLiteral(floatLeft->GetDouble() <=
				   floatRight->GetDouble());
    }
    break;
  case GTEQ:
    if (intLeft != 0 && intRight != 0) {
      result = new COrdinalLiteral(intLeft->GetLong() >=
				   intRight->GetLong());
    } else if (floatLeft != 0 && floatRight != 0) {
      result = new COrdinalLiteral(floatLeft->GetDouble() >=
				   floatRight->GetDouble());
    }
    break;
  case '<':
    if (intLeft != 0 && intRight != 0) {
      result = new COrdinalLiteral(intLeft->GetLong() <
				   intRight->GetLong());
    } else if (floatLeft != 0 && floatRight != 0) {
      result = new COrdinalLiteral(floatLeft->GetDouble() <
				   floatRight->GetDouble());
    }
    break;
  case '>':
    if (intLeft != 0 && intRight != 0) {
      result = new COrdinalLiteral(intLeft->GetLong() >
				   intRight->GetLong());
    } else if (floatLeft != 0 && floatRight != 0) {
      result = new COrdinalLiteral(floatLeft->GetDouble() >
				   floatRight->GetDouble());
    }
    break;
  case EQUAL_COMPARE:
    if (intLeft != 0 && intRight != 0) {
      result = new COrdinalLiteral(intLeft->GetLong() ==
				   intRight->GetLong());
    } else if (floatLeft != 0 && floatRight != 0) {
      result = new COrdinalLiteral(floatLeft->GetDouble() ==
				   floatRight->GetDouble());
    }
    break;
  case NOT_EQUAL:
    if (intLeft != 0 && intRight != 0) {
      result = new COrdinalLiteral(intLeft->GetLong() !=
				   intRight->GetLong());
    } else if (floatLeft != 0 && floatRight != 0) {
      result = new COrdinalLiteral(floatLeft->GetDouble() !=
				   floatRight->GetDouble());
    }
    break;
  }
  return error;
}

//
//  Method name : Unparse
//  Description : Writes the Java code for the expression to sOut.
//
CCompileError* 
CBinaryExpression::Unparse(srcstream &sOut,
			   CCompileContext& context,
			   bool isStatement)
{
  CCompileError* error = 0;
  bool isAssignment = false;
  switch(fOperator) {
  case '=': case SHIFT_RIGHT_EQUALS: case FILL_SHIFT_RIGHT_EQUALS:
  case SHIFT_LEFT_EQUALS: case AND_EQUALS: case XOR_EQUALS:
  case OR_EQUALS: case ADD_EQUALS: case SUB_EQUALS:
  case MUL_EQUALS: case DIV_EQUALS:  case MOD_EQUALS:
    isAssignment = true;
  }
  if (isAssignment) {
    if (fOperator != '=') {
      isAssignment = false; // XXXX CHANGE THIS SO *= WORKS.
    } else {
      error = fLeft->UnparseStore(sOut,context,*fRight,isStatement);
      return error;
    }
  }
  if ((fOperator == EQUAL_COMPARE) || (fOperator == NOT_EQUAL)) {
    if (((fLeft->GetType() == CJavaTypeSignature::kObject) ||
	 (fLeft->GetType().IsParameter())) &&
	((fLeft->GetType() == CJavaTypeSignature::kObject) ||
	 (fLeft->GetType().IsParameter()))) {
      sOut.ToLine(fLineNumber,false).out()
	<< ((fOperator == NOT_EQUAL) ? "!" : "")
	<< "polyj.lang.Primitive.ee(";
      error = fLeft->Unparse(sOut,context);
      sOut.out() << ",";
      if (!error) error = fRight->Unparse(sOut,context);
      sOut.out() << ")";
      return error;
    }
  }
  if (fPrecedence <= fLeft->GetPrecedence()) 
    sOut.ToLine(fLineNumber,false).out() << '(';
  error = fLeft->Unparse(sOut, context);
  if (fPrecedence <= fLeft->GetPrecedence())
    sOut.out() << ") ";
  switch (fOperator) 
    {
    case SHIFT_RIGHT_EQUALS     : sOut.out() << " >>= ";  break;
    case FILL_SHIFT_RIGHT_EQUALS: sOut.out() << " >>>= "; break;
    case SHIFT_LEFT_EQUALS      : sOut.out() << " <<= ";  break;
    case ADD_EQUALS             : sOut.out() << " += ";   break;
    case SUB_EQUALS             : sOut.out() << " -= ";   break;
    case MUL_EQUALS             : sOut.out() << " *= ";   break;
    case DIV_EQUALS             : sOut.out() << " /= ";   break;
    case MOD_EQUALS             : sOut.out() << " %= ";   break;
    case AND_EQUALS             : sOut.out() << " &= ";   break;
    case OR_EQUALS              : sOut.out() << " |= ";   break;
    case XOR_EQUALS             : sOut.out() << " ^= ";   break;
    case BITSHIFT_RIGHT         : sOut.out() << ">>";     break;
    case FILL_SHIFT_RIGHT       : sOut.out() << ">>>";    break;
    case SHIFT_LEFT             : sOut.out() << "<<";     break;
    case INCR                   : sOut.out() << "++";     break;
    case DECR                   : sOut.out() << "--";     break;
    case AND                    : sOut.out() << "&&";     break;
    case OR                     : sOut.out() << "||";     break;
    case LTEQ                   : sOut.out() << "<=";     break;
    case GTEQ                   : sOut.out() << ">=";     break;
    case EQUAL_COMPARE          : sOut.out() << "==";     break;
    case NOT_EQUAL              : sOut.out() << "!=";     break;
    case '='                    : sOut.out() << " = ";    break;
    default                     : 
      assert(fOperator<256);
      sOut.out() << ((char) fOperator);
      break;
    }
  if (error == 0) {
    if (fPrecedence <= fRight->GetPrecedence())
      sOut.ToLine(fRight->GetLineNumber(),false).out() << " (";
    error = fRight->Unparse(sOut, context);
    if (fPrecedence <= fRight->GetPrecedence())
      sOut.out() << ')';
  }

  return error;
}
