// Copyright (c) 1995  David Engberg  All rights reserved
// $Id: Compiler.C,v 1.81 2000/07/03 18:53:31 andru Exp $
#pragma implementation
#include "config.h"
#include "find_leaks.h"
#include "unicode_string.h"
#include "Srcstream.h"
#include "Compiler.h"
#include "parser_decls.h"
#include "FilePath.h"
#include "JavaClassFile.h"
#include <iostream.h>
#include <fstream.h>
#include "IntermediateClass.h"
#include "IntermediateFunction.h"
#include "JavaTypeSignature.h"
#include "parser_decls.h"
#include "Statement.h"
#include "JavaAttribute.h"
#include "JavaExceptionsTable.h"
#include "JavaCodeAttribute.h"
#include "CompileError.h"
#include "Expression.h"
#include "CompileContext.h"
#include "Instantiator.h"
#include <cstdio>
#include <climits>
#include <algorithm>
extern "C" {
  #include <unistd.h>
  #include <time.h>
}
#include "mapped_file.h"

const unicode_string CCompiler::kConstructorName = ::StringToUnicode("<init>");
const unicode_string CCompiler::kStaticName = ::StringToUnicode("<clinit>");
const string CCompiler::kDefaultImport("java/lang");
const unicode_string CCompiler::kObjectName =
                                    ::StringToUnicode("java/lang/Object");
const unicode_string CCompiler::kThrowableName =
                                    ::StringToUnicode("java/lang/Throwable");
const unicode_string CCompiler::kErrorName =
                                    ::StringToUnicode("java/lang/Error");
const unicode_string CCompiler::kRuntimeExceptionName =
                               ::StringToUnicode("java/lang/RuntimeException");
const unsigned long CCompiler::kVersionID = 0x3002d;  // beta API

const unicode_string kDefaultIPkg = ::StringToUnicode("pjIns");

//
//  Method name : CCompiler
//  Description : Default constructor.  Assumes that all package class names
//    will be relative to the current working directory.  ( ./ )
//
CCompiler::CCompiler(bool parameterized)
  : fClassPath(),
    fCurrentFunction(0),    
    fParseError(false),
    fCurrentLocalVariable(0),
    fInClassInitializers(false),
    fFilterMode(false),
    fPrintDependencies(false),
    fOutput(0),
    fParameterized(parameterized),
    fInstantiationPackage(kDefaultIPkg),
    fParsingSource(0)
{
  fInstantiationList.SetCompilerAlias(this);
  if (parameterized) InitializePrimitiveClasses();
}

//
//  Method name : CCompiler
//  Description : Constructor.  Takes a list containing all of the root
//    directories to be searched by this compiler.  For every package name
//    used in the source, each of these directory trees is searched in the
//    provided order.
//    Each of the provided directories are checked for validity (make sure
//    they are readable directories) before use.
//    If the provided list is empty, then it will just search the current
//    working directory.  (like the default constructor)
//
CCompiler::CCompiler(const deque<string>& classPath, bool parameterized)
  : fClassPath(),
    fCurrentClass(0),
    fCurrentFunction(0),
    fParseError(false),
    fCurrentLocalVariable(0),
    fInClassInitializers(false),
    fFilterMode(false),
    fPrintDependencies(false),
    fOutput(0),
    fParameterized(parameterized),
    fInstantiationPackage(kDefaultIPkg),
    fParsingSource(0)
{
  fInstantiationList.SetCompilerAlias(this);
  for (deque<string>::const_iterator i = classPath.begin();
       !(i == classPath.end()); i++) {
    CJavaDirectory oneDirectory(*i,parameterized);
    if (oneDirectory.IsValid()) {
      fClassPath.push_back(oneDirectory);
    }
    // 'foo' in the classpath really means 'foo:foo/classes.zip'
    string zipPath = (*i) + "/classes.zip";
    CFilePath fp(zipPath);
    if (fp.IsReadable()) {
      CJavaDirectory zipDir(zipPath,parameterized);
      if (zipDir.IsValid()) {
	fClassPath.push_back(zipDir);
      }
    }
  }
  if (fClassPath.size() < 1) {
    fClassPath.push_back(CJavaDirectory(".",parameterized));
  }
  if (parameterized) InitializePrimitiveClasses();
}

//
//  Method name : ~CCompiler
//  Description : Destructor
//
CCompiler::~CCompiler()
{
  IntermediateList::iterator classes = fIntermediateClasses.begin();
  for (; classes != fIntermediateClasses.end(); classes++) {
    delete *classes;
  }
  delete fCurrentFunction;
  delete fCurrentClass;
  delete fOutput;
}

//
//  Method name : Reset
//  Description : Called in-between PartialCompileFile/GenerateClass
//    cycles to remove extraneous material from this.
//
void
CCompiler::Reset()
{

}

//
//  Method name : ReduceImportAliases
//  Description : Removes extraneous import aliases from fImportAliases, while
//     removing all those which refer to fully qualified classes.
//
void
CCompiler::ReduceImportAliases()
{
}

//
//  Method name : TranslateFiles
//  Description : This static function translates all of the files in 
//      'fileNames', storing their translations relative to 'javaDest',
//      and their instantiations relative to instDest.  It sets 
//      'insInfoStr' to a newline-separated list of new files, and 
//      generates all instantiations which this file requires.  
//    * If 'alwaysGenerate' is true, all instantiations are generated;
//      otherwise, only those whose dependencies require it are generated.
//    * Returns true iff successful.
//
bool
CCompiler::TranslateFiles(const deque<string>& fileNames,
			  const deque<string>& classPath,
			  ClassList &resultClasses,
			  const CFilePath& javaDest, 
			  const CFilePath& instDest,
			  const string& instPackage,
			  string& insInfoStr,
			  bool alwaysGenerate,
			  bool printDependencies)
{
  CCompiler compiler(classPath, true);
  compiler.fFilterMode = true;
  compiler.fPrintDependencies = printDependencies;
  if (!compiler.CheckJavaLangObject()) {
    cerr << "Could not find " << UnicodeToString(kObjectName) <<
      " in the class path." << endl;
    return false;
  }
  if (!compiler.CheckSupportClasses()) {
    return false;
  }
  compiler.fInstantiationPackage = ::UTFToUnicode(instPackage);

#if 0
  SourceFileList sourceFiles;
#endif
  bool success = true;

  ///
  // Compile phase one: Get interface information.
  /// 

  for (deque<string>::const_iterator fileName = fileNames.begin();
       fileName != fileNames.end(); ++fileName) {
    compiler.fDependencies.push_back(*fileName);
    bool isPJFile =
      (*fileName).substr((*fileName).size()-3).compare(".pj") == 0;
    CSourceFile *file = new CSourceFile();
    file->fIsPJFile = isPJFile;
    bool result = compiler.PartialCompileFile(*fileName, file);
    success &= result;
    if (result && !isPJFile) {
      compiler.HaveJavaFile(*fileName);
      //delete file;
    }
  }

  ///
  // Compile phase one': Resolve files.
  ///
  for(deque<CSourceFile*>::iterator file = compiler.fParsedFiles.begin();
      file != compiler.fParsedFiles.end(); ++file) {
    if ((*file)->fIsValid) {
      bool result = compiler.ResolveFile(*file);
      success &= result;
    }
  } 

  ///
  // Compile phase two: Generate files.
  ///

  IntermediateList intermediates;

  for(deque<CSourceFile*>::iterator file = compiler.fParsedFiles.begin();
      file != compiler.fParsedFiles.end(); ++file) {
    if (! (*file)->fIsPJFile ||
	! (*file)->fIsValid) continue;

    string package = ::UnicodeToUTF((*file)->fPackageName);

    for(unsigned int i = 0; i < package.length(); ++i)
      if (package[i]=='.') package[i] = '/';
    string fileName = (*file)->fFileName;
    string::size_type lastSlash = fileName.find_last_of('/');
    if (lastSlash == string::npos) {
      lastSlash = 0;
    } else {
      ++lastSlash;
    }
    string fileDirectory(fileName,0,lastSlash);
    string baseFileName(fileName,lastSlash);
    string outFileName(baseFileName,0,baseFileName.length()-3);
    outFileName = (javaDest.GetString()  
                   + package + "/" + outFileName + ".java");  

    if (!CFilePath::MakeDirectories(outFileName)) {
      cerr << "Could not make directories for " << outFileName << endl;
      return false;
    }
    ofstream outFile;
    outFile.open(outFileName.c_str());
    if (!outFile.is_open()) {
      cerr << ">> Could not write to " << outFileName << endl;
      exit(-1);
    }
    compiler.fOutput = new srcstream(outFile);
    
    if ((*file)->fPackageName.length() != 0) {
      compiler.fOutput->ToLine((*file)->fPackageLineNumber,false);
      compiler.fOutput->out() << "package " 
			      << ::SymbolToString((*file)->fPackageName)
			      << ';';
    }	  
    ////
    // Unparse import statements, if any.
    ////
    compiler.UnparseImports(*file);
    ////
    // Now, unparse all classes.
    ////
    unicode_string unicodeFilename = ::StringToUnicode(fileName);
    IntermediateList::iterator begin = (*file)->fIntermediateClasses.begin();
    IntermediateList::iterator end = (*file)->fIntermediateClasses.end();   
    IntermediateList::iterator classIterator;
    for (classIterator = begin; classIterator != end; ++classIterator) {
      if (! compiler.GenerateClass(*(*classIterator))) {
	success = false;
	break;
      } else {
	intermediates.push_back(*classIterator);
      }
    }
    compiler.fOutput->out() << '\n';
    compiler.GeneratedFile(CJavaTypeSignature::kVoid,outFileName); 
    if (!success) continue;
      

    ////
    // Fixtypes on Auxmethods.
    ////
    for (classIterator = begin; classIterator != end; ++classIterator) {
      assert((*classIterator)->fRealClass);
      if (! compiler.FixAMM(*(*classIterator)->fRealClass, *file)) {
	success = false;
	break;
      }
    }
  }

  ////
  // Generate where-classes
  ////
  
  for (ClassList::iterator i = compiler.fWhereClasses.begin(); 
       i != compiler.fWhereClasses.end(); ++i) {
    string outFileName = 
      javaDest.GetString() + "/" + ::UnicodeToUTF((*i)->GetClassName())
      + "$$$W.java";
    if (!CFilePath::MakeDirectories(outFileName)) {
      cerr << "Couldn't make directories for " << outFileName << endl;
      success = false;
      break;
    }
    ofstream wOut;
    wOut.open(outFileName.c_str());
    if (!wOut.is_open()) {
      cerr << ">> Couldn't open " << outFileName << endl;
    } else {
      compiler.GenerateWhereClass(*(*i),wOut);
      compiler.GeneratedFile((*i)->GetType(),outFileName);
      wOut.close();
    }
  }    
  
  ////
  // Generate IClasses
  ////
  while (! compiler.fInstantiationList.Finished()) {
    CJavaTypeSignature IType = 
      compiler.fInstantiationList.StartImplementing();
    bool shouldInstantiate = alwaysGenerate ||
      compiler.ShouldRegenerateType(IType);
    if (shouldInstantiate) {
      const CJavaClassFile* IClass = compiler.ClassOfType(IType, 0);
      if (IClass != 0) {
	success &= compiler.GenerateInstantiation(IType,
						  instDest.GetString());
      } else {
	cerr << "Couldn't find class for instantiated type "
	     << IType << endl;
	success = false;
      }
    }
  }

  ///
  // Generate TClasses.
  ///
  success &= compiler.GenerateTClasses(javaDest.GetString());

  ////
  // Extract result classes.
  ////
  for (IntermediateList::iterator classIterator = intermediates.begin(); 
       classIterator != intermediates.end(); ++classIterator) {
    if ((*classIterator)->fRealClass != 0) {
      resultClasses.push_back((*classIterator)->fRealClass);
      (*classIterator)->fRealClass = 0;
    }
  }

  insInfoStr = compiler.fJavaFiles;
  //javaDeps = compiler.fJavaDeps;
  return success;
}



//
//  Method name : PartialCompileFile
//  Description : This method takes the provided Java source file name and
//    parses it for all of its interface information.  This does not do any
//    compilation of the internals of method bodies ... this must be done
//    by a later pass.  The second parameter (resultClasses) is filled with
//    the classes that are found in this file.
//
bool
CCompiler::PartialCompileFile(const string& fileName,
			      CSourceFile*& file)
{
  unsigned long previousClassCount = fIntermediateClasses.size();
  CSourceFile *oldFile = ParsedFile(fileName);
  if (oldFile) {
    delete file;
    file = oldFile;
    return true;    
  }
  fParsedFiles.push_back(file);
  fParsingSource = file;
  ////
  // Parse the file.
  ////
  bool success = ParseFile(fileName);
  fParsingSource = 0;
  if (!success) {
    file->fIsValid = false;
    return false;
  }
  ////
  // For each class: PrepareIntermediateClass.
  ////
  unicode_string unicodeFilename = ::StringToUnicode(fileName);  
  IntermediateList::iterator begin = file->fIntermediateClasses.begin();
  IntermediateList::iterator end = file->fIntermediateClasses.end();
  IntermediateList::iterator classIterator;
  for (classIterator = begin; classIterator != end; ++classIterator) {
    PrepareIntermediateClass(*(*classIterator), unicodeFilename);
  }
#if 0
  ////
  // _NOW_ CheckImports.
  ////
  success = CheckImports(file); 
  ////
  // For each class: PrepareClassDeclarations.
  ////
  for (classIterator = begin; classIterator != end; ++classIterator) {    
    if (!PrepareClassDeclarations(*(*classIterator))) {
      success = false;
    }
  }
  if (!success) return false;
  ////
  // For each class: PreparePJInfo.
  ////
  if (fParameterized) {
    for (classIterator = begin; classIterator != end; 
	 ++classIterator) {
      if (!PreparePJInfo(*(*classIterator)->fRealClass, file)) {
	success = false;
      } else {
	resultClasses.push_back(*classIterator);
      }
    } 
  } else { // !fParameterized
    for (classIterator = begin; classIterator != end; ++classIterator) {
      resultClasses.push_back(*classIterator);
    }
  }
#endif
  file->fIsValid &= success;
  return success;
}

//
// Method name : ResolveFile
// Description : Check imports and fully expand all type names in a file.
//     Returns true on success. Sets file->fIsValid accordingly.
// Requires    : All files have been partial-compiled and this file is valid.
//
bool
CCompiler::ResolveFile(CSourceFile *file)
{
  assert(file);
  if (file->fResolved)
    return true;
  file->fResolved = true;
  ////
  // _NOW_ CheckImports.
  ////
  bool success = CheckImports(file); 
  ////
  // For each class: PrepareClassDeclarations.
  ////
  IntermediateList::iterator begin = file->fIntermediateClasses.begin();
  IntermediateList::iterator end = file->fIntermediateClasses.end();
  IntermediateList::iterator classIterator;
  
  for (classIterator = begin; classIterator != end; ++classIterator) {    
    if (!PrepareClassDeclarations(*(*classIterator))) {
      success = false;
    }
  }
  ////
  // For each class: PreparePJInfo.
  ////
  if (success && fParameterized) {
    for (classIterator = begin; classIterator != end; 
	 ++classIterator) {
      if (!PreparePJInfo(*(*classIterator)->fRealClass, file)) {
	success = false;
      }
    } 
  }
  file->fIsValid &= success;
  return success;
}

//
//  Method name : SavedPartialCompileFile
//  Description : Runs PartialCompileFile, preserving the current package 
//     name, as well as other important information.
//
bool
CCompiler::SavedPartialCompileFile(const string& fileName, 
				   IntermediateList& resultClasses)
{
  bool isPJFile =
    fileName.substr(fileName.size()-3).compare(".pj") == 0;
  CSourceFile *sourceFile = new CSourceFile();
  sourceFile->fIsPJFile = isPJFile;
  bool res = PartialCompileFile(fileName, sourceFile);
  if (res) {
    res = ResolveFile(sourceFile);
  }
  if (res)
    resultClasses = sourceFile->fIntermediateClasses;
  if (res && !isPJFile) {
    HaveJavaFile(fileName);
  }
  return res;
}

//
//  Method name : ParsedFile
//  Description : Returns the file iff this file has already been parsed by 
//    this compiler.
//
CSourceFile*
CCompiler::ParsedFile(const string& fileName) const
{
  string::size_type lastSlash = fileName.find_last_of('/');
  if (lastSlash == string::npos) {
    lastSlash = 0;
  } else {
    ++lastSlash;
  }
  string baseName(fileName,lastSlash);
  for (deque<CSourceFile*>::const_iterator file = fParsedFiles.begin();       
       !(file == fParsedFiles.end()); ++file) {
    if (baseName == (*file)->fFileName) {
      return (*file);
    }
  }
  return 0;
}

//
//  Method name : ParseFile
//  Description : This method is used during compilation to parse one input
//    file.  If this operation is successful, it return true.
//
bool
CCompiler::ParseFile(const string& fileName)
{
  FILE* infile = fopen(fileName.c_str(), "rb");
  if (infile == 0) {
    cerr << "Could not open input file: " << fileName << endl;
    return false;
  }
  int fd = fileno(infile);
  mapped_file m(fd);
  if (! m.is_valid()) {
    m.print_error();
  }
  const string &textToParse = m.get_string(); // Could be faster.

  fclose(infile);
  string::size_type start = fileName.find_last_of('/');
  start = (start == string::npos) ? 0 : start + 1;
  fParsingSource->fFileName.assign(fileName, start);
  fParsingSource->fFilePath.assign(fileName);
  if (!ImportPackage(::StringToUnicode(kDefaultImport),fParsingSource)) {
    cerr << "Cannot find required package " << kDefaultImport
	 << " in class path." << endl;
  }
  InitializeParser(textToParse, this); // Could be faster.
  int parsed = ::yyparse();
  yylineno = 0;
  bool success = parsed == 0 && !fParseError;
  ::FinishParser();

  return success;
}

//
//  Method name : AddClassImport
//  Description : This method is called during parsing to indicate a class
//    that should be imported after parsing is complete.
//
void
CCompiler::AddClassImport(const unicode_string& className)
{
  assert(fParsingSource);
  fParsingSource->fClassImports.push_back(className);
  fParsingSource->fClassImportLocations.push_back(yylineno);
}

//
//  Method name : AddPackageImport
//  Description : This method is called during parsing to indicate a package
//    that should be imported after parsing is complete.
//
void
CCompiler::AddPackageImport(const unicode_string& package)
{
  assert(fParsingSource);
  fParsingSource->fPackageImports.push_back(package);
  fParsingSource->fPackageImportLocations.push_back(yylineno);
}

//
//  Method name : CheckImports
//  Description : This method is used internally to check whether the imports
//    that were found in a java file are valid.
//
bool
CCompiler::CheckImports(CSourceFile *file)
{
  bool valid = true;
  int index = 0;
  for (deque<unicode_string>::iterator i = file->fClassImports.begin();
       !(i == file->fClassImports.end()); ++i, ++index) {
    if (ImportClass(*i,file) == 0) {
      PrintCompileError(file->fFileName, 
			"Invalid class: " + ::SymbolToString(*i),
			file->fClassImportLocations[index]);
      valid = false;
    }
  }
  index = 0;
  for (deque<unicode_string>::iterator i = file->fPackageImports.begin();
       !(i == file->fPackageImports.end()); ++i, ++index) {
    if (ImportPackage(*i,file) == 0 && (*i) != file->fPackageName) {
      PrintCompileError(file->fFileName, 
			"Warning: could not find package: " + 
			::UnicodeToUTF(*i),
			file->fPackageImportLocations[index]);
      //valid = false;
    }
  }
  return valid;
}

//
//  Method name : ImportClass
//  Description : This looks for the class with the provided qualified name
//    by searching the class path given to this compiler as an argument.  If
//    the desired class is found, it is loaded in and stored for later
//    reference and a pointer to it is returned.  If it is not found, then
//    0 is returned.
//    The 'aliasBase' argument is used to specify whether the base class
//    name should be visible to the rest of the program.  For example, the
//    qualified class sun.net.www.URLConnection would have a base name
//    'URLConnection', which should be globally visible if someone used an
//    import statement, but not visible if they used a fully qualified class
//    name as a type.
//
const CJavaClassFile*
CCompiler::ImportClass(const unicode_string& className, 
		       CSourceFile *file,
		       bool aliasBase)
{
  const CJavaClassFile* classFile = 0;
  string classString = ::UnicodeToUTF(className);
  unicode_string::size_type lastSlash = className.find_last_of('/');
  bool isSimpleClass = lastSlash == unicode_string::npos;
  ///
  // Have we already built an intermediate class for this?
  ///
  const CIntermediateClass *intermediate = 
    LookupIntermediateClass(className, file);
  if (file && intermediate && intermediate->fRealClass) {
    classFile = intermediate->fRealClass;
    unicode_string fullName = classFile->GetClassName();
    // This assignment just reduces template instantiations...
    const ImportTable* aliases = &file->fImportAliases;
    ImportTable::const_iterator entry = aliases->find(fullName);
    if (entry == aliases->end()) {
      file->fImportAliases.insert(ImportTable::value_type(fullName, 
							  classFile));
      fImportAliases.insert(ImportTable::value_type(fullName,
						    classFile));
    } else {
      classFile = (*entry).second;
    }
    if (!(fullName == className)) {
      file->fImportAliases.insert(ImportTable::value_type(className,
							  classFile));
    }
  } else if (intermediate && intermediate->fRealClass) {
    classFile = intermediate->fRealClass;
    unicode_string fullName = classFile->GetClassName();
    // This assignment just reduces template instantiations...
    const ImportTable* aliases = &fImportAliases;
    ImportTable::const_iterator entry = aliases->find(fullName);
    if (entry == aliases->end()) {
      fImportAliases.insert(ImportTable::value_type(fullName,
						    classFile));
    } else {
      classFile = (*entry).second;
    }
  }

  // ???? Moved code from here.
  
  ///
  // Did we import it already?
  ///
  ImportTable::iterator exists;
  if (file) {
    exists = file->fImportAliases.find(className);
    if (exists == file->fImportAliases.end()) {
      exists = fImportAliases.find(className);
    }
  } else {
    exists = fImportAliases.find(className);
  }
  if (classFile == 0 && !(exists == fImportAliases.end())) {
    classFile = (*exists).second;
    if (!isSimpleClass && aliasBase) {
      unicode_string shortName(className, lastSlash + 1);
      if (fImportAliases.find(shortName) == fImportAliases.end()) {
	fImportAliases.insert(ImportTable::value_type(shortName,
						      classFile));
      }
    }
  }

  ///
  // Have we loaded it before?
  ///
  if (classFile == 0) {
    classFile = FindImportedClass("", className);
    if (classFile && file) {
      file->fImportAliases.insert(ImportTable::value_type(className, 
							  classFile));
      if (!isSimpleClass) {
	bool addAlias = aliasBase;
	if (addAlias) {
	  unicode_string shortName(className, lastSlash + 1);
	  if (file->fImportAliases.find(shortName) == 
	      file->fImportAliases.end()) {
	    file->fImportAliases.insert(ImportTable::value_type(shortName,
								classFile));
	  }
	}
      }
    }
  }

  // ???? I moved this.  I think it was in the wrong place before.
  
  ///
  // If it has a simple name, check whether it comes from an imported package.
  ///
  if (isSimpleClass && file) {
    for (int n = 0; n <= 1; ++n) { // Do we look for source files?
      bool source = (n==1);
      for (ClassPathList::const_iterator i = file->fImportedPackages.begin();
	   classFile == 0 && !(i == file->fImportedPackages.end()); ++i) {
	classFile = FindImportedClass((*i).GetRelativePath(),
				      className);
	if (!classFile)
	  classFile = ImportOneClassPath(className, *i, source);
	if (classFile != 0) {
	  unicode_string fullName = classFile->GetClassName();
	  // This assignment just reduces template instantiations...
	  ImportTable* aliases = &file->fImportAliases;
	  ImportTable::iterator entry = aliases->find(fullName);
	  if (entry == aliases->end()) {
	    aliases->insert(ImportTable::value_type(fullName, 
						    classFile));
	  } else {
	    classFile = (*entry).second;
	  }
	  if (!(fullName == className) && aliasBase) {
	    aliases->insert(ImportTable::value_type(className,
						    classFile));
	  }
	  break;
	}
      }
    }
  }

  ///
  // Now try each directory in the classpath.
  ///
  for (int n = 0; n <= 1; ++n) { // Do we look for source?
    bool source = (n == 1);
    for (ClassPathList::const_iterator i = fClassPath.begin();
	 classFile == 0 && !(i == fClassPath.end()); ++i) {
      classFile = ImportOneClassPath(className, *i, source);
      if (classFile != 0 && file) {
	file->fImportAliases.insert(ImportTable::value_type(className, 
							    classFile));
	CJavaDirectory thisClassPath(*i,
			       ::UnicodeToUTF(classFile->GetPackageName()));
	if (!isSimpleClass) {
	  bool addAlias = aliasBase;
	  if (!addAlias) {
	    addAlias = true;
	    for (ClassPathList::const_iterator package =
		   file->fImportedPackages.begin();
		 addAlias && !(package == file->fImportedPackages.end()); 
		 ++package) {
	      if (*package == thisClassPath) {
		addAlias = false;
	      }
	    }
	  }
	  // XXXX Fix this.
	  if (addAlias) {
	    unicode_string shortName(className, lastSlash + 1);
	    if (file->fImportAliases.find(shortName) == 
		file->fImportAliases.end()) {
	      file->fImportAliases.insert(ImportTable::value_type(shortName,
								  classFile));
	    }
	  }
	}
      }
    }
  }

  ///
  // This is a hack to deal with claspaths that don't match our expectations. 
  //
  // If we still haven't found the file, and it has a simple name, check to
  // make sure we didn't turn it up.
  ///
  if (classFile == 0 && isSimpleClass && file) {
    for (deque<unicode_string>::const_iterator i =
	   file->fPackageImports.begin();
	 classFile == 0 && !(i == file->fPackageImports.end()); ++i) {
      classFile = FindImportedClass(::UnicodeToUTF(*i),
				    className);
    }
    if (classFile != 0) {
      unicode_string fullName = classFile->GetClassName();
      ImportTable* aliases = &file->fImportAliases;
      ImportTable::iterator entry = aliases->find(fullName);
      if (entry == aliases->end()) {
	aliases->insert(ImportTable::value_type(fullName,classFile));
      }
      if (! (fullName == className) && aliasBase) {
	aliases->insert(ImportTable::value_type(className,classFile));
      }
    }
  }

  return classFile;
}

//
//  Method name : ImportOneClassPath
//  Description : This is a helper method used by ImportClass.  It looks for
//    the Java class with the provided name relative to the provided java
//    directory.  If it is successful, the class is added to the import list
//    on this compiler and the class is returned.  Otherwise, it returns 0.
//  * If source, this method will only look for source files; otherwise, it
//    will look for class files.
//
CJavaClassFile*
CCompiler::ImportOneClassPath(const unicode_string& className,
			      const CJavaDirectory& classPath,
			      bool source)
{
#if 0
  cerr << "Looking for " << className << " in ";
  if (classPath.InZipFile()) 
    cerr << classPath.GetZipFileName() << endl;
  else
    cerr << classPath.GetRealDirectory() << endl;
#endif
  CJavaClassFile* classFile = 0;
  if (!source) {
    classFile = classPath.LoadClassFile(className);
    if (classFile && fPrintDependencies) {
	if (classPath.InZipFile()) {
	    cout << classPath.GetZipFileName() << " (" <<
		    SymbolToString(className) << ")" << endl;
	} else {
	    cout << classPath.GetRealDirectory() << className <<
		    ".class" << endl;
	}
    }
  }
  string utfClassName = ::UnicodeToUTF(className);
  if (source && classFile == 0 && !classPath.InZipFile()) {
    string exts[] = {".pj", ".java"};
    for (int exti = (fParameterized ? 0 : 1); !classFile && exti < 2; exti++) {
      string sourceName = classPath.GetRealDirectory() +
	utfClassName + exts[exti];
      CFilePath sourcePath(sourceName);
      if (sourcePath.IsFile() && sourcePath.IsReadable()) {
	IntermediateList resultClasses;
	if (SavedPartialCompileFile(sourceName, resultClasses)) {
	  for (IntermediateList::iterator i = resultClasses.begin();
	       !(i == resultClasses.end()); ++i) {
	    if (className == (*i)->GetShortName() || 
		className == (*i)->GetName()) {
	      // We should check this too, but it seems that Sun's javac 
	      // doesn't.
	      // &&(*i)->fAccessFlags.fPublic) {     
	      classFile = (*i)->fRealClass;
	      if (fPrintDependencies) cout << sourceName << endl;
	      break;
	    }
	  }
	}
      }
    }
  }
  if (classFile != 0) {
    string actualName = ::UnicodeToUTF(classFile->GetClassName());
    string reportedName = ::SymbolToString(classFile->GetClassName());
    string requiredName = utfClassName;
    string packageName = classPath.GetPackage();
    if (packageName.length() > 1) {
      requiredName = packageName + requiredName;
    }
    if (!classPath.InZipFile()) {
      string depend =
	classPath.GetRealDirectory() + utfClassName + ".java";
      if (access(depend.c_str(), F_OK) == 0) {
	fDependencies.push_back(depend);
      }
    }
    if (!(actualName == requiredName)) {
      //cerr << "Invalid class " << reportedName
      //     << " found in " << classPath.GetRealDirectory() << endl;
      // delete classfile;

      // XXXX fix this
      fImportedClasses.push_back(classFile);
      fImportAliases.insert(ImportTable::value_type(classFile->GetClassName(),
						    classFile));      
      classFile = 0;
    } else {
      fImportedClasses.push_back(classFile);
      fImportAliases.insert(ImportTable::value_type(classFile->GetClassName(),
						    classFile));
      WarnDeprecatedClass(*classFile);
    }
    if (classFile && 
	classFile->SeemsParameterized() &&
	! classFile->HasPInfo()) {
      cerr << "Warning: Class " << reportedName
	   << " may be missing a .pjaux file." << endl
	   << "         Trying anyway..." << endl << endl;
    }
  }
  return classFile;
}

//
//  Method name : ImportPackage
//  Description : This looks for the package with the provided qualified
//    name and imports all of the classes directly in that package level
//    using ImportClass.  If the package doesn't exist or can't be accessed,
//    false is returned, otherwise true is returned.  (This implies that
//    true will be returned even when some or all of the class files in
//    the package aren't loadable.)
//
bool
CCompiler::ImportPackage(const unicode_string& packageName,
			 CSourceFile *file)
{
  string packageString = ::UnicodeToUTF(packageName);
  bool found = false;
  for (ClassPathList::const_iterator i = fClassPath.begin();
       !(i == fClassPath.end()); i++) {
    CJavaDirectory relativePath(*i, packageString);
    if (relativePath.IsValid()) {
      file->fImportedPackages.push_back(relativePath);
      found = true;
    }
  }
  return found;
}

//
//  Method name : SetPackage
//  Description : Specifies what the package is for the compilation unit that
//    is currently being compiled.
//
void
CCompiler::SetPackage(const unicode_string& packageName)
{
  assert(fParsingSource);
  fParsingSource->fPackageName = packageName;
  fParsingSource->fPackageLineNumber = yylineno;
  //ImportPackage(packageName, fParsingSource);
  AddPackageImport(packageName);
}

//
//  Method name : NoPackage
//  Description : Specifies that the current compilation unit has no package
//    specified.
//
void
CCompiler::NoPackage()
{
  SetPackage(unicode_string());
}

//
//  Method name : StartClass
//  Description : When the parser hits the beginning of a class declaration,
//    it uses this call to tell the compiler object the important class
//    header information and whatnot.  This keeps track of a current class
//    of evaluation that all future intra-class operations will be applied to.
//
void
CCompiler::StartClass(const unicode_string& className,
		      CJavaAccessFlags* adoptModifiers,
		      CJavaTypeSignature* adoptExtends,
		      deque<CJavaTypeSignature>* adoptInterfaces,      
		      deque<unicode_string>* adoptParameters,
		      deque<CPJavaConstraint*>* adoptConstraints,
		      bool deprecated)
{
  assert(fCurrentClass == 0);
  assert(fCurrentFunction == 0);
  assert(fParameterized  || adoptParameters  == 0); // parser should catch
  assert(fParameterized  || adoptConstraints == 0); // parser should catch
  unicode_string fullName;  
  if (fParsingSource->fPackageName.size() > 0) {
    fullName = fParsingSource->fPackageName;
    fullName += (unicode_char)'/';
    fullName += className;
  } else {
    fullName = className;
  }
  fCurrentClass = new CIntermediateClass(fParsingSource,
					 fParsingSource->fFileName, 
					 fullName, adoptModifiers,
					 adoptExtends, adoptInterfaces,
					 deprecated, yylineno);  
  if (fParameterized) {
    fCurrentClass->fPClassInfo = 
      new CPJavaClassInfo(fullName, adoptParameters, adoptExtends, 
			  adoptInterfaces, adoptConstraints);
  }
}

//
//  Method name : StartInterface
//  Description : This function signals the beginning of an interface
//    declaration in the file that is currently being parsed.  The compiler
//    object sets things up so that future fields will be accepted and
//    understood.
//
void
CCompiler::StartInterface(const unicode_string& className,
			  CJavaAccessFlags* adoptModifiers,
			  deque<CJavaTypeSignature>* adoptInterfaceExtends,
			  deque<unicode_string>* adoptParameters,
			  deque<CPJavaConstraint*>* adoptConstraints,
			  bool deprecated)
{
  StartClass(className, adoptModifiers, 0, adoptInterfaceExtends,
	     adoptParameters, adoptConstraints, deprecated);
}

//
//  Method name : EndClass
//  Description : This method is used by the parser to tell the compiler
//    that the current class scope has ended.  This allows the compiler
//    to clean up and prepare for the next class or interface in the file.
//
void
CCompiler::EndClass()
{
  assert(fCurrentFunction == 0);
  assert(fCurrentClass != 0);
  fCurrentClass->fEndLineNumber = yylineno;
  if (fParameterized)
    ClearParameterNames();
  if (fCurrentClass->fStaticInitializer != 0 ||
      fCurrentClass->fStaticDeclarations.size() > 0) {
    static const deque<CJavaTypeSignature> noArguments;
    static const CJavaTypeSignature staticReturnType(CJavaTypeSignature::Void);
    static const CJavaMethodSignature staticSignature(staticReturnType,
						   kStaticName, noArguments);
    CJavaAccessFlags modifiers;
    modifiers.fStatic = 1;
    CIntermediateFunction* staticFunction = 
      new CIntermediateFunction(staticSignature, modifiers, 0, 0, yylineno);
    if (fCurrentClass->fStaticInitializer != 0) {
      staticFunction->fBlock = fCurrentClass->fStaticInitializer;
      fCurrentClass->fStaticInitializer = 0;
    } else {
      staticFunction->fBlock = new CCompoundStatement(new StatementList);
    }
    staticFunction->fMaxLocalVariables = fCurrentClass->fStaticLocalVariables;
    fCurrentClass->fFunctions.push_back(staticFunction);
  }
  fIntermediateClasses.push_back(fCurrentClass);
  fParsingSource->fIntermediateClasses.push_back(fCurrentClass);
  fCurrentClass = 0;
}

//
//  Method name : EndBadClass
//  Description : This method is used by the parser to tell the compiler
//    that the current class scope has ended, and that the current class,
//    if any, is a lost cause.  This allows the compiler
//    to clean up and prepare for the next class or interface in the file.
//
void
CCompiler::EndBadClass()
{
  assert(fCurrentFunction == 0);

  if (fParameterized)
    ClearParameterNames();

  if (fCurrentClass == 0)
    return;

  fCurrentClass = 0;
}


//
//  Method name : StartConstructor
//  Description : This method signifies the beginning of a constructor
//    function, which is specified slightly differently than a normal
//    function, which has return types and whatnot.
//
void
CCompiler::StartConstructor(const CJavaAccessFlags& modifiers,
			    deque<CJavaFieldSignature>* adoptParameters,
			    deque<CJavaTypeSignature>* adoptThrows,
			    bool deprecated)
{
  static CJavaTypeSignature type(CJavaTypeSignature::Void);
  StartFunction(type, kConstructorName, modifiers, adoptParameters, 
		adoptThrows, deprecated);
}

//
//  Method name : StartFunction
//  Description : Used by the parser to signal the beginning of a new method
//    in a class.  All of the provided information is grabbed out of the
//    function header before the body is parsed, and the compiler has a
//    chance here to set up the appropriate scopes and other information.
//
void
CCompiler::StartFunction(const CJavaTypeSignature& returnType,
			 const unicode_string& functionName,
			 const CJavaAccessFlags& modifiers,
			 deque<CJavaFieldSignature>* adoptParameters,
			 deque<CJavaTypeSignature>* adoptThrows,
			 bool deprecated)
{
  assert(fCurrentClass != 0);
  if (fCurrentFunction != 0)
    EndBadFunction();
  assert(fLocalVariables.size() == 0);
  assert(fVariableScopes.size() == 0);
  assert(fCurrentLocalVariable == 0);
  CJavaAccessFlags newModifiers = modifiers;
  if (modifiers.fStatic == 0) {
    fCurrentLocalVariable = 1;
  }

  if (GetCurrentClass()->fAccessFlags.fInterface != 0) {
    newModifiers.fAbstract = 1;
    newModifiers.fPublic = 1;
  }
  PushLocalScope();
  deque<CJavaTypeSignature> parameterTypes;
  if (adoptParameters != 0) {
    for (deque<CJavaFieldSignature>::iterator i = adoptParameters->begin();
	 i != adoptParameters->end(); ++i) {
      AddLocalVariable(*i);
      parameterTypes.push_back((*i).GetType());
    }
  }
  CJavaMethodSignature signature(returnType, functionName, parameterTypes);
  fCurrentFunction = new CIntermediateFunction(signature, newModifiers,
		       adoptParameters, adoptThrows, yylineno, deprecated);
}


//
//  Method name : EndFunction
//  Description : This is used by the parser to signal the end of a Java
//    function.  The compiler uses this to clean up any global state and
//    close out the last scope.
//
void
CCompiler::EndFunction(CCompoundStatement* statementBlock)
{
  assert(fCurrentFunction != 0);
  assert(fCurrentClass != 0);
  if (fVariableScopes.size() > 0)
    PopLocalScope();
  fVariableScopes.erase(fVariableScopes.begin(), fVariableScopes.end());
  fLocalVariables.erase(fLocalVariables.begin(), fLocalVariables.end());
  fCurrentFunction->fMaxLocalVariables = fCurrentLocalVariable;
  fCurrentFunction->fBlock = statementBlock;
  fCurrentFunction->fEndLineNumber = yylineno;
  fCurrentClass->fFunctions.push_back(fCurrentFunction);
  fCurrentFunction = 0;
  fCurrentLocalVariable = 0;
}

//
//  Method name : EndBadFunction
//  Description : This is used by the parser to signal the end of a Java
//    function that isn't worth knowing about.  The compiler uses this to 
//    clean up any global state and close out the last scope.
//
void
CCompiler::EndBadFunction(CCompoundStatement* statementBlock)
{
  assert(fCurrentClass != 0);
  while (fVariableScopes.size() > 0)
    PopLocalScope();
  fVariableScopes.erase(fVariableScopes.begin(), fVariableScopes.end());
  fLocalVariables.erase(fLocalVariables.begin(), fLocalVariables.end());
  if (fCurrentFunction == 0) 
    return;
  fCurrentFunction->fMaxLocalVariables = fCurrentLocalVariable;
  fCurrentFunction->fBlock = statementBlock;
  fCurrentFunction->fEndLineNumber = yylineno;
  fCurrentClass->fFunctions.push_back(fCurrentFunction);
  fCurrentFunction = 0;
  fCurrentLocalVariable = 0;
}


//
//  Method name : PushLocalScope
//  Description : This function is used to signify a new local scope in an
//    active method.  This allows new variables to be declared that will be
//    removed from visibility after the scope exits.
//
void
CCompiler::PushLocalScope()
{
  fVariableScopes.push_front(deque<unicode_string>());
}


//
//  Method name : PopLocalScope
//  Description : This is the converse of PushLocalScope ... it tells the
//    compiler that the parser is leaving a local variable scope so it
//    should remove all local variables that were in the current scope.
//
void
CCompiler::PopLocalScope()
{
  assert(fVariableScopes.size() > 0);
  deque<unicode_string>::iterator end = fVariableScopes.front().end();
  for (deque<unicode_string>::iterator i = fVariableScopes.front().begin();
       i != end; i++) {
    LocalVariableTable::iterator variable = fLocalVariables.find(*i);
    assert(variable != fLocalVariables.end());
    fLocalVariables.erase(variable);
  }
  fVariableScopes.pop_front();
}


//
//  Method name : AddLocalVariable
//  Description : Adds a new local variable to the current scope with the
//    provided name.  The compiler assigns it the next available variable
//    index.  (Doesn't bother re-using out-of-scope ones since Sun's doesn't
//    seem to do this, either.)
//    Returns the index of the new variable.
//
unsigned short
CCompiler::AddLocalVariable(const CJavaFieldSignature& signature)
{  
  unicode_string name = signature.GetFieldName();
  (fVariableScopes.front()).push_front(name);
  fLocalVariables.insert(LocalVariableTable::value_type(name,
		   ::make_pair(signature, fCurrentLocalVariable)));
  unsigned short index = fCurrentLocalVariable;
  fCurrentLocalVariable += signature.GetType().GetWidth();
  return index;
}

//
//  Method name : AddUnnamedVariable
//  Description : This method is used to reserve space for a local variable
//    entry which is never referenced by name.  This is needed for subroutine
//    handling used by the 'finally' part of a 'try' statement, which stores
//    an entry in a local variable index.
//    Returns the index of the new local variable.
//
unsigned short
CCompiler::AddUnnamedVariable(unsigned long width)
{
  unsigned short index = fCurrentLocalVariable;
  fCurrentLocalVariable += width;
  return index;
}

//
//  Method name : LookupLocalVariable
//  Description : Goes into the local variable table and checks to see if
//    the variable with the provided name has been declared in the current
//    scope.  If it has, a pair encapsulating the local variable index for
//    code generation and the signature of the variable is returned.
//    If not, 0 is returned.
//
const pair<CJavaFieldSignature, CCompiler::LocalVariableIndex>*
CCompiler::LookupLocalVariable(const unicode_string& name) const
{
  LocalVariableTable::const_iterator match = fLocalVariables.find(name);
  if (match != fLocalVariables.end()) {
    return &((*match).second);
  } else {
    return 0;
  }
}

//
//  Method name : AddField
//  Description : This method is used to add a data field to the current
//    class or interface.  The declaration is examined to see whether or not
//    it is static to determine where the code goes.
//
void
CCompiler::AddField(CDeclarationStatement* declaration)
{
  assert(fCurrentClass != 0);
  assert(declaration != 0);
  assert(declaration->fDeclarations != 0);
  assert(declaration->fModifiers != 0);
  if (GetCurrentClass()->fAccessFlags.fInterface != 0) {
    declaration->fModifiers->fStatic = 1;
    declaration->fModifiers->fFinal = 1;
  }
  if (declaration->fModifiers->fStatic != 0) {
    fCurrentClass->fStaticDeclarations.push_back(declaration);
    if (fCurrentClass->IsParameterized()) {
      // We need auxmethods for static fields.
      CAuxMethodMap &AMM = *fCurrentClass->fPClassInfo->fAMM;
      for (deque<CVariableDeclaration*>::iterator i =
	     declaration->fDeclarations->begin(); 
	   i != declaration->fDeclarations->end(); ++i) {
	if (declaration->fModifiers->fPrivate) {
	  PrintCompileError(fCurrentClass->fSourceFileName,
	 "Warning: field declared as private will be translated as default",
			    declaration->GetLineNumber());
	}
	unicode_string name((*i)->GetSignature().GetFieldName());
	AMM.RequireMethod(CAuxMethodSignature::FetchStatic,
			  (*i)->GetType(), &name);
	AMM.RequireMethod(CAuxMethodSignature::SetStatic,
			  (*i)->GetType(), &name);	
      }
    }
  } else {
    fCurrentClass->fNonStaticDeclarations.push_back(declaration);
  }
}

//
//  Method name : AddStaticCode
//  Description : This adds some static code to the current class, which will
//    be added to the <clinit> pseudo-method and executed at class load time.
//
void
CCompiler::AddStaticCode(CCompoundStatement* statement)
{
  assert(statement != 0);
  assert(fCurrentClass != 0);
  if (statement->fChildren->size() > 0) {
    if (fCurrentClass->fStaticInitializer != 0) {
      fCurrentClass->fStaticInitializer->fChildren->push_back(statement);
      fCurrentClass->fStaticLocalVariables += fCurrentLocalVariable;
    } else {
      StatementList* sList = new StatementList();
      sList->push_back(statement);
      fCurrentClass->fStaticInitializer = 
	new CCompoundStatement(sList);
      fCurrentClass->fStaticLocalVariables = fCurrentLocalVariable;
    }
  } else {
    delete statement;
  }
  fCurrentLocalVariable = 0;
}

//
//  Method name : PrepareIntermediateClass
//  Description : After parsing is complete, this call is used by the compiler
//    to prepare the intermediate class so that it can be queried for class
//    signature information like any imported class.  The compiler must be
//    able to do this before the intermediate class is completely resolved
//    to allow two classes in the same file to safely reference each other.
//
void
CCompiler::PrepareIntermediateClass(CIntermediateClass& intermediate,
				    const unicode_string& fileName)
{
  delete intermediate.fRealClass;
  intermediate.fRealClass = new CJavaClassFile;
  intermediate.fRealClass->fVersion = kVersionID;
  intermediate.fRealClass->fAccessFlags = intermediate.fAccessFlags;
  intermediate.fRealClass->fSourceFile =
    new CJavaSourceFileAttribute(fileName);
  intermediate.fRealClass->fThisClassName = intermediate.fName;
  if (intermediate.fExtends != 0) {
    intermediate.fRealClass->fSuperclassName = *intermediate.fExtends;
  } else if (!(intermediate.fName == kObjectName)) {
    intermediate.fRealClass->fSuperclassName = kObjectName;
    if (intermediate.fPClassInfo) {
      intermediate.GetPInfo()->fSuper = new CJavaTypeSignature(kObjectName);
    }
  } 
  intermediate.fRealClass->fInterfaces = intermediate.fInterfaces;
  intermediate.fRealClass->fDeprecated = intermediate.fDeprecated;
  if (fParameterized && (intermediate.fPClassInfo != 0)) {
    intermediate.fRealClass->fPClassInfo = intermediate.fPClassInfo;
  }
}

//
//  Method name : PrepareClassDeclarations
//  Description : This method is used to set up the fields and methods
//    declared on one class being compiled.  If the class was declared to
//    have a field with type like 'String,' for example, this is fixed to
//    be 'java.lang.String' before insertion.
//    If no problems occur during this preparation, this method returns true.
//
bool
CCompiler::PrepareClassDeclarations(CIntermediateClass& intermediate)
{

  bool result = true;
  CJavaClassFile* realClass = intermediate.fRealClass;
  CSourceFile* file = intermediate.GetSource();
  assert(realClass != 0);

  ////
  // Check constraints.
  ////
  if (realClass->HasPInfo()) {
    string *errMsg = realClass->GetPInfo().CheckConstraints();
    if (errMsg) {
      PrintCompileError(intermediate.fSourceFileName, *errMsg,
			intermediate.fStartLineNumber);
      delete errMsg;
      result = false;
    }
  }
  ////
  // Fixtypes on static fields, then insert them.
  ////
  for (StatementList::iterator i = intermediate.fStaticDeclarations.begin();
       !(i == intermediate.fStaticDeclarations.end()); ++i) {
    CDeclarationStatement* declaration =
      DYNAMIC_CAST(CDeclarationStatement, *i);
    assert(declaration != 0 && declaration->fDeclarations != 0 &&
	   declaration->fModifiers != 0);
    deque<CVariableDeclaration*>::iterator field =
      declaration->fDeclarations->begin();
    CJavaTypeSignature fieldClassType;
    for (; !(field == declaration->fDeclarations->end()); ++field) {
      CJavaFieldSignature oldSignature = (*field)->GetSignature();
      unicode_string fieldName = oldSignature.GetFieldName();
      CJavaFieldInfo existingFieldInfo;
      const CJavaClassFile* fieldAlreadyOn =
	FindField(fieldName, realClass->GetType(), 
		  existingFieldInfo, fieldClassType,
		  file);
      if (fieldAlreadyOn != 0) {
	string message = "Field name '" + ::UnicodeToString(fieldName) +
	  "' already in use.";
	PrintCompileError(intermediate.fSourceFileName, message,
			  declaration->GetLineNumber());
	result = false;
      }
      if (result) {
	CJavaFieldSignature newSignature(FixType(oldSignature.GetType(), file),
					 fieldName);
	CJavaFieldInfo* fieldInfo =
	  new CJavaFieldInfo(*declaration->fModifiers, newSignature, 0,
			     declaration->fDeprecated);
	realClass->fFields.insert(
			  FieldTable::value_type(newSignature, fieldInfo));
      }
    }
  }
  ////
  // Fix types on, then insert non-static fields.
  ////
  for (StatementList::iterator i = intermediate.fNonStaticDeclarations.begin();
       !(i == intermediate.fNonStaticDeclarations.end()); ++i) {
    CDeclarationStatement* declaration =
      DYNAMIC_CAST(CDeclarationStatement, *i);
    assert(declaration != 0 && declaration->fDeclarations != 0 &&
	   declaration->fModifiers != 0);
    deque<CVariableDeclaration*>::iterator field =
      declaration->fDeclarations->begin();
    CJavaTypeSignature fieldClassType;
    for (; !(field == declaration->fDeclarations->end()); ++field) {
      CJavaFieldSignature oldSignature = (*field)->GetSignature();
      unicode_string fieldName = oldSignature.GetFieldName();
      CJavaFieldInfo existingFieldInfo;
      const CJavaClassFile* fieldAlreadyOn =
	FindField(fieldName, realClass->GetType(), 
		  existingFieldInfo, fieldClassType,
		  file);
      if (fieldAlreadyOn != 0) {
	string message = "Field name '" + ::UnicodeToString(fieldName) +
	  "' already in use.";
	PrintCompileError(intermediate.fSourceFileName, message,
			  declaration->GetLineNumber());
	result = false;
      }
      if (result) {
	CJavaFieldSignature newSignature(FixType(oldSignature.GetType(), file),
					 fieldName);
	CJavaFieldInfo* fieldInfo =
	  new CJavaFieldInfo(*declaration->fModifiers, newSignature, 0,
			     declaration->fDeprecated);
	realClass->fFields.insert(
			  FieldTable::value_type(newSignature, fieldInfo));
      }
    }
  }
  ////
  // Fixtypes on, then insert, all methods.
  ////
  deque<CIntermediateFunction*>::iterator functions =
    intermediate.fFunctions.begin();
  bool foundConstructor = false;
  for (; functions != intermediate.fFunctions.end(); ++functions) {
    CJavaMethodInfo* info = new CJavaMethodInfo();
    FixMethodSignature((*functions)->fSignature, info->fSignature, file);
    info->fTranslatedSignature = info->fSignature;
    if (info->fSignature.GetName() == kConstructorName) {
      foundConstructor = true;
    }
    info->fDeprecated = (*functions)->fDeprecated;
    if ((*functions)->fThrows.size() > 0) {
      info->fExceptions = new CJavaExceptionsTable();
      for (deque<unicode_string>::const_iterator throws =
	     (*functions)->fThrows.begin();
	   !(throws == (*functions)->fThrows.end()); ++throws) {
	unicode_string throwClassName;
	FixType(*throws, file).GetBaseClassName(throwClassName);
	info->fExceptions->AddException(throwClassName);
      }
    }
    info->fAccessFlags = (*functions)->fAccessFlags;

    realClass->InsertMethod(info);
  }
  ////
  // Add a default constructor, if no constructor was provided.
  ////
  if (!foundConstructor && intermediate.fAccessFlags.fInterface == 0) {
    deque<CJavaTypeSignature> noArguments;
    CJavaMethodSignature constructorSignature(CJavaTypeSignature::kVoid,
			 kConstructorName, noArguments);
    CJavaAccessFlags flags;
    flags.fPrivate = intermediate.fAccessFlags.fPrivate;
    flags.fPublic = intermediate.fAccessFlags.fPublic;
    flags.fProtected = intermediate.fAccessFlags.fProtected;
    CIntermediateFunction* defaultConstructor =
      new CIntermediateFunction(constructorSignature, flags, 0, 0, yylineno);
    defaultConstructor->fMaxLocalVariables = 1;
    defaultConstructor->fBlock = new CCompoundStatement(0);
    intermediate.fFunctions.push_back(defaultConstructor);
    CJavaMethodInfo* info = new CJavaMethodInfo;
    info->fSignature = constructorSignature;
    info->fAccessFlags = flags;
    realClass->InsertMethod(info);
  }

  if (realClass->HasPInfo()) {
    ////
    // Make sure that all supertypes are real!
    ////
    if (realClass->GetPInfo().fSuper) {
      CCompileError *error = CheckType(*realClass->GetPInfo().fSuper,
				       *realClass, file);
      if (error) {
	PrintCompileError(intermediate.fSourceFileName, error->GetMessage(),
			  intermediate.fStartLineNumber);
	delete error;
	result = false;
	*realClass->GetPInfo().fSuper = CJavaTypeSignature::kObject;
	realClass->fSuperclassName = kObjectName;
	if (intermediate.fExtends)
	  *intermediate.fExtends = kObjectName;
      }
    }
    if (realClass->GetPInfo().fInterfaces) {
      bool badInterface = false;
      for(deque<CJavaTypeSignature>::const_iterator i =
	    realClass->GetPInfo().InterfacesBegin();
	  i != realClass->GetPInfo().InterfacesEnd(); ++i) {
	CCompileError *error = CheckType(*i, *realClass, file);
	if (error) {
	  PrintCompileError(intermediate.fSourceFileName, error->GetMessage(),
			    intermediate.fStartLineNumber);
	  delete error;
	  result = false;
	  badInterface = true;
	}
      }
      if (badInterface) {
	intermediate.fInterfaces.erase(intermediate.fInterfaces.begin(),
				       intermediate.fInterfaces.end());
	realClass->fInterfaces.erase(realClass->fInterfaces.begin(),
				     realClass->fInterfaces.end());
	delete realClass->GetPInfo().fInterfaces;
	realClass->GetPInfo().fInterfaces = new deque<CJavaTypeSignature>;
      }
    }
  }
  ////
  // Fixtype on parent, and look for unimplemented methods.
  ////
  if (realClass->fSuperclassName.length() > 0) {
    const CJavaClassFile* parent = LookupClass(realClass->fSuperclassName,
					       file);
    CJavaTypeSignature parentType;
    if (realClass->fPClassInfo && realClass->fPClassInfo->fSuper) {
      parentType = FixType(*realClass->GetPInfo().fSuper, file);
      *realClass->GetPInfo().fSuper = parentType;
    } else if (parent) {
      parentType = parent->GetType();
    }
    if (parent == 0) {
      cerr << "Invalid class name used as parent of "
	   << ::UnicodeToString(intermediate.fName) << ": "
	   << ::UnicodeToString(realClass->fSuperclassName) << endl;
      result = false;
    } else if (! ValidParent(realClass->GetType(), parentType, file)) {
      cerr << ::UnicodeToString(intermediate.fName) << " cannot extend "
	   << ::UnicodeToString(realClass->fSuperclassName) << endl;
      result = false;
    } else {
      if (!(parent->GetClassName() == realClass->fSuperclassName)) {
	realClass->fSuperclassName = parent->GetClassName();
      }
      if (realClass->GetAccessFlags().fAbstract == 0 &&
	  realClass->GetAccessFlags().fInterface == 0) {
	MethodList MethodList;
	if (UnimplementedMethods(realClass->GetType(), MethodList,file) != 0) {
	  cerr << ::UnicodeToString(intermediate.fName)
	       << " must be declared abstract or implement:" << endl;
	  for (MethodList::const_iterator i = MethodList.begin();
	       !(i == MethodList.end()); ++i) {
	    cerr << "   " << (*i) << endl;
	  }
	  result = false;
	}
      }
    }
  }
  ////
  // Check validity of interfaces
  ////
  if (realClass->HasPInfo() && realClass->GetPInfo().fInterfaces) {
    for (deque<CJavaTypeSignature>::iterator i = 
	   realClass->fPClassInfo->fInterfaces->begin();
	 i != realClass->fPClassInfo->fInterfaces->end(); ++i) {
      *i = FixType(*i, file);
      const CJavaClassFile* interfaceClass = ClassOfType(*i, file);
      if (interfaceClass == 0) {
	cerr << "Invalid type used as interface for "
	     << ::UnicodeToString(intermediate.fName) << ": "
	     << (*i) << endl;
	result = false;
      } else if (! ValidInterface(realClass->GetType(),*i, file)) {
	cerr << (*i) << " cannot serve as an interface for "
	     << ::UnicodeToString(intermediate.fName) << endl;
	result = false;
      }
    }
  } 
  // We go through realClass->fInterfaces anyway, to fix the names.
  for (deque<unicode_string>::iterator i = realClass->fInterfaces.begin();
       i != realClass->fInterfaces.end(); ++i) {
    const CJavaClassFile* interface = LookupClass(*i, file);
    if (interface == 0) {
      cerr << "Invalid class name used as interface for "
	   << ::UnicodeToString(intermediate.fName) << ": "
	   << ::UnicodeToString(*i) << endl;
      result = false;
    } else if (! interface->IsParameterized() && 
	   ! ValidInterface(realClass->GetType(),interface->GetType(),file)) {
      cerr << ::UnicodeToString(*i) << " cannot serve as an interface for "
	   << ::UnicodeToString(intermediate.fName) << endl;
      result = false;
    } else if (!(*i == interface->GetClassName())) {
      *i = interface->GetClassName();
    }
  }
  return result;
}

//
//  Method name : GenerateClass
//  Description : This is the canonical 'second-pass' of the compiler.  It
//    takes one of the intermediate classes that the parser produced and
//    does the necessary checks and transformations on it to produce a valid
//    class file.  If this is successful, the class file is returned to
//    the user.  Otherwise, 0 is returned to signify an error.
//
bool
CCompiler::GenerateClass(CIntermediateClass& intermediate)
{
  unicode_string fileName = ::UTFToUnicode(intermediate.fSourceFileName);
  CJavaClassFile* realClass = intermediate.fRealClass;
  CSourceFile* file = intermediate.GetSource();
  bool result = true;
  ////
  // Check for filename/classname mismatch.
  ////
  if (intermediate.fAccessFlags.fPublic) {
    unicode_string baseFile(fileName, 0, fileName.find_last_of('.'));
    if (!(baseFile == intermediate.GetShortName())) {
      cerr << ::UnicodeToString(fileName)
           << ":"
           << intermediate.GetStartLineNumber()
           << ":"
           << ::UnicodeToString(intermediate.fName)
	   << " may not be declared public in a file named "
	   << ::UnicodeToString(fileName) << endl;
      result = false;
    }
  }
  if (!result) return result;
  ////
  // Find all constants.
  ////
  result = GenerateFieldConstants(intermediate);
  if (!result) return false;
  ////
  // Unparse the class header.
  ////
  if (fFilterMode) {
    result = UnparseClassHeader(intermediate);
  }
  CCompileContext context(this,intermediate.fRealClass,0,
			  intermediate.fSourceFile,0,false);
  CCompileContext staticcontext(this,intermediate.fRealClass,0,
				intermediate.fSourceFile,0,true);
  bool stinit = intermediate.fRealClass->HasPInfo() &&
    intermediate.fRealClass->GetPInfo().fHasStinit;
  deque<CIntermediateFunction*>::iterator function =
    intermediate.fFunctions.begin();
  deque<CIntermediateFunction*>::iterator func_end =
    intermediate.fFunctions.end();
  StatementList::iterator field_iter =
    intermediate.fNonStaticDeclarations.begin();
  StatementList::iterator field_end =
    intermediate.fNonStaticDeclarations.end();
  StatementList::iterator sfield_iter =
    intermediate.fStaticDeclarations.begin();
  StatementList::iterator sfield_end =
    intermediate.fStaticDeclarations.end();
  
  CIntermediateFunction* staticFn = 0; // Pointer to static initializer.
  ////
  // Find static initializer and parent class.
  ////
  if (fFilterMode) {
    for (; function != intermediate.fFunctions.end(); ++function) {
      if ((*function)->fSignature.GetName() == kStaticName)
	staticFn = (*function);
    }
    if (staticFn != 0 && staticFn->fBlock->fChildren != 0)
      fStaticBlockIter = staticFn->fBlock->fChildren->begin();
    function = intermediate.fFunctions.begin();
  }
  const CJavaClassFile* parentClass =
    LookupClass(intermediate.fRealClass->GetSuperclassName(), file);
  CJavaTypeSignature parentType;
  if (intermediate.fRealClass->HasPInfo() &&
      intermediate.fRealClass->GetPInfo().fSuper) {
    parentType = * intermediate.fRealClass->GetPInfo().fSuper;
  } else if (parentClass) {
    parentType = parentClass->GetType();
  }
  // $stinit must go first, in case it converts itself to 
  //         static {super.$stinit}
  if (fFilterMode && fParameterized) {
    // StInit
    CCompileError* error = UnparseStinit(*intermediate.fRealClass,
					 staticFn,
					 intermediate.fStaticDeclarations,
					 file);
    if (error != 0) {
      string message("In the static initializer for ");
      message += ::UnicodeToString(intermediate.fName) +
	": " + ::UnicodeToString(error->GetMessage());
      if (error->GetLine() == CCompileError::kNoLineNumber) {
	PrintCompileError(intermediate.fSourceFileName, message);
      } else {
	PrintCompileError(intermediate.fSourceFileName, message,
			  error->GetLine());
      }
      result = false;
    }
  }
  // Do we write out static fields?
  bool writeStatics = !intermediate.fRealClass->IsParameterized();
  bool writeInitializers = writeStatics;
  ////
  // Now, generate all elements of the class.
  ////
  while ((function != func_end) || 
	 (field_iter != field_end) ||
	 (sfield_iter != sfield_end)) {
    // Which comes first?
    CCompileError* error = 0;
    #define GZ(x) (((x)>0) ? (x) : 1)
    unsigned long field_line = (field_iter != field_end) ?
      GZ((*field_iter)->fLineNumber) : 0;
    unsigned long sfield_line = (sfield_iter != sfield_end) ?
      GZ((*sfield_iter)->fLineNumber) : 0;
    unsigned long fn_line = (function != func_end) ?
      GZ((*function)->fStartLineNumber) : 0;
    string whereStr;
    if (field_line 
	&& (!fn_line || field_line <= fn_line)
	&& (!sfield_line || field_line <= sfield_line)) {
      ////
      // The field comes first.
      ////
      if (error == 0 && !stinit) {
	whereStr = "static initializer";
	error = UnparseStaticBefore(intermediate, staticFn,
				    (*field_iter)->fLineNumber);
      }
      if (error == 0) {
	whereStr = "field declaration";
	if (writeInitializers) {
	  error = (*field_iter)->Unparse(*fOutput, context);
	} else {
	  CDeclarationStatement* decl = DYNAMIC_CAST(CDeclarationStatement,
						     *field_iter);
	  error = decl->UnparseNoInit(*fOutput,context);
	}
      }
      ++field_iter;
    } else if (sfield_line 
	       && (!fn_line || sfield_line <= fn_line)
	       && (!field_line || sfield_line <= field_line)) {
      ////
      // The static field comes first.
      ////
      if (error == 0 && !stinit) {
	whereStr = "static initializer";
	error = UnparseStaticBefore(intermediate, staticFn,
				    (*sfield_iter)->fLineNumber);
      }
      if (error == 0) {
	whereStr = "static field declaration";
	if (writeStatics) 
	  error = (*sfield_iter)->Unparse(*fOutput, context);
      }
      ++sfield_iter;
    } else if (fn_line 
	       && (!field_line || fn_line <= field_line)
	       && (!sfield_line || fn_line <= sfield_line)) {
      ////
      // The function comes first.
      ////
      if (!stinit) {
	whereStr = "static initializer";
	error = UnparseStaticBefore(intermediate, staticFn,
				    (*function)->fStartLineNumber);
      }
      // Is the override valid?
      if (!error) {
	if ((*function)->fSignature.GetName() == kConstructorName) {
	  whereStr = "a constructor";
	} else {
	  whereStr = 
	    ::UnicodeToString((*function)->fSignature.Disassemble());
	}	
	if (!((*function)->fSignature.GetName() == kConstructorName)) {
	  if (parentClass != 0) {
	    error = CheckValidOverride(parentType, *(*function), file);
	  }
	  if (intermediate.fRealClass->HasPInfo() &&
	      intermediate.fRealClass->GetPInfo().fInterfaces) {
	    for (deque<CJavaTypeSignature>::const_iterator interfaces =
		   intermediate.fRealClass->GetPInfo().InterfacesBegin();
		 error == 0 && interfaces != 
		   intermediate.fRealClass->GetPInfo().InterfacesEnd();
		 ++interfaces) {
	      if (parentClass != 0) {
		error = CheckValidOverride(*interfaces, *(*function), file);
	      }
	    }
	  } else {
	    for (CJavaClassFile::InterfaceList::const_iterator interfaces =
		   intermediate.fRealClass->GetInterfaceBegin();
		 error == 0 &&
		   interfaces != intermediate.fRealClass->GetInterfaceEnd();
		 ++interfaces) {
	      parentClass = LookupClass(*interfaces, file);
	      if (parentClass != 0) {
		error = CheckValidOverride(parentClass->GetType(), 
					   *(*function), file);
	      }
	    }
	  }
	}
	// Okay, the override's valid.
	if (error == 0) {
	  error = GenerateMethod(intermediate, *(*function));	
	}
      }
      ++function;
    } else {
      assert(0);
    }
    ////
    // Now, write out any errors we may have found.
    ////
    if (error != 0) {
      string message("In ");
      message += whereStr;
      message += " in class " + intermediate.Disassemble() +
	": " + ::UnicodeToString(error->GetMessage());
      if (error->GetLine() == CCompileError::kNoLineNumber) {
	PrintCompileError(intermediate.fSourceFileName, message);
      } else {
	PrintCompileError(intermediate.fSourceFileName, message,
			    error->GetLine());
      }
      result = false;
      delete error;
    }
  }
  ////
  // Generate $stinit, $cast, $instanceof, and $W
  ////
  if (fParameterized && fFilterMode) {
    UnparseTC1(*intermediate.fRealClass,
	       intermediate.fRealClass->GetType(),
	       file);
    if (result && ! intermediate.fAccessFlags.fInterface) {
      if (intermediate.fPClassInfo && intermediate.fPClassInfo->fInterfaces) {
	// UnparsePMatch for interfaces.
	for (deque<CJavaTypeSignature>::const_iterator inter = 
	       intermediate.fPClassInfo->InterfacesBegin();
	     inter != intermediate.fPClassInfo->InterfacesEnd(); ++inter) {
	  // Don't provide one the parent did.
	  if (! intermediate.fPClassInfo->GetSuper() ||
	      ! DescendsFrom(*intermediate.fPClassInfo->GetSuper(),
			     *inter, file)) {			     
	    UnparseTC1(*intermediate.fRealClass, *inter, file);
	  }
	}
      }
      // $W field
      if (intermediate.fPClassInfo && 
	  intermediate.fPClassInfo->IsParameterized()) {
	fOutput->out() << "private ";
	intermediate.fRealClass->GetType().UnparseWType(*fOutput);
	fOutput->out() << " $W;";
	UnparseSFMethods(*intermediate.fRealClass);
      }
    }
  }
  if (fFilterMode) {
    fOutput->PopIndent(2).Next().ToLine(intermediate.fEndLineNumber);
    fOutput->out() << "}";
  }
  return result;
}

//
//  Method name : UnparseImports
//  Description : Unparse the class's import statements. Returns true iff 
//                successful.
//
bool
CCompiler::UnparseImports(CSourceFile *file)
{
  // Class imports
  deque<unicode_string>::iterator c_iter = file->fClassImports.begin();
  deque<unicode_string>::iterator c_end  = file->fClassImports.end();
  int c_idx = 0;

  // Package imports
  deque<unicode_string>::iterator p_iter = file->fPackageImports.begin();
  deque<unicode_string>::iterator p_end = file->fPackageImports.end();
  int p_idx = 0;
  
  while((c_iter != c_end) || (p_iter != p_end)) {
    if ((c_iter != c_end) && 
	((p_iter == p_end) ||
	 (file->fClassImportLocations[c_idx] 
	  <= file->fPackageImportLocations[p_idx]))) {
      fOutput->ToLine(file->fClassImportLocations[c_idx]);
      fOutput->out() << "import "
		     << ::SymbolToString(*c_iter)
		     << ';';
      c_iter++; c_idx++;
    } else {
      if ((*p_iter).length() != 0) {
	fOutput->ToLine(file->fPackageImportLocations[p_idx]);
	fOutput->out() << "import "
		       << ::SymbolToString(*p_iter)
		       << ".*;";
      }
      p_iter++; p_idx++;
    }
  }
  return true;    
}

//
//  Method name : UnparseClassHeader
//  Description : Unparse the class's header. Returns true iff successful.
//
bool
CCompiler::UnparseClassHeader(CIntermediateClass &intermediate)
{
  assert(fFilterMode);
  CSourceFile *file = intermediate.GetSource(); 
  
  // Put the class declaration on a line before the bracket, if possible.
  fOutput->ToLineF(intermediate.fStartLineNumber-1);

  ////
  // Write out the header.
  ////
  fOutput->out() << intermediate.fAccessFlags.Unparse()
		 << (intermediate.fAccessFlags.fInterface ? 
		     "interface " : "class ")
                 << ::SymbolToString(intermediate.fRealClass->GetShortName());
  ////
  // Write out supertype information.
  ////
  if (intermediate.fExtends) {
    fOutput->out() << " extends ";
    if (intermediate.fPClassInfo) {
      assert(intermediate.fPClassInfo->fSuper);
      CJavaTypeSignature t = *intermediate.fPClassInfo->fSuper;
      t.Unparse(*fOutput);
    } else {
      fOutput->out() << ::SymbolToString(*intermediate.fExtends);
    }
  }
  if (! intermediate.fInterfaces.empty()) {
    fOutput->out() << (intermediate.fAccessFlags.fInterface ?
		       " extends" : " implements");
    bool firstTime = true;
    if (intermediate.fPClassInfo) {
      assert(intermediate.fPClassInfo->fInterfaces);
      deque<CJavaTypeSignature>::const_iterator iter = 
	intermediate.fPClassInfo->InterfacesBegin();
      for (; iter != intermediate.fPClassInfo->InterfacesEnd(); ++iter) {
	CJavaTypeSignature t = *iter;
	CheckType(t, *intermediate.fRealClass, file);
	fOutput->out() << (firstTime ? " " : ", ");
	t.Unparse(*fOutput); 
	firstTime = false;
      }
    } else {
      deque<unicode_string>::iterator iter = intermediate.fInterfaces.begin();
      for (; iter != intermediate.fInterfaces.end(); ++iter) {
	fOutput->out() << (firstTime ? " " : ", ")
		       << ::SymbolToString(*iter);
	firstTime = false;
      }
    }
  }
  fOutput->ToLine(intermediate.fStartLineNumber);
  fOutput->out() << '{';
  fOutput->PushIndent(2);

  return true;
}

//
//  Method name : RollType
//  Description : Sets setType a type signature for the "simplest" type that
//      can be used to refer to oldType.  This should be called on
//      all types that will be used with static methods or fields.
//      Returns true if oldType is "this".
//  Requires    : CheckType and FixType have been called on oldType
//  Checks      : oldType is a class type.
//
bool
CCompiler::RollType(CJavaTypeSignature& setType,
		    const CJavaTypeSignature& oldType,
		    const CCompileContext& context) const
{
  assert(oldType.GetBaseType() == CJavaTypeSignature::Class);

  setType = oldType;
  unicode_string classname;
  oldType.GetBaseClassName(classname);
  unicode_string shortname = classname;
  unicode_string::size_type lastSlash = classname.find_last_of('/');
  if (lastSlash != unicode_string::npos) {
    shortname.assign(classname, lastSlash + 1, unicode_string::npos);
  }  
  unicode_string package(classname, 0, lastSlash);

  // Simplest case: will "this" work?
  if (classname == context.GetClass().GetClassName()) {
    deque<CJavaTypeSignature>* params = 0;
    if (oldType.GetParameters())
      params = new deque<CJavaTypeSignature>(*oldType.GetParameters());
    setType = CJavaTypeSignature(shortname,
				 oldType.GetArrayBounds(),
				 params, false);
    return true;
  }

  // Is oldType in this package?
  if (package == context.GetClass().GetPackageName()) {
    deque<CJavaTypeSignature>* params = 0;
    if (oldType.GetParameters())
      params = new deque<CJavaTypeSignature>(*oldType.GetParameters());
    setType = CJavaTypeSignature(shortname,
				 oldType.GetArrayBounds(),
				 params, false);
    return false;
  }

  // Base case: no rolling needed.
  return false;
}

//
//  Method name : ShouldRegenerateType
//  Description : Returns false iff there is a class file for the 
//     instantiation of this type, and that class file is at least as new
//     as every class on which it depends.
//
bool
CCompiler::ShouldRegenerateType(const CJavaTypeSignature &t)
{
  unicode_string typeName;
  unicode_string IClassName;  
  unicode_string package = InstantiationPackage(t);
  if (package.length() > 0) { 
    IClassName = package;
    IClassName += (unicode_char) '/';
  }

  bool res = t.GetBaseClassName(typeName);
  assert(res);
  IClassName += ::SlashToDollar(typeName);
  IClassName += (unicode_char) '$';
  IClassName += (unicode_char) '$'; 
  IClassName += t.EncodeParams();
  
  const CJavaClassFile *IClass = LookupClass(IClassName, 0);
  if (!IClass) return true;
  
  time_t generatedAt = IClass->fDate;

  const CJavaClassFile *Class = ClassOfType(t, 0);
  if (!Class) return true;
  time_t mostRecentChange = (time_t) Class->fDate;
  
  for (deque<CJavaTypeSignature>::const_iterator 
	 p = t.GetParameters()->begin();
       p != t.GetParameters()->end(); ++p) {
    if ((*p).GetBaseType() != CJavaTypeSignature::Class)
      continue;
    Class = ClassOfType(*p, 0);
    if (!Class) return true;
    if (mostRecentChange < Class->fDate)
      mostRecentChange = Class->fDate;
  }
  return (mostRecentChange > generatedAt);
}

//
//  Method name: GeneratedFile
//  Description: Registers the fact that we have generated the file
//     'fileName' for the type 'forType'.
//
void 
CCompiler::GeneratedFile(const CJavaTypeSignature &forType, 
			 const string &fileName)
{ 
  fJavaFiles += fileName + "\n";
}
