// Copyright (c) 1995  David Engberg  All rights reserved
// $Id: Compiler.h,v 1.46 1998/09/07 21:26:14 andru Exp $
#ifndef _Compiler_h
#define _Compiler_h
#pragma interface

#include "JavaFieldSignature.h"
#include "JavaMethodSignature.h"
#include "unicode_string.h"
#include "parser_decls.h"
#include "JavaDirectory.h"
#include "PJavaConstraint.h"
#include "InstantiationList.h"
#include "Statement.h"
#include "SourceFile.h"
#include <string>
#include <deque>
#include <vector>
#include <map>
#include <list>
#include <set>
class CJavaMethodInfo;
class CJavaClassFile;
class CJavaAccessFlags;
class CIntermediateClass;
class CIntermediateFunction;
class CCompoundStatement;
class CDeclarationStatement;
class CCompileError;
class CJavaFieldInfo;
class CStatement;
class CPJavaConstraint;
class srcstream;
class CFilePath;

typedef set<unicode_string, less <unicode_string> > string_set;
typedef deque<CJavaTypeSignature> typeDeque;

//
//  Class name : CCompiler
//  Description : This class is used to encapsulate the operations used in
//    compiling a Java source file.  This manages all of the table-driven
//    parser code produced by flex and bison.
//    To use this interface, call the public static member function
//    'CompileValue' which takes in a string and parses it into a list of
//    CJavaClassFiles, if possible.
//
class CCompiler : public CTypeResolver {
public:
  typedef vector<CJavaClassFile*> ClassList;

  static bool TranslateFiles(const deque<string>& fileNames,
			     const deque<string>& classPath,
			     ClassList &resultClasses,
			     const CFilePath& javaDest,
			     const CFilePath& instDest,
			     const string& instPackage,
			     string& insInfoStr,
			     bool alwaysGenerate,
			     bool printDependencies);


  const CJavaClassFile* ImportClass(const unicode_string& className,
				    CSourceFile *file = 0,
				    bool aliasBase = true);
  bool ImportPackage(const unicode_string& packageName,
		     CSourceFile *file = 0);
  void SetPackage(const unicode_string& packageName);
  void NoPackage();

  void AddClassImport(const unicode_string& className);
  void AddPackageImport(const unicode_string& package);
  void StartClass(const unicode_string& className,
		  CJavaAccessFlags* adoptModifiers = 0,
		  CJavaTypeSignature* adoptExtends = 0,
		  deque<CJavaTypeSignature>* adoptInterfaces = 0,
		  deque<unicode_string>* adoptParameters = 0,
		  deque<CPJavaConstraint*>* adoptConstraints = 0,
		  bool deprecated = false);
  void StartInterface(const unicode_string& className,
		      CJavaAccessFlags* adoptModifiers = 0,
		      deque<CJavaTypeSignature>* adoptInterfaceExtends = 0,
		      deque<unicode_string>* adoptParameters = 0,
		      deque<CPJavaConstraint*>* adoptConstraints = 0,
		      bool deprecated = false);
  void EndClass();
  void EndBadClass();
  void StartFunction(const CJavaTypeSignature& returnType,
		     const unicode_string& functionName,
		     const CJavaAccessFlags& modifiers,
		     deque<CJavaFieldSignature>* adoptParameters = 0,
		     deque<CJavaTypeSignature>* adoptThrows = 0,
		     bool deprecated = false);
  void StartConstructor(const CJavaAccessFlags& modifiers,
			deque<CJavaFieldSignature>* adoptParameters = 0,
			deque<CJavaTypeSignature>* adoptThrows = 0,
			bool deprecated = false);
  void EndFunction(CCompoundStatement* adoptStatementBlock = 0);
  void EndBadFunction(CCompoundStatement* adoptStatementBlock = 0);
  bool InStatementBlock() const { return fVariableScopes.size() > 0; }
  void PushLocalScope();
  void PopLocalScope();
  unsigned short AddLocalVariable(const CJavaFieldSignature& signature);
  typedef unsigned short LocalVariableIndex;
  const pair<CJavaFieldSignature, LocalVariableIndex>*
    LookupLocalVariable(const unicode_string& name) const;
  unsigned short AddUnnamedVariable(unsigned long width = 1);
  void AddField(CDeclarationStatement* adoptDeclaration);
  void AddStaticCode(CCompoundStatement* adoptStatement);
  void ParseError(unsigned long lineNumber, const string& errorMessage,
		  const string& input);
  void ParseWarning(unsigned long lineNumber, const string& errorMessage,
		    const string& input);

  const CIntermediateClass* GetCurrentClass() const { return fCurrentClass; }

  const CJavaClassFile* LookupClass(const unicode_string& name,
				    CSourceFile *file) const;
  bool ValidParent(const CJavaTypeSignature& childType,
		   const CJavaTypeSignature& parentType,
		   CSourceFile *file) const;
  bool ValidInterface(const CJavaTypeSignature& childType,
		      const CJavaTypeSignature& interfaceType,
		      CSourceFile *file) const;
  bool DescendsFrom(const CJavaTypeSignature& childType,
		    const CJavaTypeSignature& ancestorType,
		    CSourceFile *file,
		    bool postInstantiate = false) const;
  bool AssignableSubtype(const CJavaTypeSignature& childType,
			 const CJavaTypeSignature& ancestorType,
			 CSourceFile *file,
			 bool postInstantiate = false) const;
  bool CastableType(const CJavaTypeSignature& currentType,
		    const CJavaTypeSignature& toType,
		    CSourceFile *file) const;
  bool SameType(const CJavaTypeSignature& first,
		const CJavaTypeSignature& second,
		CSourceFile *file,
		bool postInstantiate = false) const; 
  bool SameType(const CJavaMethodSignature& first,
		const CJavaMethodSignature& second,
		CSourceFile *file) const;
  bool ImplicitCastTo(const CJavaTypeSignature& from,
		      const CJavaTypeSignature& to,
		      CSourceFile *file) const;
  bool SameField(const CJavaFieldSignature& first,
		 const CJavaFieldSignature& second,
		 CSourceFile *file) const;
  CCompileError* CheckType(const CJavaTypeSignature& type,
			   const CJavaClassFile& inClass,
			   CSourceFile *file) const;
  CCompileError* CheckMethodSignature(CJavaMethodSignature& method,
				      const CJavaClassFile& inClass,
				      CSourceFile *file) const;
  virtual unicode_string InstantiationPackage(const CJavaTypeSignature &t) 
    const;
  CJavaTypeSignature FixType(const CJavaTypeSignature& oldType,
			     CSourceFile *file) const;
  unicode_string NameClassConstant(const CJavaTypeSignature& type,
				   CSourceFile* file) const;
  const CJavaClassFile* FindField(const unicode_string& field,
				  const CJavaTypeSignature& fromClassType,
				  CJavaFieldInfo& setFieldInfo,
				  CJavaTypeSignature& setClassType,
				  CSourceFile *file) const;
  virtual bool IsVarProtected(const CJavaTypeSignature &onType,
			      const unicode_string &varName,
			      bool &setProt) const;
  CCompileError* CheckValidOverride(const CJavaTypeSignature& parentType,
				    const CIntermediateFunction& method,
				    CSourceFile *file);
  CCompileError* CheckExceptionCovar(const typeDeque &child,
				     const typeDeque &parent,
				     CSourceFile *file) const;
  CCompileError* FindConstantField(CExpression*& result,
      const unicode_string& className, const CJavaFieldSignature& field,
      const CCompileContext& context, unsigned long lineNumber) const;
  CCompileError* FindIntermediateConstantField(CExpression*& result,
				      const CIntermediateClass& intermediate,
				      const CJavaFieldInfo& fieldInfo) const;
  static CExpression* CreateLiteralFromConstant(unsigned short index,
		const CJavaClassFile& onClass, const CJavaTypeSignature& type);
  static unsigned short CreateConstantFromLiteral(const CExpression* literal,
						  CJavaClassFile& onClass);
  pair<const CJavaClassFile*, const CJavaMethodInfo*>*
    MatchMethod(const unicode_string& methodName,
  		const CJavaTypeSignature& fromType,
		const deque<CJavaTypeSignature>& arguments,
		string*& errorString,
		CJavaTypeSignature& setType,
		CSourceFile *file) const;
  const CJavaMethodInfo*
    MatchClassMethod(const CJavaTypeSignature& onType,
		     const unicode_string& name,
		     const deque<CJavaTypeSignature>& arguments,
		     string*& errorString,
		     CSourceFile *file) const;
  pair<const CJavaClassFile*, const CJavaMethodInfo*>*
    ExactMatchMethod(const CJavaMethodSignature& method,
		     const CJavaTypeSignature& fromType,
		     CSourceFile *file,
		     CJavaTypeSignature /*OUT*/ &returnType);
  const CJavaMethodInfo*
    ExactMatchClassMethod(const CJavaMethodSignature& method,
			  const CJavaTypeSignature& onTypem,
			  CSourceFile *file,
			  CJavaTypeSignature /*OUT*/ &returnType) const;
  virtual bool GetDeclaredMethodType(const CJavaTypeSignature &onType,
				     const CJavaMethodSignature &findMethod,
				     CJavaMethodSignature &setMethod,
				     CJavaMethodSignature &transMethod) const;
  bool MoreSpecific(const CJavaMethodSignature& method1,
		    const CJavaMethodSignature& method2,
		    CSourceFile *file) const;

  typedef list<CJavaMethodSignature> MethodList;
  unsigned long UnimplementedMethods(const CJavaTypeSignature& onType,
				     MethodList& methods,
				     CSourceFile *file);

  bool IsThrowable(const CJavaTypeSignature& type,
		   CSourceFile *file) const;
  bool InClassInitializers() const { return fInClassInitializers; }

  typedef vector<unsigned long> FinallyHandlerStack;
  void PushFinallyHandler(unsigned long instruction);
  unsigned long PopFinallyHandler();
  FinallyHandlerStack::iterator GetHandlerBegin();
  FinallyHandlerStack::iterator GetHandlerEnd();
  
  void PushStatementContext(CStatement* statement);
  void PopStatementContext();
  StatementList::iterator GetContextBegin();
  StatementList::iterator GetContextEnd();

  void WarnDeprecatedField(const CJavaFieldInfo& field,
			   const CJavaClassFile& fieldClass,
			   const CCompileContext& context,
			   unsigned long line = 0) const;
  void WarnDeprecatedMethod(const CJavaMethodInfo& method,
			   const CJavaClassFile& methodClass,
			   const CCompileContext& context,
			   unsigned long line = 0) const;
  void WarnDeprecatedClass(const CJavaClassFile& usedClass) const;
  void WarnUnenforcedCast(const CJavaTypeSignature& type,
			  const CCompileContext& context,
			  unsigned long line = 0) const;

  static const unsigned long kVersionID;
  static const unicode_string kConstructorName;
  static const unicode_string kStaticName;
  static const unicode_string kObjectName;
  static const unicode_string kThrowableName;
  static const unicode_string kErrorName;
  static const unicode_string kRuntimeExceptionName;
  static const string kDefaultImport;

  // Public methods for PolyJ
  bool IsParameterized() const { return fParameterized; }
  void AddParameter(const unicode_string *adoptParam);
  bool IsAParameter(const unicode_string *str);
  CCompileError* CheckInstantiation(const CJavaTypeSignature& type,
				    const CJavaClassFile& inClass,
				    bool& setPackage,
				    CSourceFile *file) const;
  string* TypeSatisfiesConstraint(const CJavaTypeSignature& type,
				  const CPJavaConstraint& constraint,
				  const CJavaClassFile& inClass,
				  bool& setPackage,
				  CSourceFile *file,
				  bool verifyVisibility = true) const;
  const CPJavaConstraint::ConstraintMethod*
    MatchConstraintMethod(const CJavaMethodSignature& method,
			  const CPJavaConstraint& onConstraint) const;
  bool MethodVisibleFrom(const CJavaAccessFlags& methodFlags,
			 const CJavaClassFile& inClass,
			 const CJavaClassFile& fromClass,
			 bool& setPackage) const;

  const CJavaClassFile *ClassOfType(const CJavaTypeSignature& type,
				    CSourceFile *file) const;
  void FixMethodSignature(const CJavaMethodSignature& m,
			  CJavaMethodSignature& set,
			  CSourceFile *file) const;
  bool RollType(CJavaTypeSignature &setType,
		const CJavaTypeSignature& oldType,
		const CCompileContext& context) const;
  CJavaFieldSignature FixFieldSignature(const CJavaFieldSignature& f,
					CSourceFile *file) const;

  CInstantiationList& GetIList()
    { return fInstantiationList; }
  void PrintCompileError(const string& fileName, const string& message,
			 unsigned long line = 0) const;
  void PrintCompileError(const string& fileName, const unicode_string& message,
			 unsigned long line = 0) const;

protected:
  CCompiler(bool parameterized);
  CCompiler(const deque<string>& classPath, bool parameterized);
  virtual ~CCompiler();

  void Reset();
  void ReduceImportAliases();

  bool CheckJavaLangObject();
  bool CheckSupportClasses();
  CSourceFile* ParsedFile(const string& fileName) const;
  bool ParseFile(const string& fileName);
  bool CheckImports(CSourceFile *file);

  bool PartialCompileFile(const string& fileName,
			  CSourceFile*& file);
  bool ResolveFile(CSourceFile* file);
  bool SavedPartialCompileFile(const string& fileName,
			  IntermediateList& resultClasses);

  CJavaClassFile* ImportOneClassPath(const unicode_string& className,
				     const CJavaDirectory& classPath,
				     bool source);

  void PrepareIntermediateClass(CIntermediateClass& intermediate,
				const unicode_string& fileName);
  bool PrepareClassDeclarations(CIntermediateClass& intermediate);
  bool GenerateClass(CIntermediateClass& intermediate);
  bool UnparseImports(CSourceFile *source);
  bool UnparseClassHeader(CIntermediateClass& intermediate);
  CCompileError* UnparseMethodHeader(CIntermediateFunction& intermediate,
				     CIntermediateClass& Class,
				     bool implemented);
  CCompileError* UnparseStaticBefore(CIntermediateClass& Class, 
				     CIntermediateFunction* function,
				     unsigned long line);
  bool GenerateFieldConstants(CIntermediateClass& intermediate);
  CCompileError* UnparseMethod(CIntermediateClass& intermediateClass,
			       CIntermediateFunction& intermediateMethod,
			       CJavaMethodInfo &realMethod);
  CCompileError* GenerateMethod(CIntermediateClass& intermediateClass,
				CIntermediateFunction& intermediateMethod);
  CCompileError* InsertSuperCall(CIntermediateClass& intermediateClass,
				 CIntermediateFunction& intermediateMethod);

  const CIntermediateClass*
                LookupIntermediateClass(const unicode_string& name,
					const CSourceFile *file) const;
  const CJavaClassFile*
    FindImportedClass(const string &relPath, const unicode_string& name) const;

  // Protected methods for PolyJ
  void ClearParameterNames();
  bool PreparePJInfo(CJavaClassFile &Class,
		     CSourceFile *file);
  bool FixAMM(CJavaClassFile &Class, CSourceFile *file);
  void AdoptPJMethods(CJavaClassFile &Class, 
		      const CJavaTypeSignature &parentType,
		      CSourceFile *file);
  CCompileError* RenamePJMethods(CJavaClassFile &Class);
  bool FindConflictingMethods(CJavaClassFile &Class, CSourceFile *file);
  CCompileError* UnparseStinit(CJavaClassFile& Class, 
			       CIntermediateFunction* function,
			       StatementList &sDeclarations,
			       CSourceFile *file);
  void UnparseTC1(CJavaClassFile& inClass,
		  const CJavaTypeSignature &forType,
		  CSourceFile *file);

  void UnparseSFMethods(CJavaClassFile& Class);
  void GenerateWhereClass(CJavaClassFile& Class, ostream& os);
  bool GenerateInstantiation(const CJavaTypeSignature& IType,
			     const string& basePath);
  bool GenerateTClasses(const string &basePath);
  bool GenerateTClass(const unicode_string &package,
		      const string &basePath);
  void ReconcileOverride(const CJavaTypeSignature& classType,
			 const CJavaMethodSignature& method,
			 CJavaMethodSignature& setMethod);
  void InitializePrimitiveClasses();
  CJavaTypeSignature TypeForConstraint(const CJavaTypeSignature &t) const;

  void GeneratedFile(const CJavaTypeSignature &forType, 
		     const string &fileName);
  void HaveJavaFile(const string &fileName)
    { GeneratedFile(CJavaTypeSignature::kVoid, fileName); }

  bool ShouldRegenerateType(const CJavaTypeSignature &t);

  SourceFileList fParsedFiles;
  ClassPathList fClassPath;
  ClassList fImportedClasses;

  // Map from fully qualified name -> class.
  ImportTable fImportAliases;

  IntermediateList fIntermediateClasses;
  CIntermediateClass* fCurrentClass;
  CIntermediateFunction* fCurrentFunction;
  bool fParseError;
  unsigned short fCurrentLocalVariable;
  deque<deque<unicode_string> > fVariableScopes;
  typedef map<unicode_string, pair<CJavaFieldSignature, LocalVariableIndex>,
             less<unicode_string> > LocalVariableTable;
  LocalVariableTable fLocalVariables;
  FinallyHandlerStack fFinallyHandlers;
  StatementList fStatementContexts;
  bool fInClassInitializers;
  deque<string> fDependencies;

  bool fFilterMode;
  // Added for filtering
  srcstream *fOutput;
  StatementList::iterator fStaticBlockIter;
  // Added for PolyJ
  bool fParameterized;
  string_set fParameterNames;
  CInstantiationList fInstantiationList;
  ClassList fWhereClasses;

  string_set fTClassPackages;
  unicode_string fInstantiationPackage;

  CSourceFile* fParsingSource;
  deque<CSourceFile*> fParsedPJFiles;

  string fJavaFiles;
  bool fPrintDependencies;
};

#endif
