// #pragma implementation
#include "unicode_string.h"
#include "find_leaks.h"
#include "Srcstream.h"
#include "Compiler.h"
#include "parser_decls.h"
#include "FilePath.h"
#include "JavaClassFile.h"
#include <iostream.h>
#include <fstream.h>
#include "IntermediateClass.h"
#include "IntermediateFunction.h"
#include "JavaTypeSignature.h"
#include "parser_decls.h"
#include "Statement.h"
#include "JavaAttribute.h"
#include "JavaExceptionsTable.h"
#include "JavaCodeAttribute.h"
#include "CompileError.h"
#include "Expression.h"
#include "CompileContext.h"
#include "Instantiator.h"
#include <cstdio>
#include <climits>
#include <algorithm>
#include <set>
extern "C" {
  #include <unistd.h>
}
#include "mapped_file.h"

const unicode_string kInitializedName = ::UTFToUnicode("$initialized");

//
//  Method name : UnparseMethodHeader
//  Description : This method unparses a header for a method.
//    Returns true on success.  If "implemented", unparses through
//    the first brace.
//
CCompileError*
CCompiler::UnparseMethodHeader(CIntermediateFunction& intermediate, 
			       CIntermediateClass& Class,
			       bool implemented)
{
  CSourceFile *file = Class.GetSource();
  fOutput->ToLine(intermediate.fStartLineNumber);
  
  CJavaMethodSignature method = intermediate.fSignature;
  CJavaAccessFlags fl = intermediate.fAccessFlags;
  bool isConstructor = method.GetName() == kConstructorName;
  bool isStatic = intermediate.fAccessFlags.fStatic;
  int methodidx = 0;

  ////
  // Find the method index.
  ////
  if (Class.fRealClass->HasPInfo()) {
    assert(Class.fRealClass->GetPInfo().fMethods);
    FixMethodSignature(method,method,file);        
    MethodNameMap::iterator found = 
      Class.fRealClass->GetPInfo().fMethods->find(method);   
    if (found != Class.fRealClass->GetPInfo().fMethods->end()) {
      methodidx = (*found).second.fIndex;
      method = (*found).second.fTranslatedSignature;
    } 
  }
  
  ////
  //  Write the method name and protections.
  ////
  fOutput->out() << fl.Unparse();
  if (isConstructor) {
    fOutput->out() << ::SymbolToString(Class.fRealClass->GetShortName());
  } else {
    CJavaTypeSignature t = method.GetType();
    t.Unparse(*fOutput); 
    fOutput->out() << " "
		   << ::SymbolToString(method.GetName());
    if (methodidx > 0)
      fOutput->out() << "$" << methodidx;
  }

  ////
  // Unparse the arguments.
  ////
  fOutput->out() << "(";
  bool useW = Class.fRealClass->IsParameterized() &&
              (isStatic || isConstructor);
  if (Class.fRealClass->IsParameterized() && useW) {
    if (method.GetName() == ::StringToUnicode("main")) {
      PrintCompileError(Class.fSourceFileName,
       "Main method in parameterized class can't be called from command line.",
			intermediate.fStartLineNumber);
    }
    Class.fRealClass->GetType().UnparseWType(*fOutput);    
    fOutput->out() << " $W";
    if (intermediate.fParameters.size() > 0)
      fOutput->out() << ", ";
  } 

  deque<CJavaTypeSignature>::const_iterator typeiter = 
    method.ParametersBegin();
  if (useW) 
    ++typeiter;
  CCompileError *err = 0;
  for (deque<CJavaFieldSignature>::iterator i = 
	 intermediate.fParameters.begin();
       err == 0 && i != intermediate.fParameters.end(); ++typeiter) {
    assert(typeiter != method.ParametersEnd());
    CJavaTypeSignature t = FixType(*typeiter, file);
    bool unwrap = !t.IsPrimitive() && (*i).GetType().IsPrimitive();
    bool cast = (t.IsObject() &&
		 !(*i).GetType().IsParameter() &&
		 !(*i).GetType().IsObject());
    if (unwrap || cast) {
      unicode_string name = ::UTFToUnicode("$_") + (*i).GetFieldName();
      CJavaFieldSignature f(t, name);
      f.Unparse(*fOutput);
    } else {
      CJavaFieldSignature f(t, (*i).GetFieldName());
      f.Unparse(*fOutput); 
    }
    i++;
    if (i != intermediate.fParameters.end()) {
      fOutput->out() << ", ";
    }
  }
  if (isConstructor && methodidx >0) {
    if (intermediate.fParameters.size() > 0)
      fOutput->out() << ", ";
    for (int i = 0; i < methodidx; ++i) {
      fOutput->out() << "int $_dummy" << i;
      if (i < methodidx-1) fOutput->out() << ", ";
    }
  } 
  fOutput->out() << ")";

  if (! intermediate.fThrows.empty()) 
    fOutput->out() << " throws ";
  for (deque<unicode_string>::iterator i= intermediate.fThrows.begin();
       i != intermediate.fThrows.end(); ) {
    fOutput->out() << ::SymbolToString(*i);
    i++;
    if (i != intermediate.fThrows.end())
      fOutput->out() << ", ";
  }
  if (implemented) {
    fOutput->Next().PushIndent(2).out() << "{";
    deque<CJavaTypeSignature>::const_iterator typeiter = 
      method.ParametersBegin();
    if (useW) 
      ++typeiter;
    for (deque<CJavaFieldSignature>::iterator i = 
	   intermediate.fParameters.begin();
	 i != intermediate.fParameters.end(); ++typeiter, ++i) {
      bool unwrap = !(*typeiter).IsPrimitive() && (*i).GetType().IsPrimitive();
      bool cast = ((*typeiter).IsObject() &&
		   !(*i).GetType().IsParameter() &&
		   !(*i).GetType().IsObject());
      if (unwrap || cast) {
	CJavaTypeSignature t = FixType((*i).GetType(), file);
	CJavaFieldSignature f(t, (*i).GetFieldName());
	f.Unparse(*fOutput);
	fOutput->out() << " = ((";
	if (unwrap)
	  fOutput->out() << ::UnicodeToUTF((*i).GetType().GetWrapperName());
	else 
	  (*i).GetType().Unparse(*fOutput);
	fOutput->out() << ")$_" 
		       << ::UnicodeToUTF((*i).GetFieldName())
		       << ")";
	if (unwrap)
	  fOutput->out() << ".value;";
	else
	  fOutput->out() << ";";
      }  
    }
  }

  return err;
}

//
//  Method name : GenerateFieldConstants
//  Description : This method is used to figure out if any of the fields on
//    the provided intermediate class have a constant value that should be
//    pre-calculated in their class file.  This method iterates through all
//    of the field declarations and tries to find final fields for constant
//    insertion.
//    This method returns true if there were no errors encountered during
//    constant generation.
//
bool
CCompiler::GenerateFieldConstants(CIntermediateClass& intermediateClass)
{
  bool success = true;
  assert(intermediateClass.fRealClass != 0);
  fInClassInitializers = true;
  FieldTable::iterator endFields = intermediateClass.fRealClass->fFields.end();
  FieldTable::iterator field = intermediateClass.fRealClass->fFields.begin();
  for (; field != endFields; ++field) {
    const CJavaFieldInfo* fieldInfo = (*field).second;
    assert(fieldInfo != 0);
    if (fieldInfo->GetModifiers().fFinal != 0) {
      CExpression* result = 0;
      CCompileError* error =
	FindIntermediateConstantField(result, intermediateClass, *fieldInfo);
      if (error != 0) {
	PrintCompileError(intermediateClass.fSourceFileName,
			  "Invalid field initialization: " +
			  ::UnicodeToString(error->GetMessage()),
			  error->GetLine());
	delete error;
	success = false;
      } else if (result != 0) {
	assert(result->IsLiteral());
	JavaConstantIndex constant =
	  CreateConstantFromLiteral(result, *intermediateClass.fRealClass);
	CJavaFieldInfo* newFieldInfo = new CJavaFieldInfo(*fieldInfo);
	newFieldInfo->SetConstantIndex(constant);
	delete fieldInfo;
	(*field).second = newFieldInfo;
	delete result;
      }
    }
  }
  fInClassInitializers = false;
  return success;
}

//
//  Method name : GenerateMethod
//  Description : This method is used to generate the final method
//    representation for a java method.  This generates all of the
//    needed information for a CJavaMethodInfo and inserts it into the provided
//    class.  If this operation is unsuccessful for any reason, an error
//    is returned, otherwise, 0 is returned.
//
CCompileError*
CCompiler::GenerateMethod(CIntermediateClass& intermediateClass,
			  CIntermediateFunction& intermediateMethod)
{
  CCompileError* error = 0;
  CSourceFile *file = intermediateClass.GetSource();
  MethodTable::iterator methodPosition =
    intermediateClass.fRealClass->fMethods.begin();
  for (; !(methodPosition == intermediateClass.fRealClass->fMethods.end());
       ++methodPosition) {
    if (SameType(intermediateMethod.fSignature, (*methodPosition).first, 
		 file)) {
      break;
    }
  }
  if (methodPosition == intermediateClass.fRealClass->fMethods.end()) {    
    string errStr = "Internal error: Lost track of the method for ";
    errStr += ::UnicodeToUTF(intermediateMethod.fSignature.Disassemble());
    return new CCompileError(errStr,
			     intermediateMethod.fStartLineNumber);    
  }
  CJavaMethodInfo* method = (CJavaMethodInfo*)((*methodPosition).second);
  ////
  // Check valid throws.
  ////
  if (error == 0) {
    for (deque<unicode_string>::iterator i= intermediateMethod.fThrows.begin();
	 i != intermediateMethod.fThrows.end(); ++i) {
      const CJavaClassFile* thrown = 
	LookupClass(*i, intermediateClass.GetSource());
      if (thrown == 0) {
	method = 0;
	string errorString = "Non-existent type ";
	errorString += ::UnicodeToString(*i);
	errorString += " thrown by ";
	unicode_string disassembled =
	  intermediateMethod.fSignature.Disassemble();
	errorString += ::UnicodeToString(disassembled);
	error = new CCompileError(errorString,
				  intermediateMethod.fStartLineNumber);
	break;
      } else {
	if (! error) {
	  if (! IsThrowable(thrown->GetType(), file)) {
	    method = 0;
	    string errorString = "Non-Throwable type ";
	    errorString += ::UnicodeToString(*i);
	    errorString += " thrown by ";
	    errorString +=
	      ::UnicodeToString(intermediateMethod.fSignature.Disassemble());
	    error = new CCompileError(errorString,
				      intermediateMethod.fStartLineNumber);
	  } else if (!(thrown->GetClassName() == *i)) {
	    *i = thrown->GetClassName();
	  }
	}
      }
    }
  }
  ////
  // Check all types.
  ////
  if (error == 0) {
    error = CheckType(method->fSignature.GetType(),
		      *intermediateClass.fRealClass,
		      intermediateClass.GetSource());
    deque<CJavaTypeSignature>::const_iterator arg =
      method->fSignature.ParametersBegin();
    deque<CJavaTypeSignature>::const_iterator end =
      method->fSignature.ParametersEnd();
    for (; error == 0 && arg != end; ++arg) {
      error = CheckType(*arg,*intermediateClass.fRealClass,
			intermediateClass.GetSource());
      if (error) error->SetLine(intermediateMethod.fStartLineNumber);
    }
  }
  if (error == 0 && fFilterMode) {
    bool staticMethod = intermediateMethod.fSignature.GetName() == kStaticName;
    error = UnparseMethod(intermediateClass, intermediateMethod, *method);
    if (error == 0 &&
	intermediateClass.fAccessFlags.fInterface != 0) {
      if (intermediateMethod.fAccessFlags.fNative != 0 ||
	  intermediateMethod.fAccessFlags.fSynchronized != 0 ||
	  intermediateMethod.fAccessFlags.fFinal != 0 ||
	  intermediateMethod.fAccessFlags.fPrivate != 0 ||
	  intermediateMethod.fAccessFlags.fProtected != 0 ||
	  (intermediateMethod.fAccessFlags.fStatic != 0 && !staticMethod)) {
	error = new CCompileError("Interface methods can't be native, static, "
				  "synchronized, final, private, or protected",
				  intermediateMethod.fStartLineNumber);
      } else if (intermediateMethod.fBlock != 0 &&
		 !(method->fSignature.GetName() == kStaticName)) {
	error = new CCompileError("Invalid code in interface",
				  intermediateMethod.fStartLineNumber);
      }
    }	       
  } 
  return error;
}

//
// Method name : InsertSuperCall
// Description : Add a 'super' call (possibly implicit) to intermediateMethod
//     in intermediateClass.  If this isn't possible (because the supertype
//     constructor isn't accessible) returns an error message.  Returns 0
//     on success, or if an explicit constructor call is already found.
// Checks : This method is nonempty.
//
CCompileError* 
CCompiler::InsertSuperCall(CIntermediateClass& intermediateClass,
			   CIntermediateFunction& intermediateMethod)
{
  assert(intermediateMethod.fBlock);
  assert(intermediateMethod.fBlock->fChildren);
  // Check for a constructor call.
  if (intermediateMethod.fBlock->fChildren->size() != 0) {
    CStatement *s =
      intermediateMethod.fBlock->fChildren->front();
    if (DYNAMIC_CAST(CExplicitConstructorCall,s))
      return 0;
  }
  // No constructor call found; add one.
  CExplicitConstructorCall *constructor
    = new CExplicitConstructorCall(CExplicitConstructorCall::kSuper,0);
  constructor->fLineNumber = intermediateMethod.fBlock->fLineNumber;
  intermediateMethod.fBlock->fChildren->push_front(constructor);
  return 0;
}

//
//  Method name : UnparseStaticBefore
//  Description : Unparses those lines of a "block" which start on or before 
//    line, and have not yet been unparsed.  Keeps track of which are which
//    by keeping the iterator fStaticBlockIter pointing to the next un-unparsed
//    line.

CCompileError*
CCompiler::UnparseStaticBefore(CIntermediateClass& Class,
			       CIntermediateFunction* function,
			       unsigned long line)
{

  if (function == 0 || function->fBlock->fChildren == 0 || 
      function->fBlock->fChildren->empty() ||
      fStaticBlockIter == function->fBlock->fChildren->end())
    return 0;

  if ((*fStaticBlockIter)->GetLineNumber() > line)
    return 0;

  CCompileError* error = 0;

  MethodTable::iterator methodPosition = Class.fRealClass->fMethods.begin();

  for (; !(methodPosition == Class.fRealClass->fMethods.end());
       ++methodPosition) {
    if (SameType(function->fSignature, (*methodPosition).first, 
		 Class.GetSource())) {
      break;
    }
  }
  assert(methodPosition != Class.fRealClass->fMethods.end());
  CJavaMethodInfo* method = (CJavaMethodInfo*)((*methodPosition).second);

  CCompileContext context(this,Class.fRealClass,method,
			  Class.fSourceFile,0,true);

  while(fStaticBlockIter != function->fBlock->fChildren->end() && 
	((*fStaticBlockIter)->GetLineNumber() <= line)) {
    fOutput->ToLineBefore((*fStaticBlockIter)->GetLineNumber(),1);
    fOutput->out() << "static";
    error = (*fStaticBlockIter)->Unparse(*fOutput,context);
    ++fStaticBlockIter;
  }
  
  return error;
}

//
//  Method name : UnparseMethod
//  Description : This method writes the implementation for a method to
//     the output stream.
//
CCompileError*
CCompiler::UnparseMethod(CIntermediateClass& intermediate,
			CIntermediateFunction& intermediateMethod,
			CJavaMethodInfo& method)
{
  CCompileError* error = 0;
  CCompoundStatement* body = intermediateMethod.fBlock;
  bool staticMethod = intermediateMethod.fSignature.GetName() == kStaticName;
  bool constructor = 
    intermediateMethod.fSignature.GetName() == kConstructorName;
  CJavaAccessFlags flags = method.fAccessFlags;
  if (flags.fAbstract) {
    if (body != 0) 
      error = new CCompileError("Invalid method body on abstract method",
				intermediateMethod.fStartLineNumber);
    else if (flags.fStatic)
      error = new CCompileError("Attempt to declare static method as abstract",
				intermediateMethod.fStartLineNumber);
  } else if (flags.fNative) {
    if (body != 0)
      error = new CCompileError("Invalid method body on native method",
				intermediateMethod.fStartLineNumber);
  } else {
    if (body == 0)
      error = new CCompileError("Invalid empty method body",
				intermediateMethod.fStartLineNumber);
  } 
  if (error) 
    return error;

  if (constructor)
    error = InsertSuperCall(intermediate, intermediateMethod);

  if (error) return error;
  if (!staticMethod) {
    error = UnparseMethodHeader(intermediateMethod, intermediate,
				intermediateMethod.fBlock != 0);
  }         
  if (intermediateMethod.fBlock == 0) {
    fOutput->out() << ';';
  } else {
    CCompileContext context(this,intermediate.fRealClass,
			    &method,
			    intermediate.fSourceFile, 0,
			    intermediateMethod.fAccessFlags.fStatic!=0);
    for (deque<unicode_string>::const_iterator i =
	   intermediateMethod.fThrows.begin();
	 !(i == intermediateMethod.fThrows.end()); ++i) {
      context.PushThrowable(CJavaTypeSignature(*i));
    }
    deque<CJavaTypeSignature>::const_iterator arg = 
      method.GetSignature().ParametersBegin();
    deque<CJavaTypeSignature>::const_iterator argsEnd = 
      method.GetSignature().ParametersEnd();
    unsigned long variableIndex = 0;
    if (method.GetModifiers().fStatic == 0) {
      context.InitializeVariable(variableIndex++);
    }
    for (; arg != argsEnd; ++arg) {
      context.InitializeVariable(variableIndex);
      variableIndex += (*arg).GetWidth();
    }
    if (!staticMethod) {
      if (constructor && intermediate.fRealClass->IsParameterized()) {
	// Constructors need "this.$W = $W;".
	CStatement* first = intermediateMethod.fBlock->fChildren->front();
	CExplicitConstructorCall* ccFirst = 
	  first ? DYNAMIC_CAST(CExplicitConstructorCall,first) : 0;
	bool isThis = ccFirst && 
	  ccFirst->GetType() == CExplicitConstructorCall::kThis;
	if (ccFirst) {
	  intermediateMethod.fBlock->fChildren->pop_front();
	  error = first->Unparse(*fOutput,context);
	  if (!isThis)
	    fOutput->out() << "  this.$W = $W;";
	}
	// Constructors must then have all of the variable initializers.
	if (!isThis) {
	  for(StatementList::iterator i = 
		intermediate.fNonStaticDeclarations.begin();
	      !error && i != intermediate.fNonStaticDeclarations.end(); ++i) {
	    CDeclarationStatement *decl = 
	      DYNAMIC_CAST(CDeclarationStatement,*i);
	    assert(decl);
	    error = decl->UnparseInitializers(*fOutput,context);
	  }
	}
      }
      if (!error) 
	error = intermediateMethod.fBlock->UnparseNoBrackets(*fOutput,
							     context);
      fOutput->PopIndent(2).Next().out() << "}";
    }  
    if (error == 0) {
      bool needsReturn = method.fSignature.GetType() 
	!= CJavaTypeSignature::kVoid;
      if (needsReturn && context.IsReachable()) {
	string err = "Return statement needed at the end of ";
	err += ::UnicodeToString(method.GetSignature().Disassemble());
	error = new CCompileError(err, intermediateMethod.fEndLineNumber);
      }
    }  
  }
  return error;
}
  

//
//  Method name : LookupIntermediateClass
//  Description : This checks the list of classes to see if any of them have 
//     the given class name.  If any do, a pointer to that class is returned,
//     otherwise 0 is returned.
//   * Note: The 'file' argument is used to determine which package context
//     to search in.
//
const CIntermediateClass*
CCompiler::LookupIntermediateClass(const unicode_string& name,
				   const CSourceFile* file) const
{
  bool baseName = name.find((unicode_char)'/') == unicode_string::npos;
  unicode_string fullname;
  if (baseName && file && file->fPackageName.length() > 0) {
    fullname = file->fPackageName;
    fullname += (unicode_char) '/';
    fullname += name;
  } else {
    fullname = name;
  }
#if 0
  cerr << "LookupIntermediateClass " << name;
  if (file) {
    cerr << " in " << file->fFilePath;
  }
  cerr << endl;
#endif
  IntermediateList::const_iterator i = fIntermediateClasses.begin();
  for (; !(i == fIntermediateClasses.end()); ++i) {
    if ((*i)->fName == fullname) {
      return *i;
    }
  } 

#if 0
  // else if (baseName) {
  unicode_string const &ICName = (*i)->fName;
  unicode_string::size_type lastSlash = ICName.find_last_of('/');
  unicode_string afterSlash(ICName, lastSlash + 1);
  if (lastSlash != unicode_string::npos && afterSlash == name) {
    return *i;
  }
#endif
  return 0;
}

//
//  Method name : FindImportedClass
//  Description : This checks the list of classes in the same file to see if
//     any of them have the given class name.  If any do, a pointer to that
//     class is returned, otherwise 0 is returned.
//
const CJavaClassFile*
CCompiler::FindImportedClass(const string &relPath, 
			     const unicode_string& name) const
{  
  unicode_string className = ::UTFToUnicode(relPath);
  unicode_string::size_type slash = className.find_last_of('/');
  if (className.size() > 0 && slash != className.size()-1) {
    className += (unicode_char) '/';
  }
  className += name;
  for (ClassList::const_iterator i = fImportedClasses.begin();
       i != fImportedClasses.end(); ++i) {   
    if ((*i)->GetClassName() == className) {
      return *i;
    }
  }
  return 0;
}

//
//  Method name : LookupClass
//  Description : This can be used to see if a given class name is visible
//    inside of this file.  This includes globally-qualified class names that
//    include their package and simple names that are visible thanks to some
//    previous import statement or visible because they are in this file.
//    If one is found, that entry is returned, otherwise 0 is returned.
//    NOTE:  Since this can potentially return a half-built entry from this
//      same file, you should use this class for its interface only ... the
//      method internals and constants may not yet be valid.
//
const CJavaClassFile*
CCompiler::LookupClass(const unicode_string& name, CSourceFile* file) const
{
#if 0
  cerr << "Lookup " << name;
  if (file) {
    cerr << " in " << file->fFilePath;
  }
  cerr << endl;
#endif

  if (name.length() == 0) {
    return 0;
  }
  ////
  // See if we already have an entry for this.
  ////
  ImportTable::const_iterator entry = fImportAliases.find(name);
  if (!(entry == fImportAliases.end())) {
    return (*entry).second;
  }
  if (file) {
    entry = file->fImportAliases.find(name);
    if (!(entry == file->fImportAliases.end())) {
      return (*entry).second;
    }
  }
  ////
  // Then search the files we've already parsed.
  ////
  const CIntermediateClass* intermediate = LookupIntermediateClass(name,
								   file);
  if (intermediate != 0 && intermediate->fRealClass != 0) {
    return intermediate->fRealClass;
  }
  // Casting away 'const' to allow caching of the result.
  return ((CCompiler*)this)->ImportClass(name, file, false);
}

//
//  Method name : ValidParent
//  Description : Returns true if this 'parent' type can legally be
//    extended by 'child.'
//
bool
CCompiler::ValidParent(const CJavaTypeSignature& childType,
		       const CJavaTypeSignature& parentType,
		       CSourceFile *file) const 
{
  const CJavaClassFile* child = ClassOfType(childType, file);
  const CJavaClassFile* parent = ClassOfType(parentType, file);
  if (!child || !parent)
    return false;

  CJavaAccessFlags childAccess = child->GetAccessFlags();
  CJavaAccessFlags parentAccess = parent->GetAccessFlags();
  bool result = (parentAccess.fPublic != 0) ||
    (parent->GetPackageName() == child->GetPackageName());
  result &= parentAccess.fFinal == 0;
  result &= parentAccess.fInterface == 0;
  result &= childAccess.fInterface == 0 ||
    parent->GetClassName() == kObjectName;
  return result;
}

//
//  Method name : ValidInterface
//  Description : Returns true if the 'interface' object can legally be
//    extended by 'child.'
//
bool
CCompiler::ValidInterface(const CJavaTypeSignature& childType,
			  const CJavaTypeSignature& interfaceType,
			  CSourceFile *file) const
{
  const CJavaClassFile* child = ClassOfType(childType, file);
  const CJavaClassFile* interface = ClassOfType(interfaceType, file);
  if (!child || !interface)
    return false;

  CJavaAccessFlags interfaceFlags = interface->GetAccessFlags();
  bool result = interfaceFlags.fInterface != 0;
  result &= (interfaceFlags.fPublic != 0) ||
    (interface->GetPackageName() == child->GetPackageName());
  return result;
}

//
//  Method name : DescendsFrom
//  Description : Returns true if the provided child descends from the
//     ancestor object, as an extension class or interface.
//     Executes a depth-first search up the inheritance tree, classes
//     first, then interfaces.
//   * If postInstantiate, returns true if, for some instantiation of
//     child and ancestor, the above holds true.
//
bool
CCompiler::DescendsFrom(const CJavaTypeSignature& childType,
			const CJavaTypeSignature& ancestorType,
			CSourceFile *file,
			bool postInstantiate) const
{  
  // XXXX doesn't currently deal with circularity ... assumes well-formed tree.
  if (postInstantiate && (childType.IsParameter() || 
			  ancestorType.IsParameter()))
    return true;

  const CJavaClassFile *ancestor = ClassOfType(ancestorType, file);
  const CJavaClassFile *child = ClassOfType(childType, file);
  if (ancestor && ancestor->GetClassName() == kObjectName) {
    if (childType.IsParameter())
      return true;
  }
  if (!child || !ancestor) 
    return false;
  
  const CJavaClassFile* parent;
  const CJavaTypeSignature* parentType;
  bool result = false;
  if (child->fAccessFlags.fInterface == 0) {
    if (child->fPClassInfo != 0) {
      parentType = child->fPClassInfo->fSuper;
      if (parentType) {
	CInstantiator ins(*child, childType.GetParameters());
	CJavaTypeSignature trueParent;
	ins.Instantiate(trueParent,*parentType);
	result = SameType(ancestorType, trueParent, file, postInstantiate) ||
	  DescendsFrom(trueParent, ancestorType, file, postInstantiate);
      }
    } else {
      parent = LookupClass(child->fSuperclassName, file);
      if (parent) {
	CJavaTypeSignature parentType = parent->GetType();
	result = ancestorType == parentType || 
	  DescendsFrom(parentType, ancestorType, file, postInstantiate);
      }
    }
  } else {
    result = ancestor->GetClassName() == kObjectName;
  }
  if (!result) {
    if (child->fPClassInfo) {
      if (child->fPClassInfo->fInterfaces) {
	CInstantiator ins(*child, childType.GetParameters());
	CJavaTypeSignature trueParent;
	for (deque<CJavaTypeSignature>::const_iterator i = 
	       child->fPClassInfo->InterfacesBegin();
	     !result && i != child->fPClassInfo->InterfacesEnd(); ++i) {
	  ins.Instantiate(trueParent,*i);
	  result = SameType(ancestorType,trueParent, file, postInstantiate) ||
	    DescendsFrom(trueParent, ancestorType, file, postInstantiate);
        } 
      }
    } else {
      for (deque<unicode_string>::const_iterator i = 
	     child->fInterfaces.begin();
	   !result && i != child->fInterfaces.end(); ++i) {
	parent = LookupClass(*i, file);
	result = ancestor == parent ||
	  (parent != 0 && DescendsFrom(parent->GetType(), 
				       ancestorType, file, postInstantiate));
      }
    }
  }
  return result;
}

//
//  Method name : AssignableSubtype
//  Description : This function returns true if the childType can legally
//    be assigned to a variable of type ancestorType, but is childType !=
//    ancestorType.  Only works with classes and arrays ... not primitive
//    types.
//   * If postInstantiate, returns true if, for some instantiation of
//     child and ancestor, the above holds true.
//
bool
CCompiler::AssignableSubtype(const CJavaTypeSignature& childType,
			     const CJavaTypeSignature& ancestorType,
			     CSourceFile *file,
			     bool  postInstantiate) const
{
  bool result = false;
  if (childType.IsArray()) {
    if (ancestorType.GetArrayBounds() == childType.GetArrayBounds()) {
      CJavaTypeSignature baseChildType(childType);
      baseChildType.SetArrayBounds(0);
      CJavaTypeSignature baseAncestorType(ancestorType);
      baseAncestorType.SetArrayBounds(0);
      result = AssignableSubtype(baseChildType, baseAncestorType, file,
				 postInstantiate);
    } else {
      static const CJavaTypeSignature kObjectType(kObjectName);
      result = SameType(kObjectType, ancestorType, file, postInstantiate);
    }
  } else if (postInstantiate && (childType.IsParameter() ||
				 ancestorType.IsParameter())) {
    return true;
  } else {
    result = (childType == CJavaTypeSignature::kNullType) &&
      (!(ancestorType.IsParameter() && ancestorType.GetArrayBounds() == 0));
    if (!result) {
      result = DescendsFrom(childType, ancestorType, file, postInstantiate);
    }
  }
  return result;
}

//
//  Method name : CastableType
//  Description : This function returns true if the currentType can legally
//    be casted to an instance of type toType.
//    Only works with classes and arrays ... not primitive types.
//
bool
CCompiler::CastableType(const CJavaTypeSignature& currentType,
			const CJavaTypeSignature& toType,
			CSourceFile *file) const
{
  unicode_string currentClassName;
  bool result = currentType.GetBaseClassName(currentClassName);
  if (result) {
    unicode_string toClassName;
    if (toType.GetBaseClassName(toClassName)) {
      const CJavaClassFile* toClassFile = LookupClass(toClassName, file);
      const CJavaClassFile* currentClassFile = 
	LookupClass(currentClassName, file);
      result = (currentType == CJavaTypeSignature::kNullType ||
	 (toClassFile && toClassFile->GetAccessFlags().fInterface != 0 ) ||
	 (currentClassFile && 
                     currentClassFile->GetAccessFlags().fInterface != 0 ) ||
		AssignableSubtype(currentType, toType, file, true) ||
		AssignableSubtype(toType, currentType, file, true) ||
		SameType(toType, currentType, file, true));
    } else {
      result = toType.IsArray() && currentClassName == kObjectName;
    }
  }
  return result;
}

//
//  Method name : FindField
//  Description : This method is used to try to find a field with a given name
//    by starting with 'fromClass' and working up the inheritance tree until
//    it runs out of parents (and returns 0) or find a class that implements
//    the desired field.  If this is found, the third parameter is set to
//    the type signature of the desired field.
//
const CJavaClassFile*
CCompiler::FindField(const unicode_string& field,
		     const CJavaTypeSignature& fromClassType,
		     CJavaFieldInfo& setFieldInfo,
		     CJavaTypeSignature& setClassType,
		     CSourceFile *file) const
{
  const CJavaClassFile* fromClass = ClassOfType(fromClassType, file);
  if (!fromClass)
    return 0;

  const CJavaClassFile* found = 0;
  const CJavaFieldInfo* fieldInfo = fromClass->LookupField(field);
  if (fieldInfo != 0) {
    setFieldInfo = *fieldInfo;
    found = fromClass;
    setClassType = fromClassType;
  } else {
    const CJavaClassFile* parent = 0;
    CInstantiator ins(*fromClass, fromClassType.GetParameters());
    if (fromClass->fAccessFlags.fInterface == 0) {
      if (fromClass->fPClassInfo) {
	if (fromClass->fPClassInfo->fSuper) {
	  CJavaTypeSignature trueParent;
	  ins.Instantiate(trueParent, * fromClass->fPClassInfo->fSuper);
	  found = FindField(field, trueParent, setFieldInfo, setClassType,
			    file);
	}
      } else {
	parent = LookupClass(fromClass->fSuperclassName, file);
	if (parent != 0) {
	  found = FindField(field, parent->GetType(), 
			    setFieldInfo, setClassType, file);
	}
      }
    }
    if (fromClass->fPClassInfo) {
      if (fromClass->fPClassInfo->fInterfaces) {
	deque<CJavaTypeSignature>::const_iterator i = 
	  fromClass->fPClassInfo->InterfacesBegin();
	CJavaTypeSignature trueParent;
	for(; found == 0 && i != fromClass->fPClassInfo->InterfacesEnd();++i) {
	  ins.Instantiate(trueParent,*i);
	  found = FindField(field, trueParent, setFieldInfo, setClassType,
			    file);
	}
      }
    } else {
      deque<unicode_string>::const_iterator i = fromClass->fInterfaces.begin();
      for (; found == 0 && i != fromClass->fInterfaces.end(); ++i) {
	parent = LookupClass(*i, file);
	if (parent != 0) {
	  found = FindField(field, parent->GetType(), 
			    setFieldInfo, setClassType, file);
	}
      }
    }
  }
  return found;
}


//
//  Method name : IsVarProtected
//  Description : This method is used by AuxMethodSignature to determine
//    whether a given variable (probably static) is non-public.
//    Sets setProt to true iff onType.varName is non-public, and returns
//    true iff the variable is found.
//  * If the name is '$initialized', always returns true and sets 'setProt'
//    to true.
//
bool 
CCompiler::IsVarProtected(const CJavaTypeSignature &onType,
	       const unicode_string &varName,
	       bool &setProt) const
{
  if (varName == kInitializedName) {
    setProt = true;
    return true;
  }
  CJavaFieldInfo fieldInfo;
  CJavaTypeSignature foundOnType;
  const CJavaClassFile *c = FindField(varName, onType,
				      fieldInfo, foundOnType, 0);
  if (!c) { return false; }
  
  setProt = ! fieldInfo.GetModifiers().fPublic;
  return true;
}


//
//  Method name : MatchMethod
//  Description : This is used to search for a matching method with the given
//    name that can be called with arguments of the provided type.  The
//    search starts at 'fromClass' and works its way up the inheritance tree
//    until a match is found (in which case a pair is created and returned to
//    the user to indicate the matched class and method info) or no match can
//    be found. (in which case 0 is returned.)
//  * Sets 'setType' to the type on which this method was found.
//
pair<const CJavaClassFile*, const CJavaMethodInfo*>*
CCompiler::MatchMethod(const unicode_string& methodName,
		       const CJavaTypeSignature& fromType,
		       const deque<CJavaTypeSignature>& arguments,
		       string*& errorString,
		       CJavaTypeSignature& setType,
		       CSourceFile* file) const
{
  const CJavaClassFile* fromClass = ClassOfType(fromType, file);
  if (!fromClass) return 0;

  pair<const CJavaClassFile*, const CJavaMethodInfo*>* match = 0;
  const CJavaMethodInfo* methodInfo =
    MatchClassMethod(fromType, methodName, arguments, errorString, file);
  if (methodInfo != 0) {
    match = new pair<const CJavaClassFile*, const CJavaMethodInfo*>(fromClass,
								 methodInfo);
    setType = fromType;
  } else {
    if (fromClass->fPClassInfo) {
      CInstantiator ins(*fromClass,fromType.GetParameters());
      CJavaTypeSignature parentType;
      if (fromClass->fPClassInfo->fSuper) {
	ins.Instantiate(parentType,*fromClass->fPClassInfo->fSuper);
	match = MatchMethod(methodName, parentType, arguments, errorString, 
			    setType, file);
      }
      if (match == 0) {
	if (fromClass->fPClassInfo->fInterfaces) {
	  deque<CJavaTypeSignature>::const_iterator i =
	    fromClass->fPClassInfo->InterfacesBegin();
	  deque<CJavaTypeSignature>::const_iterator end =
	    fromClass->fPClassInfo->InterfacesEnd();
	  for(; match == 0 && i != end; ++i) {
	    ins.Instantiate(parentType,*i);
	    match = MatchMethod(methodName,parentType,arguments,errorString,
				setType, file);
	  }
	}
      }
    } else { 
      const CJavaClassFile *parent;
      parent = LookupClass(fromClass->fSuperclassName, file);
      if (parent != 0) {
	match = MatchMethod(methodName, parent->GetType(), 
			    arguments, errorString, setType, file);
      }
      if (match == 0) {
	deque<unicode_string>::const_iterator i = 
	  fromClass->fInterfaces.begin();
	for (; match == 0 && i != fromClass->fInterfaces.end(); ++i) {
	  parent = LookupClass(*i, file);
	  if (parent != 0) {
	    match = MatchMethod(methodName, parent->GetType(), 
				arguments, errorString, setType, file);
	  }
	}
      }
    }
  }

  return match;
}

//
//  Method name : MatchClassMethod
//  Description : If the provided class contains a method with the given name
//    and number of arguments, and each provided argument can be assigned to
//    its corresponding argument in the signature, then the matching method
//    info structure is returned.  Otherwise, 0 is returned.
//    Currently implemented as a linear search.
//    Sorry for the naming confusion with MatchMethod.  The difference is that
//    MatchMethod is called to try to search a class and all of its parents
//    whereas MatchClassMethod is used by MatchMethod to just search one
//    class.
//
const CJavaMethodInfo*
CCompiler::MatchClassMethod(const CJavaTypeSignature& onType,
			    const unicode_string& name,
			    const deque<CJavaTypeSignature>& arguments,
			    string*& errorString,
			    CSourceFile *file) const
{
  const CJavaClassFile* onClass = ClassOfType(onType, file);
  if (! onClass) return 0;
  // Might be the 'null instantiator'
  CInstantiator ins(*onClass, onType.GetParameters());
  const CJavaMethodInfo* match = 0;
  CJavaMethodSignature signature;
  for (MethodTable::const_iterator i = onClass->fMethods.begin();
       i != onClass->fMethods.end(); ++i) {
    ins.Instantiate(signature,(*i).first);
    // Grrr! Why must argument and parameter be synonymous?
    if (signature.GetName() == name &&
	arguments.size() == signature.ParameterCount()) {
      deque<CJavaTypeSignature>::const_iterator parameter =
	signature.ParametersBegin();
      deque<CJavaTypeSignature>::const_iterator parametersEnd =
	signature.ParametersEnd();
      deque<CJavaTypeSignature>::const_iterator argument = arguments.begin();
      bool valid = true;
      for (; valid && parameter != parametersEnd; ++parameter, ++argument) {
	if (SameType(*parameter, *argument, file)) {
	  valid = true;
	} else {
	  valid = ImplicitCastTo(*argument, *parameter, file);
	}
      }
      if (valid) {
	if (match == 0 || MoreSpecific((*i).first, (*match).GetSignature(),
				       file)) {
	  match = (*i).second;
	} else if (!MoreSpecific((*match).GetSignature(), (*i).first, file)) {
	  errorString = new string("Ambiguous method call");
	  match = 0;
	}
      } 
    }
  }
  return match;
}

//
//  Method name : ExactMatchMethod
//  Description : This is used to search for a matching method with the given
//    type signature. The search starts at 'fromClass' and works its way up
//    the inheritance tree until a match is found (in which case a pair
//    is created and returned to the caller to indicate the matched class and
//    method info) or no match can be found. (in which case 0 is returned.)
//    The type of the matching method (if any) is placed into "returnType",
//    rewritten appropriately for the chain of instantiations.
//
pair<const CJavaClassFile*, const CJavaMethodInfo*>*
CCompiler::ExactMatchMethod(const CJavaMethodSignature& method,
			    const CJavaTypeSignature& fromType,
			    CSourceFile *file,
			    CJavaTypeSignature& returnType)
{
  const CJavaClassFile* fromClass = ClassOfType(fromType, file);
  if (!fromClass) return 0;
  pair<const CJavaClassFile*, const CJavaMethodInfo*>* match = 0;
  const CJavaMethodInfo* methodInfo = ExactMatchClassMethod(method, fromType,
							    file, returnType);
  if (methodInfo != 0) {
    match = new pair<const CJavaClassFile*, const CJavaMethodInfo*>(
						    fromClass, methodInfo);
  } else {
    if (fromClass->fPClassInfo) {
      CJavaTypeSignature parentType;
      CInstantiator ins(*fromClass,fromType.GetParameters());
      if (fromClass->fPClassInfo->fSuper) {
	ins.Instantiate(parentType,*fromClass->fPClassInfo->fSuper);
	match = ExactMatchMethod(method, parentType, file, returnType);
      }
      if (match == 0 && fromClass->fPClassInfo->fInterfaces) {
	deque<CJavaTypeSignature>::const_iterator i =
	  fromClass->fPClassInfo->InterfacesBegin();
	deque<CJavaTypeSignature>::const_iterator end =
	  fromClass->fPClassInfo->InterfacesEnd();
	for(; match == 0 && i != end; ++i) {
	  ins.Instantiate(parentType,*i);
	  match = ExactMatchMethod(method, parentType, file, returnType);
	}
      }
    } else {
      const CJavaClassFile* parent = 
	LookupClass(fromClass->fSuperclassName, file);
      if (parent != 0) {
	match = ExactMatchMethod(method, parent->GetType(), file, returnType);
      }
      if (match == 0) {
	deque<unicode_string>::const_iterator i = 
	  fromClass->fInterfaces.begin();
	for (; match == 0 && i != fromClass->fInterfaces.end(); ++i) {
	  parent = LookupClass(*i, 0);
	  if (parent != 0) {
	    match = ExactMatchMethod(method, parent->GetType(), file, returnType);
	  }
	}
      }
    }
  }
  return match;
}

//
//  Method name : ExactMatchClassMethod
//  Description : If the provided class contains a method with the given
//    signature, but possibly a different return type, the matching method
//    info structure is returned, and the instantiated return type is placed
//    in returnType.  Otherwise, 0 is returned.
//    Currently implemented as a linear search.
//
const CJavaMethodInfo*
CCompiler::ExactMatchClassMethod(const CJavaMethodSignature& method,
				 const CJavaTypeSignature& onType,
				 CSourceFile *file,
				 CJavaTypeSignature /*OUT*/ &returnType) const
{
  const CJavaClassFile* onClass = ClassOfType(onType, file);
  if (!onClass) return 0;
  // Might be the 'null instantiator'
  CInstantiator ins(*onClass, onType.GetParameters()); 
  const CJavaMethodInfo* match = 0;
  
  for (MethodTable::const_iterator i = onClass->fMethods.begin();
       match == 0 && i != onClass->fMethods.end(); ++i) {
    if (ins.EqualsIgnoreReturn((*i).first,method)) {
      match = (*i).second;
      ins.Instantiate(returnType,(*i).first.GetType());
    }
  }
  return match;
}

//
//  Method name : GetDeclaredMethodType
//  Description : Finds the method matching findMethod on type 'onType',
//    using MatchMethod.  Sets setMethod to the signature of the method,
//    as first declared.  Sets setMethod to the translated signature of the
//    method.  Returns true iff the method was found.
//
bool
CCompiler::GetDeclaredMethodType(const CJavaTypeSignature &onType,
				 const CJavaMethodSignature &findMethod,
				 CJavaMethodSignature &setMethod,
				 CJavaMethodSignature &transMethod) const
{
  pair<const CJavaClassFile*, const CJavaMethodInfo*> *match;
  string *errStr = 0;
  CJavaTypeSignature setType;

  match = MatchMethod(findMethod.GetName(),
		      onType,
		      findMethod.GetArguments(),
		      errStr,
		      setType,
		      0);
  if (!match) {
    delete errStr;
    return false;
  }
  assert(match->second);
  setMethod = match->second->GetSignature();
  transMethod = match->second->GetTranslatedSignature();
  delete match;

  return true;
}

//
//  Method name : MoreSpecific
//  Description : Returns true if the first method signature is more
//    specific than the second.
//  Requires    : The signatures have the same number of arguments.
//
bool
CCompiler::MoreSpecific(const CJavaMethodSignature& first,
			const CJavaMethodSignature& second,
			CSourceFile *file) const
{
  bool specific = true;
  deque<CJavaTypeSignature>::const_iterator firstParameters =
    first.ParametersBegin();
  deque<CJavaTypeSignature>::const_iterator firstEnd =
    first.ParametersEnd();
  deque<CJavaTypeSignature>::const_iterator secondParameters =
    second.ParametersBegin();
  for (; specific && firstParameters != firstEnd;
       ++firstParameters, ++secondParameters) {
    if (!ImplicitCastTo(*firstParameters, *secondParameters, file)) {
      specific = false;
    }
  }
  return specific;
}

//
//  Method name : SameType
//  Description : Returns true if the provided types are identical in the eyes
//    of the compiler.  This should be used whenever there's a chance that
//    both types are class names.  The problem is, because of importing, 
//    'String' and 'java.lang.String' may both be valid names for the same
//    type, and only the compiler knows this.
//  * If postInstantiate, returns true if, for some instance of first and
//    second, the above may hold true.
//
bool
CCompiler::SameType(const CJavaTypeSignature& first,
		    const CJavaTypeSignature& second,
		    CSourceFile *file,
		    bool postInstantiate) const
{
  unicode_string firstName, secondName;
  if (first.IsParameter() || second.IsParameter()) {
    if (postInstantiate)
      return true;
    else
      return (first == second);
  }
  bool result = (first == second) ||
    (first.GetArrayBounds() == second.GetArrayBounds() &&
     first.GetBaseType() == second.GetBaseType() &&
     (!first.GetBaseClassName(firstName) ||
      !second.GetBaseClassName(secondName) ||
      LookupClass(firstName, file) == LookupClass(secondName, file)));
  if (result && first.GetParameters() != 0) {
    if (second.GetParameters() == 0 || 
	first.GetNParameters() != second.GetNParameters()) {
      result = false;
    } else {
      deque<CJavaTypeSignature>::const_iterator first_params =
	first.GetParameters()->begin();
      deque<CJavaTypeSignature>::const_iterator second_params =
	second.GetParameters()->begin();
      deque<CJavaTypeSignature>::const_iterator end =
	first.GetParameters()->end();
      for(; result && first_params != end; ++first_params, ++second_params) {
	result = SameType(*first_params, *second_params, file,postInstantiate);
      }
    }
  }
  return result;
}

//
//  Method name : ImplicitCastTo
//  Description : Returns true if implicit type conversion is allowed from
//    a value of type 'from' to a value of time 'to.'
//
bool
CCompiler::ImplicitCastTo(const CJavaTypeSignature& from,
			  const CJavaTypeSignature& to,
			  CSourceFile *file) const
{
  bool allowed = SameType(from, to, file) ||
    (from.IsReference() && to.IsReference() &&AssignableSubtype(from,to,file));
  if (!allowed) {
    if (from.IsNumeric() && to.IsNumeric()) {
      CJavaTypeSignature toType = to.GetBaseType();
      switch (from.GetBaseType()) {
      case CJavaTypeSignature::Byte:
	if (toType == CJavaTypeSignature::Short) {
	  return true;
	}
      case CJavaTypeSignature::Short:
      case CJavaTypeSignature::Character:
	if (toType == CJavaTypeSignature::Integer) {
	  return true;
	}
      case CJavaTypeSignature::Integer:
	if (toType == CJavaTypeSignature::LongInteger) {
	  return true;
	}
      case CJavaTypeSignature::LongInteger:
	if (toType == CJavaTypeSignature::Float) {
	  return true;
	}
      case CJavaTypeSignature::Float:
	if (toType == CJavaTypeSignature::Double) {
	  return true;
	}
      default:
	break;
      }
    }
  }
  return allowed;
}

//
//  Method name : SameType
//  Description : Returns true if the provided method signatures are identical
//    in the eyes of the compiler.  This is needed because two signatures
//    may be lexically different but semantically identical.  For example:
//         void foo(Object o)
//         void foo(java.lang.Object o)
//
bool
CCompiler::SameType(const CJavaMethodSignature& first,
		    const CJavaMethodSignature& second,
		    CSourceFile *file) const
{
  bool result = first.GetName() == second.GetName() &&
    SameType(first.GetType(), second.GetType(), file);
  if (result) {
    deque<CJavaTypeSignature>::const_iterator firstParameter =
      first.ParametersBegin();
    deque<CJavaTypeSignature>::const_iterator secondParameter =
      second.ParametersBegin();
    for (; result &&
	   firstParameter != first.ParametersEnd() &&
  	   secondParameter != second.ParametersEnd();
	 ++firstParameter, ++secondParameter) {
      result = SameType(*firstParameter, *secondParameter, file);
    }
    if (result) {
      result = firstParameter == first.ParametersEnd() &&
	secondParameter == second.ParametersEnd();
    }
  }
  return result;
}

//
//  Method name : NameClassConstant
//  Description : Java really, really screwed up the class constant format in
//    their bytecode.  If you are talking about a base (non-array) class type,
//    then you enter a string like "java/lang/String" into the constant pool,
//    but if you're talking about an array type (which is sorta a class type),
//    then you would put in a type signature like "[I" or "[LFoo;"
//    So, because this is annoying and inconsistent, I added a function here to
//    do the translation, and in the process it changes underqualified names
//    like 'String' to fully qualified ones like 'java/lang/String'.
//    It also assumes that any class names used in this type really do exist,
//    so you'd better check them using LookupClass before calling this.
//
unicode_string
CCompiler::NameClassConstant(const CJavaTypeSignature& type,
			     CSourceFile *file) const
{
  unicode_string result;
  assert(type.IsReference());
  if (! type.IsArray()) {
    const CJavaClassFile* classFile = ClassOfType(type, file);
    assert(classFile != 0);
    result = classFile->GetClassName();
  } else {
    result = ::UTFToUnicode(type.Compile());
  }
  return result;
}

const unicode_string kErrorStr = ::StringToUnicode("*ERROR*");
const unicode_string kBlankStr = ::StringToUnicode("");

//
//  Method name : CheckType
//  Description : This method takes the provided type signature and verifies
//    that it actually denotes a usable type, when considered from within
//    inClass.  If it does, 0 is returned, otherwise a compile error is 
//    returned that explains why the type is not usable.  This indicates 
//    that a class name is used that cannot be found, or that a class's
//    parameters are unsuitable.
//  * The fInsClass field is set in this type.
//
CCompileError*
CCompiler::CheckType(const CJavaTypeSignature& type,
		     const CJavaClassFile& inClass,
		     CSourceFile *file) const
{
  CCompileError* error = 0;
  bool setPackage = false;
  if (type.IsReference() && !type.IsParameter()) {
    unicode_string className;
    if (type.GetBaseClassName(className)) {
      if (LookupClass(className, file) == 0) {
	string errorString("Class for type ");
	errorString += ::UnicodeToUTF(type.Disassemble());
	errorString += " could not be found.";
	error = new CCompileError(errorString);
      }
      if (error == 0) {
	error = CheckInstantiation(type,inClass,setPackage,file);
      }
    }
  }
  return error;
}

//
//  Method name : CheckMethodSignature
//  Description : This method takes the provided method signature and verifies
//    that it actually denotes a usable type, when considered from within
//    inClass.  If it does, 0 is returned, otherwise a compile error is 
//    returned that explains why the type is not usable.  This indicates 
//    that a class name is used that cannot be found, or that a class's
//    parameters are unsuitable.
//
CCompileError* 
CCompiler::CheckMethodSignature(CJavaMethodSignature& method,
				const CJavaClassFile& inClass,
				CSourceFile *file) const
{
  CJavaTypeSignature ret = method.GetType();
  CCompileError* err = CheckType(ret,inClass,file);
  deque<CJavaTypeSignature> args;
  for(deque<CJavaTypeSignature>::const_iterator i = method.ParametersBegin();
      !err && i != method.ParametersEnd(); ++i) {
    CJavaTypeSignature type = *i;
    err = CheckType(type,inClass,file);
    args.push_back(type);
  }
  if (!err)
    method = CJavaMethodSignature(ret, method.GetName(), args);
  return err;
}

//
//  Method name : InstantiationPackage
//  Description : Returns the package in which to generate the WClasses
//      for a given type.
//  Requires    : This type has already passed a call to CheckType.
//
unicode_string 
CCompiler::InstantiationPackage(const CJavaTypeSignature &type) const
{
  bool needPackage = false;
  unicode_string package;
  assert(type.GetBaseType() == CJavaTypeSignature::Class);
  if (! type.IsParameterized()) 
    return fInstantiationPackage;

  const CJavaClassFile *Class = ClassOfType(type, 0);
  const CPJavaClassInfo &PCInfo = Class->GetPInfo();
  CInstantiator ins(*Class, type.GetParameters());
  deque<unicode_string>::const_iterator p =
    PCInfo.fParams->begin();
  deque<CJavaTypeSignature>::const_iterator t = type.GetParameters()->begin();
  for ( ; needPackage == false && t != type.GetParameters()->end(); 
	++p, ++t) {
    const CJavaClassFile *c = ClassOfType(*t, 0);
    if (c) {
      if (! c->GetAccessFlags().fPublic) {
	needPackage = true;
	package = c->GetPackageName();
      }
    }
    const CPJavaConstraint* constraint = PCInfo.LookupConstraint(*p);
    assert(constraint);
    CPJavaConstraint fillConstraint;
    ins.Instantiate(fillConstraint, *constraint);
    bool spackage = false;
    TypeSatisfiesConstraint(*t, fillConstraint, *Class, spackage, 0, false);
    if (spackage) {
      needPackage = true;
      package = c->GetPackageName();
    }
  }
  // cerr << "IP of " << type << " is " << package << endl;
  
  if (needPackage)
    return package;
  
  return fInstantiationPackage;
}

//
//  Method name : FixType
//  Description : If this type signature uses any class names that should be
//    fully qualified (i.e. 'String' instead of 'java/lang/String'), this
//    function will create a valid one and return it to the caller. Otherwise,
//    the original type will be returned.
//
CJavaTypeSignature
CCompiler::FixType(const CJavaTypeSignature& oldType,
		   CSourceFile *file) const
{
  unicode_string className;
  const CJavaClassFile* classFile;
  if (oldType.GetBaseType() == CJavaTypeSignature::Class &&
      oldType.GetBaseClassName(className) &&
      (classFile = LookupClass(className, file)) != 0) {
    if (oldType.GetNParameters() > 0) {
      deque<CJavaTypeSignature> fixedParams;
      deque<CJavaTypeSignature>::const_iterator i 
	= oldType.GetParameters()->begin();
      deque<CJavaTypeSignature>::const_iterator end
	= oldType.GetParameters()->end();
      for(; i != end; ++i) {
	fixedParams.push_back(FixType(*i, file));
      }
#if 0
      cerr << "Fixing " << oldType << " to "
	   << CJavaTypeSignature(classFile->GetClassName(),
				oldType.GetArrayBounds(),
				 &fixedParams) << endl;
#endif
      return CJavaTypeSignature(classFile->GetClassName(),
				oldType.GetArrayBounds(),
				&fixedParams);
    } else {
#if 0
      cerr << "Fixing " << oldType << " to "
	   << CJavaTypeSignature(classFile->GetClassName(),
				 oldType.GetArrayBounds()) << endl;
#endif
      return CJavaTypeSignature(classFile->GetClassName(),
				oldType.GetArrayBounds());
    }
  } else {
#if 0
    cerr << "Fixing " << oldType << " to (no change) "
	 << oldType << endl;
#endif
    return oldType;
  }
}


//
//  Method name : PushFinallyHandler
//  Description : This method is used to indicate the beginning of a scope
//    governed by a 'finally' clause which must be invoked regardless how
//    control flow leaves the body of a 'try' statement.
//    The instruction that is provided indicates the code instruction offset
//    from the beginning of the code for the current method.  This instruction
//    should be invoked as a subroutine to execute the appropriate 'finally'
//    code.
//
void
CCompiler::PushFinallyHandler(unsigned long instruction)
{
  fFinallyHandlers.push_back(instruction);
}

//
//  Method name : PopFinallyHandler
//  Description : This is the inverse of PushFinallyHandler.  It removes the
//    last-pushed finally instruction index from the list.
//
unsigned long
CCompiler::PopFinallyHandler()
{
  unsigned long back = fFinallyHandlers.back();
  fFinallyHandlers.pop_back();
  return back;
}

//
//  Method name : GetHandlerBegin
//  Description : This is used to get an iterator pointing to the first (most
//    recently pushed) handler instruction in the stack.  This iterator can
//    be used to cycle through all of the handlers.
//
CCompiler::FinallyHandlerStack::iterator
CCompiler::GetHandlerBegin()
{
  return fFinallyHandlers.begin();
}

//
//  Method name : GetHandlerEnd
//  Description : This method is used to get an iterator pointing past the
//    last (least recently pushed) handler instruction in the stack.
//
CCompiler::FinallyHandlerStack::iterator
CCompiler::GetHandlerEnd()
{
  return fFinallyHandlers.end();
}

//
//  Method name : PushStatementContext
//  Description : This method is used to declare that one statement is part
//    of the context of subsequent statements.  This is needed to implement
//    non-local 'break' and 'continue', which need to know where they are
//    located to properly implement their branches.
//
void
CCompiler::PushStatementContext(CStatement* statement)
{
  fStatementContexts.push_front(statement);
}

//
//  Method name : PopStatementContext
//  Description : This method is the converse of PushStatementContext ... it
//    removes the last-pushed statement from the context of future statements.
//
void
CCompiler::PopStatementContext()
{
  fStatementContexts.pop_front();
}

//
//  Method name : GetContextBegin
//  Description : This method allows some part of the compilation process
//    to get an iterator to the first statement in the enclosing context.
//
StatementList::iterator
CCompiler::GetContextBegin()
{
  return fStatementContexts.begin();
}

//
//  Method name : GetContextEnd
//  Description : This method allows some part of the compilation process
//    to get an iterator past the last statement in the enclosing context.
//
StatementList::iterator
CCompiler::GetContextEnd()
{
  return fStatementContexts.end();
}

//
//  Method name : IsThrowable
//  Description : Returns true if the provided type is a valid type for use
//    in a 'throw' statement.
//
bool
CCompiler::IsThrowable(const CJavaTypeSignature& type,
		       CSourceFile *file) const
{
  static const CJavaTypeSignature throwableType(kThrowableName);
  return type.IsReference() && ! type.IsParameterized() &&
    (SameType(type, throwableType, file) ||
     AssignableSubtype(type, throwableType, file));
}

//
//  Method name : ParseError
//  Description : This call tells the compiler that the parser hit some
//    error on the provided line number and input.
//
void
CCompiler::ParseError(unsigned long lineNumber, const string& errorMessage,
		      const string& input)
{
  PrintCompileError(fParsingSource->fFileName, errorMessage, lineNumber);
  cerr << "  on input: " << input << endl;
  fParseError = true;
}

//
//  Method name : ParseWarning
//  Description : This call tells the compiler that the parser hit some
//    error that should be reported as a non-fatal warning.
//
void
CCompiler::ParseWarning(unsigned long lineNumber, const string& errorMessage,
			const string& input)
{
  PrintCompileError(fParsingSource->fFileName, errorMessage, lineNumber);
}

//
//  Method name : FindConstantField
//  Description : This method asks the compiler to try to find a field on the
//    provided class name with the given field signature.  If it can find an
//    appropriate field with a constant value, a new literal expression is
//    created and returned in the 'result' parameter, otherwise 'result' will
//    be set to 0.  If any errors are found during compilation, they are
//    returned, otherwise the function returns 0.
//    This is a little bit of a weird function because of the potential
//    ordering difficulties in trying to evaluate final fields.  For
//    example, the following is legal, and all three fields should have an
//    associated constant value of 5:
//      class t1 { final static int i = t3.i; }
//      class t2 { final static int i = 5; }
//      class t3 { final static int i = t2.i; }
//    This poses problems, since there is no guarantee that a field on a class
//    in the same file as this one has been evaluated yet.  So... this needs
//    to both check for fully-evaluated classes and half-finished ones.  In
//    the process, it may have to finish type evaluation and constant folding
//    on the other class, which may produce the compile error I am returning.
//
CCompileError*
CCompiler::FindConstantField(CExpression*& result,
   const unicode_string& className, const CJavaFieldSignature& fieldSignature,
   const CCompileContext& context, unsigned long lineNumber) const
{
  CCompileError* error = 0;
  result = 0;
  const CJavaClassFile* classFile = 0;
  const CIntermediateClass* intermediate = LookupIntermediateClass(className,
								   0);
  if (intermediate != 0 && intermediate->fRealClass != 0) {
    classFile = intermediate->fRealClass;
    const CJavaFieldInfo* fieldInfo = classFile->LookupField(fieldSignature);
    if (fieldInfo != 0 && fieldInfo->GetModifiers().fFinal != 0) {
      error =
	FindIntermediateConstantField(result, *intermediate, *fieldInfo);
      classFile = 0;
    }
  } else {
    classFile = LookupClass(className, 0);
    if (classFile != 0) {
      const CJavaFieldInfo* fieldInfo = classFile->LookupField(fieldSignature);
      if (fieldInfo != 0 && fieldInfo->GetModifiers().fFinal != 0) {
	if (fieldInfo->IsConstant()) {
	  JavaConstantIndex index = fieldInfo->GetConstantIndex();
	  result = CreateLiteralFromConstant(index, *classFile,
					     fieldSignature.GetType());
	  WarnDeprecatedField(*fieldInfo, *classFile, context, lineNumber);
	}
	classFile = 0;
      }
    }
  }
  if (classFile != 0 && error == 0 && result == 0) {
    error = FindConstantField(result, classFile->fSuperclassName,
			      fieldSignature, context, lineNumber);
    deque<unicode_string>::const_iterator i = classFile->fInterfaces.begin();
    for (; error == 0 && result == 0 && i != classFile->fInterfaces.end();
	 ++i) {
      error = FindConstantField(result, *i, fieldSignature,
				context, lineNumber);
    }
  }
  return error;
}

//
//  Method name : FindIntermediateConstantField
//  Description : This method is used to check an intermediate class to see
//    if it provides a specified field with a constant value.
//
CCompileError*
CCompiler::FindIntermediateConstantField(CExpression*& result,
				       const CIntermediateClass& intermediate,
				       const CJavaFieldInfo& fieldInfo) const
{
  CCompileError* error = 0;
  const CJavaClassFile* classFile = intermediate.fRealClass;
  CSourceFile* sourceFile = intermediate.fSourceFile;
  const StatementList* declarations =
    fieldInfo.GetModifiers().fStatic != 0 ?
    &intermediate.fStaticDeclarations :
    &intermediate.fNonStaticDeclarations;
  bool found = false;
  for (StatementList::const_iterator i = declarations->begin();
       !found && !(i == declarations->end()); ++i) {
    CDeclarationStatement* declaration =
      DYNAMIC_CAST(CDeclarationStatement, *i);
    assert(declaration != 0);
    deque<CVariableDeclaration*>::iterator field =
      declaration->fDeclarations->begin();
    for (; !found && !(field == declaration->fDeclarations->end());
	 ++field) {
      if (SameField((*field)->GetSignature(), 
		    fieldInfo.GetSignature(), sourceFile)) {
	bool wasInClass = fInClassInitializers;
	CJavaMethodInfo dummyInfo;
	// casting away const
	CCompiler* nonConstThis = (CCompiler*)this;
	nonConstThis->fInClassInitializers = true;
	CCompileContext context(nonConstThis, (CJavaClassFile*)classFile,
				&dummyInfo, 
				intermediate.fSourceFile, 0, false);
	error = (*field)->GetConstantValue(result, context);
	found = true;
	nonConstThis->fInClassInitializers = wasInClass;
      }
    }
  }
  return error;
}

//
//  Method name : SameField
//  Description : This method returns true if the two field signatures match.
//
bool
CCompiler::SameField(const CJavaFieldSignature& first,
		     const CJavaFieldSignature& second, 
		     CSourceFile *file) const
{
  return first.GetFieldName() == second.GetFieldName() &&
    SameType(first.GetType(), second.GetType(), file);
}

//
//  Method name : CreateLiteralFromConstant
//  Description : This little static method is used to lookup an index in a
//    classes constant pool and then attempt to interpret that as a literal
//    expression value.  If this is possible, a new literal expression is
//    created and returned to the caller, otherwise 0 is returned.
//
CExpression*
CCompiler::CreateLiteralFromConstant(unsigned short index,
		const CJavaClassFile& onClass, const CJavaTypeSignature& type)
{
  CExpression* result = 0;
  const CJavaConstant* constant = onClass.LookupConstant(index);
  if (constant != 0) {
    const CJavaStringConstant* stringConstant =
      DYNAMIC_CAST_CONST(CJavaStringConstant, constant);
    if (stringConstant != 0) {
      constant = onClass.LookupConstant(stringConstant->GetStringIndex());
      const CJavaAscizConstant* ascizConstant =
	DYNAMIC_CAST_CONST(CJavaAscizConstant, constant);
      if (ascizConstant != 0) {
	result = new CStringLiteral(ascizConstant->GetUnicodeString());
      }
    } else {
      const CJavaIntegerConstant* integerConstant =
	DYNAMIC_CAST_CONST(CJavaIntegerConstant, constant);
      if (integerConstant != 0) {
	result = new COrdinalLiteral(integerConstant->GetInteger(), type);
      } else {
	const CJavaFloatConstant* floatConstant =
	  DYNAMIC_CAST_CONST(CJavaFloatConstant, constant);
	if (floatConstant != 0) {
	  result = new CFloatLiteral(floatConstant->GetFloat());
	} else {
	  const CJavaDoubleConstant* doubleConstant =
	    DYNAMIC_CAST_CONST(CJavaDoubleConstant, constant);
	  if (doubleConstant != 0) {
	    result = new CFloatLiteral(doubleConstant->GetDouble());
	  } else {
	    const CJavaLongConstant* longConstant =
	      DYNAMIC_CAST_CONST(CJavaLongConstant, constant);
	    if (longConstant != 0) {
	      result = new COrdinalLiteral(longConstant->GetLong());
	    }
	  }
	}
      }
    }
  }
  return result;
}

//
//  Method name : CreateConstantFromLiteral
//  Description : Given an expression that is assumed to be a literal value of
//    some type, this method creates an entry in the constant pool of the
//    provided class that can be used to mark a constant value. The index
//    into the constant pool is returned. 
//
unsigned short
CCompiler::CreateConstantFromLiteral(const CExpression* literal,
				     CJavaClassFile& onClass)
{
  unsigned short index;
  const COrdinalLiteral* intLiteral = 
    DYNAMIC_CAST_CONST(COrdinalLiteral, literal);
  if (intLiteral != 0) {
    index = intLiteral->AddToConstantPool(onClass);
  } else {
    const CStringLiteral* stringLiteral =
      DYNAMIC_CAST_CONST(CStringLiteral, literal);
    if (stringLiteral != 0) {
      index = stringLiteral->AddToConstantPool(onClass);
    } else {
      const CFloatLiteral* floatLiteral = 
	DYNAMIC_CAST_CONST(CFloatLiteral, literal);
      if (floatLiteral != 0) {
	index = floatLiteral->AddToConstantPool(onClass);
      } else {
	assert(0);
      }
    }
  }
  return index;
}

//
//  Method name : PrintCompileError
//  Description : Dumps out the provided message and line number in a standard
//    error format, to standard error.
//    If 'line' is 0, this assumes that no line number information is
//    available.
//
void
CCompiler::PrintCompileError(const string& fileName,
			     const unicode_string& message,
			     unsigned long line) const
{
  PrintCompileError(fileName, ::UnicodeToString(message), line);
}

//
//  Method name : PrintCompileError
//  Description : Dumps out the provided message and line number in a standard
//    error format, to standard error.
//    If 'line' is 0, this assumes that no line number information is
//    available.
//
void
CCompiler::PrintCompileError(const string& fileName, const string& message,
			     unsigned long line) const
{
  string::size_type lastSlash = fileName.find_last_of('/');
  string shortName(fileName, lastSlash + 1);
  cerr << shortName << ":";
  if (line != 0) {
    cerr << line << ": ";
  }
  cerr << message << endl;
}

//
//  Method name : CheckValidOverride
//  Description : Checks to see if a method with the provided signature can
//    be used in a subclass of 'parentClass.'  If this is not allowed for some
//    reason, an informative error is returned, otherwise this method returns 0
//
CCompileError*
CCompiler::CheckValidOverride(const CJavaTypeSignature& parentType,
			      const CIntermediateFunction& method,
			      CSourceFile *file)
{
  CCompileError* error = 0;
  CJavaTypeSignature returnType;
  pair<const CJavaClassFile*, const CJavaMethodInfo*>* match =
    ExactMatchMethod(method.fSignature, parentType, file, returnType);
  if (match != 0) {
    assert(match->first != 0 && match->second != 0);
    const CJavaMethodInfo* otherMethod = match->second;    
    CJavaAccessFlags parentModifiers = otherMethod->GetModifiers();
    CInstantiator ins(*ClassOfType(parentType, file), 
		      parentType.GetParameters());
    if (parentModifiers.fFinal != 0) {
      string message =
	"Invalid override of final method ";
      message += ::SymbolToString(method.fSignature.Disassemble());
      message += " on class ";
      message += ::SymbolToString(match->first->GetClassName());
      error = new CCompileError(message, method.fStartLineNumber);
    } else if (!SameType(method.fSignature.GetType(), 
			 returnType, file)) {
      string message("Cannot override method ");
      message += ::UnicodeToUTF(otherMethod->GetSignature().Disassemble());
      message += " with different return type.";
      error = new CCompileError(message, method.fStartLineNumber);
    } else if (method.fAccessFlags.MorePrivateThan(parentModifiers)) {
      string message("Cannot override method to be more private: ");
      message += ::UnicodeToUTF(otherMethod->GetSignature().Disassemble());
      error = new CCompileError(message, method.fStartLineNumber);
    } else {
      error = CheckExceptionCovar(otherMethod->GetThrows(),
				  method.GetThrows(), file);
      if (error) {
	string message = ::UnicodeToUTF(error->GetMessage());
	message += " when overriding " + 
	  ::UnicodeToUTF(otherMethod->GetSignature().Disassemble());
	error = new CCompileError(message, method.fStartLineNumber);
      }
    }
    delete match;
  }
  return error;
}

//
//  Method name : CheckExceptionCovar
//  Description : Checks the exceptions in child and parent for covariance:
//     every exception in parent must either be unchecked, or have an
//     exception in child of which it is a subtype.  Returns an error on 
//     failure; 0 on success.
//
CCompileError* 
CCompiler::CheckExceptionCovar(const typeDeque &child,
			       const typeDeque &parent,
			       CSourceFile *file) const
{
  static const CJavaTypeSignature runtimeException(kRuntimeExceptionName);
  static const CJavaTypeSignature errorType(kErrorName);
  for (typeDeque::const_iterator p = parent.begin(); p != parent.end(); ++p) {
    if (SameType(*p, errorType, file) ||
	AssignableSubtype(*p, errorType, file) ||
	SameType(*p, runtimeException, file) ||
	AssignableSubtype(*p, runtimeException, file)) continue;
    bool found = false;
    for (typeDeque::const_iterator c = child.begin(); 
	 !found && c != child.end(); ++c) {
      if (SameType(*p,*c,file)||
	  AssignableSubtype(*p,*c,file)) found = true;
    }
    if (!found) {      
      string eStr = "Exception " + ::UnicodeToUTF((*p).Disassemble()) 
	+ " not found";
      return new CCompileError(eStr);
    }
  }
  return 0;
}

//
//  Method name : UnimplementedMethods
//  Description : This call is used to find out how many methods are visible,
//    but not defined, on this class.  This includes all abstract methods
//    inherited from parents.  A list of all method signatures that are not
//    implemented is given as the second argument and is filled over the course
//    of this call to contain all of the unimplemented method signatures.  This
//    list may contain duplicates if an undefined method signature is
//    inherited in multiple ways.
//
unsigned long
CCompiler::UnimplementedMethods(const CJavaTypeSignature& onType,
				CCompiler::MethodList& methods,
				CSourceFile *file)
{
  MethodList unimplementedParents;  
  const CJavaClassFile *onClass = ClassOfType(onType, file);
  if (onClass == 0) {
    cerr << "***Couldn't find class for type " << onType << endl;
  }
  CInstantiator ins(*onClass, onType.GetParameters());

  if (onClass->fPClassInfo) {
    CJavaTypeSignature parent;
    if (onClass->fPClassInfo->fSuper) {
      ins.Instantiate(parent, *onClass->fPClassInfo->fSuper);
      UnimplementedMethods(parent, unimplementedParents, file);
    }
    if (onClass->fPClassInfo->fInterfaces) {
      for (deque<CJavaTypeSignature>::iterator i 
	     = onClass->fPClassInfo->fInterfaces->begin();
	   i != onClass->fPClassInfo->fInterfaces->end(); ++i) {
	ins.Instantiate(parent, *i);
	UnimplementedMethods(parent, unimplementedParents, file);
      }
    }
  } else {
    const CJavaClassFile* parent = LookupClass(onClass->fSuperclassName, file);
    if (parent != 0) {
      UnimplementedMethods(parent->GetType(), unimplementedParents, file);
    }
    for (deque<unicode_string>::const_iterator i = 
	   onClass->fInterfaces.begin();
	 i != onClass->fInterfaces.end(); ++i) {
      const CJavaClassFile* interface = LookupClass(*i, file); 
      if (interface != 0) {
	UnimplementedMethods(interface->GetType(), unimplementedParents, file);
      }
    }
  }
  // Might be null instantiator
  CJavaMethodSignature m;
  for (MethodList::const_iterator i = unimplementedParents.begin();
       !(i == unimplementedParents.end()); ++i) {
    CJavaTypeSignature dummy;
    pair<const CJavaClassFile*, const CJavaMethodInfo*>* matchedMethod =
      ExactMatchMethod(*i, onType, file, dummy);
    if (matchedMethod == 0 || 
	matchedMethod->second->GetModifiers().fAbstract != 0) {
      methods.push_back(*i);
    }
    delete matchedMethod;
  }
  for (MethodTable::const_iterator method = onClass->fMethods.begin();
       !(method == onClass->fMethods.end()); ++method) {
    if ((*method).second->GetModifiers().fAbstract != 0) {
      ins.Instantiate(m,(*method).first);
      CJavaMethodSignature fixed;
      FixMethodSignature(m,fixed,file);
      methods.push_back(fixed);
    }
  }
  return methods.size();
}

//
//  Method name : CheckJavaLangObject
//  Description : Checks to see if this compiler can find java.lang.Object
//    in the current class path.  Returns true if it can.
//
bool
CCompiler::CheckJavaLangObject()
{
  return fImportAliases.find(kObjectName) != fImportAliases.end() ||
    ImportClass(kObjectName) != 0;
}

//
//  Method name : CheckSupportClasses
//  Description : Checks to see if this compiler can find polyj.lang.Primitive
//    in the current class path.  Returns true if it can.
//
bool
CCompiler::CheckSupportClasses()
{
  static const unicode_string kPrimitiveName = 
    ::UTFToUnicode("polyj/lang/Primitive");
  bool res = (fImportAliases.find(kPrimitiveName) != fImportAliases.end()) 
    || (ImportClass(kPrimitiveName) != 0);

  if (!res)
    cerr << "Error: could not find polyj.lang.Primitive in the class path."
	 << endl;
  return res;
}

//
//  Method name : WarnDeprecatedField
//  Description : If the provided field is deprecated, this will emit
//    a warning message to the user.
//    If line is '0', this assumes that no line number information
//    is available.
//
void
CCompiler::WarnDeprecatedField(const CJavaFieldInfo& field,
			       const CJavaClassFile& fieldClass,
			       const CCompileContext& context,
			       unsigned long line) const
{
  if (field.IsDeprecated()) {
    string fileName =
      ::UnicodeToString(context.GetClass().fSourceFile->GetFileName());
    string message = "Warning: field " +
      ::SymbolToString(fieldClass.GetClassName()) + "." +
      ::UnicodeToString(field.GetSignature().GetFieldName()) +
      " has been deprecated.";
    PrintCompileError(fileName, message, line);
  }
}

//
//  Method name : WarnDeprecatedMethod
//  Description : If the provided method is deprecated, this will emit
//    a warning message to the user.
//    If line is '0', this assumes that no line number information
//    is available.
//
void
CCompiler::WarnDeprecatedMethod(const CJavaMethodInfo& method,
			       const CJavaClassFile& methodClass,
			       const CCompileContext& context,
			       unsigned long line) const
{
  if (method.IsDeprecated()) {
    string fileName =
      ::UnicodeToString(context.GetClass().fSourceFile->GetFileName());
    string message = "Warning: method " +
      ::UnicodeToString(method.GetSignature().Disassemble()) + " on class " +
      ::SymbolToString(methodClass.GetClassName()) + " has been deprecated.";
    PrintCompileError(fileName, message, line);
  }
}

//
//  Method name : WarnDeprecatedClass
//  Description : If the provided class is deprecated, this will emit a
//    warning message to the user.
//    If line is '0', this assumes that no line number information
//    is available.
//
void
CCompiler::WarnDeprecatedClass(const CJavaClassFile& usedClass) const
{
  if (usedClass.IsDeprecated()) {
    cerr << "Warning: class " +
      ::SymbolToString(usedClass.GetClassName()) + " has been deprecated.";
  }
}

//
//  Method name : WarnUnenforcedCast
//  Description : Warns the user if a cast to "type" will not be enforced at 
//    run-time.
//  * If line is '0', assume that no line number information
//    is available.
//
void 
CCompiler::WarnUnenforcedCast(const CJavaTypeSignature& type,
			      const CCompileContext& context,
			      unsigned long line = 0) const
{
  if (type.GetNParameters() > 0 && type.IsArray()) {
    string fileName =
      ::UnicodeToString(context.GetClass().fSourceFile->GetFileName());
    string message = "Warning: type checking for " +
      ::UnicodeToString(type.Disassemble()) + 
      " may not be fully enforced at runtime.";
    PrintCompileError(fileName, message, line);
  }
}


//
//  Method name : AddParameter
//  Description : Adds the given unicode string to the list of identifiers to
//      be considered as being of type "parameter".
//
void
CCompiler::AddParameter(const unicode_string *adoptStr)
{
  fParameterNames.insert(*adoptStr);
  delete adoptStr;
}

//
//  Method name : IsAParameter
//  Description : Returns true iff the given unicode string represents a 
//     parameter in the current code.
//
bool 
CCompiler::IsAParameter(const unicode_string *str)
{
  return fParameterized && 
    (fParameterNames.find(*str) != fParameterNames.end());
}

//
//  Method name : ClearParameterNames
//  Description : Empties fParameterNames. Called at the end of a PClass 
//       compilaton to prepare for the next class.
//
void
CCompiler::ClearParameterNames()
{
  fParameterNames.erase(fParameterNames.begin(),fParameterNames.end());
}

//
//  Method name : CheckInstantiation
//  Description : Checks whether "type" is a valid instantiation. (Satisfies
//      where clauses) Returns 0 on success, and an error on failure.
//    * Sets setPackage to true iff the instantiation for "type" needs to
//      be generated in the same package as inClass.  Otherwise sets
//      setPackage to false.  (If this method returns an error, the
//      value of setPackage is undefined.)
// 
CCompileError *
CCompiler::CheckInstantiation(const CJavaTypeSignature& type,
			      const CJavaClassFile& inClass,
			      bool& setPackage,
			      CSourceFile *file) const
{
  assert(type.GetBaseType() == CJavaTypeSignature::Class);
  const CJavaClassFile *Class = ClassOfType(type, file);
  setPackage = false;
  ////
  // Check to see whether we have a valid (possibly parameterized) class.
  ////
  if (! Class) {
    string err("Couldn't find class ");
    err += ::UnicodeToString(type.Disassemble());
    return new CCompileError(err);
  }
  if (! type.IsParameterized()) {
    if (Class->IsParameterized()) {
      string err("No parameters given for parameterized class ");
      err += ::UnicodeToString(type.Disassemble());
      return new CCompileError(err);
    } else {
      return 0;
    }
  }
  // So we've got a PType...
  if (! Class->IsParameterized()) {
    string err("Attempt to parameterize non-parameterized class "); 
    err += ::UnicodeToString(type.Disassemble());
    return new CCompileError(err);
  }
  const CPJavaClassInfo &PCInfo = Class->GetPInfo();
  if ((unsigned int) PCInfo.fParams->size() != 
      (unsigned int) type.GetNParameters()) {
    string err("Wrong number of parameters on "); 
    err += ::UnicodeToString(type.Disassemble());
    return new CCompileError(err);
  }
  // So the # of parameters matches... First checktype on the parameters,
  //   then make sure they satisfy their constraints.  
  deque<CJavaTypeSignature>::const_iterator t =
    type.GetParameters()->begin();
  deque<CJavaTypeSignature>::const_iterator end =
    type.GetParameters()->end();
  for( ; t != end; ++t) {
    CCompileError* err = CheckType(*t, inClass, file);
    if (err)
      return err;
    if ((*t).GetBaseType() == CJavaTypeSignature::Void) {
      err = new CCompileError("Attempt to parameterize on void type");
      return err;
    }
  }
  string *result = 0;
  CInstantiator ins(*Class, type.GetParameters());
  deque<unicode_string>::const_iterator p =
    PCInfo.fParams->begin();
  t = type.GetParameters()->begin();
  CPJavaConstraint fillConstraint;  
  for( ; (result == 0) && t != end; ++p, ++t) {
    bool spackage = false;
    // Check the type.
    const CJavaClassFile* c = ClassOfType(*t, file);
    if (c) {
      // Same package?
      if (! (Class->GetPackageName() == c->GetPackageName())) {
	// Is c public?
	if (! c->GetAccessFlags().fPublic) {
	  setPackage = true;
	}
      }
    }
    // Now check constraints
    const CPJavaConstraint* constraint = PCInfo.LookupConstraint(*p);
    assert(constraint);
    ins.Instantiate(fillConstraint, *constraint);
    result = TypeSatisfiesConstraint(*t, fillConstraint, 
				     inClass, spackage, file);
    if (spackage && inClass.GetPackageName() != Class->GetPackageName())
      setPackage = true;
  }
  if (result) {
    *result += " on type ";
    *result += ::UnicodeToString(type.Disassemble());
    return new CCompileError(*result);
    delete result;
  }

  return 0;
}

static const unicode_string kEqualsName = ::UTFToUnicode("equals");
static const unicode_string kCloneName  = ::UTFToUnicode("clone");

//
// Method Name: TypeSatisfiesConstraint
// Description: Checks to see whether the type "type" satisfies the 
//     instantiated constraint "constraint." Returns 0 on success; a newly 
//     allocated informative string otherwise.
//   * If the instantiation class for this constraint satisfied
//     must be declared in the same package as inClass, sets setPackage
//     to true.  Otherwise, setPackage is untouched.
//   * If verifyVisibility is false, we do _not_ check to ensure that
//     all methods are visible from the appropriate class.
//
string *
CCompiler::TypeSatisfiesConstraint(const CJavaTypeSignature& type,
				   const CPJavaConstraint& constraint,
				   const CJavaClassFile& inClass,
				   bool& setPackage,
				   CSourceFile *file,
				   bool verifyVisibility) const
{  
  deque<CPJavaConstraint::ConstraintMethod*>::const_iterator m = 
    constraint.MethodsBegin();
  bool fakeClass = type.IsPrimitive() || type.IsArray();
  CJavaTypeSignature t = TypeForConstraint(type);
  for( ; m != constraint.MethodsEnd(); ++m) {
    const typeDeque *constraintExcp = (*m)->fThrows;
    const typeDeque *methodExcp = 0;
    if (type.IsParameter()) {
      ////
      // Is it a parameter type?
      ////
      unicode_string name;
      if (!type.GetBaseClassName(name))
	assert(0);  // What's a nameless parameter doing here?
      if (inClass.fPClassInfo) {
	const CPJavaConstraint* constraint
	  = inClass.fPClassInfo->LookupConstraint(name);
	assert(constraint);
	const CPJavaConstraint::ConstraintMethod *found;
	found = MatchConstraintMethod(* (*m)->fMethod,
				      *((CPJavaConstraint*)constraint));
	if (found == 0) {
	  string* matchErrorString = 0;
	  CJavaTypeSignature resType;
	  pair<const CJavaClassFile*, const CJavaMethodInfo*>* JLOmatch =
	    MatchMethod((*m)->fMethod->GetName(), 
			CJavaTypeSignature::kObject,
			(*m)->fMethod->fArgumentTypes,
			matchErrorString, 
			resType, file);
	  delete matchErrorString;
	  // Don't try this optimization with "equals()".
	  bool onObject = (JLOmatch != 0 && 
	       JLOmatch->second->GetSignature().GetName() != kEqualsName &&
               JLOmatch->second->GetSignature().GetName() != kCloneName);  
	  delete JLOmatch;
	  if (onObject) 
	    return 0;
	}
	if (found == 0 ||
	    (found->fStatic != (*m)->fStatic) ||
	    (!ImplicitCastTo(found->fMethod->GetType(), 
			     (*m)->fMethod->GetType(), file))) {
	  string *errorString = new string("Parameter ");
	  *errorString += ::UnicodeToString(name);
	  *errorString += " failed constraint ";
	  *errorString += 
	    ::UnicodeToString((*m)->fMethod->Disassemble());
	  return errorString;
	}
	methodExcp = new deque<CJavaTypeSignature>(*found->fThrows);
      }
    } else {
      ////
      // No; it's either a primitive type, an array type, or a class type.
      ////
      pair<const CJavaClassFile*, const CJavaMethodInfo*>* res;
      string *str = 0;
      CJavaTypeSignature resType;
      res = MatchMethod((*m)->fMethod->GetName(),
			t,
			(*m)->fMethod->GetArguments(),
			str, resType, file);
      CJavaMethodSignature methodSig;
      if (res && res->first && res->second) {
	CInstantiator ins(*res->first, resType.GetParameters());
	ins.Instantiate(methodSig,res->second->GetSignature());
      }
      if (res == 0 || res->second == 0 ||
	  res->second->GetModifiers().fStatic != (*m)->fStatic ||
	  !ImplicitCastTo(methodSig.GetType(),
			  (*m)->fMethod->GetType(), file)) {
	// We didn't find the method, or the return value was wrong.
	if (str != 0) {
	  string *errorString = new string("Type ");
	  *errorString += ::UnicodeToString(type.Disassemble());
	  *errorString += " failed constraint: ";
	  *errorString += *str;
	  delete str;
	  return errorString;
	} else {
	  string *errorString = new string("Type ");
	  *errorString += ::UnicodeToString(type.Disassemble());
	  *errorString += " failed constraint ";
	  *errorString += ::UnicodeToString((*m)->fMethod->
					    Disassemble());
	  return errorString;
	}	
      } else {
	if (!fakeClass) {
	  // Is the method visible?
	  if (verifyVisibility) {
	    if (!MethodVisibleFrom(res->second->GetModifiers(),
				   *res->first, 
				   inClass,
				   setPackage)) {
	      delete res;
	      string *errorString = new string("Type ");
	      *errorString += ::UnicodeToString(type.Disassemble());
	      *errorString += " failed constraint ";
	      *errorString += ::UnicodeToString((*m)->fMethod->
						Disassemble());
	      *errorString += " (method wasn't visible from ";
	      *errorString += ::UnicodeToString(inClass.GetShortName());
	      *errorString += ")";
	      return errorString;
	    }
	  }
	  if (! res->second->GetModifiers().fPublic)
	    setPackage = true;
	  methodExcp = new deque<CJavaTypeSignature>(res->second->GetThrows());
	}
      }
    }
    if (!fakeClass) {
      // Now, check exception covariance.
      assert(constraintExcp && methodExcp);
      CCompileError *err = CheckExceptionCovar(*constraintExcp,
					       *methodExcp, file);
      if (err) {
	string *errorString = new string("Type ");
	*errorString += ::UnicodeToString(type.Disassemble());
	*errorString += " failed constraint ";
	*errorString += ::UnicodeToString((*m)->fMethod->
					  Disassemble());
	*errorString += " (" + ::UnicodeToUTF(err->GetMessage()) + ")";
	delete err;
	return errorString;
      }
    }
    delete methodExcp;
  }
  return 0;
}

//
// Method Name : ClassOfType
// Description : Returns the class of type.  If type is a primitive or 
//    parameter type, or if the class cannot be found, 0 is returned.
//
const CJavaClassFile *
CCompiler::ClassOfType(const CJavaTypeSignature& type,
		       CSourceFile *file) const
{
  unicode_string name;
  if (type.GetBaseType() != CJavaTypeSignature::Class)
    return 0;
  if (! type.GetBaseClassName(name))
    return 0;
  return LookupClass(name, file);
}

//
//  Method name : MatchConstraintMethod
//  Description : This is used to search a constraint for a matching method 
//    with the given signature that can be called with arguments of the 
//    provided type.  Returns the true signature of the matching method (if
//    one matches), otherwise returns 0.
//
const CPJavaConstraint::ConstraintMethod*
CCompiler::MatchConstraintMethod(const CJavaMethodSignature& method,
				 const CPJavaConstraint& onConstraint) const
{
  const CPJavaConstraint::ConstraintMethod* match = 0;

  deque<CPJavaConstraint::ConstraintMethod*>::const_iterator i =
    onConstraint.MethodsBegin();
  deque<CPJavaConstraint::ConstraintMethod*>::const_iterator end =
    onConstraint.MethodsEnd();
  for (; i != end; ++i) {
    // Grrr! Are argument and parameter synonymous now?
    if (((*i)->fMethod->GetName() == method.GetName()) &&
	(method.ParameterCount() == (*i)->fMethod->ParameterCount())) {
      deque<CJavaTypeSignature>::const_iterator parameter =
	(*i)->fMethod->ParametersBegin();
      deque<CJavaTypeSignature>::const_iterator parametersEnd =
	(*i)->fMethod->ParametersEnd();
      deque<CJavaTypeSignature>::const_iterator argument = 
	method.ParametersBegin();
      bool valid = true;
      for (; valid && parameter != parametersEnd; ++parameter, ++argument) {
	if (SameType(*parameter, *argument, 0)) {
	  valid = true;
	} else {
	  valid = ImplicitCastTo(*argument, *parameter, 0);
	}
      }
      if (valid) {
	if (match == 0 || MoreSpecific(* (*i)->fMethod, *match->fMethod, 0)) {
	  match = (*i);
	} else if (!MoreSpecific(*match->fMethod, *(*i)->fMethod, 0)) {
	  // Ambiguous signature.
	  return 0;
	}
      }      
    }

  }
  return match;
}


//
// Method name : MethodVisibleFrom
// Description : Return true iff a method with protection flags "methodFlags",
//     declared in "inClass" is visible from within "fromClass". 
//   * If the instantiation class for a constraint satisfied by "methodFlags"
//     must be declared in the same package as fromClass, sets setPackage
//     to true.  Otherwise, setPackage is untouched.
//   * Used for the purposes of constraint satisfaction.
//
bool 
CCompiler::MethodVisibleFrom(const CJavaAccessFlags& methodFlags,
			     const CJavaClassFile& inClass,
			     const CJavaClassFile& fromClass,
			     bool& setPackage) const
{
  bool samePackage = (inClass.GetPackageName() == fromClass.GetPackageName());
  if (! inClass.GetAccessFlags().fPublic) {
    if (! samePackage)
      return false;
    setPackage = true;
  }
  if (methodFlags.fPublic) {
    // public
    return true;
  } else if (methodFlags.fPrivate) {
    // private or private protected
    return false;
  } else {
    // protected or default
    if (samePackage) {
      setPackage = true;
    }
    // Really, we should check for SUBTYPE(fromClass,inClass) here for 
    // protected methods, and act appropriately.  The acting appropriately 
    // would make things too tricky, though: instead of needing at most one
    // layers of  implementation, we could end up needing a layer of 
    // implementation for each constraint method.
    // We're better off pretending we don't know how to handle it.
    return samePackage;
  }  
}

//
// Method name : PreparePJInfo
// Description : Processes parameterized information on the provided class,
//    if necessary.  Returns true on success.
//
bool
CCompiler::PreparePJInfo(CJavaClassFile &Class,
			 CSourceFile *file)
{
  string fileName;
  if (Class.GetSourceFile()) {
    fileName = ::UnicodeToString(Class.GetSourceFile()->GetFileName());
  } else {
    fileName = "*UNKNOWN*";
  }
  if (!Class.fPClassInfo || Class.fPClassInfo->IsFinished()) {
    return true;
  }
  ////
  // Have we tried to prepare this before?
  //// 
  CPJavaClassInfo &PJInfo = *Class.fPClassInfo;
  if (PJInfo.IsProcessing()) {
    PrintCompileError(fileName,
		      "Circular inheritance on class " +
		      ::SymbolToString(Class.GetClassName()) +
		      ", or confused by earlier errors.");
    return false;
  }
  // Good. This is the first time we've tried to PreparePJInfo this class.
  Class.fPClassInfo->BeginProcessing();  
  bool needPInfo = false;
  ////
  // Prepare our superclass.
  ////
  if (PJInfo.fSuper) {
    const CJavaClassFile *parent = ClassOfType(*PJInfo.fSuper, file);
    if (!parent) {
      PrintCompileError(fileName,
			"Couldn't find class for: " +
			::UnicodeToString(PJInfo.fSuper->Disassemble()));
      return false;
    }
    if (! PreparePJInfo(* (CJavaClassFile *)parent, file)) // discard const
      return false;
    if (parent->HasPInfo())
      needPInfo = true;
  }
  ////
  // Prepare our interfaces.
  ////
  if (PJInfo.fInterfaces) {
    for(deque<CJavaTypeSignature>::iterator i = PJInfo.fInterfaces->begin();
	i != PJInfo.fInterfaces->end(); ++i) {
      const CJavaClassFile *parent = ClassOfType(*i, file);
      if (!parent) {
	PrintCompileError(fileName,
			  "Couldn't find class for: " +
			  ::UnicodeToString((*i).Disassemble()));
	return false;
      }
      if (! PreparePJInfo(*(CJavaClassFile *)parent, file)) // discard const
	return false;
      if (parent->HasPInfo())
	needPInfo = true;
    }
  }
  if (PJInfo.fParams && PJInfo.fParams->size() > 0)
    needPInfo = true;
  const CIntermediateClass* intermediate = 
    LookupIntermediateClass(Class.GetClassName(), file);
  ////
  // Do our fields need a PInfo?
  /////
  for (FieldTable::iterator i = Class.fFields.begin();
	!needPInfo && i != Class.fFields.end(); ++i) {    
    // We should test for private-ness, but we don't really have to.
    if ((*i).first.GetType().IsParameterized()) 
        needPInfo = true;
  }
  ////
  // Do our methods need a PInfo?
  ////
  for(MethodTable::iterator i = Class.fMethods.begin(); 
      !needPInfo && i != Class.fMethods.end(); ++i) {
    if ((*i).first.IsParameterized()) {
      needPInfo = true;
    }
  }
  if (! needPInfo) {
    // Our parents have no PInfo fields, and we have no parameters. None of
    // our fields or methods have parameters either.  Therefore remove our 
    // PInfo.
    delete Class.fPClassInfo;
    Class.fPClassInfo = 0;
    if (intermediate)
      ((CIntermediateClass*)intermediate)->fPClassInfo = 0;
    return true;
  }
  // Our constraints have already been checked.

  ////
  // Set fHasStinit: If we have any <stinit> code, or any initialized
  //    static fields, set fHasStinit to true.
  ////
  PJInfo.fHasStinit = false;
  if (PJInfo.IsParameterized()) {
    if (PJInfo.fSuper) { 
      const CJavaClassFile* superClass = ClassOfType(*PJInfo.fSuper, file);
      if (superClass && superClass->fPClassInfo && 
	  superClass->fPClassInfo->fHasStinit) {
	PJInfo.fHasStinit = true;
	PJInfo.GetAMM().RequireMethod(CAuxMethodSignature::GetWhere,
				      *PJInfo.fSuper);
      }
    }
    bool needsTClass = false;
    if (intermediate) {
      for(deque<CIntermediateFunction*>::const_iterator i = 
	    intermediate->fFunctions.begin(); 
	  !PJInfo.fHasStinit && i != intermediate->fFunctions.end(); ++i) {
	if ((*i)->fSignature.GetName() == kStaticName && (*i)->fBlock &&
	    (*i)->fBlock->fChildren && (*i)->fBlock->fChildren->size() > 0)
	  PJInfo.fHasStinit = true;
      }	  
      for(StatementList::const_iterator i = 
	    intermediate->fStaticDeclarations.begin();
	  !PJInfo.fHasStinit && 
	    i != intermediate->fStaticDeclarations.end(); 
	  ++i) {
        CDeclarationStatement *ds = DYNAMIC_CAST(CDeclarationStatement, (*i));
	assert(ds);
	if (ds->HasInitializer())
	  PJInfo.fHasStinit = true;
	if (!ds->fModifiers->fPublic)
	  needsTClass = true;
      }
    }
    if (PJInfo.fHasStinit) {
      PJInfo.fAMM->RequireMethod(CAuxMethodSignature::FetchStatic,
				 CJavaTypeSignature::kBoolean,
				 &kInitializedName);
      PJInfo.fAMM->RequireMethod(CAuxMethodSignature::SetStatic,
				 CJavaTypeSignature::kBoolean,
				 &kInitializedName);
      needsTClass = true;
    }
  }
  fTClassPackages.insert(Class.GetPackageName());
  ////
  // Generate where-type
  ////
  if (PJInfo.IsParameterized()) 
    fWhereClasses.push_back(&Class);
  ////
  // Fix types on constraints
  ////
  assert(PJInfo.fPTable); 
  bool foundBadConstraint = false;
  for(ParamTable::iterator i = PJInfo.fPTable->begin();
      i != PJInfo.fPTable->end(); ++i) {
    // For each constraint
    assert((*i).second);
    for(deque<CPJavaConstraint::ConstraintMethod*>::iterator cm =
	  (*i).second->fMethods.begin();
	cm != (*i).second->fMethods.end(); ++cm) {      
      // For each constraint method
      CCompileError *err = 0;
      assert((*cm)->fMethod);
      assert((*cm)->fThrows);
      FixMethodSignature(*(*cm)->fMethod,*(*cm)->fMethod,file);
      for(deque<CJavaTypeSignature>::iterator excpt = (*cm)->fThrows->begin();
	  excpt != (*cm)->fThrows->end(); ++excpt) {
	*excpt = FixType(*excpt, file); 
	err = CheckType(*excpt, Class, file);
	if (!err && ! IsThrowable(*excpt, file)) {
	  string errorString = "Non-Throwable type ";
	  errorString += ::UnicodeToString((*excpt).Disassemble());
	  err = new CCompileError(errorString);
	}
      }
      if (!err)
	err = CheckMethodSignature(*(*cm)->fMethod, Class, file);
      if (err) {
	PrintCompileError(fileName,
			  "On constraint for " +
			  ::UnicodeToString((*i).second->GetParameter()) +
			  " on class " + 
			  ::SymbolToString(Class.GetClassName()) +
			  ": " +
			  ::UnicodeToString(err->GetMessage()));
	delete err;
	foundBadConstraint = true;
      }
    }
  }
  if (foundBadConstraint) 
    return false;

  ////
  // CheckType/FixType on parents and interfaces
  ////
  if (PJInfo.fSuper) {
    CCompileError *err = CheckType(*PJInfo.fSuper,Class,file);
    if (err) {
      PrintCompileError(fileName, err->GetMessage());
      return false;
    } else {      
      *PJInfo.fSuper = FixType(*PJInfo.fSuper,file);
      if (PJInfo.fSuper->IsInstantiable()) {
	CheckType(*PJInfo.fSuper,Class,file);
	fInstantiationList.AddType(*PJInfo.fSuper, file);
      }
    }
  }
  deque<CJavaTypeSignature> newInterfaces;
  if (PJInfo.fInterfaces) {
    for(deque<CJavaTypeSignature>::iterator i = PJInfo.fInterfaces->begin();
	i != PJInfo.fInterfaces->end(); ++i) {
      CCompileError *err = CheckType((*i),Class,file);
      if (err) {
	PrintCompileError(fileName, err->GetMessage());
	return false;
      } else {
	(*i) = FixType(*i,file);
	if ((*i).IsInstantiable()) {
	  CheckType((*i),Class,file);
	  fInstantiationList.AddType(*i, file);
	}
      }
      // "The interface of my interface is my interface."
      const CJavaClassFile *c = ClassOfType(*i, file);
      if (c->HasPInfo() && c->GetPInfo().fInterfaces) {
	CInstantiator ins(*c, (*i).GetParameters());
	for (deque<CJavaTypeSignature>::iterator in = 
	       c->GetPInfo().fInterfaces->begin(); 
	     in != c->GetPInfo().fInterfaces->end(); ++in) {
	  CJavaTypeSignature t;
	  ins.Instantiate(t,*in);
	  newInterfaces.push_back(t);
	}
      }
    }
  }
  for (deque<CJavaTypeSignature>::iterator i =
	 newInterfaces.begin();
       i != newInterfaces.end(); ++i) {
    bool found = false;
    for(deque<CJavaTypeSignature>::iterator j = 
	  PJInfo.fInterfaces->begin();
	j != PJInfo.fInterfaces->end(); ++j) {
      if (SameType(*j, *i, file))
	found = true;
    }
    if (!found) 
      PJInfo.fInterfaces->push_back(*i);
  }

  ////
  // Adopt from parents/Interfaces
  ////
  if (PJInfo.fSuper) {
    const CJavaClassFile *parent = ClassOfType(*PJInfo.fSuper, file);
    if (!parent) {
      PrintCompileError(fileName,
			  "Couldn't find class for: " +
			  ::UnicodeToString(PJInfo.fSuper->Disassemble()));
      return false;
    } else {
      if (parent->fPClassInfo) {
	AdoptPJMethods(Class, *PJInfo.fSuper, file);
      }
    }
  }
  if (PJInfo.fInterfaces) {
    for(deque<CJavaTypeSignature>::iterator i = PJInfo.fInterfaces->begin();
	i != PJInfo.fInterfaces->end(); ++i) {
      const CJavaClassFile *interface = ClassOfType(*i, file);
      if (!interface) {
	PrintCompileError(fileName,
			  "Couldn't find class for: " +
			  ::UnicodeToString((*i).Disassemble()));
	return false;
      } else {
	if (interface->fPClassInfo) {
	  AdoptPJMethods(Class, *i, file);
	}
      }
    }
  }

  ////
  // Register constraints.
  ////
  PJInfo.RegisterConstraints();

  ////
  // Name AuxMethods
  ////
  PJInfo.fAMM->NameMethods();  

  ////
  // Look for conflicting methods
  //// 
  if (!FindConflictingMethods(Class, file))
    return false;  


  ////
  // Register & mangle method names
  ////
  CCompileError *err = RenamePJMethods(Class);
  if (err) {
    PrintCompileError(fileName, err->GetMessage());
    return false;
  }

  // So now we're done!
  Class.fPClassInfo->Finish();
       
  return true;
}

//
// Method name : FixAMM
// Description : FixType's all auxmethod signatures (if any) in Class.
//    Returns true iff successful.      
// Requires    : Class has already been unparsed, so that we know about
//    all AuxMethods.
//
bool 
CCompiler::FixAMM(CJavaClassFile &Class,
		  CSourceFile* file) 
{
  if (!Class.HasPInfo()) return true;
  CPJavaClassInfo &PJInfo = Class.GetPInfo();

  assert(PJInfo.fAMM);
  CAuxMethodMap *newAMM = new CAuxMethodMap(Class.GetClassName());
  for(amm_rep::iterator am = PJInfo.fAMM->fMap.begin();
      am != PJInfo.fAMM->fMap.end(); ++am) {
    // For each auxmethod
    // Fix Class
    CJavaTypeSignature cl = FixType((*am).first.fClass, file);
    unicode_string *n = 0;
    CJavaMethodSignature *m = 0;
    CJavaMethodSignature *fm = 0;    
    CJavaTypeSignature *fc = 0;
    deque<CJavaTypeSignature> *th = 0;
    // Copy Name.
    if ((*am).first.fName) {
      n = new unicode_string(*(*am).first.fName);
    }
    // Fix Method
    if ((*am).first.fMethod) {
      m = new CJavaMethodSignature();
      FixMethodSignature(*(*am).first.fMethod,*m,file);
    }
    // Fix firstMethod
    if ((*am).first.fFirstMethod) {
      fm = new CJavaMethodSignature();
      FixMethodSignature(*(*am).first.fFirstMethod,*fm,file);
    }    
    // Fix Throws
    if ((*am).first.fThrows) {
      th = new deque<CJavaTypeSignature>;
      for (deque<CJavaTypeSignature>::iterator i = 
	     (*am).first.fThrows->begin(); 
	   i != (*am).first.fThrows->end(); ++i) {
	CJavaTypeSignature t;
	t = FixType(*i, file);
	th->push_back(t);
      }
    }
    // Fix FirstClass
    if ((*am).first.fFirstClass) {
      fc = new CJavaTypeSignature();
      *fc = FixType(*(*am).first.fFirstClass,file);
    }
    CAuxMethodSignature newAMS((*am).first.fType, cl, n, m, th,
			       (*am).first.fWType, (*am).first.fWidth,
			       (*am).first.fDim, fm, fc);
    newAMM->fMap.insert(amm_rep::value_type(newAMS, (*am).second));
    (*am).second = 0;
  }
  delete PJInfo.fAMM;
  PJInfo.fAMM = newAMM;
  return true;
}

//
// Method name: FixMethodSignature
// Description: FixType's all types in m's arguments and returns; sets
//        "set" to fixed type.
//
void 
CCompiler::FixMethodSignature(const CJavaMethodSignature &m,
			      CJavaMethodSignature &set,
			      CSourceFile *file) const
{
  unicode_string methodName = m.GetName();
  CJavaTypeSignature returns = FixType(m.GetType(), file);
  deque<CJavaTypeSignature> args;
  deque<CJavaTypeSignature>::const_iterator arg = m.ParametersBegin();
  deque<CJavaTypeSignature>::const_iterator end = m.ParametersEnd();
  for (; arg != end; ++arg) {
    args.push_back(FixType(*arg,file));
  }
  CJavaMethodSignature signature(returns, methodName, args);
  set = signature;
}

//
// Method name: FixFieldSignature
// Description: FixType's the type of a field signature f.
//
CJavaFieldSignature 
CCompiler::FixFieldSignature(const CJavaFieldSignature& f,
			     CSourceFile *file) const
{
  CJavaTypeSignature t = FixType(f.GetType(),file);
  return CJavaFieldSignature(t, f.GetFieldName());
}

//
// Method name: AdoptPJMethods
// Description: Class.fPClassInfo->fMethods (if it exists) adopts all method 
//    renaming from parent[ins] (if it exists.) Should be called 
//    before RenamePJMethods.
//
void 
CCompiler::AdoptPJMethods(CJavaClassFile &Class, 
			  const CJavaTypeSignature &parentType,
			  CSourceFile *file)
{
  assert(Class.fPClassInfo);
  const CJavaClassFile *parent = ClassOfType(parentType, file);
  assert(parent);
  CInstantiator ins(*parent, parentType.GetParameters());
  if (! parent->fPClassInfo)
    return;
  if (! Class.fPClassInfo->fMethods)
    Class.fPClassInfo->fMethods = new MethodNameMap;

  for (MethodNameMap::iterator i = parent->fPClassInfo->fMethods->begin();
       i != parent->fPClassInfo->fMethods->end(); ++i) {
    CJavaMethodSignature method;
    ins.Instantiate(method, (*i).first);
    bool insert = true;
    if ((*i).first.GetName() == kConstructorName || 
	(*i).first.GetName() == kStaticName ||
	(*i).second.fAccessFlags.fPrivate ||
	(*i).second.fAccessFlags.fStatic) { 
      insert = false;
    }
    for(MethodNameMap::iterator j = Class.fPClassInfo->fMethods->begin();
	insert && j != Class.fPClassInfo->fMethods->end(); ++j) {
      if (method == (*j).first) {
	// We already have this mapping.
	insert = false;
	if ((*j).second.fIndex != (*i).second.fIndex) {
	  // Eventually, this would be a sign that we need proxy objects.
	  // We don't do 'em now.
	  cerr << "Method renaming adoption conflict: " << endl << "    "
	       << ::UnicodeToString(method.Disassemble())
	       << "->" << (*i).second.fIndex << endl << " vs "
	       << ::UnicodeToString((*j).first.Disassemble())
	       << "->" << (*j).second.fIndex << endl;
	}      
      }
    }
    if (insert) {
      TranslatedMethod trans;
      trans.fIndex = (*i).second.fIndex;
      trans.fTranslatedSignature = (*i).second.fTranslatedSignature;
      trans.fAccessFlags = (*i).second.fAccessFlags;
      MethodNameMap::value_type v(method,trans);
      Class.fPClassInfo->fMethods->insert(v);
    }
  }
}

//
// Method name : FindConflictingMethods
// Description : Find methods which, if instantiated badly, will have 
//    conflicting signatures.  Returns true on success (none found),
//    prints errors and returns false on failure.
//
bool
CCompiler::FindConflictingMethods(CJavaClassFile &Class,
				  CSourceFile *file)
{
  bool foundConflict = false;
  for(MethodTable::iterator i = Class.fMethods.begin();
      i != Class.fMethods.end(); ++i) {
    MethodTable::iterator j = i;
    j++;
    for (; j!=Class.fMethods.end(); ++j) {
      if (((*i).first.GetName() == (*j).first.GetName()) &&
	  ((*i).first.ParameterCount() == (*j).first.ParameterCount())) {
	deque<CJavaTypeSignature>::const_iterator t1 = 
	  (*i).first.ParametersBegin();
	deque<CJavaTypeSignature>::const_iterator t2 =
	  (*j).first.ParametersBegin();
	// Allmatch: Have all arguments of (*i) and (*j) matched 
	// (postinstantiation) so far?
	bool allMatch = true;
	for(; allMatch && t1 != (*i).first.ParametersEnd(); ++t1, ++t2) {
	  if (!SameType(*t1,*t2,file)) { 
	    if ((*t1).IsParameter() && (*t2).IsParameter()) {
	      unicode_string n1,n2;
	      (*t1).GetBaseClassName(n1);
	      (*t2).GetBaseClassName(n2);
	      if (n1 == n2 && (*t1).GetArrayBounds() != (*t2).GetArrayBounds())
		allMatch = false;
	    } else if (!((*t1).IsParameter()) &&
		       !((*t2).IsParameter())) {
	      allMatch = false;
	    }
	  }
	}
	if (allMatch) {
	  foundConflict = true;
	  string message = "Methods " 
	    + ::SymbolToString((*i).first.Disassemble())
	    + " and " + ::SymbolToString((*j).first.Disassemble())
	    + " could conflict after instantiation";
	  PrintCompileError(
	   ::UnicodeToString(Class.GetSourceFile()->GetFileName()), message);
	}
      }
    }
  }
  return !foundConflict;
}


//
// Method Name: RenamePJMethods
// Description: Adds method renaming entries to Class.fPClassInfo.  
//    Assumes we've already called AdoptPJMethods on parents and interfaces.
//
CCompileError*
CCompiler::RenamePJMethods(CJavaClassFile &Class)
{
  assert(Class.fPClassInfo);
  if (! Class.fPClassInfo->fMethods) {
    Class.fPClassInfo->fMethods = new MethodNameMap;
  }  
  for(MethodTable::iterator i = Class.fMethods.begin();
      i != Class.fMethods.end(); ++i) {
    if ((*i).first.GetName() == kStaticName) continue;
    // Is this method an override?
    MethodNameMap::iterator found = 
      Class.fPClassInfo->fMethods->find((*i).first);
    if (found != Class.fPClassInfo->fMethods->end()) {
      // If so, update the method's translated signature
      if (! ((*found).second.fTranslatedSignature == 
	     (*i).second->fTranslatedSignature)) {
	(*i).second->SetTranslatedSignature(
				       (*found).second.fTranslatedSignature);
      }
      continue;
    }
    // No, not an override... so what's the next assignable index?
    int highestFound = -1;
    MethodNameMap::iterator conflictingMethod = 
      Class.fPClassInfo->fMethods->end();
    bool isStatic = ((*i).second) &&  
      (((*i).second->GetModifiers().fStatic) || (*i).first.IsConstructor());
    for(MethodNameMap::iterator j = Class.fPClassInfo->fMethods->begin();
	j != Class.fPClassInfo->fMethods->end(); ++j) {
      if ((*i).first.PostTransCompare((*j).first, false) == 0) {
	if ((*j).second.fIndex > highestFound) {
	  highestFound = (*j).second.fIndex;
	  conflictingMethod = j;
	}
      }
    }
    CJavaMethodSignature transType = (*i).first;
    if (isStatic && Class.IsParameterized()) {
      transType.SObjectify(Class.GetClassName());
    } else {
      transType.Objectify();
    }
    // Now do index for constructor case...
    if ((*i).first.IsConstructor() && highestFound >= 0) {
      for(int k = 0; k <= highestFound; ++k) {
	transType.AddArgument(CJavaTypeSignature::kInteger);
      }
      bool found;
      do {
	found = false;
	for(MethodNameMap::iterator j = Class.fPClassInfo->fMethods->begin();
	    !found && j != Class.fPClassInfo->fMethods->end(); ++j) {
	  if (transType.Compare((*j).second.fTranslatedSignature) == 0) {
	    transType.AddArgument(CJavaTypeSignature::kInteger);
	    highestFound++;
	    found = true;
	  }
	}
      } while (found);
    }
    TranslatedMethod trans;
    trans.fIndex = highestFound+1;
    trans.fTranslatedSignature = transType;
    trans.fAccessFlags = (*i).second->GetModifiers();
    MethodNameMap::value_type v((*i).first, trans);
    Class.fPClassInfo->fMethods->insert(v);
    (*i).second->SetTranslatedSignature(transType);
  }
  return 0;
}

//
// Method name: UnparseStinit
// Description: Unparses the $stinit function for the given class to *fOutput,
//    given the class's static initializer function, and static variable
//    declarations.
//
CCompileError* 
CCompiler::UnparseStinit(CJavaClassFile& Class, 
			 CIntermediateFunction* function,
			 StatementList& sDeclarations,
			 CSourceFile *file)
{
  if (!Class.HasPInfo()) {
    return 0;
  }
  const CJavaClassFile* superClass = 0;
  if (Class.GetPInfo().fSuper) {
    superClass = ClassOfType(*Class.GetPInfo().fSuper,file);
  }

  if (!Class.GetPInfo().fHasStinit) {
    if (superClass && superClass->HasPInfo() && 
	superClass->GetPInfo().fHasStinit) {
      fOutput->out() << "static { ";
      superClass->GetType().Unparse(*fOutput);
      fOutput->out() << ".$stinit(";
      Class.GetPInfo().fSuper->UnparseInstantiation(*fOutput,*this);
      fOutput->out() << ".getInstance()); }";
    }
    return 0;
  }

  // static protected void $stinit(Type$$$W $W)
  // {
  //   if (! $W.FetchInitialized(token.the_instance)) return;
  //   $W.StoreInitialized(true,token.the_instance);

  fOutput->Next();
  fOutput->out() << "static public void $stinit(";
  Class.GetType().UnparseWType(*fOutput);
  fOutput->out() << " $W)";
  fOutput->Next().out() << "{";
  fOutput->PushIndent(2).Next().out()
    << "if (";
  
  // Find auxmethod for $initialized [FETCH]
  unicode_string fetch = Class.GetPInfo().GetAMM().
    AuxMethodName(CAuxMethodSignature::FetchStatic, 
		  CJavaTypeSignature::kBoolean, &kInitializedName);
  // Find auxmethod for $initialized [STORE]
  unicode_string store = Class.GetPInfo().GetAMM().
    AuxMethodName(CAuxMethodSignature::SetStatic, 
		  CJavaTypeSignature::kBoolean, &kInitializedName);

  fOutput->out() << "$W." << ::SymbolToString(fetch) << "(";
  Class.GetType().UnparseTokenClass(*fOutput);
  fOutput->out() << ".the_instance)) return;";
    
  fOutput->Next().out() 
    << "$W." << ::SymbolToString(store) << "(true, "; 
  Class.GetType().UnparseTokenClass(*fOutput);
  fOutput->out() << ".the_instance);";
  
  // Initialize supertype:
  // super.$stinit($W);
  if (Class.GetPInfo().fSuper) {
    const CJavaClassFile* superClass =
      ClassOfType(*Class.GetPInfo().fSuper, file);
    assert(superClass);
    if (superClass->HasPInfo() && superClass->GetPInfo().fHasStinit) {
      unicode_string getWhere = Class.GetPInfo().GetAMM().
	AuxMethodName(CAuxMethodSignature::GetWhere, 
		      *Class.GetPInfo().fSuper);
      superClass->GetType().Unparse(*fOutput);
      fOutput->Next().out() << ".$stinit($W."
			    << ::SymbolToString(getWhere) << "());";
    }
  }
  
  // Unparse static variable initializers.
  CCompileContext context(this, &Class, 0, 
			  file, 0, true);  
  CCompileError *error = 0;
  for(StatementList::iterator i = sDeclarations.begin();
      !error && i != sDeclarations.end(); ++i) {
    CDeclarationStatement *ds =
      DYNAMIC_CAST(CDeclarationStatement, (*i));
    assert(ds);
    error = ds->UnparseInitializers(*fOutput, context);
  }
  // Unparse static code block.
  if (!error && function && function->fBlock) {
    error = function->fBlock->Unparse(*fOutput, context);
  }
  fOutput->PopIndent(2).Next().out() << "}";
  
  return error;
}

//
// Method name: UnparseTC1
// Description: Unparses first-level typechecking functions ($instanceof) for 
//     the given class to *fOutput, given the class's parameters (if any).
//
void
CCompiler::UnparseTC1(CJavaClassFile& inClass,
		      const CJavaTypeSignature& forType,
		      CSourceFile *file)
{
  unicode_string name;
  bool res = forType.GetBaseClassName(name);
  assert(res);
  const CJavaClassFile *Class = ClassOfType(forType,file);
  assert(Class);
  if (! Class->HasPInfo() || !Class->GetPInfo().IsParameterized())
    return;
  CPJavaClassInfo &PInfo = Class->GetPInfo();
  int n_params = PInfo.fParams->size();
  ////
  // Instanceof method
  ////
  fOutput->Next().out() << "// Typechecking methods for " << forType;  
  fOutput->Next().out() << "public boolean $instanceof$"
			<< ::SymbolToString(::SlashToDollar(name))
		        << "(";
  int index = 1;
  for (; index <= n_params; ++index) {
    fOutput->out() << "String $" << index;
    if (index < n_params) fOutput->out() << ", ";
  }
  fOutput->out() << ")";
  if (inClass.GetAccessFlags().fInterface) {
    fOutput->out() << ";";
    fOutput->Next();
    return;
  }
  deque<CJavaTypeSignature>::const_iterator i =
    forType.GetParameters()->begin();
  deque<CJavaTypeSignature>::const_iterator t_end =
    forType.GetParameters()->end();
  CAuxMethodMap &AMM = inClass.GetPInfo().GetAMM();
  index = 1;
  fOutput->out() << " {";
  fOutput->PushIndent(2).Next().out()
    << "return ";
  if (i == t_end) fOutput->out() << "true";
  for (; i != t_end; ++index) {
    if ((*i).IsInstantiable()) {
      fOutput->out() << "$" << index << ".equals(\""
		     << (*i).Compile() << "\")";
    } else {
      unicode_string auxMethod =
	AMM.AuxMethodName(CAuxMethodSignature::PSig, *i);
      fOutput->out() << "$W." << ::UnicodeToUTF(auxMethod)
		     << "().equals($" << index << ")";
    }
    if (++i != t_end) fOutput->out() << " && ";      
  }
  fOutput->out() << "; }";
  fOutput->PopIndent(2).Next();

}

//
//  Method name : UnparseSFMethods
//  Description : Write set and fetch methods for this class's protected
//     static variables.
void 
CCompiler::UnparseSFMethods(CJavaClassFile& Class)
{
  if (!Class.HasPInfo()) return;
  CPJavaClassInfo &PInfo = Class.GetPInfo();
  for(FieldTable::iterator i = Class.fFields.begin();
      i != Class.fFields.end(); ++i) {
    assert((*i).second);
    if ((*i).second->GetModifiers().fProtected &&
	(*i).second->GetModifiers().fStatic) {
      // Write the methods.
      unicode_string fieldname = (*i).first.GetFieldName();
      unicode_string setname = PInfo.GetAMM().
	AuxMethodName(CAuxMethodSignature::SetStatic,
		      (*i).first.GetType(), &fieldname);
      unicode_string fetchname = PInfo.GetAMM().
	AuxMethodName(CAuxMethodSignature::FetchStatic,
		      (*i).first.GetType(), &fieldname);
      // Set method.
      fOutput->out() << "protected static ";
      if ((*i).first.GetType().IsPrimitive())
	(*i).first.GetType().Unparse(*fOutput);
      else
	fOutput->out() << "Object";      
      fOutput->out() << " $set_" 
		     << ::SymbolToString((*i).first.GetFieldName())
		     << "(";
      Class.GetType().UnparseWType(*fOutput);
      fOutput->out() << " $W, ";
      if ((*i).first.GetType().IsPrimitive())
	(*i).first.GetType().Unparse(*fOutput);
      else
	fOutput->out() << "Object";
      fOutput->out() << " val) {";
      fOutput->PushIndent(4).Next().out() 
	<< "return $W." << ::SymbolToString(setname) 
	<< "(val, $token.the_instance);";
      fOutput->PopIndent(4).Next().out() << "}";
      // Fetch method.
      fOutput->out() << "protected static ";
      if ((*i).first.GetType().IsPrimitive())
	(*i).first.GetType().Unparse(*fOutput);
      else
	fOutput->out() << "Object";
      fOutput->out() << " $fetch_" 
		     << ::SymbolToString((*i).first.GetFieldName())
		     << "(";
      Class.GetType().UnparseWType(*fOutput);
      fOutput->out() << " $W) { ";
      fOutput->PushIndent(4).Next().out() 
	<< "return $W." << ::SymbolToString(fetchname) 
	<< "($token.the_instance);";
      fOutput->PopIndent(4).Next().out() << "}";
    }	
  }
}


//
// Method name : GenerateWhereClass
// Description : Writes the code necessary to implement Class's 
//     where-type to os.
//
void 
CCompiler::GenerateWhereClass(CJavaClassFile& Class, ostream& os)
{
  assert(Class.HasPInfo());

  srcstream sOut(os);
  TypeSet instantiationList;
  string cName = ::SymbolToString(Class.GetShortName());
  string fullName = ::SymbolToString(Class.GetClassName());
  string wName = cName + "$$$W";
  string pName = ::SymbolToString(Class.GetPackageName());
  
  if (pName.length() >= 1) {
    sOut.out() << "package " << pName << ";";
    sOut.Next();
  }

  CJavaAccessFlags fl = Class.GetAccessFlags();

  sOut.out() << "public abstract class " << wName;

  sOut.Next().PushIndent(2).out() << "{";
  sOut.Next();

  if (!fl.fInterface) {
    ////
    // Signatures for auxmethods.
    ////
    for (amm_rep::const_iterator am = Class.GetPInfo().GetAMM().Begin();
	 am != Class.GetPInfo().GetAMM().End(); ++am) {
      assert((*am).second);
      (*am).first.Unparse(*(*am).second, sOut, instantiationList, *this,
			  Class.GetType(), true);
    }
    fInstantiationList.AddTypes(instantiationList, 0);

    ////
    // Code for static initializer.
    ////
    if (Class.GetPInfo().fHasStinit) {
      sOut.Next().Next().out() << "public static void $stinit("
			       << wName << " $W)";
      sOut.PushIndent(2).Next().out() << "{ " << cName << ".$stinit($W); }";
      sOut.PopIndent(2).Next();
    }
  }
  
  ////
  // Method for $cast
  ////
  CJavaTypeSignature type = Class.GetType();
  int n_params = type.GetNParameters();
  sOut.Next().out() << "public static " << cName << " $cast(Object o";
  for (int i = 1; i <= n_params; ++i) {
    sOut.out() << ", String $" << i;
  }
  sOut.PushIndent(2).out() << ") {";
  sOut.Next().out() << cName << " res = (" << cName << ") o;";
  sOut.Next().out() << "if (res != null && ! res.$instanceof$"
		    << ::UnicodeToUTF(::SlashToDollar(Class.GetClassName()))
		    << "(";
  for (int i = 1; i <= n_params; ++i) {
    sOut.out() << "$" << i;
    if (i < n_params) sOut.out() << ",";
  }
  sOut.out() << "))";
  sOut.PushIndent(2).Next().out() << "throw new ClassCastException();";
  sOut.PopIndent(2).Next().out() << "return res;";
  sOut.PopIndent(2).Next().out() << "}";    
  ////
  // Method for $instanceof
  ////
  sOut.Next().out() << "public static boolean $instanceof(Object o";
  for (int i = 1; i <= n_params; ++i) {
    sOut.out() << ", String $" << i;
  }
  sOut.PushIndent(2).out() << ") {";
  sOut.Next().out() << "return (o instanceof " << cName << ") && (("
		    << cName << ") o).$instanceof$"
		    << ::UnicodeToUTF(::SlashToDollar(Class.GetClassName()))
		    << "(";
  for (int i = 1; i <= n_params; ++i) {
    sOut.out() << "$" << i;
    if (i < n_params) sOut.out() << ",";
  }
  sOut.out() << ");";
  sOut.PopIndent(2).Next().out() << "}";    
 
  sOut.PopIndent(2).Next().out() << "}";
  sOut.Next();
}

//
//  Method name : GenerateInstantiation
//  Description : Creates the ".java" files for the instantiation of IType.
//      Returns true on success. Prints error messages if needed.
//
bool
CCompiler::GenerateInstantiation(const CJavaTypeSignature& IType,
				 const string& basePath)
{
  assert(IType.IsInstantiable());
  const CJavaClassFile* Class = ClassOfType(IType, 0);
  assert(Class && Class->HasPInfo());  
  CPJavaClassInfo &PInfo = Class->GetPInfo();
  CInstantiator ins(*Class, IType.GetParameters());
  ofstream os;
  TypeSet instantiationList;
  CJavaTypeSignature parentType;
  assert(PInfo.GetSuper());
  ins.Instantiate(parentType, *PInfo.GetSuper());

  ////
  // Find names for classes.
  ////
  string pkgName    = ::SymbolToString(Class->GetPackageName());
  string IpkgName   = ::SymbolToString(InstantiationPackage(IType));
  string fullName   = ::SymbolToString(Class->GetClassName());
  string slashName  = ::SymbolToString(::SlashToDollar(Class->GetClassName()));
  string shortName  = ::SymbolToString(Class->GetShortName());
  string WName      = shortName + "$$$W";
  string fullWName  = fullName + "$$$W";
  string params     = ::SymbolToString(IType.EncodeParams());
  string IName      = ::SymbolToString(::SlashToDollar(Class->GetClassName()))
                        + "$$" + params;
  string fullIName  = IpkgName + (IpkgName.size() > 0 ? "." : "") + IName;

  string pathPrefix = basePath + UnicodeToUTF(InstantiationPackage(IType));
  if (pathPrefix.length() > 0) pathPrefix += "/";

  ////
  // Find and make paths for WClass; open sOut.
  ////
  string fileName = pathPrefix + IName  + ".java";

  if (!CFilePath::MakeDirectories(fileName)) {
    cerr << ">> Couldn't create directory for " << fileName << endl;
    return false;
  }
  os.open(fileName.c_str());
  if (!os.is_open()) {
    cerr << ">> Couldn't open " << fileName << endl;
    return false;
  }
  srcstream sOut(os);

  GeneratedFile(IType, fileName);
  
  ////
  // Find any non-public statics
  ////
  bool hasProtStatic = false;
  for (FieldTable::const_iterator i = Class->fFields.begin();
       i != Class->fFields.end(); ++i) {
    assert((*i).second);
    if ((*i).second->GetModifiers().fStatic &&
	!(*i).second->GetModifiers().fPublic)
      hasProtStatic = true;
  }
  
  ////
  // Generate head of file, and class header.
  ////
  sOut.out() << 
    "/* Instantiation file for " << IType << "\n"
    " * (Automatically generated by PolyJ -- do not edit manually.)\n" 
    " */\n";
  if (IpkgName.length() > 0) {
    sOut.out() << "package " << IpkgName << ";";
    sOut.Next();
  }

  // Import anything that's in the null package
  if (pkgName.size() == 0) {
    sOut.out() << "import " << fullWName << ";";
    sOut.Next();
    if (hasProtStatic) {
      sOut.out() << "import $token;";
      sOut.Next();
    }
  }
  for (deque<CJavaTypeSignature>::const_iterator i = 
	 IType.GetParameters()->begin();
       i != IType.GetParameters()->end(); ++i) {
    unicode_string className;
    if ((*i).GetBaseClassName(className)) {
      if (className.find_first_of((unicode_char) '/') ==unicode_string::npos) {
	sOut.out() << "import " << className << ";"; 
	sOut.Next();	
      }
    }   
  }

  sOut.out() << "public final class " << IName
	     << " extends " << fullWName;
  sOut.Next().PushIndent(2).out() << "{";
  
  ////
  // Make constructor code.
  ////
  sOut.Next().out() << "private " << IName << "() {}";
  sOut.Next().out() << "private static " << fullWName
		    << " the_instance = new " << IName << "();";
  sOut.Next().out() << "public static " << fullWName
		    << " getInstance() { return the_instance; }";
  
  ////
  // Call $stinit.
  ////
  if (PInfo.fHasStinit) {
    sOut.Next().Next().out() << "static { "
			     << fullWName << ".$stinit(the_instance); }";
  }
  
  
  ////
  // Generate static fields
  ////
  sOut.Next().Next().out() << "// static fields";
  sOut.Next();
  for (FieldTable::const_iterator i = Class->fFields.begin();
       i != Class->fFields.end(); ++i) {
    assert((*i).second);
    if ((*i).second->GetModifiers().fStatic) {
      sOut.out() << "private static ";
      if ((*i).first.GetType().IsPrimitive()) {
	(*i).first.Unparse(sOut);
      } else {
	sOut.out() << "Object " << ::UnicodeToUTF((*i).first.GetFieldName());
      }
      sOut.out() << ';';
      sOut.Next();
    }
  }
  if (PInfo.fHasStinit) {
    sOut.out() << "boolean $initialized;";
    sOut.Next();
  }

  ////
  // Generate AuxMethods
  ////
  sOut.Next().out() << "// AuxMethods";
  sOut.Next();
  for (amm_rep::const_iterator i = PInfo.GetAMM().Begin();
       i != PInfo.GetAMM().End(); ++i) {
    CAuxMethodSignature ams;
    ins.Instantiate(ams, (*i).first);
    assert((*i).second);
    ams.Unparse(*(*i).second, sOut, instantiationList, *this, IType,
		false);
  }

  sOut.PopIndent(2).Next().out() << "}\n";
  os.close();
  fInstantiationList.AddTypes(instantiationList, 0);  
  return true;
}
  
//
//  Method name : GenerateTClasses
//  Description : Generate a TClass for every package that needs one,
//      relative to basePath.
//
bool
CCompiler::GenerateTClasses(const string &basePath)
{
  bool success = true;
  for (string_set::iterator i = fTClassPackages.begin();
       i != fTClassPackages.end(); ++i) {
    success &= GenerateTClass(*i, basePath);
  }
  return success;
}

//
//  Method name : GenerateTClass
//  Description : Generate a TClass for 'package', relative to basePath.
//
bool
CCompiler::GenerateTClass(const unicode_string &package,
			  const string &basePath)
{
  string fileName = basePath + ::UnicodeToUTF(package) + "/$token.java";
  string symPkg = ::SymbolToString(package);
  if (!CFilePath::MakeDirectories(fileName)) {
    cerr << ">> Couldn't create directory for " << fileName << endl;
    return false;
  }
  ofstream os;
  os.open(fileName.c_str());
  if (!os.is_open()) {
    cerr << ">> Couldn't open " << fileName << endl;
    return false;
  }
  os 
    << "// Token class for " << symPkg << endl
    << "// (Generated automatically by PolyJ. Do not edit manually.)" 
    << endl << endl;
  if (symPkg.length() > 0)
    os << "package " << symPkg << ";" << endl;
  os << "public final class $token {" << endl
     << "  static final $token the_instance = new $token();" << endl
     << "  private $token() {};" << endl
     << "}" << endl;
  os.close();

  GeneratedFile(CJavaTypeSignature::kVoid, fileName);
    
  return true;
}

//
//  Method name : InitializePrimitiveClasses
//  Description : Add the interface classes for primitive types.
//
void 
CCompiler::InitializePrimitiveClasses()
{
  CJavaClassFile *c;
#define DECL_PRIMCLASS(sig, name) \
  c = ::CreateClassForType(CJavaTypeSignature:: sig, ::UTFToUnicode(name)); \
  fImportAliases.insert(ImportTable::value_type(::UTFToUnicode(name),c)); \
  fImportedClasses.push_back(c);

  DECL_PRIMCLASS(kByte,      " byte");
  DECL_PRIMCLASS(kShort,     " short");
  DECL_PRIMCLASS(kInteger,   " int");
  DECL_PRIMCLASS(kLong,      " long");
  DECL_PRIMCLASS(kFloat,     " float");
  DECL_PRIMCLASS(kDouble,    " double");
  DECL_PRIMCLASS(kCharacter, " char");
  DECL_PRIMCLASS(kBoolean,   " boolean");

  c = ::CreateClassForType(CJavaTypeSignature::kVoid,
			   ::UTFToUnicode(" array"));
  fImportAliases.insert(
         ImportTable::value_type(::UTFToUnicode(" array"), c));
  fImportedClasses.push_back(c); 
}

//
// Method name : TypeForConstraint
// Description : Returns the type needed to perform a constraint check on
//    type t, possibly translating primitive types to primitive classes.
//
CJavaTypeSignature
CCompiler::TypeForConstraint(const CJavaTypeSignature &t) const
{
  if (t.IsArray()) {
    CJavaTypeSignature member(t);
    member.SetArrayBounds(member.GetArrayBounds()-1);
    deque<CJavaTypeSignature> params;
    params.push_back(member);
    return CJavaTypeSignature(::UTFToUnicode(" array"), 0, &params);
  } else {
    switch (t.GetBaseType())
      {
      case CJavaTypeSignature::Class:
      case CJavaTypeSignature::Parameter:
	return t;
      case CJavaTypeSignature::Byte:
        return CJavaTypeSignature(::UTFToUnicode(" byte"));
      case CJavaTypeSignature::Character:
        return CJavaTypeSignature(::UTFToUnicode(" char"));
      case CJavaTypeSignature::Double:
        return CJavaTypeSignature(::UTFToUnicode(" double"));
      case CJavaTypeSignature::Float:
        return CJavaTypeSignature(::UTFToUnicode(" float"));
      case CJavaTypeSignature::Integer:
        return CJavaTypeSignature(::UTFToUnicode(" int"));
      case CJavaTypeSignature::LongInteger:
        return CJavaTypeSignature(::UTFToUnicode(" long"));
      case CJavaTypeSignature::Short:
        return CJavaTypeSignature(::UTFToUnicode(" short"));
      case CJavaTypeSignature::Boolean:
        return CJavaTypeSignature(::UTFToUnicode(" boolean"));
      case CJavaTypeSignature::Void:
        return CJavaTypeSignature(::UTFToUnicode(" void"));
      }
  }       
  assert(0);
}


