// Copyright (c) 1995  David Engberg  All rights reserved
// $Id: Expression.h,v 1.23 2000/07/03 18:53:31 andru Exp $
#ifndef _Expression_h
#define _Expression_h
#pragma interface

#include "JavaTypeSignature.h"
#include "dynamic_cast.h"
#include "parser_decls.h"
#include "JavaConstant.h"
#include "JavaFieldInfo.h"
class CJavaFieldSignature;
class CCompiler;
class CJavaClassFile;
class CCompileError;
class CJavaMethodInfo;
class CCompileContext;

//
//  Class name : CExpression
//  Description : An expression is used during compilation to represent a
//    value expression in Java.  CExpression is an abstract base class that
//    provides a common interface to the various types of expressions used
//    in the language.
//
class CExpression {
  DynamicCastDeclarations;
public:
  virtual ~CExpression();
  const CJavaTypeSignature& GetType() const { return fType; }
  bool IsLValue() const { return fLValue; }
  void SetLValue(bool isLValue) { fLValue = isLValue; }
  virtual bool IsLiteral() const;
  virtual bool MayBeType() const;
  virtual CJavaTypeSignature* ToType() const;
  virtual CExpression* Clone() const = 0;

  static CCompileError* EvaluateType(CExpression*& expressionPointer,
			 CCompileContext& context, CJavaTypeSignature& setType,
			 bool asLValue = false);
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
  virtual CCompileError* UnparseStore(srcstream &sOut,
				      CCompileContext& context,
				      CExpression& rValue,
				      bool isStatement = false);
  void InheritLineNumber(const CExpression& from);
  unsigned long GetLineNumber() const { return fLineNumber; }
  int GetPrecedence() const { return fPrecedence; }
  bool IsLValueAlias() const { return fLValAlias; }
  static CCompileError* UnparseWrap(const CJavaTypeSignature &wrapAs,
				    CExpression& wrapped,
				    srcstream& sOut,
 				    CCompileContext &context);
  static CCompileError* UnparseArgument(const CJavaTypeSignature &argType,
					const CJavaTypeSignature &transType,
					CExpression& wrapped,
					srcstream& sOut,
					CCompileContext &context);
  static void UnparseWClassCode(const CJavaTypeSignature &type,
			 CCompileContext &context,
			 srcstream& sOut);
protected:
  CExpression(const CJavaTypeSignature& type, bool isLValue = false);
  CExpression(const CExpression& source);
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);

  virtual CCompileError* Resolve(CCompileContext& context,
				 CExpression*& setExpression);
  
  CCompileError* MakeError(const unicode_string& message) const;
  CCompileError* MakeError(const string& message) const;

  static bool UnaryPromote(CExpression*& promotePointer,
			   CCompileContext& context);
  static bool BinaryPromote(CExpression*& left, CExpression*& right,
			   CCompileContext& context);
  void SetType(const CJavaTypeSignature& type);
  CJavaTypeSignature fType;
  bool fTypeHasBeenEvaluated;
  bool fLValue;
  unsigned long fLineNumber;
  int fPrecedence;
  bool fLValAlias; // Can this expression be a subexpression?
		   // (Will only be accurate after 
		   //  HandleEvaluateType.)
};

//
//  Class name : COrdinalLiteral
//  Description : Represents a literal of an ordinal type.  This includes
//     integers, characters, boolean, and long values.  Bytes and shorts
//     aren't included since they can't be expressed as a literal in a Java
//     program.
//
class COrdinalLiteral : public CExpression {
  DynamicCastDeclarations;
public:
  COrdinalLiteral(JavaUInt integerLiteral,
		  const CJavaTypeSignature& type);
  COrdinalLiteral(JavaUInt integerLiteral);
  COrdinalLiteral(JavaSInt integerLiteral);
  COrdinalLiteral(JavaUChar characterLiteral, bool isChar = true);
  COrdinalLiteral(JavaSShort shortLiteral);
  COrdinalLiteral(JavaUByte byteLiteral);
  COrdinalLiteral(JavaSByte byteLiteral);
  COrdinalLiteral(bool booleanLiteral);
  COrdinalLiteral(JavaULong longLiteral);
  COrdinalLiteral(JavaSLong longLiteral);
  COrdinalLiteral(const COrdinalLiteral& source);
  virtual ~COrdinalLiteral();
  virtual CExpression* Clone() const { return new COrdinalLiteral(*this); }

  virtual bool IsLiteral() const;
  JavaUInt GetInteger() const { return fOrdinalLiteral; }
  bool GetBoolean() const { return fOrdinalLiteral != 0; }
  JavaUChar GetCharacter() const { return fOrdinalLiteral; }
  JavaUShort GetShort() const { return fOrdinalLiteral; }
  JavaUByte GetByte() const { return fOrdinalLiteral; }
  JavaULong GetLong() const { return fOrdinalLiteral; }

  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  JavaConstantIndex AddToConstantPool(CJavaClassFile& inClass) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
private:
  unsigned long long fOrdinalLiteral;
};

//
//  Class name : CFloatLiteral
//  Description : Represents a literal of an floating-point type.
//     This includes float and double values.
//
class CFloatLiteral : public CExpression {
  DynamicCastDeclarations;
public:
  CFloatLiteral(float floatLiteral,string *optAdoptString=0);
  CFloatLiteral(double doubleLiteral,string *optAdoptString=0);
  CFloatLiteral(const CFloatLiteral& source);
  virtual ~CFloatLiteral();
  virtual CExpression* Clone() const { return new CFloatLiteral(*this); }
  
  virtual bool IsLiteral() const;
  float GetFloat() const { return fValue; }
  double GetDouble() const { return fValue; }
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  JavaConstantIndex AddToConstantPool(CJavaClassFile& inClass) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
private:
  double fValue;
  string *fString;
};


//
//  Class name : CStringLiteral
//  Description : Represents a string literal used in a Java program.
//
class CStringLiteral : public CExpression {
  DynamicCastDeclarations;
public:
  CStringLiteral(const unicode_string& value);
  CStringLiteral(const CStringLiteral& source);
  virtual ~CStringLiteral();
  virtual bool IsLiteral() const;
  virtual CExpression* Clone() const { return new CStringLiteral(*this); }
  unicode_string GetString() const { return fValue; }
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  JavaConstantIndex AddToConstantPool(CJavaClassFile& inClass) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
private:
  unicode_string fValue;
};


//
//  Class name : CNull
//  Description : Represents a reference expression with a value of null.
//
class CNull : public CExpression {
  DynamicCastDeclarations;
public:
  CNull();
  CNull(const CNull& other);
  virtual ~CNull();  
  virtual CExpression* Clone() const { return new CNull(*this); }
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
};

//
//  Class name : CCastExpression
//  Description : Represents an expression casting one value to a different
//    type.
//
class CCastExpression : public CExpression {
  DynamicCastDeclarations;
public:
  CCastExpression(const CJavaTypeSignature& toType, 
		  CExpression* adoptCastee,
		  bool fast = false);
  CCastExpression(const CCastExpression& source);
  virtual ~CCastExpression();
  virtual CExpression* Clone() const { return new CCastExpression(*this); }
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;

  static CCompileError* CastPointer(CExpression*& expressionPointer,
				const CJavaTypeSignature& toType,
				CCompileContext& context);
  static CCompileError* ImplicitCastPointer(CExpression*& expressionPointer,
					    const CJavaTypeSignature& toType,
					    CCompileContext& context);
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
private:
  CCompileError* MakeBadCastMessage(const CJavaTypeSignature& castFrom) const;
  CJavaTypeSignature fCastToType;
  CExpression* fCastee;
  bool fFastCast; // Don't add type-checking code.  Unused as yet.
};


//
//  Class name : CNewArray
//  Description : Represents an expression dynamically creating a new array
//    of values using the Java 'new' keyword.
//
class CNewArray : public CExpression {
  DynamicCastDeclarations;
public:
  CNewArray(const CJavaTypeSignature& type, ExpressionList* adoptArraySizes);
  CNewArray(CExpression* adoptBase, ExpressionList* adoptArraySizes,
	    int optDims);
  CNewArray(const CNewArray& source);
  virtual ~CNewArray();
  virtual CExpression* Clone() const { return new CNewArray(*this); }
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
protected:
  virtual CCompileError* Resolve(CCompileContext& context,
				 CExpression*& setExpression);
private:
  ExpressionList* fArraySizes;
  // Fields for ambiguous case
  CExpression* fBase;
  int fOptDims;
};

//
//  Class name : CNewObject
//  Description : Represents an expression dynamically creating a new instance
//    of a class using the Java 'new' keyword.
//
class CNewObject : public CExpression {
  DynamicCastDeclarations;
public:
  CNewObject(const unicode_string& className, ExpressionList* adoptArgs = 0);
  CNewObject(const CJavaTypeSignature& type, ExpressionList* adoptArgs = 0);
  CNewObject(const CNewObject& source);
  virtual ~CNewObject();
  virtual CExpression* Clone() const { return new CNewObject(*this); }
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
private:
  friend class CCompiler;
  ExpressionList* fConstructorArguments;
};

//
//  Class name : CTrinaryExpression
//  Description : Represents an expression composed of three parts.  Currently,
//    the only instance of this in Java is the  A ? B : C  syntax.
//
class CTrinaryExpression : public CExpression {
  DynamicCastDeclarations;
public:
  CTrinaryExpression(CExpression* leftExpression,
		     CExpression* middleExpression,
		     CExpression* rightExpression);
  CTrinaryExpression(const CTrinaryExpression& source);
  virtual ~CTrinaryExpression();
  virtual CExpression* Clone() const { return new CTrinaryExpression(*this); }
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
private:
  bool UnifyNumeric(CExpression*& first, CExpression*& second,
		    CCompileContext& context, CJavaTypeSignature& setType);

  CExpression* fLeft;
  CExpression* fMiddle;
  CExpression* fRight;
};

//
//  Class name : CUnaryExpression
//  Description : Represents an expression composed of a unary operator
//    applied to another expression.  Stores the child expression, the
//    operator, and some sense of whether it was prefix or postfix.
//
class CUnaryExpression : public CExpression {
  DynamicCastDeclarations;
public:
  typedef enum { kPrefix, kPostfix } Order;
  CUnaryExpression(CExpression* childExpression, unsigned short unaryOperator,
		   Order side, bool isLValue);
  CUnaryExpression(const CUnaryExpression& source);
  virtual ~CUnaryExpression();
  virtual CExpression* Clone() const { return new CUnaryExpression(*this); }
  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
private:
  CExpression* fChild;
  unsigned short fOperator;
  Order fSide;
};

//
//  Class name : CLocalVariableExpression
//  Description : A variable expression is used to mark the usage of a local
//    variable in a program.  These are distinct from non-local variables
//    (class fields) in that they can be uniquely bound during the first pass
//    of compilation.  (i.e. local variables must be declared before use.)
//
class CLocalVariableExpression : public CExpression {
  DynamicCastDeclarations;
public:
  CLocalVariableExpression(const CJavaFieldSignature& name,
			   unsigned short index,
			   bool hasParameterName = false);
  CLocalVariableExpression(const CLocalVariableExpression& source);
  virtual ~CLocalVariableExpression();
  virtual CExpression* Clone() const
     { return new CLocalVariableExpression(*this); }
  unsigned short GetVariableIndex() const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
  virtual CCompileError* UnparseStore(srcstream &sOut,
				      CCompileContext& context,
				      CExpression& rValue,
				      bool isStatement = false);
  virtual bool MayBeType() const;
  virtual CJavaTypeSignature* ToType() const;
private:
  unicode_string fName;
  unsigned short fLocalVariableIndex;
  bool fPName;
};

//
//  Class name : CClassFieldExpression
//  Description : This type of expression represents the use of a variable
//    in a function which isn't declared as part of that function.  The
//    first pass of the compiler just assumes for the time being that it
//    must be a valid field, and doesn't even bother looking it up, since
//    use-before-set is legal in Java.
//    The semantic analyzer checks it later for validity and figures out
//    whether it is a static field like FooClass.x or whether it is a
//    member field off of 'this.'  At that point, it constructs an expression
//    of the _real_ type and holds it internally.
//
class CClassFieldExpression : public CExpression {
  DynamicCastDeclarations;
public:
  CClassFieldExpression(const unicode_string& label,
			bool hasParameterName = false,
			bool isDefault = false);
  CClassFieldExpression(CExpression* adoptBase, const unicode_string& label);
  CClassFieldExpression(CJavaTypeSignature* adoptType,
			const unicode_string& label);
  CClassFieldExpression(const CClassFieldExpression& source);
  virtual ~CClassFieldExpression();
  virtual CExpression* Clone() const
     { return new CClassFieldExpression(*this); }
  
  bool IsStatic() const;
  bool IsArrayLength() const;
  unicode_string GetFieldString() const { return fLabel; }

  virtual bool MayBeType() const;
  virtual CJavaTypeSignature* ToType() const;

  virtual CCompileError* CreateConstantExpression(CExpression*& result,
			  CCompileContext& context) const;
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
  virtual CCompileError* UnparseStore(srcstream &sOut,
				      CCompileContext& context,
				      CExpression& rValue,
				      bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
  CCompileError* EvaluateBaseType(CCompileContext& context,
				  CJavaTypeSignature& setType);
  virtual CCompileError* Resolve(CCompileContext& context,
				 CExpression*& setExpression);

private:
  unicode_string fLabel;
  CJavaFieldInfo fFieldInfo;
  const CJavaClassFile* fClassAlias;
  CExpression* fBase;
  CJavaTypeSignature* fBaseType;
  bool fNeedCast;
  bool fNeedWrap;
  bool fHasPName;
  bool fIsDefault;
};


//
//  Class name : CArrayInitializer
//  Description : This type of expression is only used in initialization of
//    arrays using a special hokey syntax.  It's basically a wrapper for
//    a list of expressions that should all match the type of the array.
//
class CArrayInitializer : public CExpression {
  DynamicCastDeclarations;
public:
  CArrayInitializer(ExpressionList* adoptExpressions = 0);
  CArrayInitializer(const CArrayInitializer& source);
  virtual ~CArrayInitializer();
  virtual CExpression* Clone() const { return new CArrayInitializer(*this); }
  void SetArrayType(const CJavaTypeSignature& type);
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
  CCompileError* UnparseAssignments(const CExpression *assignTo,
				    srcstream &sOut,
				    CCompileContext &context,
				    bool allLevels);
protected:
  friend class CVariableDeclaration;
private:
  ExpressionList fExpressions;
};

//
//  Class name : CArrayIndex
//  Description : This represents an expression made using the array element
//    notation to extract an entry from an array value.
//
class CArrayIndex : public CExpression {
  DynamicCastDeclarations;
public:
  CArrayIndex(CExpression* baseExpression, CExpression* index);
  CArrayIndex(const CArrayIndex& source);
  virtual ~CArrayIndex();
  virtual CExpression* Clone() const { return new CArrayIndex(*this); }
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
  virtual CCompileError* UnparseStore(srcstream &sOut,
				      CCompileContext& context,
				      CExpression& rValue,
				      bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
private:
  CExpression* fBase;
  CExpression* fIndex;
};

//
//  Class name : CMethodCall
//  Description : This expression represents the invocation of a method
//    upon an object.  This combines the target object, the name of the
//    method, and a list of the arguments.
//
class CMethodCall : public CExpression {
  DynamicCastDeclarations;
public:
  CMethodCall(CExpression* adoptTarget, const unicode_string& methodName,
	     ExpressionList* adoptArguments = 0, bool forceNonvirtual = false);
  CMethodCall(CJavaTypeSignature* adoptTargetType, 
	      const unicode_string& methodName,
	      ExpressionList* adoptArguments = 0, 
	      bool forceNonvirtual = false);
  CMethodCall(const unicode_string& methodName,
	     ExpressionList* adoptArguments = 0, bool forceNonvirtual = false);
  CMethodCall(const CMethodCall& source);
  virtual ~CMethodCall();
  virtual CExpression* Clone() const { return new CMethodCall(*this); }
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
  CCompileError* EvaluatePrefixType(unicode_string prefix,
                                    CCompileContext& context);
  virtual CCompileError* Resolve(CCompileContext& context,
				 CExpression*& setExpression);
  bool IsUniversal(CCompileContext& context,
		   const deque<CJavaTypeSignature> &argumentTypes,
		   CJavaTypeSignature &setType);

private:
  CExpression* fTarget;
  unicode_string fMethodName;
  ExpressionList fArguments;
  CJavaTypeSignature* fTargetType;
  const CJavaClassFile* fClassAlias;
  const CJavaMethodInfo* fMethodAlias;
  const CPJavaConstraint::ConstraintMethod* fConstraintMethodAlias;
  bool fForceNonvirtual;
  bool fOnObject;
  bool fIsArrayClone;
  bool fIsPArrayClone;
};

//
//  Class name : CSpecialExpression
//  Description : This is kind of a catch-all for expressions that require
//    special handling, and are generally inconsistent throughout.
//
class CSpecialExpression : public CExpression {
  DynamicCastDeclarations;
public:
  typedef enum { kThis, kSuper } Type;
  CSpecialExpression(Type expressionType);
  CSpecialExpression(const CSpecialExpression& source);
  virtual ~CSpecialExpression();
  virtual CExpression* Clone() const { return new CSpecialExpression(*this); }
  Type GetType() const { return fType; }

  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
protected:
  virtual CCompileError* HandleEvaluateType(CCompileContext& context,
					    CJavaTypeSignature& setType);
private:
  Type fType;
};

//
//  Class name : CInstanceof
//  Description : This is used for the expression  (foo instanceof String)
//    which checks if an object descends from a given base.
//
class CInstanceof : public CExpression {
  DynamicCastDeclarations;
public:
  CInstanceof(CExpression* adoptBase, const CJavaTypeSignature& toType);
  CInstanceof(const CInstanceof& source);
  virtual ~CInstanceof();
  virtual CExpression* Clone() const { return new CInstanceof(*this); }
  virtual CCompileError* Unparse(srcstream &sOut,
				 CCompileContext& context,
				 bool isStatement = false);
private:
  CExpression* fBase;
  CJavaTypeSignature fToType;
};

//
//  Class name  : CTypeOrExpression
//  Description : This is basically a protected, tagged, expression/type
//     union.
//
class CTypeOrExpression {
public:
  CTypeOrExpression(CExpression *adoptExp);
  CTypeOrExpression(CJavaTypeSignature *adoptType);
  CTypeOrExpression(const CTypeOrExpression& source);
  bool IsType() const { return fIsType; };
  CExpression *GetExpression() const;
  CJavaTypeSignature *GetType() const;
  ~CTypeOrExpression();
private:
  bool fIsType;
  union { 
    CJavaTypeSignature* t;
    CExpression* e;
  } fVal;
};

typedef deque<CTypeOrExpression*> TypeOrExpList;

//
//  Class name  : CAmbiguousAccess
//  Description : This represents a (possibly ambiguous) expression of
//      the form Name[ ExpressionList ].
//
class CAmbiguousAccess : public CExpression {
  DynamicCastDeclarations;
public:
  CAmbiguousAccess(unicode_string* adoptName, 
		   CExpression* adoptBase,
		   TypeOrExpList* adoptEList,
		   bool hasPName);
  CAmbiguousAccess(const CAmbiguousAccess& source);
  virtual ~CAmbiguousAccess();
  virtual CAmbiguousAccess* Clone() const  
              { return new CAmbiguousAccess(*this); }
  virtual bool MayBeType() const;
  virtual CJavaTypeSignature* ToType() const;
  bool MayBeIndex() const;
  CArrayIndex* ToIndex() const;
  
  CCompileError* ToTypeAndDimension(CJavaTypeSignature& setType,
				    CExpression*& setExpression);

  CCompileError* ResolveAE(CCompileContext& context,
			   CExpression*& setExpression,
			   CJavaTypeSignature& setType);
protected:
  virtual CCompileError* Resolve(CCompileContext& context,
				 CExpression*& setExpression);
private:
  unicode_string* fBaseString;
  CExpression* fBaseExpression;
  TypeOrExpList* fEList;
  bool fHasPName;
};

#endif
