// ------------- Things to fix ------------
// NOTE 0: The interface to 'compiler' only parses one file at a time.  
//  * This means that, if we see multiple .pj files on the command line,
//    we still won't parse them ahead of time if they aren't on the classpath.
// NOTE 1: We currently keep _no_ metadata.
//  * This means that we cannot 'Sync' properly, or remove unneeded ITypes.

#include "config.h"
#include "find_leaks.h"
#include "JavaTypeSignature.h"
#include <map>
#include <set>
#include <deque>
#include "unicode_string.h"
#include "mapped_file.h"
#include "CommandLine.h"
#include "FilePath.h"
#include "JavaDirectory.h"
#include "Compiler.h"
#include "JavaConstant.h"
#include "Expression.h"
#include "parser_decls.h"
#include "parser.h"
#include "JavaClassFile.h"
#include <string>
#include <iostream.h>
#include <fstream.h>
extern "C" {
  #include <stdlib.h>
  #include <unistd.h>
  #include <stdio.h>
}

// Define this if we're on MIT's Athena system
// #define ATHENA

// Define this if we want running times.
// #define PROFILE

#ifdef PROFILE
extern "C" {
  #include <sys/time.h>
}
long gStartSec;
long gStartuSec;
#endif

/////////
// PolyJMain: main file for translator/compiler wrapper.
/////////

///////
// Globals:  // (Set by "ParseCmdLine")
///////

// Command names
string gJavacName;

// File and names
string gInsFileName;
string gDepFileName;
string gRootPath;
string gClassDest;
string gJavaDest;
deque<string> gClassPath;
string gClassPathStr;
bool gVerbose;
bool gVeryVerbose;
bool gDebugFlag;
bool gOptFlag;
bool gSync;
bool gForceRecompile;
bool gPrintDependencies;
string gInsPkg;
deque<string> gPJSourceFiles;
deque<string> gJavaSourceFiles;

/////
// Prototypes
/////

void Usage();
bool ParseCommandLine(const CCommandLine& arguments);
void AddSlash(string &s);
void FixClassPath();
bool ParseClassPath(const string& pathString, deque<string>& pathList);
string PackageOfFile(const string &fname);
bool TranslateFiles();
void ReduceJavaFiles();
void CompileFiles();

//
//  Function name: Usage
//  Description  : Print a short usage message, then exit.
//
void Usage() {
  cerr 
    << "Usage: polyj [-r rootdir] [-c classdest] [-t javadest]"
       " [-classpath classpath]"
    << endl
    << "             [-I insPackage] [-sync] [-force] [-g] [-O] [-v] [-V]"
    << endl 
    << "             [SourceFile ...]" 
    << endl << endl
    << "See http://www.pmg.lcs.mit.edu/polyj for more information about PolyJ."
    << endl;
  exit(1);
}

//
//  Function name : AddSlash
//  Description   : Adds a slash to s, iff s is a non-empty string which does
//     not already have a slash.  Used to fix pathnames.
void AddSlash(string &s)
{
  if ((s.length() != 0)
      && (s[s.length()-1] != '/')) {
    s += '/';
  }
}

// 
//  Function name: ParseCommandLine
//  Description  : Set the global variables, from the command line arguments,
//     from the environment, or from the defaults.  Returns true on success.
//     
bool
ParseCommandLine(const CCommandLine& arguments)
{
  //////
  // Set defaults:
  //////

  gVerbose = gVeryVerbose = gDebugFlag = gOptFlag = 
    gSync = gForceRecompile = gPrintDependencies = false;

  /////
  // DEFAULTS HERE
  /////

  gJavacName = "javac";
  gRootPath = ".";
  gInsPkg = "pjIns";

  string classPathStr;

  ////
  // Now try the environment
  ////
  char *envP; 
  if ((envP = getenv("POLYJROOT"))) {
    gRootPath = envP;
  }
  if ((envP = getenv("POLYJCLASSTARGET"))) {
    gClassDest = envP;
  }
  if ((envP = getenv("POLYJJAVATARGET"))) {
    gJavaDest = envP;
  }
  if ((envP = getenv("CLASSPATH"))) {
    classPathStr = envP;
  }
  if ((envP = getenv("JAVACOMPILER"))) {
    gJavacName = envP;
  }
  if ((envP = getenv("PJINSPACKAGE"))) {
    gInsPkg = envP;
  }

  ////
  // Now, parse the command line
  ///
  for (CCommandLine::ArgumentIteratorType arg = arguments.GetArgumentBegin();
       arg != arguments.GetArgumentEnd(); ++arg) {
    unsigned long optionLength = (*arg).length();
    if (*arg == "-classpath") {
      ++arg;
      if (arg == arguments.GetArgumentEnd()) {
	Usage();
      } else {
	classPathStr = *arg;
      }
    } else if (*arg == "-r") {
      ++arg;
      if (arg == arguments.GetArgumentEnd()) {
	Usage();
      } else {
	gRootPath = *arg;
      }
    } else if (*arg == "-c") {
      ++arg;
      if (arg == arguments.GetArgumentEnd()) {
	Usage();
      } else {
	gClassDest = *arg;
      }
    } else if (*arg == "-t") {
      ++arg;
      if (arg == arguments.GetArgumentEnd()) {
	Usage();
      } else {
	gJavaDest = *arg;
      }
    } else if (*arg == "-I") {
      ++arg;
      if (arg == arguments.GetArgumentEnd()) {
	Usage();
      } else {
	gInsPkg = *arg;
      }
    } else if (*arg == "-sync") {
      gSync = true;
    } else if (*arg == "-v") {
      gVerbose = true;
    } else if (*arg == "-V") {
      gVeryVerbose = gVerbose = true;
    } else if (*arg == "-M") {
      gPrintDependencies = true;
    } else if (*arg == "-force") {
      gForceRecompile = true;
    } else if (*arg == "-g") {
      gDebugFlag = true;
    } else if (*arg == "-O") {
      gOptFlag = true;
    } else if (optionLength > 0 && (*arg)[0] == '-') {
      cerr << ">> Invalid compiler flag: " << *arg << endl;
      Usage();
    } else {
      CFilePath filePath(*arg);
      if (!filePath.IsFile()) {
	cerr << ">> Invalid argument (not a plain file): " << *arg << endl;
	Usage();
      } else if (optionLength > 5 &&
		 (*arg).compare(".java",
				optionLength - 5) == 0) {
	gJavaSourceFiles.push_back(*arg);
      } else if	(optionLength > 3 &&
		 (*arg).compare(".pj", 
				optionLength - 3) == 0) {
	gPJSourceFiles.push_back(*arg);
      } else {
	cerr << ">> Invalid source file (bad suffix): " << *arg << endl;
	Usage();
      }
    }
  }

  if (gVerbose) {
    cerr << "polyj version " VERSION << endl;
  }
  if (gJavaSourceFiles.size() == 0 &&
      gPJSourceFiles.size() == 0) {
    cerr << "No source files specified." << endl;
    Usage();
  }
   

  /////
  // Now, set the un-set paths.
  /////
  if (gClassDest.length() == 0) {
    gClassDest = gRootPath;
    gClassDest += "/classes";
  }
  if (gJavaDest.length() == 0) {
    gJavaDest = gRootPath;
    gJavaDest += "/translated";
  }
  AddSlash(gClassDest);
  AddSlash(gJavaDest);
  if (!CFilePath::MakeDirectories(gClassDest)) {
    cerr << ">> Could not create path to " << gClassDest << endl;
    exit(1);
  }
  if (!CFilePath::MakeDirectories(gJavaDest)) {
    cerr << ">> Could not create path to " << gJavaDest << endl;
    exit(1);
  }

  gClassPathStr = classPathStr;
  FixClassPath();
  if (!ParseClassPath(gClassPathStr, gClassPath)) {
    cerr << "Could not parse class path " << gClassPathStr;
  }
  gDepFileName = gClassDest;
  gDepFileName += "/.PolyJ";  
  return true;
}

//
//  Function name : ParseClassPath
//  Description   : Set pathList to contain the paths listed in pathString,
//      which should be a colon-separated list of paths.
//
bool ParseClassPath(const string& pathString, 
		    deque<string>& pathList)
{
  string::size_type startPosition = 0;
  string::size_type colonPosition;
  do {
    colonPosition = pathString.find(PATH_DELIM, startPosition);
    if (colonPosition > pathString.size()) {
      colonPosition = pathString.size();
    }
    string directoryName(pathString, startPosition,
			 colonPosition - startPosition);
    directoryName = CJavaDirectory::Convert(directoryName);
    CJavaDirectory path(directoryName, true);
    if (path.IsValid()) {
      pathList.push_back(directoryName);
    } else {
      cerr << ">> Warning: Invalid classpath entry \"" << directoryName << "\""
	    << endl;
    }
    startPosition = colonPosition + 1;
  } while(colonPosition < pathString.size());
  return !pathList.empty();
}

//
//  Method name : FixClassPath
//  Description : Add extra material to classpath.
//
void FixClassPath()
{
  if (gClassPathStr.length() > 0) {
    gClassPathStr = gClassDest + PATH_DELIM + gClassPathStr + PATH_DELIM + gRootPath;
  } else {
    gClassPathStr = gClassDest + PATH_DELIM + gRootPath;
  }
}

//
//  Method name : TranslateFiles
//  Description : Translates all PolyJ files.  Aborts on failure.
//
bool TranslateFiles()
{
  cerr << "Translating:" << endl;
  string IFiles;
  CCompiler::ClassList resultClasses;
  for(deque<string>::iterator i = gJavaSourceFiles.begin();
      i != gJavaSourceFiles.end(); ++i) {
    gPJSourceFiles.push_back(*i);
  }  
  bool success = CCompiler::TranslateFiles(gPJSourceFiles, gClassPath, 
					   resultClasses,
					   gJavaDest, gJavaDest,
					   gInsPkg, IFiles, 
					   gForceRecompile,
					   gPrintDependencies);
  if (!success) {
    cerr << "   Translation failed." << endl;
    return false;
  }
  ////
  // Make pjaux files.
  ////
  for (CCompiler::ClassList::iterator i = resultClasses.begin();
       i != resultClasses.end(); ++i) {
    string className = ::UnicodeToString((*i)->GetShortName());
    if (gVerbose)
      cerr << "   [Writing pjaux file for " << className << "]" << endl;
    string outFileName = gClassDest + 
      ::UnicodeToString((*i)->GetClassName()) + ".pjaux";
    if ((*i)->HasPInfo()) {
      if (!CFilePath::MakeDirectories(outFileName)) {
	cerr << ">> Couldn't make path for " << outFileName << endl;
	exit(1);
      }
      ofstream outFile;
      outFile.open(outFileName.c_str());
      if (!outFile.is_open()) {
	cerr << ">> could not open output file " << outFileName << endl;
	exit(1);
      }
      outFile << (*i)->GetPInfo().WriteToString(*(*i));
      outFile.close();
    }
  }
  cerr << "   Translation succeeded." << endl;

  ////
  // Delete classfile objects.
  ////
  for (CCompiler::ClassList::iterator i = resultClasses.begin();
       i != resultClasses.end(); ++i) {
    delete *i;
  }

  ////
  // Read IFiles string.
  ////
  for (string::size_type pos = 0;
       pos < IFiles.size(); ) {
    string::size_type eol = IFiles.find_first_of('\n',pos);
    if (pos != eol) {
      string oneFile(IFiles,pos,eol-pos);
      if (gVerbose)
	cerr << "   [Generated " << oneFile << "]" << endl;
      gJavaSourceFiles.push_back(oneFile);
    }
    pos = eol+1;
  }  
  return success;
}

//
//  Method name : ReduceJavaFiles
//  Description : Makes sure that all filenames in gJavaSourceFiles are unique.
//
void ReduceJavaFiles()
{
  set<unicode_string, less<unicode_string> > allFiles;
  deque<string>::iterator end = gJavaSourceFiles.end();
  for(deque<string>::iterator i = gJavaSourceFiles.begin();
      i != end; ++i) {
    unicode_string fname = ::UTFToUnicode(*i);
    if (allFiles.find(fname) == allFiles.end()) {
      allFiles.insert(fname);
      gJavaSourceFiles.push_back(*i);
    }
  }
  // This prevents the symbol tables from exploding under the DEC linker.
  gJavaSourceFiles.erase(gJavaSourceFiles.begin(),end);
}

#define convert CJavaDirectory::Convert

//
//  Method name : CompileFiles
//  Description : Invoke javac on the files.
//
void CompileFiles()
{
  cerr << "Compiling:" << endl;
  string cmd = "'" + convert(gJavacName) + "'"
    + " -nowarn -d '" + convert(gClassDest) + "'"
    + " -classpath '" + convert(gClassPathStr) + PATH_DELIM + convert(gJavaDest) + "'";
  if (gDebugFlag) cmd += " -g";
  if (gOptFlag) cmd += " -O";
    
  for(deque<string>::iterator i = gJavaSourceFiles.begin();
      i != gJavaSourceFiles.end(); ++i) {
    cmd += " ";
    cmd += "'";
    cmd += convert(*i);
    cmd += "'";
  }
  if (gVerbose)
    cerr << "   [Executing " << cmd << "]" << endl;
  
  int ret = system(cmd.c_str());
  if (ret != 0) {
    cerr << "   Compilation failed." << endl;
    exit(1);
  } else {
    cerr << "   Compilation succeeded." << endl;
  }
}

int main(int argc, const char** argv)
{
  CCommandLine arguments;
#ifdef PROFILE
  struct timeval tv;
  gettimeofday(&tv,NULL);
  gStartSec = tv.tv_sec;
  gStartuSec = tv.tv_usec;
#endif
  
  arguments.ParseArgumentVector(argc, argv);  
  ParseCommandLine(arguments);

  bool success = TranslateFiles();
  if (!success) return 1;
  
  ReduceJavaFiles();
#ifdef PROFILE
  gettimeofday(&tv,NULL);
  if (tv.tv_usec<gStartuSec) {tv.tv_usec+=1000000; tv.tv_sec-=1; }
  cerr << "   (" << tv.tv_sec-gStartSec << " s, "
       << tv.tv_usec-gStartuSec << " us)." << endl;
  gStartSec = tv.tv_sec;
  gStartuSec = tv.tv_usec;
#endif

  if (gJavaSourceFiles.size() > 0)
    CompileFiles();

#ifdef PROFILE
  gettimeofday(&tv,NULL);
  if (tv.tv_usec<gStartuSec) {tv.tv_usec+=1000000; tv.tv_sec-=1; }
  cerr << "   (" << tv.tv_sec-gStartSec << " s, "
       << tv.tv_usec-gStartuSec << " us)." << endl;
#endif
  
  return 0;
}

#ifndef NDEBUG
void ds(const unicode_char *s) {
    cerr << "\"";
    while (*s != 0) {
	cerr << (char)(*s);
	s++;
    }
    cerr << "\"" << endl;
}
#endif
