// Copyright (c) 1995  David Engberg  All rights reserved
#pragma implementation
#include "find_leaks.h"
#include "Statement.h"
#include "Expression.h"
#include "VariableDeclaration.h"
#include "JavaAccessFlags.h"
#include "parser_decls.h"
#include "CompileError.h"
#include "JavaClassFile.h"
#include "JavaMethodInfo.h"
#include "JavaCodeAttribute.h"
#include "Compiler.h"
#include "CompileContext.h"
#include "Srcstream.h"
#include <climits>

//
//  Method name : CStatement
//  Description : Default constructor.  Read the comment for CExpression() for
//    my lame attempt at justifying accessing the 'yylineno' variable like
//    this.
//
CStatement::CStatement()
  : fLineNumber(yylineno)
{
}

//
//  Method name : ~CStatement
//  Description : Destructor
//
CStatement::~CStatement()
{
}

//
//  Method name : MakeError
//  Description : This provides a convenient hook for children of CStatement
//    to create compiler errors from just a string.
//
CCompileError*
CStatement::MakeError(const unicode_string& message)
{
  return new CCompileError(message, fLineNumber);
}

//
//  Method name : MakeError
//  Description : This provides a convenient hook for children of CStatement
//    to create compiler errors from just a string.
//
CCompileError*
CStatement::MakeError(const string& message)
{
  return new CCompileError(::StringToUnicode(message), fLineNumber);
}

//
//  Method name : PrintError
//  Description : This static method prints a given error message.  Used when
//     one statment gives more than one error message, and we need to print 
//     the first and return the second.
void 
CStatement::PrintError(const CCompileError* error, 
		       const CCompileContext &context)
{
  string message("In ");
  CJavaMethodSignature m = context.GetMethod().GetSignature();
  if (m.GetName() == CCompiler::kConstructorName)
    message += "a constructor";
  else if (m.GetName() == CCompiler::kStaticName)
    message += "static code";
  else 
    message += ::UnicodeToString(m.Disassemble());
  message += " in class " 
    + ::SymbolToString(context.GetClass().GetClassName()) +
    ": " + ::UnicodeToString(error->GetMessage());
  if (error->GetLine() == CCompileError::kNoLineNumber) {
    context.GetCompiler().PrintCompileError(context.GetSourceFileName(),
					    message);
  } else {
    context.GetCompiler().PrintCompileError(context.GetSourceFileName(), 
					    message,error->GetLine());
  }
}

//
//  Method name : MergeError
//  Description : If newErr != 0, prints oldErr (if present) 
//     and sets oldErr to newErr.
//
void 
CStatement::MergeError(CCompileError*& oldErr,
		       CCompileError* newErr,
		       const CCompileContext& context)
{
  if (oldErr && newErr) {
    PrintError(oldErr, context);
    delete oldErr;
    oldErr = 0;
  }
  if (newErr)
    oldErr = newErr;
}

 
//
//  Method name : Unparse
//  Description : Tries to unparse the given statement to sOut.  In addition,
//     it updates initialization and reachability flags in 'context'.
//     Returns an error on failure, 0 on success.  If more than one error 
//     occurs, it may print all but the lexically final error, and return
//     the last.
//
CCompileError*
CStatement::Unparse(srcstream &sOut,
		    CCompileContext &context)
{
  CCompileError *error = 0;
  if (context.IsReachable()) {
    error = HandleUnparse(sOut,context);
  } else {
    error = MakeError("Code cannot be reached");
  }
  return error;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CStatement::HandleUnparse(srcstream &sOut,
		    CCompileContext &context)
{
  assert(0);
  return 0; 
}


//=========================== CExpressionStatement ============================

//
//  Method name : CExpressionStatement
//  Description : Constructs a CExpressionStatement out of an adopted
//    expression.
//
CExpressionStatement::CExpressionStatement(CExpression* adoptExpression)
  : fExpression(adoptExpression)
{
  fLineNumber = fExpression->GetLineNumber();
}

//
//  Method name : ~CExpressionStatement
//  Description : Destructor.
//
CExpressionStatement::~CExpressionStatement()
{
  delete fExpression;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CExpressionStatement::HandleUnparse(srcstream &sOut,
			      CCompileContext &context)
{
  assert(fExpression != 0);
  CJavaTypeSignature type;
  CCompileError* error = 
    CExpression::EvaluateType(fExpression,context,type);
  if (error == 0)
    error = fExpression->Unparse(sOut,context,true);
  sOut.out() << ';';
  return error;
}


//============================ CCompoundStatement =============================

//
//  Method name : CCompoundStatement
//  Description : Constructs a CCompoundStatement from an adopted list of
//    sub-statements.
//
CCompoundStatement::CCompoundStatement(StatementList* adoptStatements)
  : fChildren(adoptStatements)
{
  fLastLineNumber = fLineNumber;
  if (fChildren == 0) {
    fChildren = new StatementList;
  }
  if (! fChildren->empty() && fChildren->front() != 0)
    fLineNumber = fChildren->front()->GetLineNumber();
}

//
//  Method name : ~CCompoundStatement
//  Description : Destructor.
//
CCompoundStatement::~CCompoundStatement()
{
  for (StatementList::iterator i = fChildren->begin();
       i != fChildren->end(); ++i) {
    delete *i;
  }
  delete fChildren;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CCompoundStatement::HandleUnparse(srcstream &sOut,
			    CCompileContext &context)
{
  sOut.ToLineF(fLineNumber-1);
  sOut.out() << "{";
  sOut.PushIndent(2);
  
  CCompileError *error = UnparseNoBrackets(sOut,context);

  sOut.PopIndent(2);
  sOut.ToLine(fLastLineNumber);
  sOut.out() << "}";

  return error; 
}

//
//  Method name : UnparseNoBrackets
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message. (No brackets are used.)
//    Used by CCompoundStatement::Unparse and CForStatement::Unparse
//
CCompileError*
CCompoundStatement::UnparseNoBrackets(srcstream &sOut,
				      CCompileContext &context)
{
  assert(fChildren != 0);
  CCompileError* error = 0;
  CCompileError* newError = 0;
  // Invariant: error!=0 only if newError!=0.  Either error==newError==0,
  //            or newError!=error.
  for (StatementList::iterator i = fChildren->begin();
       i != fChildren->end(); ++i) {
    assert(*i != 0);
    newError = (*i)->Unparse(sOut, context);
    MergeError(error, newError, context);
  }
  return error;
}


//=========================== CDeclarationStatement ===========================

//
//  Method name : CDeclarationStatement
//  Description : Constructs a CDeclarationStatement out of a list of single
//    variable declarations.
//
CDeclarationStatement::CDeclarationStatement(
		       deque<CVariableDeclaration*>* adoptDeclarations,
		     CJavaAccessFlags* modifiers, bool deprecated)
   : fDeclarations(adoptDeclarations),
     fModifiers(modifiers),
     fDeprecated(deprecated)
{
  if (fDeclarations == 0) {
    fDeclarations = new deque<CVariableDeclaration*>;
  }
  if (fModifiers == 0) {
    fModifiers = new CJavaAccessFlags;
  }
  if (!fDeclarations->empty()) {
    fLineNumber = fDeclarations->front()->GetLineNumber();
  }
}

//
//  Method name : ~CDeclarationStatement
//  Description : Destructor.
//
CDeclarationStatement::~CDeclarationStatement()
{
  for (deque<CVariableDeclaration*>::iterator i = fDeclarations->begin();
       i != fDeclarations->end(); ++i) {
    delete *i;
  }
  delete fDeclarations;
  delete fModifiers;
}

//
//  Method name : IsInitialized
//  Description : returns true iff one of the variables declared in this
//     statement is initialized.
//
bool 
CDeclarationStatement::HasInitializer() const
{
  assert(fDeclarations != 0);
  for (deque<CVariableDeclaration*>::iterator i = fDeclarations->begin();
       i != fDeclarations->end(); ++i) {
    if ((*i)->HasInitializer())
      return true;
  }
  return false;
}

//
// Method name : HandleUnparse
// Description : Write the statement to sOut, in canonical Java code.
//
CCompileError*
CDeclarationStatement::HandleUnparse(srcstream& sOut,
				     CCompileContext& context)
{
  assert(fDeclarations != 0 && fModifiers != 0);
  CCompileError* error = 0;
  for (deque<CVariableDeclaration*>::iterator i = fDeclarations->begin();
       i != fDeclarations->end(); ++i) {
    assert(*i != 0);
    sOut.ToLine((*i)->GetLineNumber(),false).out() << fModifiers->Unparse();
    CCompileError *newError = (*i)->Unparse(sOut,context);
    MergeError(error, newError, context);
  }       
  return error;
}

//
// Method name : UnparseInitializers
// Description : Write the statement's initializers to sOut, without the
//    declarations.
//
CCompileError*
CDeclarationStatement::UnparseInitializers(srcstream& sOut,
					   CCompileContext& context)
{
  assert(fDeclarations != 0 && fModifiers != 0);
  CCompileError* error = 0;
  for (deque<CVariableDeclaration*>::iterator i = fDeclarations->begin();
       error == 0 && i != fDeclarations->end(); ++i) {
    assert(*i != 0);
    error = (*i)->UnparseInitializer(sOut,context);
  }       
  return error;
}

//
// Method name : UnparseNoInit
// Description : Write the statement's declarations to sOut, without the
//    initializers
//
CCompileError*
CDeclarationStatement::UnparseNoInit(srcstream& sOut,
					   CCompileContext& context)
{
  assert(fDeclarations != 0 && fModifiers != 0);
  CCompileError* error = 0;
  for (deque<CVariableDeclaration*>::iterator i = fDeclarations->begin();
       error == 0 && i != fDeclarations->end(); ++i) {
    assert(*i != 0); 
    sOut.ToLine((*i)->GetLineNumber(),false).out() << fModifiers->Unparse();
    CCompileError *newError = (*i)->UnparseNoInitializer(sOut,context);
    MergeError(error, newError, context);
  }       
  return error;
}


//=============================== CIfStatement ================================

//
//  Method name : CIfStatement
//  Description : Constructs a CIfStatement from its parts
//
CIfStatement::CIfStatement(CExpression* condition, CStatement* thenClause,
			   CStatement* elseClause)
   : fCondition(condition),
     fThenClause(thenClause),
     fElseClause(elseClause)
{
  if (condition != 0)
    fLineNumber = condition->GetLineNumber();
}

//
//  Method name : ~CIfStatement
//  Description : Destructor.
//
CIfStatement::~CIfStatement()
{
  delete fCondition;
  delete fThenClause;
  delete fElseClause;
}


//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CIfStatement::HandleUnparse(srcstream &sOut,
			    CCompileContext &context)
{
  assert(fCondition != 0);
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fCondition, context, type);
  if (error == 0 && type != CJavaTypeSignature::kBoolean) {
    error = MakeError("Non-boolean value used as 'if' conditional.");
  }
  if (error == 0) {
    sOut.ToLine(fLineNumber).out()
      << "if (";
    error = fCondition->Unparse(sOut, context);
    sOut.out() << ")";
  }
  CCompileContext elseContext = context;
  if (error == 0) { 
    sOut.PushIndent(2);
    if (fThenClause != 0) {     
      error = fThenClause->Unparse(sOut,context);
    } else {
      sOut.out() << ";";
    }
    sOut.PopIndent(2);
  }
  if (error == 0 && fElseClause != 0) {
    sOut.out() << " else ";
    sOut.PushIndent(2);
    error = fElseClause->Unparse(sOut,elseContext);
    sOut.PopIndent(2);
  }
  context.Merge(elseContext);

  return error; 
}


//============================== CWhileStatement ==============================

//
//  Method name : CWhileStatement
//  Description : Constructs a CWhileStatement
//
CWhileStatement::CWhileStatement(CStatement* adoptLoop,
				 CExpression* adoptCondition)
   : fCondition(adoptCondition),
     fLoopStatement(adoptLoop)     
{
  if (adoptCondition != 0)
    fLineNumber = adoptCondition->GetLineNumber();
}

//
//  Method name : ~CWhileStatement
//  Description : Destructor.
//
CWhileStatement::~CWhileStatement()
{
  delete fLoopStatement;
  delete fCondition;
}

//
//  Method name : RegisterContinueBranch
//  Description : This method tells the loop construct that a continue
//    statement was found in its body and gives the loop the location of the
//    'goto' code in the instruction block.  The loop is then responsible for
//    adjusting the target of the 'goto' to jump to the right location.
//
void
CWhileStatement::RegisterContinueBranch(const CNonlocalBranch& instruction)
{
  fNestedContinues.push_front(instruction);
}

//
//  Method name : RegisterBreakBranch
//  Description : This method tells the loop construct that a break
//    statement was found in its body and gives the loop the location of the
//    'goto' code in the instruction block.  The loop is then responsible for
//    adjusting the target of the 'goto' to jump to the right location.
//
void
CWhileStatement::RegisterBreakBranch(const CNonlocalBranch& instruction)
{
  fNestedBreaks.push_front(instruction);
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CWhileStatement::HandleUnparse(srcstream &sOut,
			       CCompileContext &context)
{
  assert(fCondition != 0);
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fCondition, context, type);
  if (error == 0 && type != CJavaTypeSignature::kBoolean) {
    return MakeError("Non-boolean value used as 'while' conditional.");
  }
  COrdinalLiteral* boolLiteral = DYNAMIC_CAST(COrdinalLiteral, fCondition);
  bool reachableBody = true;
  if (boolLiteral != 0) {
    if (!boolLiteral->GetBoolean()) {
      reachableBody = false;
    }
  }
  if (error == 0) {
    sOut.ToLine(fLineNumber).out()
      << "while (";
    error = fCondition->Unparse(sOut, context);
    sOut.out() << ")";
  }
  if (error == 0) {
    CCompileContext bodyContext = context;
    bodyContext.GetCompiler().PushStatementContext(this); 
    sOut.PushIndent(2);
    if (fLoopStatement != 0) {
      error = fLoopStatement->Unparse(sOut,bodyContext);
    } else {
      sOut.out() << ";";
    }
    bodyContext.GetCompiler().PopStatementContext();
    sOut.PopIndent(2);
  }
  if (boolLiteral != 0) {
    if (boolLiteral->GetBoolean())
      context.SetUnreachable();
  }
  for (list<CNonlocalBranch>::iterator i = fNestedBreaks.begin();
       !(i == fNestedBreaks.end()); ++i) {
    context.Merge((*i).GetContext());
  }  
  return error; 
}


//============================== CDoStatement ==============================

//
//  Method name : CDoStatement
//  Description : Constructs a CDoStatement
//
CDoStatement::CDoStatement(CStatement* adoptLoop,
			   CExpression* adoptCondition)
   : fLoopStatement(adoptLoop),
     fCondition(adoptCondition)
{
  if (adoptLoop != 0)
    fLineNumber = adoptLoop->GetLineNumber();
}

//
//  Method name : ~CDoStatement
//  Description : Destructor.
//
CDoStatement::~CDoStatement()
{
  delete fLoopStatement;
  delete fCondition;
}

//
//  Method name : RegisterContinueBranch
//  Description : This method tells the loop construct that a continue
//    statement was found in its body and gives the loop the location of the
//    'goto' code in the instruction block.  The loop is then responsible for
//    adjusting the target of the 'goto' to jump to the right location.
//
void
CDoStatement::RegisterContinueBranch(const CNonlocalBranch& instruction)
{
  fNestedContinues.push_front(instruction);
}

//
//  Method name : RegisterBreakBranch
//  Description : This method tells the loop construct that a break
//    statement was found in its body and gives the loop the location of the
//    'goto' code in the instruction block.  The loop is then responsible for
//    adjusting the target of the 'goto' to jump to the right location.
//
void
CDoStatement::RegisterBreakBranch(const CNonlocalBranch& instruction)
{
  fNestedBreaks.push_front(instruction);
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CDoStatement::HandleUnparse(srcstream &sOut,
			    CCompileContext &context)
{
  assert(fCondition != 0);
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fCondition, context, type);
  if (error == 0 && type != CJavaTypeSignature::kBoolean) {
    error = MakeError("Non-boolean value used as 'do' conditional.");
  }
  if (error == 0) {
    if (fLoopStatement != 0) {
      sOut.ToLineBefore(fLoopStatement->GetLineNumber(),1);
    } else {
      sOut.ToLineBefore(fCondition->GetLineNumber(),2);
    }
    sOut.out() << "do ";
    sOut.PushIndent(2);
    if (fLoopStatement != 0) {
      context.GetCompiler().PushStatementContext(this);
      error = fLoopStatement->Unparse(sOut,context);
      context.GetCompiler().PopStatementContext();
    } else {
      sOut.ToLineF(fCondition->GetLineNumber()-1);
      sOut.out() << ";";
    }
    sOut.PopIndent(2);
  }
  if (error == 0) {
    sOut.ToLine(fCondition->GetLineNumber()).out()
      << "while (";
    error = fCondition->Unparse(sOut,context);
    sOut.out() << ");";
  }
  if (!fNestedBreaks.empty()) {
    for(list<CNonlocalBranch>::iterator i = fNestedBreaks.begin();
	!(i == fNestedBreaks.end()); ++i) {
      context.Merge((*i).GetContext());
    }
  }
  return error; 
}

//============================== CReturnStatement =============================

//
//  Method name : CReturnStatement
//  Description : Constructs a CReturnStatement that will return the given
//    value (if one is provided).
//
CReturnStatement::CReturnStatement(CExpression* returnValue)
   : fValue(returnValue)
{
  if (returnValue)
    fLineNumber = returnValue->GetLineNumber();
}

//
//  Method name : ~CReturnStatement
//  Description : Destructor.
//
CReturnStatement::~CReturnStatement()
{
  delete fValue;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CReturnStatement::HandleUnparse(srcstream &sOut,
				CCompileContext &context)
{
  CCompileError* error = 0;
  CJavaTypeSignature returnType = context.GetMethod().GetSignature().GetType();
  CJavaTypeSignature transType = 
    context.GetMethod().GetTranslatedSignature().GetType();
  error = context.GetCompiler().CheckType(returnType,
				 context.GetClass(),context.GetSource());
  if (error != 0)
    return error;

  if (fValue == 0) {
    if (returnType != CJavaTypeSignature::kVoid) {
      error = MakeError("Invalid empty return from non-void function.");
    } else {
      sOut.ToLine(fLineNumber).out() << "return;";
    }
  } else {
    CJavaTypeSignature type;
    error = CExpression::EvaluateType(fValue, context, type);
    if (error == 0) {
      if (!context.GetCompiler().SameType(returnType, type,
					  context.GetSource())) {
	error =
	  CCastExpression::ImplicitCastPointer(fValue, returnType, context);
      }
      if (error == 0) {
	// Do we need to wrap?
	bool needWrap = returnType.IsPrimitive() && ! transType.IsPrimitive();
	sOut.ToLine(fValue->GetLineNumber()).out() <<
	  "return ";
	if (needWrap) {
	  error = CExpression::UnparseWrap(returnType, *fValue, sOut, context);
	} else {
	  error = fValue->Unparse(sOut,context);
	}
	sOut.out() << ';';
      }
      if (error == 0 && returnType.GetBaseType() ==CJavaTypeSignature::Void) {
	    error =
	      MakeError("Invalid value returned from method declared 'void'");
      }
    }
  }
  context.SetUnreachable();
  return error;
}


//============================== CBranchStatement =============================

//
//  Method name : CBranchStatement
//  Description : Constructs a CBranchStatement
//
CBranchStatement::CBranchStatement(CBranchStatement::Type type,
				   unicode_string* label)
   : fBranchType(type),
     fToLabel(label)
{
}

//
//  Method name : ~CBranchStatement
//  Description : Destructor.
//
CBranchStatement::~CBranchStatement()
{
  delete fToLabel;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CBranchStatement::HandleUnparse(srcstream &sOut,
				CCompileContext &context)
{
  sOut.ToLine(fLineNumber);
  StatementList::iterator contextPtr =
    context.GetCompiler().GetContextBegin();
  StatementList::iterator contextEnd =
    context.GetCompiler().GetContextEnd();
  CCompileError *error = 0;
  for(; error == 0 && contextPtr != contextEnd; ++contextPtr) {
    CNonlocalBranch thisBranch(0,context);
    if (fToLabel != 0) {
      CLabelStatement *label = DYNAMIC_CAST(CLabelStatement, *contextPtr);
      if (label != 0 && label->GetLabel() != 0 &&
	  *label->GetLabel() == *fToLabel) {
	if (fBranchType == kBreak) {
	  label->RegisterBreakBranch(thisBranch);
	  break;
	} else {
	  CStatement* labelChild = label->GetChildStatement();
	  assert(labelChild);
	  CForStatement* forS = DYNAMIC_CAST(CForStatement, labelChild);
	  if (forS) 
	    forS->RegisterContinueBranch(thisBranch);
	  break; // We don't register continues on non-for stuff.
	}
      }
    } else {
      if (fBranchType == kBreak) {
	CSwitch* switchStatement = DYNAMIC_CAST(CSwitch, *contextPtr);
	if (switchStatement) {
	  switchStatement->RegisterBreakBranch(thisBranch);
	  break;
	}
      }
      CWhileStatement* whileS = DYNAMIC_CAST(CWhileStatement, *contextPtr);
      if (whileS) {
	if (fBranchType == kBreak) {
	  whileS->RegisterBreakBranch(thisBranch);
	} // We don't need to register continue.
	break;
      }
      CDoStatement* doS = DYNAMIC_CAST(CDoStatement, *contextPtr);
      if (doS) {
	if (fBranchType == kBreak) {
	  doS->RegisterBreakBranch(thisBranch);
	} // We don't need to register continue.
	break;
      }
      CForStatement* forS = DYNAMIC_CAST(CForStatement, *contextPtr);
      if (forS) {
	if (fBranchType == kBreak) 
	  forS->RegisterBreakBranch(thisBranch);
	else
	  forS->RegisterContinueBranch(thisBranch);
	break;
      }
    }
  }
  if (error == 0 && contextPtr == contextEnd) {
    if (fBranchType == kBreak)
      error = MakeError("Invalid 'break' statement without matching loop.");
    else 
      error = MakeError("Invalid 'continue' statement without matching loop.");
  }
  context.SetUnreachable();
  if (fBranchType == kBreak) {
    sOut.out() << "break";
  } else {
    sOut.out() << "continue";
  }
  if (fToLabel != 0) 
    sOut.out() << ' ' << ::SymbolToString(*fToLabel);
  sOut.out() << ';';
  return error; 
}

//=============================== CForStatement ===============================

//
//  Method name : CForStatement
//  Description : Constructs a CForStatement
//
CForStatement::CForStatement(CStatement* initializer, CExpression* conditional,
			     ExpressionList* incrementor, CStatement* body)
   : fInitializer(initializer),
     fConditional(conditional),
     fIncrementor(incrementor),
     fBody(body)
{
  if (initializer != 0)
    fLineNumber=initializer->GetLineNumber();
  else if (conditional != 0)
    fLineNumber=conditional->GetLineNumber();
  else if (incrementor != 0 && incrementor->front() != 0)
    fLineNumber=incrementor->front()->GetLineNumber();
  else if (body != 0)
    fLineNumber=body->GetLineNumber();
}

//
//  Method name : ~CForStatement
//  Description : Destructor.
//
CForStatement::~CForStatement()
{
  delete fInitializer;
  delete fConditional;
  delete fBody;
  if (fIncrementor != 0) {
    for (ExpressionList::iterator i = fIncrementor->begin();
	 i != fIncrementor->end(); ++i) {
      delete *i;
    }
    delete fIncrementor;
  }
}

//
//  Method name : RegisterContinueBranch
//  Description : This method tells the loop construct that a continue
//    statement was found in its body and gives the loop the location of the
//    'goto' code in the instruction block.  The loop is then responsible for
//    adjusting the target of the 'goto' to jump to the right location.
//
void
CForStatement::RegisterContinueBranch(const CNonlocalBranch& instruction)
{
  fNestedContinues.push_front(instruction);
}

//
//  Method name : RegisterBreakBranch
//  Description : This method tells the loop construct that a break
//    statement was found in its body and gives the loop the location of the
//    'goto' code in the instruction block.  The loop is then responsible for
//    adjusting the target of the 'goto' to jump to the right location.
//
void
CForStatement::RegisterBreakBranch(const CNonlocalBranch& instruction)
{
  fNestedBreaks.push_front(instruction);
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CForStatement::HandleUnparse(srcstream &sOut,
			     CCompileContext &context)
{
  CCompileError* error = 0;
  sOut.ToLine(fLineNumber);
    
  sOut.out() << "for (";
  sOut.PushIndent(5);
 
  if (fInitializer != 0) {
    CCompoundStatement *list = DYNAMIC_CAST(CCompoundStatement, fInitializer);
    if (list != 0) {
      error = list->UnparseNoBrackets(sOut,context);
    } else {
      error = fInitializer->Unparse(sOut, context);
    }
  } else {
    sOut.out() << "; ";
  }
  if (error == 0 && fConditional != 0) {
    CJavaTypeSignature type;
    error = CExpression::EvaluateType(fConditional, context, type);
    if (error == 0 && type != CJavaTypeSignature::kBoolean) {
      error = MakeError("Non-boolean value used as 'for' conditional.");
    }    
  }
  sOut.out() << ' ';
  if (error == 0 && fConditional != 0) {
    error = fConditional->Unparse(sOut,context);
  }
  sOut.out() << "; ";
  if (error == 0 && fIncrementor != 0) {
    CCompileContext incrContext = context;
    // XXXX This is a hack; we need to unparse the incrementor before
    //      the body, but we don't know the incrementor's context until we
    //      unparse the body.    
    if (fBody && error == 0) {
      srcstream &nullStr = srcstream::GetNullStream();
      incrContext.GetCompiler().PushStatementContext(this);
      error = fBody->Unparse(nullStr,incrContext);
      incrContext.GetCompiler().PopStatementContext();
      if (!fNestedContinues.empty()) {
	for(list<CNonlocalBranch>::iterator i = fNestedContinues.begin();
	    !(i == fNestedContinues.end()); ++i) {
	  incrContext.Merge((*i).GetContext());
	}
      }
    }    
    CJavaTypeSignature type;
    for (ExpressionList::iterator i = fIncrementor->begin();
	 error == 0 && i != fIncrementor->end(); ) {
      error = CExpression::EvaluateType((*i), incrContext, type);
      if (error == 0) {
	error = (*i)->Unparse(sOut,incrContext,true);
	++i;
	if (i != fIncrementor->end()) 
	  sOut.out() << ", ";
      }
    }
  }
  sOut.out() << ')';
  sOut.PopIndent(3);

  CCompileContext bodyContext = context;
  if (error == 0 && fBody != 0) {
    bodyContext.GetCompiler().PushStatementContext(this);
    error = fBody->Unparse(sOut,bodyContext);
    bodyContext.GetCompiler().PopStatementContext();
  } else if (error == 0) {
    sOut.out() << " ;";
  }
  sOut.PopIndent(2);
  if (!fNestedContinues.empty()) {
    for(list<CNonlocalBranch>::iterator i = fNestedContinues.begin();
	!(i == fNestedContinues.end()); ++i) {
    bodyContext.Merge((*i).GetContext());
    }
  }

  for (list<CNonlocalBranch>::iterator i = fNestedBreaks.begin();
       !(i == fNestedBreaks.end()); ++i) {
    context.Merge((*i).GetContext());
  }
  return error; 
}

//============================== CThrowStatement =============================

//
//  Method name : CThrowStatement
//  Description : Constructs a CThrowStatement that will throw the given
//    value.
//
CThrowStatement::CThrowStatement(CExpression* throwValue)
   : fValue(throwValue)
{
  if (fValue)
    fLineNumber = fValue->GetLineNumber();
}

//
//  Method name : ~CThrowStatement
//  Description : Destructor.
//
CThrowStatement::~CThrowStatement()
{
  delete fValue;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CThrowStatement::HandleUnparse(srcstream &sOut,
			       CCompileContext &context)
{
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fValue, context, type);
  if (error == 0 && !context.GetCompiler().IsThrowable(type,
						       context.GetSource())) {
    error = MakeError("Attempt to throw a non-throwable value.");
  } else if (!context.Throwable(type)) {
    unicode_string message = type.Disassemble();
    message += ::UTFToUnicode(" must be caught or declared in this method's "
			      " 'throws' list.");
    error = MakeError(message);
  } else if (type.IsParameterized()) {
    error = MakeError("Attempted to throw parameterized type.");
  }
  if (error == 0) {
    sOut.ToLine(fValue->GetLineNumber()).out() << "throw ";
    error = fValue->Unparse(sOut, context);
  }   
  sOut.out() << ';';
  context.SetUnreachable();

  return error; 
}

//=============================== CSynchronized ===============================

//
//  Method name : CSynchronized
//  Description : Constructs a CSynchronized statement block.
//
CSynchronized::CSynchronized(CExpression* adoptCondition,
			     CStatement* adoptBlock,
			     unsigned short synchronizeVariable)
   : fCondition(adoptCondition),
     fBlock(adoptBlock),
     fSynchronizeVariable(synchronizeVariable)
{
  if (adoptCondition != 0)
    fLineNumber = adoptCondition->GetLineNumber();
}

//
//  Method name : ~CSynchronized
//  Description : Destructor.
//
CSynchronized::~CSynchronized()
{
  delete fCondition;
  delete fBlock;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CSynchronized::HandleUnparse(srcstream &sOut,
			     CCompileContext &context)
{
  assert(fCondition != 0 && fBlock != 0);
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fCondition, context, type);
  if (error == 0 && !type.IsReference()) {
    error = MakeError("'synchronized' used with non-reference expression");
  }
  sOut.ToLine(fLineNumber).out() << 
    "synchronized (";
  if (error == 0) {
    error = fCondition->Unparse(sOut,context);
  }
  sOut.out() << ")";
  if (error == 0) {
    error = fBlock->Unparse(sOut,context);
  }

  return error; 
}

//============================== CLabelStatement ==============================

//
//  Method name : CLabelStatement
//  Description : Constructs a CLabelStatement for a labelled jump.
//
CLabelStatement::CLabelStatement(unicode_string* adoptLabel,
				 CStatement* adoptStatement)
  : fLabel(adoptLabel),
    fCaseExpression(0),
    fStatement(adoptStatement),
    fType(kLabel)
{
  if (adoptStatement != 0)
    fLineNumber = adoptStatement->GetLineNumber();
}

//
//  Method name : CLabelStatement
//  Description : Constructs a CLabelStatement for a 'default' statement
//
CLabelStatement::CLabelStatement(CStatement* adoptDefault)
  : fLabel(0),
    fCaseExpression(0),
    fStatement(adoptDefault),
    fType(kDefault)
{
}

//
//  Method name : CLabelStatement
//  Description : Constructs a CLabelStatement for a 'case' statement
//
CLabelStatement::CLabelStatement(CExpression* caseExpression,
				 CStatement* adoptCase)
  : fLabel(0),
    fCaseExpression(caseExpression),
    fStatement(adoptCase),
    fType(kCase)
{
}

//
//  Method name : ~CLabelStatement
//  Description : Destructor.
//
CLabelStatement::~CLabelStatement()
{
  delete fLabel;
  delete fCaseExpression;
  delete fStatement;
}

//
//  Method name : GetLabel
//  Description : Retrieves the label that is applied using this label
//    statement and returns it to the user.  If this is a case or default
//    label, 0 is returned.
//
const unicode_string*
CLabelStatement::GetLabel() const
{
  return fLabel;
}

//
//  Method name : RegisterBreakBranch
//  Description : This method tells the label construct that a break
//    statement was found in its body and gives this statement the location
//    of the 'goto' code in the instruction block.  The label is then
//    responsible for adjusting the target of the 'goto' to jump to the right
//    location.
//
void
CLabelStatement::RegisterBreakBranch(const CNonlocalBranch& instruction)
{
  fNestedBreaks.push_front(instruction);
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CLabelStatement::HandleUnparse(srcstream &sOut,
			       CCompileContext &context)
{
  // We don't indent labels very well.
  CCompileError* error = 0;
  CJavaTypeSignature type;

  sOut.ToLineBefore(fLineNumber,1);

  switch(fType) {
  case kLabel: 
    assert(fLabel != 0);
    sOut.out() << ::SymbolToString(*fLabel) << ':';
    break;
  case kCase:
    assert(fCaseExpression != 0);
    sOut.out() << "case ";
    error = CExpression::EvaluateType(fCaseExpression, context, type);
    if (error == 0)
      error = fCaseExpression->Unparse(sOut,context);
    sOut.out() << ':';
    break;
  case kDefault:
    sOut.out() << "default:";
    break;
  default:
    assert(0);
  }

  if (error == 0) {
    if (fStatement != 0) {
      context.GetCompiler().PushStatementContext(this);
      error = fStatement->Unparse(sOut,context);
      context.GetCompiler().PopStatementContext();
    } else {
      sOut.ToLine(fLineNumber).out() << ";";
    }
  }
  if (!fNestedBreaks.empty()) {
    for(list<CNonlocalBranch>::iterator i = fNestedBreaks.begin();
	!(i == fNestedBreaks.end()); ++i) {
      context.Merge((*i).GetContext());
    }
  }
  return error; 
}

//============================== CTryStatement ==============================

//
//  Method name : CTryStatement
//  Description : Constructs a CTryStatement
//
CTryStatement::CTryStatement(CCompoundStatement* TryBlock,
		     deque<CCatchClause*>* catchClauses,
		     CCompoundStatement* finally,
		     unsigned short finallyHandlerVariable,
		     unsigned short finallySubroutineVariable)
   : fTryBlock(TryBlock),
     fCatchList(catchClauses),
     fFinally(finally),
     fFinallyHandlerVariable(finallyHandlerVariable),
     fFinallySubroutineVariable(finallySubroutineVariable),
     fFinallySubroutineLocation(0)
{
  if (fCatchList == 0) {
    fCatchList = new deque<CCatchClause*>;
  }
  if (TryBlock)
    fLineNumber = TryBlock->GetLineNumber(); 
}

//
//  Method name : ~CTryStatement
//  Description : Destructor.
//
CTryStatement::~CTryStatement()
{
  delete fTryBlock;
  delete fFinally;
  for (deque<CCatchClause*>::iterator i = fCatchList->begin();
       i != fCatchList->end(); ++i) {
    delete *i;
  }
  delete fCatchList;
}

//
//  Method name : CCatchClause
//  Description : Constructs a catch clause for a Try statement
//
CCatchClause::CCatchClause(unsigned short variableIndex,
			   const CJavaTypeSignature& catchType,
			   const CJavaFieldSignature& catchDecl,
			   CCompoundStatement* block)
   : fExceptionVariable(variableIndex),
     fCatchType(catchType),
     fCatchDecl(catchDecl),
     fBlock(block)
{
}

//
//  Method name : ~CCatchClause
//  Description : Destructor.
//
CCatchClause::~CCatchClause()
{
  delete fBlock;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CTryStatement::HandleUnparse(srcstream &sOut,
			     CCompileContext &context)
{
  assert(fTryBlock != 0 && fCatchList != 0);
  CCompileError *error = 0;
  // We don't deal with PushFinallyHandler
  CCompileContext inContext = context;
  sOut.ToLineF(fLineNumber-1).out() << "try";
  context.GetCompiler().PushStatementContext(this);
  for (deque<CCatchClause*>::iterator i = fCatchList->begin();
       !(i == fCatchList->end()); ++i) {
    context.PushThrowable(context.GetCompiler().FixType((*i)->fCatchType,
							context.GetSource()));
  }
  error = fTryBlock->Unparse(sOut,context);
  for (unsigned i = 0; i < fCatchList->size(); ++i) {
    context.PopThrowable();
  }
  if (fCatchList->size() > 0) {
    for (deque<CCatchClause*>::iterator i = fCatchList->begin();
	 !(i == fCatchList->end()); ++i) {
      assert((*i) != 0 && (*i)->fBlock != 0);
      assert((*i)->fCatchType.IsReference());
      CCompileContext catchContext = inContext;
      CCompileError *newError = 0;
      if ((*i)->fCatchType.IsParameterized() || 
	  (*i)->fCatchType.IsParameter()) {
	newError = MakeError("Tried to catch parameterized type");
	MergeError(error, newError, context);
      } else {
	sOut.ToLineF((*i)->fBlock->GetLineNumber()-1).out() << "catch (";
	CJavaTypeSignature tp = (*i)->fCatchDecl.GetType();
	newError = context.GetCompiler().CheckType(tp,context.GetClass(),
						   context.GetSource());
	MergeError(error, newError, context);
	CJavaFieldSignature fld(tp, (*i)->fCatchDecl.GetFieldName());
	fld.Unparse(sOut);
	sOut.out() << ")";
	catchContext.InitializeVariable((*i)->fExceptionVariable);     
	newError = (*i)->fBlock->Unparse(sOut,catchContext);
	MergeError(error, newError, context);
	context.Merge(catchContext);
      }
    }
  }
  if (error == 0 && fFinally != 0) {
    sOut.out() << "finally";
    error = fFinally->Unparse(sOut,inContext);
    if (!inContext.IsReachable())
      context.SetUnreachable();
  }
  return error; 
}

//================================== CSwitch ==================================

//
//  Method name : CSwitch
//  Description : Constructs a CSwitch statement from the expression that
//    forms the basis for the jump and the block of statements to jump into.
//
CSwitch::CSwitch(CExpression* adoptSwitch, CCompoundStatement* block)
  : fSwitchExpression(adoptSwitch),
    fBlock(block)
{
  if (adoptSwitch)
    fLineNumber = adoptSwitch->GetLineNumber();
}

//
//  Method name : ~CSwitch
//  Description : Destructor.
//
CSwitch::~CSwitch()
{
  delete fSwitchExpression;
  delete fBlock;
}

//
//  Method name : ScanStatementForLabels
//  Description : This is a little helper function used by CSwitch to go
//    through a statement to try to find the top-level 'case' and 'default'
//    labels which it needs to know about.  In the process, it figures out
//    the highest and lowest 'case' labels as well as a count of all case
//    labels found.  If this evaluation leads to an error, this method returns
//    a newly-allocated compile error.
//
CCompileError*
CSwitch::ScanStatementForLabels(CStatement& statement,
			   vector<unsigned long>& keys, bool& foundDefault,
			   long& highKey, long& lowKey,
			   CCompileContext& context)
{
  CCompileError* error = 0;
  CLabelStatement* label = DYNAMIC_CAST(CLabelStatement, &statement);
  if (label != 0) {
    switch (label->fType) {
    case CLabelStatement::kCase:
      {
	CJavaTypeSignature caseType;
	error = CExpression::EvaluateType(label->fCaseExpression,
					  context, caseType);
	if (error == 0) {
	  COrdinalLiteral* key =
	    DYNAMIC_CAST(COrdinalLiteral, label->fCaseExpression);
	  if (key == 0) {
	    error = new CCompileError("Invalid 'case' label.",
				      label->fCaseExpression->GetLineNumber());
	  } else {
	    long keyValue = key->GetInteger();
	    for (vector<unsigned long>::iterator i = keys.begin();
		i != keys.end(); ++i) {
	      if (*i == (unsigned long)keyValue) {
		error = new CCompileError("Duplicate 'case' label.",
				      label->fCaseExpression->GetLineNumber());
		break;
	      }
	    }
	    if (error == 0) {
	      keys.push_back((unsigned long)keyValue);
	      highKey = (keyValue > highKey) ? keyValue : highKey;
	      lowKey = (keyValue < lowKey) ? keyValue : lowKey;
	      error = ScanStatementForLabels(*label->fStatement, keys,
				     foundDefault, highKey, lowKey, context);
	    }
	  }
	}
      }
      break;
    case CLabelStatement::kDefault:
      if (foundDefault) {
	error = MakeError("Two 'default' tags found in body of 'switch'");
      } else {
	foundDefault = true;
	error = ScanStatementForLabels(*label->fStatement, keys,
				       foundDefault, highKey, lowKey, context);
      }
      break;
    default:
      break;
    }
  }
  return error;
}
 
//
//  Method name : RegisterBreakBranch
//  Description : This method tells the switch construct that a break
//    statement was found in its body.  This gives the switch the location of
//    the 'goto' code in the instruction block.  The loop is then responsible
//    for adjusting the target of the 'goto' to jump to the right location.
//
void
CSwitch::RegisterBreakBranch(const CNonlocalBranch& instruction)
{
  fNestedBreaks.push_front(instruction);
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CSwitch::HandleUnparse(srcstream &sOut,
		       CCompileContext &context)
{
  // No typechecking on case statements.
  assert(fSwitchExpression != 0 && fBlock != 0 && fBlock->fChildren != 0);
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fSwitchExpression, context, type);
  context.GetCompiler().PushStatementContext(this);
  if (error == 0) {
    sOut.ToLine(fLineNumber).out() << "switch (";
    error = fSwitchExpression->Unparse(sOut, context);
    sOut.out() << ") {";
  }
  if (error == 0) {    
    bool foundDefault = false;
    CCompileContext currentContext = context;
    currentContext.SetUnreachable();
    sOut.PushIndent(2);    
    for (StatementList::iterator i = fBlock->fChildren->begin();
	 i != fBlock->fChildren->end(); ++i) {
      CLabelStatement* statement = DYNAMIC_CAST(CLabelStatement, *i);
      if (statement) {
	switch (statement->fType) {
	case CLabelStatement::kDefault:
	  foundDefault = true;
	case CLabelStatement::kCase:
	  currentContext.Merge(context);
	  break;
	default: break;
	}
      }  
      CCompileError* newError = (*i)->Unparse(sOut,currentContext);
      MergeError(error, newError, context);
    }
    sOut.PopIndent(2).Next().out() << "}";
    if (!foundDefault) {
      context.Merge(currentContext);
    } else {
      context = currentContext;
    }
    if (!fNestedBreaks.empty()) { 
      for(list<CNonlocalBranch>::iterator i = fNestedBreaks.begin();
	  !(i == fNestedBreaks.end()); ++i) {
	context.Merge((*i).GetContext());
      }
    }
  } 
  context.GetCompiler().PopStatementContext();
  return error; 
}


//========================= CExplicitConstructorCall ==========================

//
//  Method name : CExplicitConstructorCall
//  Description : Constructs a CExplicitConstructorCall
//
CExplicitConstructorCall::CExplicitConstructorCall(Type type,
					   ExpressionList* adoptArguments)
  : fType(type)
{
  CExpression* target = (type == kThis) ?
    new CSpecialExpression(CSpecialExpression::kThis) :
    new CSpecialExpression(CSpecialExpression::kSuper);
  fExpression =
    new CMethodCall(target, CCompiler::kConstructorName, adoptArguments, true);
}

//
//  Method name : ~CExplicitConstructorCall
//  Description : Destructor.
//
CExplicitConstructorCall::~CExplicitConstructorCall()
{
  delete fExpression;
}

//
//  Method name : HandleUnparse
//  Description : Tries to unparse the given statment into the srcstream sOut.
//    returns 0 on success, otherwise an error message.
//
CCompileError*
CExplicitConstructorCall::HandleUnparse(srcstream &sOut,
				 CCompileContext &context)
{
  assert(fExpression != 0);
  CJavaTypeSignature type;
  CCompileError* error =
    CExpression::EvaluateType(fExpression, context, type);
  if (error == 0)
    error = fExpression->Unparse(sOut, context, true);
  sOut.out() << ';';
  return error;
}



